/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// attldcl.c
///  ATTLIST declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "prsstate.h"
#include "quantity.h"
#include "dlmstr.h"
#include "sgmldcl.h"
#include "prsmode.h"
#include "rsvname.h"
#include "sort.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "attldcl.h"
#include "notadcl.h"

/*STATIC-----------------------------------------------inst_new_declared_value-+
| This procedure is called by parse_name_group for each declared value         |
| occuring in a group.                                                         |
+-----------------------------------------------------------------------------*/
static void inst_new_declared_value(void *pvBase)
{
   ((AttlistDcl *)pvBase)->new_declared_value();
}

/*STATIC---------------------------------------------------inst_new_attlist_gi-+
| This procedure is called by parse_name_group for each associated element     |
| name, or associated notation name for this ATTLIST.                          |
+-----------------------------------------------------------------------------*/
static void inst_new_attlist_gi(void *pvBase)
{
   ((AttlistDcl *)pvBase)->new_attlist_gi();
}

/*-----------------------------------------------------------------------parse-+
| Process ATTLIST MD (Attribute Definition List Declaration)                   |
|                                                                              |
| Enter here when ATTLIST has been decoded,                                    |
| Returns with curr_char (supposed to ) == MDC                                 |
|                                                                              |
| [141] attribute definition list declaration =                                |
|         mdo, "ATTLIST", ps+,                                                 |
|         (associated element type | associated notation name), ps+            |
|         attribute definition list,                                           |
|         ps*, mdc)                                                            |
|                                                                              |
|                                                                              |
| [142] attribute definition list = attribute definition,                      |
|                 (ps+, attribute definition)*                                 |
|                                                                              |
| [143] attribute definition =                                                 |
|                 attribute name, ps+, declared value, ps+, default value      |
|                                                                              |
| [144] attribute name = name                                                  |
|                                                                              |
| [145] declared value = "CDATA" | "ENTITY" | "ENTITIES" |                     |
| [146]                  "ID" | "IDREF" | "IDREFS" |                           |
|                        "NAME" | "NAMES" | "NMTOKEN" | "NMTOKENS" |           |
|                        "NUMBER" | "NUMBERS" | "NUTOKEN" | "NUTOKENS" |       |
|                        ("NOTATION", ps+, name group) | name token group      |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
|                                                                              |
|  `att.iListCount' is the field to examine to know if a significant default   |
|  value has been specified: if so, `att.iListCount' is >= 1.                  |
|                                                                              |
|  It's not safe to check the field `att.ofsLit': it can be non-zero, even     |
|  when no default AVS (as in <!ATTLIST LI type (ord|unord|std) #REQUIRED>).   |
|  `att.ofsLit == 0' means "no AVS specified" only when there also is no       |
|  "choice list" -- group of name or name token in the declared value --,      |
|  as in <!ATTLIST XY ENTITIES #REQUIRED>.                                     |
+-----------------------------------------------------------------------------*/
void AttlistDcl::parse(const e_LinkKind lkkindArg)
{
   static const VarntChoiceEntry dclv_choices[] = {
      { CDATA_VRN_ix,     CDATA_ADV    }, // character data
      { ENTITY_VRN_ix,    ENTITY_ADV   }, // general entity name
      { ENTITIES_VRN_ix,  ENTITIES_ADV },
      { ID_VRN_ix,        ID_ADV       }, // id value
      { IDREF_VRN_ix,     IDREF_ADV    }, // id reference value
      { IDREFS_VRN_ix,    IDREFS_ADV   },
      { NAME_VRN_ix,      NAME_ADV     }, // name
      { NAMES_VRN_ix,     NAMES_ADV    },
      { NMTOKEN_VRN_ix,   NMTOKEN_ADV  }, // name token
      { NMTOKENS_VRN_ix,  NMTOKENS_ADV },
      { NOTATION_VRN_ix,  NOTATION_ADV }, // nota name (+ nm grp)
      { NUMBER_VRN_ix,    NUMBER_ADV   }, // number
      { NUMBERS_VRN_ix,   NUMBERS_ADV  },
      { NUTOKEN_VRN_ix,   NUTOKEN_ADV  }, // number token
      { NUTOKENS_VRN_ix,  NUTOKENS_ADV },
      { NOT_FOUND_VRN_ix, INVALID_ADV  }
   };
   static const VarntChoiceEntry defv_choices[] = { // attribute is */
      { FIXED_VRN_ix,     FIXED_AFV    },  // a fixed attribute
      { REQUIRED_VRN_ix,  REQUIRED_AFV },  // a required attribute
      { CURRENT_VRN_ix,   CURRENT_AFV  },  // a current attribute
      { CONREF_VRN_ix,    CONREF_AFV   },  // a content ref attr
      { IMPLIED_VRN_ix,   IMPLIED_AFV  },  // an impliable attr
      { NOT_FOUND_VRN_ix, INVALID_AFV  }
   };

   /*-----------+
   | Initialize |
   +-----------*/
   lkkindPV = lkkindArg;
   ppPatch = (void **)&ppPatch;          // patch chain
   iCountAttcntPV = 0;                   // count attribute name + name tokens
   isNotationAdl = FALSE;

   /*-----------------------------------------------------+
   | Decode associated element / notation:                |
   | [149.1] associated notation name =                   |
   |    rni, "NOTATION, ps+, (notation name | name group) |
   | [72] associated element type =                       |
   |    generic identifier | name group                   |
   +-----------------------------------------------------*/
   pPrsstat->skip_ps();
   if (pPrsstat->get_curr_token() == RNI_TK) {   // Associated Notation?
      pSyntax->match_rsvname_or_die(pEh, NOTATION_VRN_ix);
      isNotationAdl = TRUE;
      pPrsstat->skip_ps();
   }
   switch (pPrsstat->get_curr_token()) {
   case GRPO_TK:
      pPrsstat->parse_name_group(inst_new_attlist_gi, this);
      break;
   case NOT_FOUND_TK:
      new_attlist_gi();                         // install elemt/notation name
      break;
   default:
      errorh(INVNAMGRP, pEh, ERH_T_SEV);
      break;
   }
   /*
   | If no GenericIdBlock were referenced, set up invalid flag
   */
   boolean isDefinitivelyInvalid = (ppPatch==(void **)&ppPatch)? TRUE : FALSE;
   pPrsstat->skip_ps();

   /*--------------------------------+
   | For all attributes of this GI...|
   +--------------------------------*/
   AttributeNodeFlags isAnode;                   // Attribute flag
   memset(&isAnode, 0, sizeof isAnode);          // zero it;
   AttributeList attlist;                        // empty attribute list

   pAttlistPV = &attlist;
   pszNamePV = pchNamesBuffer;
   char *pszName = pszNamePV;

   do {                                           // For each attribute
      pSyntax->read_general_name(pEh, pszName);   // attribute name
      if (++iCountAttcntPV > ATTCNT) {            // check count
         errorh(BIGATTCNT, pEh, ERH_E_SEV, pszName);
         isDefinitivelyInvalid = TRUE;
         --iCountAttcntPV;
      }
      boolean isInvalidAttribute = isDefinitivelyInvalid;  // initialize

      /*---------------+
      | Declared Value |
      +---------------*/
      e_DeclaredValue declared_value;
      pPrsstat->skip_ps();
      switch (pPrsstat->get_curr_token()) {
      case GRPO_TK:
         declared_value = GROUP_ADV;             // is a list of valid names
         break;
      case NOT_FOUND_TK:
         declared_value = (e_DeclaredValue)pSyntax->match_rsvname_in_list(
            pEh, dclv_choices
         );
         switch (declared_value) {
         case INVALID_ADV:
            errorh(INVKWD, pEh, ERH_T_SEV);
            break;
         case ENTITY_ADV:
         case ENTITIES_ADV:
            if (isNotationAdl) {                  // See 11.4.1
               errorh(INVDCLVDA, pEh, ERH_E_SEV);
               declared_value = (declared_value & 1)? NAMES_ADV : NAME_ADV;
            }else  {
               isAnode.entity = TRUE;
            }
            break;
         case IDREF_ADV:
         case IDREFS_ADV:
            if (
               (lkkindPV != NOT_A_LINK_LKKIND) || // See 12.1.4.2
               (isNotationAdl)                    // See 11.4.1
            ) {                                   // Treat as NAME(s)
               errorh(isNotationAdl? INVDCLVDA : INVDCLVLK, pEh, ERH_E_SEV);
               declared_value = (declared_value & 1)? NAMES_ADV : NAME_ADV;
            }
            break;
         case NOTATION_ADV:
            if (isAnode.notation) {
               errorh(DUPNOTDEF, pEh, ERH_E_SEV);
               declared_value = GROUP_ADV;        // Treat as a list
            }else if (
               (lkkindPV !=NOT_A_LINK_LKKIND) ||  // See 12.1.4.2
               (isNotationAdl)                    // See 11.4.1
            ) {
               errorh(isNotationAdl? INVDCLVDA : INVDCLVLK, pEh, ERH_E_SEV);
               declared_value = GROUP_ADV;        // Treat as a list
            }else {
               isAnode.notation = TRUE;           // no more than 1!
            }
            pPrsstat->skip_ps();                  // 10/30/90: allow no ps
            if (pPrsstat->get_curr_token()!=GRPO_TK) {
               errorh(INVATTDCL, pEh, ERH_T_SEV);
            }
            break;
         case ID_ADV:
            if (isAnode.id) {
               errorh(DUPIDDEF, pEh, ERH_E_SEV);
               declared_value = NAME_ADV;         // Treat as a name
            }else if (
               (lkkindPV != NOT_A_LINK_LKKIND) || // See 12.1.4.2
               (isNotationAdl)                    // See 11.4.1
            ) {
               errorh(isNotationAdl? INVDCLVDA : INVDCLVLK, pEh, ERH_E_SEV);
               declared_value = NAME_ADV;         // Treat as a name
            }else {
               isAnode.id = TRUE;                 // no more than 1!
               /*--------------------------------------+
               | ID attribute name should be unique    |
               | (see 11.3.2, + Tech. Errata 04/30/87) |
               +--------------------------------------*/
               if (achNameOfIDAttribute[0]) {
                  if (strcmp(achNameOfIDAttribute, pszName)) {
                     errorh(DUPIDNAME, pEh, ERH_I_SEV, achNameOfIDAttribute);
                  }
               }else {                           // store first idname
                  strcpy(achNameOfIDAttribute, pszName);
               }
            }
            break;
         default:
            break;
         }
         break;
      default:
         errorh(INVATTDCL, pEh, ERH_T_SEV);
         declared_value = INVALID_ADV;
         break;
      }

      /*---------------------------------+
      | Check if attribute is duplicated |
      +---------------------------------*/
      {
         Attribute *pAtt;

         for (int ixAtt=0; (pAtt = (*pAttlistPV)[ixAtt]) != 0; ++ixAtt) {
            if (!strcmp(pszName, pAtt->name())) {
               errorh(DUPATT, pEh, ERH_E_SEV, pszName);
               isInvalidAttribute = TRUE;
               break;
            }
         }
      }
      if (isInvalidAttribute) pszName[0] = '\0';
      *(pszNamePV = pszName + (1+strlen(pszName))) = '\0';


      Attribute att(pszName, pszNamePV, declared_value, lkkindPV);
      attlist.push(att);

      /*------------------------------------------------------------+
      | Enter the new attribute into the `attlistPV' Attribute List.|
      | Create NOW the value string from the list of allowed value  |
      | These values will follow the attribute name.                |
      +------------------------------------------------------------*/
      if ((declared_value == GROUP_ADV) || (declared_value == NOTATION_ADV)) {
         pPrsstat->parse_name_group(inst_new_declared_value, this);
         ++pszNamePV;            // pszName set; point past the end
      }

      /*-----------------------------------------------------------+
      | [147] default value =                                      |
      |    ((rni, "FIXED", ps+)?, attribute value specification) | |
      |    (rni,("REQUIRED" | "CURRENT" | " CONREF" | "IMPLIED"))  |
      +-----------------------------------------------------------*/
      e_DefaultValue default_value;
      pPrsstat->skip_ps();
      switch (pPrsstat->get_curr_token()) {
      case RNI_TK:
         default_value = (e_DefaultValue)pSyntax->match_rsvname_in_list(
            pEh, defv_choices
         );
         switch (default_value) {
         case FIXED_AFV:
            pPrsstat->skip_inferable_required_ps();
            break;
         case REQUIRED_AFV:
            isAnode.required = TRUE;
            break;
         case CURRENT_AFV:
            if ((lkkindPV !=NOT_A_LINK_LKKIND)||  // See 12.1.4.2
               (isNotationAdl)) {                 // See 11.4.1
               if (lkkindPV != SIMPLE_LKKIND) {   // will be fixed later
                  errorh(INVDEFV, pEh, ERH_E_SEV);
                  default_value = IMPLIED_AFV;    // assume IMPLIED
               }
            }else {
               isAnode.current = TRUE;
            }
            break;
         case CONREF_AFV:
            if ((lkkindPV !=NOT_A_LINK_LKKIND)||  // See 12.1.4.2
               (isNotationAdl)) {                 // See 11.4.1
               if (lkkindPV != SIMPLE_LKKIND) {   // will be fixed later
                  errorh(INVDEFV, pEh, ERH_E_SEV);
                  default_value = IMPLIED_AFV;    // assume IMPLIED
               }
            }else {
               isAnode.conref = TRUE;
            }
            break;
         case IMPLIED_AFV:
            break;
         default: /* INVALID_AFV */
            errorh(INVKWD, pEh, ERH_T_SEV);
            break;
         }
         break;
      default:
         default_value = DEFLT_AFV;
         break;
      }
      if (
         (lkkindPV == SIMPLE_LKKIND) &&
         (default_value < DEFLT_AFV)              // AVS not specified:
      ) {
         errorh(NOFIXED, pEh, ERH_E_SEV);
         isInvalidAttribute = TRUE;               // no recovery
      }else if (
         (declared_value == ID_ADV) &&            // ID must be ..
         (default_value > REQUIRED_AFV)           // IMPLIED or REQUIRED
      ) {
         errorh(INVIDDEF, pEh, ERH_E_SEV);        // tell user
         if (default_value >= DEFLT_AFV) {        // AVS specified:
            Component *pCompo;
            unsigned int uiNormedLenVal;
            pPrsstat->att_val_spec(               // get rid of it
               pchLitBuffer, 0, DEF_TGF, pCompo, uiNormedLenVal
            );
         }
         default_value = IMPLIED_AFV;             // assume IMPLIED
      }else if (default_value >= DEFLT_AFV) {     // DEFAULT or FIXED
         if ((lkkindPV == SIMPLE_LKKIND) && (default_value != FIXED_AFV)) {
            errorh(NOFIXED, pEh, ERH_E_SEV);
            default_value = FIXED_AFV;            // recovery OK
         }
      }
      att.set(default_value);

      /*------------------------------------+
      | Parse the default AVS, save it.     |
      +------------------------------------*/
      if (default_value >= DEFLT_AFV) {           // DEFAULT or FIXED
         unsigned int uiNormedLenVal;
         Component *pCompo;
         unsigned int iListCount = pPrsstat->att_val_spec(
            pchLitBuffer, &att, DEF_TGF, pCompo, uiNormedLenVal
         );
         if (!isInvalidAttribute) {
            if (!iListCount) {                    // Invalid AVS
               if (lkkindPV == SIMPLE_LKKIND) {
                  isInvalidAttribute = TRUE;
               }else if (
                  (declared_value == GROUP_ADV) ||
                  (declared_value == NOTATION_ADV)
               ) {
                  att.set(REQUIRED_AFV);          // assume REQUIRED
                  isAnode.required = TRUE;
               }else {
                  att.set(IMPLIED_AFV);           // assume IMPLIED
               }
            }else {
               nbrof.dclv_text += uiNormedLenVal; // Normalized length
               if (pCompo) nbrof.components += pCompo->count();
               att.set(sk, pchLitBuffer, pCompo, iListCount);
            }
         }
      }
      if (isInvalidAttribute) {
         attlist.pop();
      }else {
         pszName = pszNamePV;                    // at next free room
      }
   }while (
      (pPrsstat->skip_ps()) && (pPrsstat->get_curr_token() == NOT_FOUND_TK)
   );

   /*---------------------------+
   | Dislocate the Attlist,     |
   | Create the Anode structure |
   +---------------------------*/
   AttributeNode *pAnode;
   if (attlist.count()) {
      attlist.dislocate(sk, pchNamesBuffer, diffptr(pszNamePV, pchNamesBuffer));
      pAnode = new AttributeNode(isAnode, attlist);
   }else {                                       // not even one
      pAnode = 0;                                // disable pAnode in GI
   }
   nbrof.attributes += attlist.count();
   nbrof.att_dclv += iCountAttcntPV - attlist.count(); // to compute AVGRPCAP

   /*------------------------------------+
   | Solve (patch) hdr pointer in gi(s). |
   | Compute how many gi's use this hdr  |
   +------------------------------------*/
   {
      void **ppPatchCurrent = ppPatch;
      int iReferenceCount = 0;
      while (ppPatchCurrent != (void **)&ppPatch) {
         void **ppPatchCurrentSave = ppPatchCurrent;
         ppPatchCurrent = (void **)(*ppPatchCurrent);
         *ppPatchCurrentSave = pAnode;
         ++iReferenceCount;
      }
      if (pAnode) pAnode->iReferenceCount = iReferenceCount;
   }
}

/*--------------------------------------------------------------new_attlist_gi-+
| Install a new associated element type name in the hash table                 |
| Create a patch chain (the AttributeItem pointer will be set after            |
| attributes are validated)                                                    |
+-----------------------------------------------------------------------------*/
void AttlistDcl::new_attlist_gi()
{
   AttributeNode **ppAnode;
   char pszName[NAMELEN_MAX+1];

   pSyntax->read_general_name(pEh, pszName);
   if (isNotationAdl) {
      ppAnode = &(
         (
            (NotationBlock *)expnhashNotation.gimme_value(
               pszName, sizeof(NotationBlock)
            )
         )->pAnode
      );
   }else {
      switch (lkkindPV) {
      case SIMPLE_LKKIND:
         {
            GenericIdBlock *pGi =
            (GenericIdBlock *)expnhashElem.search_value(pszName);
            if ((!pGi) || (pGi != pGiDocElement)) {
               errorh(INVATTSLK, pEh, ERH_E_SEV);
               return;
            }
            ppAnode = &(pGi->pAnodeLink);
         }
         break;
      case IMPLICIT_LKKIND:
      case EXPLICIT_LKKIND:
         ppAnode=&(
            (
               (GenericIdBlock *)expnhashElem.gimme_value(
                  pszName, sizeof(GenericIdBlock)
               )
            )->pAnodeLink
         );
         break;
      default:                                   // regular ATTLIST of a DTD
         ppAnode=&(
            (
               (GenericIdBlock *)expnhashElem.gimme_value(
                  pszName, sizeof(GenericIdBlock)
               )
             )->pAnode
         );
         break;
      }
   }
   if (*ppAnode) {                               // already defined:
      errorh(DUPNAM, pEh, ERH_E_SEV);            // it's an error
   }else {
      *ppAnode = (AttributeNode *)ppPatch;
      ppPatch = (void **)ppAnode;
   }
}

/*----------------------------------------------------------new_declared_value-+
| Register a new name in a name token group declared value                     |
|                                                                              |
| Build a list of allowed value: asciiz strings, ended by an extra '\0'        |
|                                                                              |
| IMPLEMENTATION NOTES:                                                        |
|                                                                              |
| >> At first time:                                                            |
|                                                                              |
|            (name)\0\0                         (name) \0X X X X \0\0          |
|           -+-------+--                       -+--------+---------+--         |
|            |       |         ====>            |        |         |           |
|    pszName-+       |                  pszName-+        |         |           |
|             pszLit-+                            pszLit-+         |           |
|          pszNamePV-+                                   pszNamePV-+           |
|                                                                              |
|                                                                              |
|                                                                              |
|                                      (name)\0X X X\0 Y Y Y \0\0              |
|                                     -+-------+---------------+--             |
|   Next time(s)   ====>               |       |               |               |
|                              pszName-+       |               |               |
|                                       pszLit-+     pszNamePV-+               |
|                                                                              |
+-----------------------------------------------------------------------------*/
void AttlistDcl::new_declared_value()
{
   if (
      (*pAttlistPV)[pAttlistPV->count() - 1]->declared_value() == NOTATION_ADV
   ) {
      pSyntax->read_general_name(pEh, pszNamePV);
   }else {
      pSyntax->read_nametoken(pEh, pszNamePV);
   }
   if (++iCountAttcntPV > ATTCNT) {              // count name tokens
      errorh(BIGATTCNT, pEh, ERH_E_SEV, pszNamePV);
      --iCountAttcntPV;
      *pszNamePV = '\0';
      return;
   }
   *(pszNamePV += 1+strlen(pszNamePV)) = '\0';   // end the list;

   Attribute *pAtt;
   for (int ixAtt=0; (pAtt = (*pAttlistPV)[ixAtt]) != 0; ++ixAtt) {
      const char *pszNameFound = pAtt->find_token(pszNamePV);
      if ((pszNameFound) && (pszNameFound != pszNamePV)) {
         errorh(DUPDCLV, pEh, ERH_E_SEV, pszNamePV, pAtt->name());
         *(pszNamePV -= 1+strlen(pszNamePV)) = '\0';
         --iCountAttcntPV;
         return;
      }
   }
}

/*=====================> A T T R I B U T E   N O D E <========================*/

/*---------------------------------------------------------------AttributeNode-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
AttributeNode::AttributeNode(
   AttributeNodeFlags isArg, AttributeList& attlistArg
) : attlist(attlistArg), is(isArg)
{ /* ... */ }

/*---------------------------------------------------------------------pile_up-+
| Pile up the attribute definition list from the anode structure, into the     |
| AttributeRep array.  Returns the index of the first piled AttributeRep.      |
|                                                                              |
| If the anode is still referenced by others (ELEM or NOTATION name group),    |
|   - turn on the is.solved bit to tell Attribute stowage done,                |
|   - put the index value in pAnode->ixAttrep.                                 |
| Next time, this value will just be returned.                                 |
| If no more use of the AttributeNode node, (ref_count == 0), delete the anode.|
| (WARNING: pAnode no more usable after this routine!)                         |
+-----------------------------------------------------------------------------*/
int AttributeNode::pile_up(
   AttributeRep *&pAttrepMP, const AttributeRep * const aAttrepTable,
   Component *&pCompoMP, const Component * const aCompoTable,
   boolean isCatenated
)
{
   if (!is.solved) {                       // if not already "solved":
      is.solved = TRUE;                    // now solved
      ixAttrep = diffptr(pAttrepMP, aAttrepTable);
      attlist.copy(pAttrepMP, pCompoMP, aCompoTable, isCatenated);
   }
   return ixAttrep;
}
/*============================================================================*/
