/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// attrib.c
///  Attribute and AttributeList classes
//   Pierre G. Richard 94/01/04 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <new.h>
#include <memory.h>
#include "gendef.h"
#include "event.h"
#include "element.h"
#include "entity.h"
#include "attrib.h"
#include "notation.h"
#include "dtddcl.h"
#include "errorh.h"
#include "syntax.h"
#include "quantity.h"
#include "prsstate.h"
#include "sort.h"

/*=====================> A T T R I B U T E   R E P <==========================*/

/*----------------------------------------------------------------AttributeRep-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
inline AttributeRep::AttributeRep(
   char *pszNameArg,
   char *pszLitArg,
   e_DeclaredValue declared_value,
   e_LinkKind lkkind,
   e_DefaultValue default_value
)
{
   pszName = pszNameArg;
   isLast = FALSE;
   if ((declared_value == GROUP_ADV) || (declared_value == NOTATION_ADV)) {
     isListed = TRUE;
   }else {
     isListed = FALSE;
   }
   isSpecified = FALSE;
   isComposite = FALSE;
   isOmitName = FALSE;
   isLitParked = FALSE;
   if (lkkind == NOT_A_LINK_LKKIND) {
      isLink = FALSE;
   }else {
      isLink = TRUE;
   }
   kind = declared_value;
   defv = default_value;
   pszLit = pszLitArg;
   iListCount = 0;               // assume no value specified
   iReferenceCount = 1;
}

/*----------------------------------------------------------------AttributeRep-+
| Copy constructor                                                             |
+-----------------------------------------------------------------------------*/
inline AttributeRep::AttributeRep(const AttributeRep& attrepSource)
{
   memcpy(this, &attrepSource, sizeof(AttributeRep));
   iReferenceCount = 1;
}

/*---------------------------------------------------------------~AttributeRep-+
| Destruct an Attribute Rep structure                                          |
+-----------------------------------------------------------------------------*/
inline AttributeRep::~AttributeRep()
{
   if (iListCount) {
      if (isComposite) {
         if (!isLitParked) free((void *)aCompo->value());
         free(aCompo);
      }else {
         if (!isLitParked) free(pszLit);
      }
   }
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries AttributeRep objects piled_up in an array               |
+-----------------------------------------------------------------------------*/
void AttributeRep::relocate_pile(
   unsigned int iMaxEntries,
   Component *aCompoTable,
   EntityDef *aEntGenTable,
   char *strbuf
)
{
   AttributeRep *pAttrep = this;

   for (int i=0; i < iMaxEntries; ++i, ++pAttrep) {

      pAttrep->pszName = strbuf + pAttrep->ofsName;
      if (pAttrep->isComposite) {
         Component *pCompo = pAttrep->aCompo = aCompoTable + pAttrep->ixCompo;

         if (pAttrep->isLitParked) {   // reset by `pile_up' for default spec
            pCompo->relocate(aEntGenTable, strbuf);
            pAttrep->isLitParked = FALSE;
         }
      }else {
         pAttrep->pszLit = strbuf + pAttrep->ofsLit;
         pAttrep->isLitParked = FALSE;
      }
      pAttrep->iReferenceCount = 1;
   }
}

/*---------------------------------------------------------------------display-+
| Display                                                                      |
+-----------------------------------------------------------------------------*/
void AttributeRep::display(ParseState *pPrsstat, int iIndent) const
{
   {
      char *pszSpecif;
      boolean isDefaulted = FALSE;

      if (isSpecified) {          // parsed
         pszSpecif = "set";
      }else {                     // default from def list
         isDefaulted = TRUE;
         pszSpecif = "dfl";
      }
      fprintf(
         stderr, "%*.s*%s* %s%c", iIndent, " ",
         pszSpecif, pszName,  isOmitName? ':' : '='
      );
      if (isDefaulted) {
         switch (defv) {
         case IMPLIED_AFV:
            fprintf(stderr, "(#IMPLIED) ");
            break;
         case CONREF_AFV:
            fprintf(stderr, "(#CONREF) ");
            break;
         case CURRENT_AFV:
            fprintf(stderr, "(#CURRENT) ");
            break;
         case FIXED_AFV:
            fprintf(stderr, "(#FIXED) ");
            break;
         default:
            break;
         }
      }
   }
   if (!iListCount) {
      fprintf(stderr, "(** unknown **)\n");
      return;
   }
   switch (kind) {
   case ENTITIES_ADV:
   case ENTITY_ADV:
      fprintf(stderr, "(entit%s):\n", kind & 1 ? "ies" : "y");
      {
         StringList strlst;
         char *psz;

         strlst.add_tokens(pszLit, pPrsstat->pSyntax->get_SPACEchar());
         for (int i=0; psz = strlst[i], psz; ++i) {
            EntityDef *pEnt = pPrsstat->find_general_entity(psz);
            if (!pEnt) {
               fprintf(stderr, "%*.s %s: *unknown*\n", iIndent+10, " ", psz);
            }else {
               pEnt->display(pPrsstat, iIndent+10);
            }
         }
      }
      break;
   case NOTATION_ADV:
      fprintf(stderr, "(notation):\n");
      {
         NotationType *pNcb;

         pNcb = (NotationType *)pPrsstat->pDtddcl->pHshNotation->search_value(
            pszLit
         );
         if (!pNcb) {
            fprintf(stderr, "%s: *unknown*\n", pszLit);
         }else {
            pNcb->display(pPrsstat, iIndent+6);
         }
      }
      break;
   default:
      if (isComposite) {
         fprintf(stderr, "(concatenation of):\n");
         aCompo->display(iIndent + 6 + 5);
      }else {
         fprintf(stderr, "%s\n", pszLit);
      }
      break;
   }
}

/*==========================> A T T R I B U T E <=============================*/
static Attribute attNULL(
   (char *)0, (char *)0, INVALID_ADV, NOT_A_LINK_LKKIND, INVALID_AFV
);

/*-------------------------------------------------------------------Attribute-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Attribute::Attribute(
   char *pszNameArg,
   char *pszLitArg,
   e_DeclaredValue declared_value,
   e_LinkKind lkkind,
   e_DefaultValue default_value
)
{
   pAttrep = new AttributeRep(
      pszNameArg, pszLitArg, declared_value, lkkind, default_value
   );
}

/*-------------------------------------------------------------------Attribute-+
| Copy constructor                                                             |
+-----------------------------------------------------------------------------*/
Attribute::Attribute(const Attribute& attSource)  // copy constructor
{
   pAttrep = attSource.pAttrep;
   ++(pAttrep->iReferenceCount);
}

/*------------------------------------------------------------------~Attribute-+
| Destructor                                                                   |
|                                                                              |
| Note: The pAttrep field might be NULL (and not freed):                       |
|       1) when this is the last entry in the `aAtt' array of Attribute.       |
|          This entry is reused (union) for reference counting.  C++ doesn't   |
|          allow to control the suppression of an array: hence, this           |
|          not-a-pAttrep entry is also deleted.  To surround this problem,     |
|          when the array is about to be deleted, the pAttrep field of the     |
|          union is zeroed (overwritting the now useless iReferenceCount).     |
|       2) when the attributes of an AttributeList of a data entity haven't    |
|          been specified, their pAttrep field are NULL -- hence, the only     |
|          non-NULL pAttrep fields are for explicitely defined attributes.     |
|                                                                              |
+-----------------------------------------------------------------------------*/
Attribute::~Attribute()
{
   if ((pAttrep) && (--pAttrep->iReferenceCount <= 0)) {
      delete pAttrep;
   }
}

/*-------------------------------------------------------------------operator=-+
| Assignement                                                                  |
+-----------------------------------------------------------------------------*/
Attribute& Attribute::operator=(const Attribute& attSource)
{
   if (pAttrep != attSource.pAttrep) {
      this->~Attribute();
      pAttrep = attSource.pAttrep;
      ++pAttrep->iReferenceCount;
   }
   return *this;
}

/*-------------------------------------------------------------------------set-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void Attribute::set(
  StringKeeper& sk,
  const char * const pchBuffer,
  Component *pCompo,
  const int iListCount
)
{
   pAttrep->iListCount = iListCount;
   if (!pCompo) {
      if (!pAttrep->isListed) {
         /*-------------------------------------------------+
         | If the default attribute was a token in a group, |
         | check_list_and_class put it at the first place.  |
         | Otherwise, malloc room for the deflt literal.    |
         | If empty string, offset is already 0.            |
         +-------------------------------------------------*/
         int iLen = strlen(pchBuffer);
         if (iLen) {
            pAttrep->ofsLit = sk.park(pchBuffer, iLen+1);
         }else {
            pAttrep->ofsLit = 0;
         }
         pAttrep->isLitParked = TRUE;
      }
   }else {
      int iLen = strlen(pchBuffer);
      pAttrep->isComposite = TRUE;      // lit is composite
      if (iLen) {
         pCompo->ofsLit = sk.park(pchBuffer, iLen+1);
      }else {
         pCompo->ofsLit = 0;         // entity reference to an empty string
      }
      pAttrep->isLitParked = TRUE;
      pAttrep->aCompo = pCompo;
   }
}

/*------------------------------------------------------------------find_token-+
| Check if a token belongs to a list (ASCIIZ strings, ended by an extra \0)    |
| Returns a NULL pointer if not found, or pointing at the string found.        |
|                                                                              |
| Note: for LISTED attributes, the pszList pointer is no more at               |
|       att.pszLit after the attribute value has been assigned:                |
|       this assignement destroys the initial value.                           |
|       To be able to call this routine even for duplicate attributes,         |
|       we use the property that attrep.pszName is followed by the list:       |
|       So, pszName + (1+strlen(pszName)) points the original list.            |
+-----------------------------------------------------------------------------*/
const char * Attribute::find_token(const char * pszTokenToFind) const
{
   if (pAttrep->isListed) {
      const char *pszList = pAttrep->pszName + (1+strlen(pAttrep->pszName));
      while (*pszList) {
         if (!strcmp(pszList, pszTokenToFind)) return pszList;  // found
         pszList += 1 + strlen(pszList);
      }
   }
   return 0;
}

/*------------------------------------------------------------------check_list-+
| Check if the AVS belongs to a list of declared value.                        |
+-----------------------------------------------------------------------------*/
boolean Attribute::check_list(
   ParseState *pPrsstat, const char *pszTokenToFind
) const
{
   if (!find_token(pszTokenToFind)) {
      errorh(INVATTSPC, pPrsstat->pEh, ERH_E_SEV);
      return FALSE;
   }else {
      return TRUE;
   }
}

/*--------------------------------------------------------check_list_and_class-+
| Check if the AVS belongs to a list of declared value.                        |
| If so, reorder the list in such a manner the specified attribute value       |
| comes first, and return TRUE.                                                |
| Return FALSE  otherwise.                                                     |
|                                                                              |
| IMPLEMENTATION NOTE: Only for parsing the DEFAULT value, in ATTL MD, and     |
| resetting a new CURRENT.  This routine will make the DEFAULT value appears   |
| first in the list of allowable attribute value spec.                         |
+-----------------------------------------------------------------------------*/
boolean Attribute::check_list_and_class(
   ParseState *pPrsstat, const char *pszTokenToFind
) const
{
   const char *pszTokenInList = find_token(pszTokenToFind);

   if (!pszTokenInList) {
      errorh(INVATTSPC, pPrsstat->pEh, ERH_E_SEV);
      return FALSE;
   }else {
      char *pszLit = pAttrep->pszLit;

      if (pszTokenInList > pszLit) {               // Not already first
         int i = 1+strlen(pszTokenToFind);
         memmove(pszLit+i, pszLit, diffptr(pszTokenInList, pszLit));
         memcpy(pszLit, pszTokenToFind, i);
      }
      return TRUE;
   }
}

/*--------------------------------------------------------------------check_id-+
| Check if an ID is unique                                                     |
+-----------------------------------------------------------------------------*/
boolean Attribute::check_id(
   ParseState *pPrsstat, const char *pszToken
) const
{
   int *piElemId = (int *)pPrsstat->ExpnhashId.gimme_value(
      pszToken, sizeof(int)
   );

   if (*piElemId == 0) {                  // not set
      *piElemId = pPrsstat->curr_elem_id;
      return TRUE;
   }
   errorh(DUPID, pPrsstat->pEh, ERH_E_SEV, pszToken, *piElemId);
   return FALSE;
}

/*=====================> A T T R I B U T E   L I S T <========================*/
/*---------------------------------------------------------------AttributeList-+
| Null Constructor                                                             |
+-----------------------------------------------------------------------------*/
AttributeList::AttributeList()
{
   iCount = 0;
   aAtt = 0;
}

/*---------------------------------------------------------------AttributeList-+
| Constructor from a dislocated attribute definition list (true copy)          |
+-----------------------------------------------------------------------------*/
AttributeList::AttributeList(
   StringKeeper& sk, const AttributeList& atlSource, unsigned int*& aOfsTab
)
{
   iCount = atlSource.iCount;

   if (iCount) {
      aAtt = new Attribute[iCount+1];  // last entry is a reference count
      aOfsTab = new unsigned int [2 * iCount];
      unsigned int *pOfs = aOfsTab;

      for (int i=0; i < iCount; ++i) {
         AttributeRep *pAttrep = atlSource.aAtt[i].pAttrep;
         pAttrep->pszName = sk.point(*pOfs++ = pAttrep->ofsName);
         if (pAttrep->isComposite) {
            pAttrep->aCompo[0].pszLit = sk.point(
               *pOfs++ = pAttrep->aCompo[0].ofsLit
            );
         }else {
            pAttrep->pszLit = sk.point(*pOfs++ = pAttrep->ofsLit);
         }
         ++(pAttrep->iReferenceCount);
         aAtt[i].pAttrep = pAttrep;
      }
      aAtt[i].iReferenceCount = 1;
   }else {
      aOfsTab = 0;
      iCount = 0;
      aAtt = 0;
   }
}

/*-------------------------------------------------------------------re_offset-+
| Re-offset the dislocated attribute definition list                           |
+-----------------------------------------------------------------------------*/
void AttributeList::re_offset(unsigned int *aOfsTable)
{
   unsigned int *pOfs = aOfsTable;

   for (int i=0; i < iCount; ++i) {
      AttributeRep *pAttrep = aAtt[i].pAttrep;

      pAttrep->ofsName = *pOfs++;
      if (pAttrep->isComposite) {
         pAttrep->aCompo[0].ofsLit = *pOfs++;
      }else {
         pAttrep->ofsLit = *pOfs++;
      }
   }
   delete [] aOfsTable;
}

/*---------------------------------------------------------------AttributeList-+
| Copy Constructor                                                             |
+-----------------------------------------------------------------------------*/
AttributeList::AttributeList(const AttributeList& attlistSource)
{
   aAtt = attlistSource.aAtt;
   iCount = attlistSource.iCount;
   if (iCount) {
      ++(aAtt[iCount].iReferenceCount);
   }
}

/*--------------------------------------------------------------~AttributeList-+
| Destruct a list of Attributes                                                |
+-----------------------------------------------------------------------------*/
AttributeList::~AttributeList()
{
   if ((iCount) && (--(aAtt[iCount].iReferenceCount) <= 0)) {
      aAtt[iCount].pAttrep = 0; // no more ref count: see note in ~Attribute()
      delete [] aAtt;
   }
}

/*-----------------------------------------------------------------------build-+
| Make an attribute list from an array of AttributeRep                         |
+-----------------------------------------------------------------------------*/
void AttributeList::build(AttributeRep *aAttrep, int iCountArg)
{
   this->~AttributeList();             // clean_up: remove current contents
   iCount = iCountArg;
   if (aAttrep) {
      AttributeRep *pAttrep;
      if (!iCount) {                   // Count the attributes, if necessary
         for (
            pAttrep = aAttrep, iCount = 1;
            !(pAttrep->isLast);
            ++iCount, ++pAttrep
         );
      }
      aAtt = new Attribute[iCount+1];  // last entry is a reference count
      pAttrep = aAttrep;
      for (int i=0; i < iCount; ++i, ++pAttrep) {
         aAtt[i].pAttrep = pAttrep;
         ++(pAttrep->iReferenceCount);
      }
      aAtt[i].iReferenceCount = 1;
   }else {
      iCount = 0;
      aAtt = 0;
   }
}

/*-------------------------------------------------------------------operator=-+
| Assignement                                                                  |
+-----------------------------------------------------------------------------*/
AttributeList& AttributeList::operator=(const AttributeList& attlistSource)
{
   if (aAtt != attlistSource.aAtt) {
      this->~AttributeList();
      aAtt = attlistSource.aAtt;
      iCount = attlistSource.iCount;
      if (iCount) {
         ++(aAtt[iCount].iReferenceCount);
      }
   }
   return *this;
}

/*----------------------------------------------------------------find_by_name-+
| Find if an attribute in the Attribute list matches a given name              |
+-----------------------------------------------------------------------------*/
Attribute *AttributeList::find_by_name(
   const char * const pszName, int * const pixAtt
) const
{
   for (int i=0; i < iCount; ++i) {
      AttributeRep *pAttrep = aAtt[i].pAttrep;
      if (
         (!pAttrep->isLink) &&          // LINKs are transparent
         (!strcmp(pszName, pAttrep->pszName))
      ) {                               // Found...
         if (pixAtt) *pixAtt = i;
         return aAtt + i;
      }
   }
   if (pixAtt) *pixAtt = -1;
   return &attNULL;
}

/*---------------------------------------------------------------find_by_value-+
| Find if an attribute in the Attribute list matches a given value             |
+-----------------------------------------------------------------------------*/
Attribute * AttributeList::find_by_value(
   const char * const pszValue, int * const pixAtt
) const
{
   for (int i=0; i < iCount; ++i) {
      AttributeRep *pAttrep = aAtt[i].pAttrep;

      if (
         (!pAttrep->isLink) &&          // LINKs are transparent
         (pAttrep->isListed) &&
         (aAtt[i].find_token(pszValue))
      ) {
         if (pixAtt) *pixAtt = i;
         return aAtt + i;
      }
   }
   if (pixAtt) *pixAtt = -1;
   return &attNULL;
}

/*------------------------------------------------------------------reset_spec-+
| Reset the specifications of attribute `ixAtt' of an AttributeList.           |
+-----------------------------------------------------------------------------*/
void AttributeList::reset_spec(
   const int ixAtt,
   const char * const pszValue,
   const int iListCount,
   Component * const pCompo,
   boolean isOmitName,
   unsigned long& iAttChCap
) const {
   if ((ixAtt < 0) || (ixAtt >= iCount)) return;

   const unsigned int iNewLen = strlen(pszValue);
   AttributeRep *pAttrepSource = aAtt[ixAtt].pAttrep;
   if (
      (pAttrepSource->defv == CURRENT_AFV) &&
      (!pAttrepSource->isSpecified)      // only if this is the definition list
   ) {
      /*
      |  If the source is the definition list, then save a new current attribute
      |  ***>NOTE: This function violates the constness of the Dtddcl!!!
      */
      if (!pAttrepSource->iListCount) {                  // first time
         iAttChCap += iNewLen;                           // update ATTCHCAP
      }else {
         int iOldLen;
         if (pAttrepSource->isComposite) {               // old was composite
            iOldLen = strlen(pAttrepSource->aCompo[0].pszLit);
            free(pAttrepSource->aCompo[0].pszLit);       // free literal
            free(pAttrepSource->aCompo);                 // and the compo array
         }else {
            iOldLen = strlen(pAttrepSource->pszLit);
            if (!pAttrepSource->isListed) {              // not a choice-list:
               free(pAttrepSource->pszLit);              // free previous
            }
         }
         if (iNewLen > iOldLen) {
            iAttChCap += (iNewLen - iOldLen);
         }
      }
      pAttrepSource->iListCount = iListCount;
      if (pCompo) {                            // new is composite
         pAttrepSource->isComposite = TRUE;
         int iSize = pCompo->iCount * sizeof(Component);
         memcpy(
            pAttrepSource->aCompo = (Component *)malloc(iSize), pCompo, iSize
         );
         memcpy(
            pAttrepSource->aCompo[0].pszLit = (char *)malloc(iNewLen+1),
            pszValue, iNewLen+1
         );
      }else {
         pAttrepSource->isComposite = FALSE;
         if (!pAttrepSource->isListed) {                 // not a choice-list:
            memcpy(
               pAttrepSource->pszLit = (char *)malloc(iNewLen+1),
               pszValue, iNewLen+1
            );
         }
      }
   }
   /*
   |  Create a copy with the new specifs
   */
   AttributeRep *pAttrep = new AttributeRep(*pAttrepSource);  // true copy

   aAtt[ixAtt].~Attribute();
   aAtt[ixAtt].pAttrep = pAttrep;

   pAttrep->isOmitName = isOmitName;
   pAttrep->isSpecified = TRUE;
   pAttrep->iListCount = iListCount;

   char *pszLit = (char *)malloc(1+iNewLen);
   memcpy(pszLit, pszValue, 1+ iNewLen);
   if (pCompo) {
      pAttrep->isComposite = TRUE;
      pCompo->pszLit = pszLit;
      pAttrep->aCompo = pCompo;
   }else {
      pAttrep->isComposite = FALSE;
      pAttrep->pszLit = pszLit;
   }
}

/*------------------------------------------------------------------------push-+
| Catenate a new attribute to the list                                         |
| Copies of the Attribute list (if any) won't reflect the addition.            |
+-----------------------------------------------------------------------------*/
void AttributeList::push(const Attribute& attSource)
{
   Attribute *aAttNew = new Attribute[iCount+2];

   for (int i=0; i < iCount; ++i) {
      aAttNew[i].pAttrep = aAtt[i].pAttrep;
      ++(aAttNew[i].pAttrep->iReferenceCount);
   }
   this->~AttributeList();
   aAtt = aAttNew;
   aAtt[i++].pAttrep = attSource.pAttrep;
   ++(attSource.pAttrep->iReferenceCount);
   ++iCount;
   aAtt[iCount].iReferenceCount = 1;
}

/*-------------------------------------------------------------------------pop-+
| Remove the last attribute from the list                                      |
| Copies of the Attribute list (if any) won't reflect the deletion.            |
+-----------------------------------------------------------------------------*/
void AttributeList::pop()
{
   if (!iCount) return;   // be nice

   Attribute *aAttNew;
   if (iCount == 1) {
      aAttNew = 0;
   }else {
      aAttNew = new Attribute[iCount];
      for (int i=0; i < iCount-1; ++i) {
         aAttNew[i].pAttrep = aAtt[i].pAttrep;
         ++(aAttNew[i].pAttrep->iReferenceCount);
      }
   }
   this->~AttributeList();
   aAtt = aAttNew;
   --iCount;
   aAtt[iCount].iReferenceCount = 1;
}

/*------------------------------------------------------------------------copy-+
| Copy components of the component array "asis".  The entity pointers must be  |
| unlocated later, when entities have been solved (unlocate_components).       |
+-----------------------------------------------------------------------------*/
void AttributeList::copy(
   AttributeRep *&pAttrepMP,
   Component *&pCompoMP, const Component * const aCompoTable,
   boolean isCatenated
) const {
   if (isCatenated) (pAttrepMP-1)->isLast = FALSE;  // reset LAST bit
   for (int i=0; i < iCount; ++i, ++pAttrepMP) {
      new (pAttrepMP) AttributeRep(*(aAtt[i].pAttrep)); // true copy
      if (pAttrepMP->isComposite) {
         pAttrepMP->ixCompo = pAttrepMP->aCompo->copy(pCompoMP, aCompoTable);
      }
   }
}

/*-------------------------------------------------------------------dislocate-+
| Dislocate (transform pointers to offsets) an attribute definition list       |
+-----------------------------------------------------------------------------*/
void AttributeList::dislocate(
    StringKeeper& sk,
    const char *pchBuffer,
    const int iBufferLen
)
{
   if (!iCount) return;

   const unsigned int ofs = sk.park(pchBuffer, iBufferLen);
   aAtt[iCount-1].pAttrep->isLast = TRUE;  // tell the last attribute
   for (int i=0; i < iCount; ++i) {
      AttributeRep *pAttrep = aAtt[i].pAttrep;
      pAttrep->ofsName = ofs + diffptr(pAttrep->pszName, pchBuffer);
      if (pAttrep->isListed) {
         pAttrep->ofsLit = ofs + diffptr(pAttrep->pszLit, pchBuffer);
         pAttrep->isLitParked = TRUE;
      }else {
         if (!pAttrep->iListCount) pAttrep->ofsLit = 0;
      }
   }
}

/*-------------------------------------------------------------------dislocate-+
| Dislocate (transform pointers to offsets) an attribute specification list    |
| into an array of AttributeRep's                                              |
+-----------------------------------------------------------------------------*/
void AttributeList::dislocate(StringKeeper& sk, const AttributeList& attlistDef)
{
   for (int i=0; i < iCount; ++i) {
      AttributeRep *pAttrepNew = aAtt[i].pAttrep;
      if (pAttrepNew->isSpecified) {
         pAttrepNew->ofsName = attlistDef.aAtt[i].pAttrep->ofsName;
         char *pszLit;
         unsigned int *pOfsLit;
         if (pAttrepNew->isComposite) {             // is described
            pszLit = pAttrepNew->aCompo[0].pszLit;
            pOfsLit = &(pAttrepNew->aCompo[0].ofsLit);
         }else {
            pszLit = pAttrepNew->pszLit;
            pOfsLit = &(pAttrepNew->ofsLit);
         }
         if (strlen(pszLit)) {
            *pOfsLit = sk.park(pszLit, 1 + strlen(pszLit));
         }else {
            *pOfsLit = 0;
         }
         free(pszLit);
         pAttrepNew->isLitParked = TRUE;
      }else {
         aAtt[i].~Attribute();
         aAtt[i].pAttrep = 0;      // see note in ~AttributeRep
      }
   }
}

/*---------------------------------------------------------------------pile_up-+
| Save an attribute specification list in an array of AttributeRep's           |
| Note: the defaulted attributes have already the component array solved!      |
+-----------------------------------------------------------------------------*/
int AttributeList::pile_up(
   AttributeRep *&pAttrepMP, const AttributeRep * const aAttrepTable,
   Component *&pCompoMP, const Component * const aCompoTable,
   int ixAttrepDef
) const {
   int ixAttrep = diffptr(pAttrepMP, aAttrepTable);
   for (int i=0; i < iCount; ++i, ++ixAttrepDef, ++pAttrepMP) {
      AttributeRep * pAttrep = aAtt[i].pAttrep;
      if (!pAttrep) {       // non-specified attrep
         new (pAttrepMP) AttributeRep(aAttrepTable[ixAttrepDef]); // true copy
         pAttrepMP->isLitParked = FALSE;   // to avoid relocating pCompo twice!
      }else {
         new (pAttrepMP) AttributeRep(*(aAtt[i].pAttrep)); // true copy
         if (pAttrepMP->isComposite) {
            pAttrepMP->ixCompo = pAttrepMP->aCompo->copy(pCompoMP, aCompoTable);
         }
      }
   }
   return ixAttrep;
}

/*---------------------------------------------------------------------display-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void AttributeList::display(ParseState *pPrsstat, int iIndent) const
{
   for (int i=0; i < iCount; ++i) {
      aAtt[i].pAttrep->display(pPrsstat, iIndent);
   }
}

/*STATIC----------------------------------------------------------compare_item-+
| Compare (for wqsort)                                                         |
| Non printable character are always "greater" than printable characters.      |
+-----------------------------------------------------------------------------*/
int AttributeList::compare_item(int ix1, int ix2, void *pvBase)
{
   AttributeList& attlist = *((AttributeList *)pvBase);

   const char *pszName1 = attlist[ix1]->name();
   const char *pszName2 = attlist[ix2]->name();

   while ((*pszName1 == *pszName2) && (*pszName1)) { // until it differs or EOS
     ++pszName1, ++pszName2;
   }
   if (((*pszName1 >= 32) ^ (*pszName2 >= 32)) &&    // if one is less than 32
      (*pszName1) && (*pszName2) &&                  // and none is an EOS
      (*pszName1 <= 126) && (*pszName2 <= 126)) {    // and one is printable
      return(*pszName2 - *pszName1);                 // reverse the result
   }else {
      return(*pszName1 - *pszName2);
   }
}

/*STATIC---------------------------------------------------------exchange_item-+
| Exchange (for wqsort)                                                        |
+-----------------------------------------------------------------------------*/
void AttributeList::exchange_item(int ix1, int ix2, void *pvBase)
{
   AttributeList& attlist = *((AttributeList *)pvBase);
   Attribute attTemp(attlist.aAtt[ix1]);
   attlist.aAtt[ix1] = attlist.aAtt[ix2];
   attlist.aAtt[ix2] = attTemp;
}

/*------------------------------------------------------------------------sort-+
| Sort the attribute list                                                      |
| Non printable character are always "greater" than printable characters.      |
+-----------------------------------------------------------------------------*/
void AttributeList::sort()
{
   Sort::quick(iCount, compare_item, exchange_item, this);
}
/*============================================================================*/
