/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// attrib.h
///  Declarations for attrib.c
//   Pierre G. Richard 94/01/04 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _ATTRIB_H
#define _ATTRIB_H 1
//******************************************************************************

#include "prsstate.h"   // is a friend
#include "notation.h"   // used by inlines
#include "compo.h"      // used by inlines

#include "attrib.eh"    // enums

class StringKeeper;
class StringList;
class ElementType;
class NotationType;
class AttributeRep;
class Attribute;
class AttributeList;
class AttSpecList;
class AttSpec;
class Component;
class PrimEvent;

/*-----------------------------------------------------------------------------+
| Attribute Rep - This is the letter class of the Attribute class              |
+-----------------------------------------------------------------------------*/
class AttributeRep {           // ATTREP: Letter for attribute data
   friend Attribute;
   friend AttributeList;

public:
   void relocate_pile(
      unsigned int iMaxEntries,
      Component *aCompoTable,
      EntityDef *aEntGenTable,
      char *strbuf
   );

private:
   union {
      char *pszName;           // Pointer to attribute name
      unsigned int ofsName;    // (offset, when unloaded)
   };
   boolean isLast      :1;     // last in chain
   boolean isListed    :1;     // belongs to a name (token?) group
   boolean isSpecified :1;     // attribute has been specified (Spec list)
   boolean isComposite :1;     // CDATA made of text, SDATA nd CDATA entities
   boolean isOmitName  :1;     // Attribute name was omitted
   boolean isLink      :1;     // Is a link attribute
   boolean isLitParked :1;     // Internal: the literal is sk.park'ed
   e_DeclaredValue kind;       // Kind of attribute
   e_DefaultValue defv;        // Default Value: see AFV list
   union {
                               // For vanilia attributes:
      char *pszLit;            // pointer to the literal
      unsigned int ofsLit;     // (offset, when unloaded)

      Component *aCompo;       // For composite literals (if isComposite)
      int ixCompo;             // (index to the compo array when unloaded)
   };
   unsigned int iListCount;    // List count: >= 1; 0 if no specification
   int iReferenceCount;

   AttributeRep(
      char *pszNameArg,
      char *pszLitArg,
      e_DeclaredValue declared_value,
      e_LinkKind lkkind,
      e_DefaultValue default_value = INVALID_AFV
   );
   AttributeRep(const AttributeRep& attrepSource); // copy constructor
   ~AttributeRep();
   void display(ParseState *pPrsstat, int iIndent) const;
};

/*-----------------------------------------------------------------------------+
| Attribute class                                                              |
+-----------------------------------------------------------------------------*/
class Attribute {
   friend AttributeRep;
   friend AttributeList;
   friend ParseState::att_val_spec(
      char * const pchBuffer,
      const Attribute *pAtt,
      e_AttSpecKind tgf,
      Component*& pCompo,
      unsigned int& uiNormedLenVal
   );
public:
   Attribute() { /*Empty*/ }   // NULL constructor (required for arrays)
   Attribute(
      char *pszNameArg,
      char *pszLitArg,
      e_DeclaredValue declared_value,
      e_LinkKind lkkind,
      e_DefaultValue default_value = INVALID_AFV
   );
   Attribute(const Attribute& attSource);              // copy operator
   ~Attribute();
   Attribute& operator=(const Attribute& attSource);   // assignment
   void set(
     StringKeeper& sk,
     const char * const pchBuffer,
     Component *pCompo,
     const int iListCount
   );
   void set(e_DefaultValue default_value);
   const char *name() const;
   const char *value() const;
   boolean is_specified() const;
   boolean Attribute::is_implied() const;
   boolean is_name_omitted() const;
   boolean is_link() const;
   boolean is_composite() const;
   e_DeclaredValue declared_value() const;
   e_DefaultValue default_value() const;
   const char * find_token(const char * pszTokenToFind) const;
   int component_array_size() const;
   Component * component() const;

private:
   union {
      AttributeRep *pAttrep;   // Pointer to the letter
      int iReferenceCount;     // for a list, last entry is a reference count
   };
   boolean check_list(
      ParseState *pPrsstat, const char *pszTokenToFind
   ) const;
   boolean check_list_and_class(
      ParseState *pPrsstat, const char *pszTokenToFind
   ) const;
   boolean check_id(
      ParseState *pPrsstat, const char *pszToken
   ) const;
};

/*--------> INLINES */
inline const char * Attribute::name() const {
   return pAttrep->pszName;
}
inline const char * Attribute::value() const {
   if (!pAttrep->iListCount) return 0;
   if (pAttrep->isComposite) return pAttrep->aCompo->value();
   return pAttrep->pszLit;
}
inline boolean Attribute::is_specified() const {
   return pAttrep->isSpecified? TRUE : FALSE;
}
inline boolean Attribute::is_name_omitted() const {
   return pAttrep->isOmitName? TRUE : FALSE;
}
inline boolean Attribute::is_link() const {
   return pAttrep->isLink? TRUE : FALSE;
}
inline boolean Attribute::is_composite() const {
   return pAttrep->isComposite? TRUE : FALSE;
}
inline e_DeclaredValue Attribute::declared_value() const {
   return pAttrep->kind;
}
inline e_DefaultValue Attribute::default_value() const {
   return pAttrep->defv;
}
inline void Attribute::set(e_DefaultValue default_value) {
   pAttrep->defv = default_value;
}
inline int Attribute::component_array_size() const {
   return pAttrep->isComposite? pAttrep->aCompo->count() : 0;
}
inline Component * Attribute::component() const {
   return pAttrep->isComposite? pAttrep->aCompo : 0;
}
inline boolean Attribute::is_implied() const {
   return (
      (!is_specified()) && (                   // not specified, and
         (default_value() == IMPLIED_AFV) ||   // implied or
         (default_value() == CONREF_AFV)       // conref
      )                                        // then it's implied
   )? TRUE : FALSE;
}


/*-----------------------------------------------------------------------------+
| AttributeList class                                                          |
|                                                                              |
| An attribute list is a finite collection of attributes:                      |
|    Attribute = AttributeList[i], where i >=0 and < Attributelist.count()     |
+-----------------------------------------------------------------------------*/
class AttributeList {
   friend AttributeRep;
public:
   AttributeList();
   AttributeList(const ElementType *pElem);
   AttributeList(const NotationType *pNcb);
   AttributeList(const EntityNotation *pNad);
   AttributeList(const AttributeList& attlistSource);  // copy
   AttributeList(const AttributeList *pAttlist, const ElementType *pElem);
   AttributeList(
       StringKeeper& sk, const AttributeList& atlSource, unsigned int*& aOfsTab
   );
   ~AttributeList();
   AttributeList& operator=(const AttributeList& attSource);   // assignment
   void build(AttributeRep *aAttrep, int iCountArg = 0);
   void sort();

   int count() const;
   Attribute * operator[](int ixAtt) const; // NULL if out of bounds
   Attribute * find_by_name(
      const char * const pszName, int * const pixAtt = 0
   ) const;
   Attribute * find_by_value(
      const char * const pszValue, int * const pixAtt = 0
   ) const;
   void reset_spec(
      const int ixAtt, const char * const pszValue,
      const int iListCount, Component * const pCompo,
      boolean isNameOmitted, unsigned long& iAttChCap
   ) const;
   void push(const Attribute& attSource);
   void pop();
   void copy(
      AttributeRep *&pAttrepMP,
      Component *&pCompoMP, const Component * const aCompoTable,
      boolean isCatenated
   ) const;
   void dislocate(
      StringKeeper& sk,
      const char *pchBuffer,
      const int iBufferLen
   );
   void AttributeList::dislocate(
      StringKeeper& sk, const AttributeList& attlistDef
   );
   int pile_up(
      AttributeRep *&pAttrepMP, const AttributeRep * const aAttrepTable,
      Component *&pCompoMP, const Component * const aCompoTable,
      int ixAttrepDef
   ) const;
   void re_offset(unsigned int *aOfsTable);
   void display(ParseState *pPrsstat, int iIndent) const;

private:
   int iCount;            // Count of attributes in the list
   Attribute *aAtt;       // Array of `iCount' attributes + a reference count
   static int compare_item(int ix1, int ix2, void *pvBase);
   static void exchange_item(int ix1, int ix2, void *pvBase);
};

/*--------> INLINES */
inline AttributeList::AttributeList(const ElementType *pElem) {
   iCount = 0;
   pElem->attribute_list(*this);
}
inline AttributeList::AttributeList(const NotationType *pNcb) {
   iCount = 0;
   pNcb->attribute_list(*this);
}
inline AttributeList::AttributeList(const EntityNotation *pNad) {
   iCount = 0;
   pNad->attribute_list(*this);
}
inline Attribute * AttributeList::operator[](int ixAtt) const {
   return ((ixAtt < 0) || (ixAtt >= iCount)) ? 0 : aAtt + ixAtt;
}
inline int AttributeList::count() const { return iCount; }
#endif /* _ATTRIB_H ***********************************************************/
