/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// bitarray.c
///  This class handles array of bits
//   Pierre G. Richard 93/10/19 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>
#include "gendef.h"
#include "bitarray.h"

/*--------------------------------------------------------------------BitArray-+
| Regular constructor -                                                        |
| MaxBitsArg can be omitted, so BitArray original(3), copy = original; works.  |
+-----------------------------------------------------------------------------*/
BitArray::BitArray(int iMaxBitsArg)
{
   iMaxBits = iMaxBitsArg;
   ais = new char[size_of()];
}

/*--------------------------------------------------------------------BitArray-+
| Copy constructor - Duplicate the bit array to a brand new one                |
+-----------------------------------------------------------------------------*/
BitArray::BitArray(const BitArray& source) {
   iMaxBits = source.iMaxBits;
   ais = new char[size_of()];
   memcpy(ais, source.ais, size_of());
}

/*-------------------------------------------------------------------operator=-+
| Copy the source BitArray to an already existing BitaArray                    |
+-----------------------------------------------------------------------------*/
BitArray& BitArray::operator= (const BitArray& source) {
   if (ais != source.ais) {                 // don't do it for identity
      if (source.size_of() != size_of()) {  // useless if same size
         delete [] ais;
         iMaxBits = source.iMaxBits;
         ais = new char[size_of()];
      }
      memcpy(ais, source.ais, size_of());
   }
   return *this;
}

/*-------------------------------------------------------------------operator=-+
| Set all bits to 0's if v==0, or 1's if v!=0                                  |
+-----------------------------------------------------------------------------*/
BitArray& BitArray::operator= (int v)
{
   if (v) v = ~0;
   memset(ais, v, size_of());
   return *this;
}

/*------------------------------------------------------------------operator==-+
| Check any hit bit is on or all bits off                                      |
+-----------------------------------------------------------------------------*/
int BitArray::operator==(int v)
{
   int max = size_of();

   while (max--) {
      if (ais[max]) {             // one bit is on
         if (v!=0) return 1;      // so any_bit_on is TRUE
         return 0;                // and all_bits_off is FALSE
      }
   }
   if (v==0) return 1;            // all_bits_off is TRUE
   return 0;                      // any_bit_on is FALSE
}
/******************************************************************************/
