/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// bitarray.h
///  Declarations for bitarray
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _BITARRAY_H
#define _BITARRAY_H 1
/*
| This class handles bit array of any number of bits
| It is assumed that the indices passed to the [], +=, -= operators,
| are within the size of the array.
| Bits are numbered from 0 to N-1 (C convention)
*/
//******************************************************************************

class BitArray {
public:
   BitArray(int iMaxBits=0);   // Create 0 or iMaxBits (uninitialized)
   BitArray(const BitArray &source); // Create from an already existing BitArray
   ~BitArray();
   BitArray& operator=(const BitArray& source); // Copy assignment
   BitArray& operator=(int v); // turn ON/OFF (v is 0/!0) *all* bits
   int operator==(int v);      // if any bit on (v!=0) or all bits off (v==0)
   int operator!=(int v);      // negate operator==
   boolean operator[](int ix); // get ix-th bit value (no lvalue)
   void operator+=(int ix);    // turn ON ix-th bit
   void operator-=(int ix);    // turn OFF ix-th bit

private:
   int iMaxBits;
   char *ais;
   int size_of() const;
};

inline BitArray::~BitArray() { delete [] ais; }
inline int BitArray::size_of() const { return (iMaxBits+7)/8; }
inline boolean BitArray::operator[](int ix) {
   return (ais[ix>>3]&(1<<(ix & 0x07)))? TRUE : FALSE;
}
inline void BitArray::operator+=(int ix) { ais[ix>>3] |= 1<<(ix & 0x07); }
inline void BitArray::operator-=(int ix) { ais[ix>>3] &= ~(1<<(ix & 0x07)); }
inline int BitArray::operator!=(int v) { return !(*this == v); }

#endif /* _BITARRAY_H *********************************************************/
