/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// capacity.c
///  Class for SGML capacities
//   Pierre G. Richard 87/06/30 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <memory.h>
#include "gendef.h"
#include "enthdlr.h"
#include "capacity.h"
#include "quantity.eh"
#include "errorh.h"
#include "rsvname.h"

const long CAPACITY_VALUE_RCS = 35000L;

#if (INT_MAX == SHRT_MAX)
const long ATTCAP_MAX     =  200000L;
const long ATTCHCAP_MAX   =  200000L;
const long AVGRPCAP_MAX   =  200000L;
const long DTACAP_MAX     =  200000L;
const long DTACHCAP_MAX   =  200000L;
const long ELEMCAP_MAX    =  200000L;
const long ENTCAP_MAX     =  200000L;
const long ENTCHCAP_MAX   =  200000L;
const long EXGRPCAP_MAX   =  200000L;
const long EXNMCAP_MAX    =  200000L;
const long GRPCAP_MAX     =  300000L;
const long IDCAP_MAX      =  200000L;
const long IDREFCAP_MAX   =  200000L;
const long LKNMCAP_MAX    =  200000L;
const long LKSETCAP_MAX   =  200000L;
const long MAPCAP_MAX     =  200000L;
const long NOTCAP_MAX     =  200000L;
const long NOTCHCAP_MAX   =  200000L;
const long TOTALCAP_MAX   =  500000L;
#else
const long ATTCAP_MAX     = 1000000L;
const long ATTCHCAP_MAX   =  500000L;
const long AVGRPCAP_MAX   =  500000L;
const long DTACAP_MAX     =  500000L;
const long DTACHCAP_MAX   =  500000L;
const long ELEMCAP_MAX    =  500000L;
const long ENTCAP_MAX     =  500000L;
const long ENTCHCAP_MAX   =  500000L;
const long EXGRPCAP_MAX   =  500000L;
const long EXNMCAP_MAX    =  500000L;
const long GRPCAP_MAX     = 1000000L;
const long IDCAP_MAX      =  500000L;
const long IDREFCAP_MAX   =  500000L;
const long LKNMCAP_MAX    =  500000L;
const long LKSETCAP_MAX   =  500000L;
const long MAPCAP_MAX     =  500000L;
const long NOTCAP_MAX     =  500000L;
const long NOTCHCAP_MAX   =  500000L;
const long TOTALCAP_MAX   = 1000000L;
#endif

/*--------------------------------------------------------------------Capacity-+
| Constructor from scratch (i.e Ref Concrete Syntax)                           |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Capacity::Capacity()
{
   auiCapacityPoint[ATTCAP_ix] =     NAMELEN_RCS;
   auiCapacityPoint[ATTCHCAP_ix] =   1;
   auiCapacityPoint[AVGRPCAP_ix] =   NAMELEN_RCS;
   auiCapacityPoint[DTACAP_ix] =     0;              // Not in ISO 8879
   auiCapacityPoint[DTACHCAP_ix] =   0;              // Not in ISO 8879
   auiCapacityPoint[ELEMCAP_ix] =    NAMELEN_RCS;
   auiCapacityPoint[ENTCAP_ix] =     NAMELEN_RCS;
   auiCapacityPoint[ENTCHCAP_ix] =   1;
   auiCapacityPoint[EXGRPCAP_ix] =   NAMELEN_RCS;
   auiCapacityPoint[EXNMCAP_ix] =    NAMELEN_RCS;
   auiCapacityPoint[GRPCAP_ix] =     NAMELEN_RCS;
   auiCapacityPoint[IDCAP_ix] =      NAMELEN_RCS;
   auiCapacityPoint[IDREFCAP_ix] =   NAMELEN_RCS;
   auiCapacityPoint[LKNMCAP_ix] =    NAMELEN_RCS;
   auiCapacityPoint[LKSETCAP_ix] =   NAMELEN_RCS;
   auiCapacityPoint[MAPCAP_ix] =     NAMELEN_RCS;
   auiCapacityPoint[NOTCAP_ix] =     NAMELEN_RCS;
   auiCapacityPoint[NOTCHCAP_ix] =   1;
   auiCapacityPoint[TOTALCAP_ix] =   0;
   for (int i=0; i < CAP_MAX_SIZE; ++i) {
      aulCapacityMax[i] = CAPACITY_VALUE_RCS;
   }
}

/*--------------------------------------------------------------------Capacity-+
| Constructor from OCS                                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Capacity::Capacity(ApplicationHandle ahlAccess)
{
   EntityHandler::read(ahlAccess, aulCapacityMax,   sizeof aulCapacityMax);
   EntityHandler::read(ahlAccess, auiCapacityPoint, sizeof auiCapacityPoint);
}

/*---------------------------------------------------------------------name_of-+
|Get the reserved name associated to a index in the capacity table.            |
+-----------------------------------------------------------------------------*/
const char * const Capacity::name_of(e_IxCapacity ix)
{
   const char * pszNameCapa;

   switch (ix) {
   case ATTCAP_ix:
      pszNameCapa = ReservedName::name_of(ATTCAP_RRN_ix);
      break;
   case ATTCHCAP_ix:
      pszNameCapa = ReservedName::name_of(ATTCHCAP_RRN_ix);
      break;
   case AVGRPCAP_ix:
      pszNameCapa = ReservedName::name_of(AVGRPCAP_RRN_ix);
      break;
   case DTACAP_ix:
      pszNameCapa = "";      // Not in ISO (YAO internal use)
      break;
   case DTACHCAP_ix:
      pszNameCapa = "";      // Not in ISO (YAO internal use)
      break;
   case ELEMCAP_ix:
      pszNameCapa = ReservedName::name_of(ELEMCAP_RRN_ix);
      break;
   case ENTCAP_ix:
      pszNameCapa = ReservedName::name_of(ENTCAP_RRN_ix);
      break;
   case ENTCHCAP_ix:
      pszNameCapa = ReservedName::name_of(ENTCHCAP_RRN_ix);
      break;
   case EXGRPCAP_ix:
      pszNameCapa = ReservedName::name_of(EXGRPCAP_RRN_ix);
      break;
   case EXNMCAP_ix:
      pszNameCapa = ReservedName::name_of(EXNMCAP_RRN_ix);
      break;
   case GRPCAP_ix:
      pszNameCapa = ReservedName::name_of(GRPCAP_RRN_ix);
      break;
   case IDCAP_ix:
      pszNameCapa = ReservedName::name_of(IDCAP_RRN_ix);
      break;
   case IDREFCAP_ix:
      pszNameCapa = ReservedName::name_of(IDREFCAP_RRN_ix);
      break;
   case LKNMCAP_ix:
      pszNameCapa = ReservedName::name_of(LKNMCAP_RRN_ix);
      break;
   case LKSETCAP_ix:
      pszNameCapa = ReservedName::name_of(LKSETCAP_RRN_ix);
      break;
   case MAPCAP_ix:
      pszNameCapa = ReservedName::name_of(MAPCAP_RRN_ix);
      break;
   case NOTCAP_ix:
      pszNameCapa = ReservedName::name_of(NOTCAP_RRN_ix);
      break;
   case NOTCHCAP_ix:
      pszNameCapa = ReservedName::name_of(NOTCHCAP_RRN_ix);
      break;
   default:  /* case TOTALCAP_ix: */
      pszNameCapa = ReservedName::name_of(TOTALCAP_RRN_ix);
      break;
   }
   return pszNameCapa;
}

/*--------------------------------------------------------------------capa_chk-+
| Check the capacities do not exceed the max allowed values                    |
+-----------------------------------------------------------------------------*/
boolean Capacity::capa_chk(unsigned long *aulCapacityCount)
{
   boolean rc = TRUE;
   unsigned long j, total;

   total = 0;
   for (int i=0; i < CAP_MAX_SIZE; ++i) {
      if (i == TOTALCAP_ix) {
         j = total;
      }else {
         j = aulCapacityCount[i] * auiCapacityPoint[i];
         total += j;
      }
      if (j > aulCapacityMax[i]) {
         errorh(
            BIGCAP, ERH_E_SEV, name_of((e_IxCapacity)i),
            j , aulCapacityMax[i]
         );
         rc = FALSE;
      }
   }
   return rc;
}

/*---------------------------------------------------------------------compute-+
| Compute the value of a capacity, given an array of points count.             |
+-----------------------------------------------------------------------------*/
unsigned long Capacity::compute(
   e_IxCapacity ix, unsigned long *aulCapacityCount
)
{
   if (ix < TOTALCAP_ix) {
      return aulCapacityCount[ix] * auiCapacityPoint[ix];
   }else if (ix == TOTALCAP_ix) {
      unsigned long total = 0;
      for (int i=0; i < CAP_MAX_SIZE; ++i) {
         if (i != TOTALCAP_ix) {
            total += aulCapacityCount[i] * auiCapacityPoint[i];
         }
      }
      return total;
   }else {
      return 0;
   }
}

/*-------------------------------------------------------------------set_value-+
| Set new capacity values                                                      |
| limit each capacity to a specific max                                        |
+-----------------------------------------------------------------------------*/
unsigned long NewCapacity::set_value(e_IxCapacity ix, unsigned long value)
{
   unsigned long max;

   switch (ix) {
   case ATTCAP_ix:    max = ATTCAP_MAX;      break;
   case ATTCHCAP_ix:  max = ATTCHCAP_MAX;    break;
   case AVGRPCAP_ix:  max = AVGRPCAP_MAX;    break;
   case ELEMCAP_ix:   max = ELEMCAP_MAX;     break;
   case ENTCAP_ix:    max = ENTCAP_MAX;      break;
   case ENTCHCAP_ix:  max = ENTCHCAP_MAX;    break;
   case EXGRPCAP_ix:  max = EXGRPCAP_MAX;    break;
   case EXNMCAP_ix:   max = EXNMCAP_MAX;     break;
   case GRPCAP_ix:    max = GRPCAP_MAX;      break;
   case IDCAP_ix:     max = IDCAP_MAX;       break;
   case IDREFCAP_ix:  max = IDREFCAP_MAX;    break;
   case LKNMCAP_ix:   max = LKNMCAP_MAX;     break;
   case LKSETCAP_ix:  max = LKSETCAP_MAX;    break;
   case MAPCAP_ix:    max = MAPCAP_MAX;      break;
   case NOTCAP_ix:    max = NOTCAP_MAX;      break;
   case NOTCHCAP_ix:  max = NOTCHCAP_MAX;    break;
   default:           max = TOTALCAP_MAX;    break;
   }
   if (value > max) value = max;
   aulCapacityMax[ix] = value;
   return value;
}

/*-----------------------------------------------------------check_consistency-+
| Verify TOTALCAP is greater than any single value                             |
+-----------------------------------------------------------------------------*/
void NewCapacity::check_consistency()
{
   unsigned long ulMaxCap = 0;
   for (int i=0; i < CAP_MAX_SIZE; ++i) {
      if (aulCapacityMax[i] > ulMaxCap) ulMaxCap = aulCapacityMax[i];
   }
   if (ulMaxCap > aulCapacityMax[TOTALCAP_ix]) {
      aulCapacityMax[TOTALCAP_ix] = ulMaxCap;
      errorh(BIGQTY, ERH_E_SEV, ulMaxCap);
   }
}

/*----------------------------------------------------------------------unload-+
| Write the capacities to disk                                                 |
+-----------------------------------------------------------------------------*/
void NewCapacity::unload(ApplicationHandle ahlAccess, int iNamelen)
{
   /*---------------------------+
   | Update the capacity points |
   +---------------------------*/

   for (int i=0; i < CAP_MAX_SIZE; ++i) {
      if (auiCapacityPoint[i] == NAMELEN_RCS) auiCapacityPoint[i] = iNamelen;
   }
   EntityHandler::write(ahlAccess, aulCapacityMax, sizeof aulCapacityMax);
   EntityHandler::write(
      ahlAccess, auiCapacityPoint, sizeof auiCapacityPoint
   );
}
/******************************************************************************/
