/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// compo.c
///  Attribute Components classes
//   Pierre G. Richard 94/02/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <stdio.h>
#include "gendef.h"
#include "entity.h"
#include "attrib.h"
#include "compo.h"

/*------------------------------------------------------------ComponentBuilder-+
| Note: this algorithm does not handle nested entities - it's made to work     |
| for CDATA and SDATA entities, where no nesting entity can occur.             |
+-----------------------------------------------------------------------------*/
ComponentBuilder::ComponentBuilder(char *pchBufferArg) : pchBuffer(pchBufferArg)
{
   iCount = 0;
   aCompo = 0;
   pchPart = pchBuffer;
}

/*-------------------------------------------------------------------record_Ee-+
| Set the length of the current (now ended) part and start a new current part. |
| Assume the new part is text (the pEnt field of the current part is NULL).    |
+-----------------------------------------------------------------------------*/
void ComponentBuilder::record_Ee(char *pchCurrent)
{
   if ((iCount) && (aCompo[iCount].pEnt != 0)) {  // end of a recorded entity
      (aCompo[iCount]).iLengthPart = diffptr(pchCurrent, pchPart);
      aCompo = (Component *)realloc(aCompo, (1 + ++iCount) * sizeof(Component));
      (aCompo[iCount]).pEnt = 0;
      pchPart = pchCurrent;
   }
}

/*------------------------------------------------------------record_ent_start-+
| If the current text-part is not empty or was created (when Ee was found),    |
| set its length.  Start a new current entity part.                            |
+-----------------------------------------------------------------------------*/
void ComponentBuilder::record_ent_start(
   char *pchCurrent, const EntityDef *pEnt
)
{
   if (!iCount) {
      if (pchCurrent > pchPart) {
         aCompo = (Component *)malloc(3 * sizeof(Component));
         (aCompo[1]).pEnt = 0;
         (aCompo[1]).iLengthPart = diffptr(pchCurrent, pchPart);
         iCount = 2;
      }else {
         aCompo = (Component *)malloc(2 * sizeof(Component));
         iCount = 1;
      }
   }else {
      if (pchCurrent > pchPart) {
         (aCompo[iCount]).iLengthPart = diffptr(pchCurrent, pchPart);
         aCompo = (Component *)realloc(
            aCompo, (1 + ++iCount) * sizeof(Component)
         );
      }
   }
   (aCompo[iCount]).pEnt = pEnt;
   pchPart = pchCurrent;
}

/*-------------------------------------------------------------------terminate-+
| If text was entered in the current part -- cannot be an entity: last Ee      |
| created a text part -- set the length of it.  Otherwise, remove the          |
| extraneous component entry created when last Ee was processed.               |
+-----------------------------------------------------------------------------*/
Component * ComponentBuilder::terminate(char *pchCurrent)
{
   if (iCount) {
      if (pchCurrent > pchPart) {
         (aCompo[iCount]).iLengthPart = diffptr(pchCurrent, pchPart);
      }else {
         aCompo = (Component *)realloc(
            aCompo, (1 + --iCount) * sizeof(Component)
         );
      }
      aCompo[0].iCount = iCount + 1;
      aCompo[0].pszLit = pchBuffer;
   }
   return aCompo;
}

/*------------------------------------------------------------------------copy-+
|                                                                              |
|                                                                              |
+-----------------------------------------------------------------------------*/
int Component::copy(
   Component *&pCompoMP, const Component * const aCompoTable
)
{
   int ixCompo = diffptr(pCompoMP, aCompoTable);
   memcpy(pCompoMP, this, iCount * sizeof(Component));
   pCompoMP += iCount;
   return ixCompo;
}

/*---------------------------------------------------------------------pile_up-+
| Unlocate entity pointers in components, replacing addresses by their resp-   |
| ective indices in the ENT array.  If it is plain text, set the index to -1.  |
+-----------------------------------------------------------------------------*/
void Component::pile_up(unsigned int iMaxEntries)
{
   Component *pCompo = this;

   while (iMaxEntries) {                   // for the whole array
      int iCount = pCompo->iCount;         // get the number of components
      iMaxEntries -= iCount;
      while (--iCount) {                   // for each component
         const EntityDef *pEnt = (++pCompo)->pEnt;
         if (pEnt) {
            pCompo->ixEnt = pEnt->ent_index();
         }else {
            pCompo->ixEnt = -1;            // native text
         }
      }
      ++pCompo;                            // next component header
   }
}

/*--------------------------------------------------------------------relocate-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void Component::relocate(const EntityDef *aEntGenTable, char *strbuf)
{
   pszLit = strbuf + ofsLit;
   int i = iCount;
   Component *pCompo = this;
   while (--i) {
      if ((++pCompo)->ixEnt == -1) {
         pCompo->pEnt = 0;
      }else {
         pCompo->pEnt = aEntGenTable + pCompo->ixEnt;
      }
   }
}

/*---------------------------------------------------------------------display-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void Component::display(int iIndent)
{
   int i = iCount;
   const char *pch = pszLit;

   Component *pCompo = this + 1;
   while (--i) {
      fprintf(
         stderr, "%*.s%-10s=>\"%.*s\"\n", iIndent, " ",
         pCompo->pEnt? pCompo->pEnt->name() : "#TEXT",
         pCompo->iLengthPart, pch
      );
      pch += (pCompo++)->iLengthPart;
   }
}
/*============================================================================*/
