/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// compo.h
///  Declarations for compo.c
//   Pierre G. Richard 94/02/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _COMPO_H
#define _COMPO_H 1
//******************************************************************************

class EntityDef;

/*-----------------------------------------------------------------------------+
| An array of `Components' describes a literal made of text, mixed with CDATA  |
|  and SDATA entities references.                                              |
| - the first structure (the component array header) hold a pointer to the     |
|   literal (pszLit) and a count of subsequent components (iCount)             |
| - each of the others hold a pointer to the EntityDef this part of the        |
|   literal is coming from, and the length of the part of literal.             |
|   If this part of the literal is just native text, the pEnt pointer is 0.    |
+-----------------------------------------------------------------------------*/
class Component {
   friend class ComponentBuilder;
   friend class AttributeList;
   friend class Attribute;   // Only for set()
public:
   int copy(Component *&pCompoMP, const Component * const aCompoTable);
   void pile_up(unsigned int iMaxEntries);
   void relocate(const EntityDef *aEntGenTable, char *strbuf);
   const char * value();
   int count();
   const char * part_value(int ix);
   const int part_length(int ix);
   const EntityDef * entity_def(int ix);
   void display(int iIndent);
private:
   union {
      char *pszLit;          // Literal (first entry)
      unsigned int ofsLit;   // (offset, when unloaded)
      const EntityDef *pEnt; // Entity or 0 when native text
      int ixEnt;             // (index to ETY array when unloaded)
   };
   union {
      unsigned int iCount;   // nbr of components (first entry)
      unsigned int iLengthPart;
   };
   Component();              // undefined
   ~Component();             // undefined
};

/*-----------------------------------------------------------------------------+
| ComponentBuilder is used to build the Component Array                        |
+-----------------------------------------------------------------------------*/
class ComponentBuilder {
public:
   ComponentBuilder(char *pchCurrent);
   void record_Ee(char *pchCurrent);
   void record_ent_start(char *pchCurrent, const EntityDef *pEnt);
   Component * terminate(char *pchCurrent);
private:
   int iCount;               // Count of components in the list
   Component *aCompo;        // Array of `iCount' components
   char * const pchBuffer;   // Buffer
   char * pchPart;           // where the current part is starting;
};

/*--------> INLINES */
inline const char * Component::value() { return pszLit; }
inline int Component::count() { return iCount; }
inline const char * Component::part_value(int ix) {
   char *pch = pszLit;
   int i = 0;
   while (i < ix) pch += this[++i].iLengthPart;
   return pch;
}
inline const int Component::part_length(int ix) {
   return this[ix + 1].iLengthPart;
}
inline const EntityDef * Component::entity_def(int ix) {
   return this[ix + 1].pEnt;
}
#endif /* _COMPO_H ************************************************************/
