/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dlmstr.c
///  Delimiter strings for fast recognition of SGML delimiters
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "enthdlr.h"
#include "dlmstr.h"
#include "refstr.h"

/*-------------------------------------------------------------DelimiterString-+
| Constructor from scratch: this is the minimal CORE Concrete Syntax           |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
DelimiterString::DelimiterString()
{
   aDlmstrRefer = ReferenceString::get_achReferDlmValues();
   memcpy(
      aDlmstrVarnt,
      ReferenceString::get_aKywdVarntDlmValues(),
      sizeof aDlmstrVarnt
   );
}

/*-------------------------------------------------------------DelimiterString-+
| Constructor from OCS                                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
DelimiterString::DelimiterString(
   ApplicationHandle ahlAccess,       // OCS file handle
   char *strbuf                       // string area of this OCS
)
{
   aDlmstrRefer = ReferenceString::get_achReferDlmValues();
   EntityHandler::read(ahlAccess, aDlmstrVarnt, sizeof aDlmstrVarnt);
   /*
   | Relocate aDlmstrVarnt
   */
   for (int i=0; i < sizeof aDlmstrVarnt / sizeof aDlmstrVarnt[0]; ++i) {
      aDlmstrVarnt[i].p = strbuf + aDlmstrVarnt[i].ofs;
   }
}

/*PROTECTED----------------------------------------------------DelimiterString-+
| Constructor for building delimiter strings from the SGML declaration         |
+-----------------------------------------------------------------------------*/
DelimiterString::DelimiterString(int)
{
   aDlmstrRefer = ReferenceString::get_achReferDlmValues();
   memset(aDlmstrVarnt, 0, sizeof aDlmstrVarnt);
}

/*----------------------------------------------------------NewDelimiterString-+
| NewDelimiterString constructor                                               |
+-----------------------------------------------------------------------------*/
NewDelimiterString::NewDelimiterString(
   StringKeeper& skArg
) : DelimiterString(0), sk(skArg)
{
}

/*--------------------------------------------------set_delimiter_string_table-+
| Set up the delimiter string table for single delimiters                      |
+-----------------------------------------------------------------------------*/
void NewDelimiterString::set_delimiter_string_table(
   const e_IxDlmitmRefer ixDli, const char *pszValue
)
{
   e_IxDlmstrVarnt ixVdl = ReferenceString::delim_index_of(ixDli);

   if (ixVdl != NOT_FOUND_VDL_ix) {  // selected single delims in aDlmstrVarnt
      (aDlmstrVarnt)[ixVdl].ofs = sk.park(pszValue, 1+strlen(pszValue));
   }
}

/*--------------------------------------------------set_delimiter_string_table-+
| Set up the delimiter string table for the double delimiters                  |
+-----------------------------------------------------------------------------*/
void NewDelimiterString::set_delimiter_string_table(
   const e_IxDDlmitmRefer ixDDli, const char *pszValue
)
{
   e_IxDlmstrVarnt ixVdl;       // selected double delims in aDlmstrVarnt

   switch (ixDDli) {
   case MDO_DSO_DDLI_ix:
      ixVdl = MDO_DSO_VDL_ix;
      break;
   case MSC_MDC_DDLI_ix:
      ixVdl = MSC_MDC_VDL_ix;
      break;
   default:
      return;
   }
   aDlmstrVarnt[ixVdl].ofs = sk.park(pszValue, 1+strlen(pszValue));
}

/*----------------------------------------------------------------------unload-+
| Write the delimiter strings to disk                                          |
+-----------------------------------------------------------------------------*/
void NewDelimiterString::unload(ApplicationHandle ahlAccess)
{
   EntityHandler::write(ahlAccess, aDlmstrVarnt, sizeof aDlmstrVarnt);
}
/******************************************************************************/
