/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dlmstr.h.h
///  Declarations for dlmstr.h
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _DLMSTR_H
#define _DLMSTR_H 1
//******************************************************************************

#include "lextoken.eh"  // enums
#include "dlmstr.eh"

class StringKeeper;

/*-----------------------------------------------------------------------------+
| Delimiter String Class                                                       |
|                                                                              |
| This class describes the delimiter strings of the syntax.                    |
+-----------------------------------------------------------------------------*/
class DelimiterString {          // Dlmstr
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   DelimiterString();                                       // Core constructor
   DelimiterString(ApplicationHandle ahlAccess, char *strbuf); //OCS constructor

   const char * value_of(e_IxDlmstrRefer ix);  // Reference value
   const char * value_of(e_IxDlmstrVarnt ix);  // Variant value

protected:
   /*----------------------+
   | Protected Definitions |
   +----------------------*/
   union {
      char *p;
      unsigned int ofs;
   }aDlmstrVarnt[VDL_MAX_SIZE];

   /*--------------------+
   | Protected Variables |
   +--------------------*/
   const char * const *aDlmstrRefer;

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   DelimiterString(int);      // when building from the SGML declaration
};

/*-----------------------------------------------------------------------------+
| NewDelimiterString class  (while building it)                                |
+-----------------------------------------------------------------------------*/
class NewDelimiterString : public DelimiterString
{
public:
   NewDelimiterString(StringKeeper& sk);
   void set_delimiter_string_table(
      const e_IxDlmitmRefer ixDli, const char *pszValue
   );
   void set_delimiter_string_table(
      const e_IxDDlmitmRefer ixDDli, const char *pszValue
   );
   void unload(ApplicationHandle ahlAccess);
private:
   StringKeeper& sk;
};

/*INLINE--------------------------------------------------------------value_of-+
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline const char * DelimiterString::value_of(e_IxDlmstrRefer ix) {
   return aDlmstrRefer[ix];
}

/*INLINE--------------------------------------------------------------value_of-+
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline const char *DelimiterString::value_of(e_IxDlmstrVarnt ix) {
   return aDlmstrVarnt[ix].p;
}
#endif /* _DLMSTR_H ***********************************************************/
