/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dlmvect.h
///  Declarations for dlmvect
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _DLMVECT_H
#define _DLMVECT_H 1
//******************************************************************************

#include "prsstate.h"   // ParseState::recognize_delimiter is a friend

#include "lextoken.eh"  // enums
#include "prsmode.eh"
#include "syntax.eh"
#include "dlmstr.eh"

class SortArray;
class NewDelimiterString;
class StringKeeper;
class NewSyntax;

struct Features;

/*-----------------------------------------------------------------------------+
| VectorItem structure:                                                        |
| This structure is used internally by the DelimiterVector class.              |
+-----------------------------------------------------------------------------*/
struct VectorItem {              // VI:
   union {                       // either:
      struct {                   // for a primary node:
         boolean isBseq;         // B sequence indicator
         char count;             // and count of consequent vectors
      }h;                        // for a secondary node, and bseq-node:
      VectorItem *next;          // pointer to next primary node, or
      VectorItem *previous;      // pointer to previous primary node
      unsigned int ixVi;         // allow relocation when loaded from disk
   };
   union {
      unsigned int bseqlen;      // bseq_node: count of minimum B-seq follows
      e_LexToken tk;             // token associated
   };
   union {                       // either
      char *str;                 // Prep Char List / Continuation String, or
      boolean must_recur;        // bseq_node: fast recognition of a recurrence
      unsigned int ofsStr;       // allow relocation when loaded from disk
   };
};

/*-----------------------------------------------------------------------------+
| DelimiterVector class:                                                       |
| The Delimiter Vector Table maps a tree for a fast delimiter recognition.     |
+-----------------------------------------------------------------------------*/
class DelimiterVector            // DV: Delimiter Vector
{
   friend e_LexToken ParseState::recognize_delimiter();
   friend boolean    ParseState::may_be_delimiter();

public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   DelimiterVector();           // minimal constructor
   DelimiterVector(ApplicationHandle ahlAccess, char *strbuf);  // OCS constructor
   ~DelimiterVector();          // destructor

protected:
   /*----------------------+
   | Protected Definitions |
   +----------------------*/
   union {                      // Selective entries in dvtab = f(mode)
      unsigned int rel;         // relative offset to the unloaded DVTAB
      VectorItem *pVi;          // ... then relocated
   }dvtbent[NBR_OF_MODES];      // Entries in DVTAB for a given mode

   /*--------------------+
   | Protected Variables |
   +--------------------*/
   VectorItem *aViTable;        // Table of vector items
   int iMaxVi;

   // "starts a delimiter" = if mode_bit on in dltab[char]
   unsigned int dltab[MAX_CHARSET];

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   DelimiterVector(int);        // when building from the SGML declaration
};

/*-----------------------------------------------------------------------------+
| NewDelimiterVector class (while building it)                                 |
+-----------------------------------------------------------------------------*/
class NewDelimiterVector : public DelimiterVector
{
public:
   /*-------------------+
   | Public Definitions |
   +-------------------*/
   struct DelimDesc {
      char *pszDelim;
      e_LexToken token;
   };

   /*-----------------+
   | Public Functions |
   +-----------------*/
   NewDelimiterVector(StringKeeper& sk, NewSyntax& syntaxNew);
   void define_delimiter(
      const char *pszValue,
      const e_LexToken token,
      unsigned int allmode,
      const boolean isSingleDelimiter
   );
   void build_delimiter_vectors(
      Features& featuresNew, NewDelimiterString& dlmstrNew
   );
   int compare_delim_desc(const DelimDesc& dlmdsc1, const DelimDesc& dlmdsc2);
   void unload(ApplicationHandle ahlAccess);

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   struct DoubleDelimiterStrings {  // combined delimiters
      char *s1;
      char *s2;
   };
   struct DelimTree {          // Threaded tree, as usual...
      boolean ltag    :1;
      boolean rtag    :1;
      boolean isBseq :1;       // indicates a BSEQ, rather than a prep char
      /**/                     // at the primary AND secondary levels
      DelimTree *llink;
      DelimTree *rlink;
      int len;
      union {
         unsigned int bseqlen; // B-SEQ minimum length (Prim Node)
         char chPrp;           // Preponderant Character of a 2nd Node
      };
      char *info;
      e_LexToken token;
   };

   /*------------------+
   | Private Variables |
   +------------------*/
   StringKeeper& sk;
   NewSyntax& syntaxNew;

   DoubleDelimiterStrings aDdlmvstr[DDLI_MAX_SIZE];
   SortArray *pSaDelim;
   DelimDesc  *aDlmdscTable;
   unsigned int iCountDlmdsc;

   unsigned int ixViPrevBseq;
   VectorItem *pViCurr;
   char *pchPCL;               // predominant character list

   /*------------------+
   | Private Functions |
   +------------------*/
   void set_delimiter_string_table(
      const e_IxDlmitmRefer ixDli, const char *pszValue
   );
   void set_delimiter_string_table(
      const e_IxDDlmitmRefer ixDDli, const char *pszValue
   );
   void check_duplicate_delimiter(unsigned int, unsigned int, int);
   void get_stub(
      unsigned int, unsigned int, unsigned int, unsigned int,
      DelimTree *, char *
   );
   void prim_postorder(DelimTree *);
   void secd_postorder(DelimTree *);
};
#endif /* _DLMVECT_H **********************************************************/
