/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dtddcl.c
///  Agglomerate for classes components of the Doctype declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include "gendef.h"
#include "hash.h"
#include "dtddcl.h"
#include "externid.h"
#include "element.h"
#include "attrib.h"
#include "entity.h"
#include "notation.h"
#include "quantity.h"
#include "syntax.h"
#include "enthdlr.h"
#include "errorh.h"

/*----------------------------------------------------------------------DTDdcl-+
| DTDdcl constructor from ODT                                                  |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
DTDdcl::DTDdcl(
   ApplicationHandle ahlAccess,
   Syntax *pSyntax           // to check OCS compatible syntax
)
{
   int ixBaseElem;            // index in ElementType array

   EntityHandler::read(ahlAccess, &ixBaseElem, sizeof ixBaseElem);
   EntityHandler::read(                    // OCS dependencies
      ahlAccess, &(iCountShortref), sizeof iCountShortref
   );
   {
      // OCS dependencies
      boolean isCompatQuantity = pSyntax->pQuantity->read_and_check_quantity(
         ahlAccess
      );
      boolean isCompatCharset = pSyntax->read_and_check_charset(ahlAccess);

      if ((!isCompatQuantity) || (!isCompatCharset)) {
         errorh(ABORTED, ERH_T_SEV);
      }
   }
   EntityHandler::read(
      ahlAccess, aulCapacityCount, sizeof(aulCapacityCount)
   );
   int iCountCnttok = EntityHandler::read_record(
      ahlAccess, (void **)&aCnttokTable
   ) / sizeof aCnttokTable[0];
   iCountElem = EntityHandler::read_record(
      ahlAccess, (void **)&aElemTable
   ) / sizeof aElemTable[0];
   int iCountEx = EntityHandler::read_record(
      ahlAccess, (void **)&aExTable
   ) / sizeof aExTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashElem);
   int iCountAtt = EntityHandler::read_record(
      ahlAccess, (void **)&aAttrepTable
   ) / sizeof aAttrepTable[0];
   EntityHandler::read_record(ahlAccess, (void **)&aCompoTable);
   iCountEntGen = EntityHandler::read_record(
      ahlAccess, (void **)&aEntGenTable
   ) / sizeof aEntGenTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashGenEnt);
   iCountEntPrm = EntityHandler::read_record(
      ahlAccess, (void **)&aEntPrmTable
   ) / sizeof aEntPrmTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashPrmEnt);
   int iCountSrm = EntityHandler::read_record(
      ahlAccess, (void **)&aSrmTable
   ) / sizeof aSrmTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashSrMap);
   int iCountDtp = EntityHandler::read_record(
      ahlAccess, (void **)&aDtpTable
   ) / sizeof aDtpTable[0];
   int iCountNad = EntityHandler::read_record(
      ahlAccess, (void **)&aNadTable
   ) / sizeof aNadTable[0];
   int iCountNcb = EntityHandler::read_record(
      ahlAccess, (void **)&aNcbTable
   ) / sizeof aNcbTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashNotation);
   int iCountExid = EntityHandler::read_record(
      ahlAccess, (void **)&aExidTable
   ) / sizeof aExidTable[0];
   int iCountSts = EntityHandler::read_record(
      ahlAccess, (void **)&aStsTable
   ) / sizeof aStsTable[0];
   EntityHandler::read_record(ahlAccess, &pvHashStem);
   int iCountRkl = EntityHandler::read_record(
      ahlAccess, (void **)&aRklTable
   ) / sizeof aRklTable[0];
   EntityHandler::read_record(ahlAccess, (void **)&strbuf);

   /*-----------------------+
   | Create the hash tables |
   +-----------------------*/
   pHshElem  = new StringizedHash(pvHashElem, aElemTable);
   pHshGenEnt = new StringizedHash(pvHashGenEnt, aEntGenTable);
   pHshPrmEnt = new StringizedHash(pvHashPrmEnt, aEntPrmTable);
   pHshSrMap  = new StringizedHash(pvHashSrMap, aSrmTable);
   pHshNotation  = new StringizedHash(pvHashNotation, aNcbTable);
   pHshStem  = new StringizedHash(pvHashStem, aStsTable);

   /*------------+
   | Relocate... |
   +------------*/
   aCnttokTable->relocate_pile(     // ContentToken tables
      iCountCnttok, aElemTable, aDtpTable
   );
   aElemTable->relocate_pile(       // ElementType tables
      iCountElem,
      aCnttokTable, aExTable, aAttrepTable, aSrmTable, aRklTable, strbuf
   );
   aExTable->relocate_pile(         // Exceptions
      iCountEx, aElemTable
   );
   aAttrepTable->relocate_pile(     // AttributeRep tables
      iCountAtt, aCompoTable, aEntGenTable, strbuf
   );
   aDtpTable->relocate_pile(        // DataTagPattern tables
      iCountDtp, strbuf
   );
   aNadTable->relocate_pile(        // EntityNotation tables
      iCountNad, aAttrepTable, aNcbTable
   );
   aNcbTable->relocate_pile(        // NotationType tables
      iCountNcb, aAttrepTable, strbuf
   );
   aStsTable->relocate_pile(        // StemValue table
      iCountSts, strbuf
   );
   aRklTable->relocate_pile(        // RankList table
      iCountRkl, aStsTable
   );
   aExidTable->relocate_pile(       // ExternId table
      iCountExid, strbuf
   );
   aEntGenTable->relocate_pile(     // General entity tables
      iCountEntGen, aNadTable, aExidTable, strbuf
   );
   aEntPrmTable->relocate_pile(     // Parameter entity tables
      iCountEntPrm, aNadTable, aExidTable, strbuf
   );
   aSrmTable->relocate_pile(        // ShortrefMap Tables
      iCountSrm, iCountShortref, aEntGenTable, strbuf
   );
   if (ixBaseElem == -1) {
      pElemBase = 0;
   }else {
      pElemBase = aElemTable + ixBaseElem;
   }
}

/*---------------------------------------------------------------------~DTDdcl-+
| DTDdcl destructor                                                            |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
DTDdcl::~DTDdcl()
{
   delete pHshElem;
   delete pHshGenEnt;
   delete pHshPrmEnt;
   delete pHshSrMap;
   delete pHshNotation;
   delete pHshStem;

   free(pvHashElem);
   free(pvHashGenEnt);
   free(pvHashPrmEnt);
   free(pvHashSrMap);
   free(pvHashNotation);
   free(pvHashStem);
   free(aElemTable);
   free(aNcbTable);
   free(aEntGenTable);
   free(aEntPrmTable);
   free(aCnttokTable);
   free(aExTable);
   free(aAttrepTable);
   free(aCompoTable);
   free(aSrmTable);
   free(aDtpTable);
   free(aNadTable);
   free(aExidTable);
   free(aStsTable);
   free(aRklTable);
   free(strbuf);
}

/*---------------------------------------------------------search_element_type-+
| Search for a possible generic identifier, directly or thru the stem+suffix   |
| mechanism.  Return a pointer to its element type, or O if not found.         |
|                                                                              |
| Note1: if the rank feature is not enabled, the suffix is -1                  |
|                                                                              |
| Note2: pszName is assumed to by NAMELEN+1 long, so the suffix can be         |
|       catenated without overflow.                                            |
+-----------------------------------------------------------------------------*/
const ElementType * DTDdcl::search_element_type(
   char *pszName
)
{
   const ElementType *pElem;

   pElem = (ElementType *)pHshElem->search_value(pszName);
   if (!pElem) {
      StemValue *pSts = (StemValue *)pHshStem->search_value(pszName);
      if (pSts) {
         if (pSts->suffix() != -1) {               // OK: catenate suffix
            sprintf(pszName + strlen(pszName), "%d", pSts->suffix());
            pElem = (ElementType *)pHshElem->search_value(pszName);
         }
      }
   }
   return pElem;
}
/******************************************************************************/
