/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dtddcl.h
///  Declarations for dtddcl
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _DTDDCL_H
#define _DTDDCL_H 1
//******************************************************************************

#include "capacity.eh"  // enums

class Hash;
class ExpandedHash;
class EntityHandler;
class Quantity;
class Syntax;

class ElementType;
class ContentToken;
class EntityDef;
class NotationType;
class AttributeRep;
class DataTagPattern;
class EntityNotation;
class Exception;
class RankList;
class StemValue;
class ShortrefMap;
class ExternId;
class Component;

/*-----------------------------------------------------------------------------+
| Class DTDdcl                                                                 |
| This class reflects the data pertaining to the DTD                           |
+-----------------------------------------------------------------------------*/
class DTDdcl
{
public:
   DTDdcl(                          // constructor from ODT
      ApplicationHandle ahlAccess,
      Syntax   *pSyntax             // to check OCS compatible syntax
   );
   ~DTDdcl();

   ElementType *pElemBase;      // Document element

   const ElementType * search_element_type(char *pszName);

// protected:
//***>NOTE: will be this when DTDdcl will be a
//***>NOTE: virtual? private class of ParseState

   Hash *pHshElem;              // Hash table for elements GI
   Hash *pHshGenEnt;           // Hash table for general entities   (also Expn)
   Hash *pHshPrmEnt;           // Hash table for parameter entities (also Expn)
   Hash *pHshSrMap;            // Hash table for SR MAPS
   Hash *pHshNotation;         // Hash table for NOTATION
   Hash *pHshStem;             // hash table for RANK

   int iCountShortref;                            // Number of Shortref defined
   unsigned long aulCapacityCount[CAP_MAX_SIZE];  // DTD capacities

private:
   ElementType      *aElemTable;
   int iCountElem;
   NotationType     *aNcbTable;
   int iCountNcb;
   EntityDef        *aEntGenTable;
   int iCountEntGen;
   EntityDef        *aEntPrmTable;
   int iCountEntPrm;
   ContentToken     *aCnttokTable;
   Exception        *aExTable;
   AttributeRep     *aAttrepTable;
   Component        *aCompoTable;
   ShortrefMap      *aSrmTable;
   DataTagPattern   *aDtpTable;
   EntityNotation   *aNadTable;
   ExternId         *aExidTable;
   StemValue        *aStsTable;
   RankList         *aRklTable;
   char             *strbuf;
   void *pvHashElem;       // Work area for Elem hash table
   void *pvHashGenEnt;    // Work area for GenEnt hash table
   void *pvHashPrmEnt;    // Work area for PrmEnt hash table
   void *pvHashSrMap;     // Work area for SrMap hash table
   void *pvHashNotation;  // Work area for Notation hash table
   void *pvHashStem;      // Work area for Stem hash table
};
#endif /* _DTDDCL_H ***********************************************************/
