/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// elemdcl
///  ELEMENT declaration
//   Pierre G. Richard 87/08/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "prsstate.h"
#include "quantity.h"
#include "dlmstr.h"
#include "sgmldcl.h"
#include "prsmode.h"
#include "rsvname.h"
#include "sort.h"
#include "skmgr.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "elemdcl.h"
#include "chkambg.h"
#include "errorh.h"


/*STATIC------------------------------------------inst_new_declared_gi_or_stem-+
| This procedure is called by parse_name_group for each element name or rank   |
| stem.  Build a list until we know if it is a stem or a element               |
+-----------------------------------------------------------------------------*/
static void inst_new_declared_gi_or_stem(void *pvBase)
{
   ((ElementDcl *)pvBase)->new_declared_gi_or_stem();
}

/*STATIC-------------------------------------------------inst_new_exception_gi-+
| This procedure is called by parse_name_group for each exception.             |
+-----------------------------------------------------------------------------*/
static void inst_new_exception_gi(void *pvBase)
{
   ((ElementDcl *)pvBase)->new_exception_gi();
}

/*STATIC------------------------------------------------------compare_template-+
| This procedure is called by Sort::quick to compare templates                 |
+-----------------------------------------------------------------------------*/
static int compare_template(int i1, int i2, void *pvBase)
{
   return ((ElementDcl *)pvBase)->template_compare(i1, i2);
}

/*STATIC-----------------------------------------------------exchange_template-+
| This procedure is called by Sort::quick to exchange templates                |
+-----------------------------------------------------------------------------*/
static void exchange_template(int i1, int i2, void *pvBase)
{
   ((ElementDcl *)pvBase)->template_exchange(i1, i2);
}

/*-----------------------------------------------------------------------parse-+
| Process ELEMENT MD.                                                          |
|                                                                              |
| [116] element declaration = mdo,"ELEMENT",                                   |
|       ps+, element_type, (ps+, omitted tag minimization)?,                   |
|       ps+, (declared_content | content_model), ps*                           |
|       mdc                                                                    |
|                                                                              |
| Enter here when current_char after "ELEMENT",                                |
| When return, current_char at mdc                                             |
+-----------------------------------------------------------------------------*/
void ElementDcl::parse()
{
   static const VarntChoiceEntry eldc_choices[] = {
      { CDATA_VRN_ix,     CDATA_TK     },
      { RCDATA_VRN_ix,    RCDATA_TK    },
      { EMPTY_VRN_ix,     EMPTY_TK     },
      { ANY_VRN_ix,       ANY_TK       },
      { NOT_FOUND_VRN_ix, NOT_FOUND_TK }
   };

   /*-----------+
   | Initialize |
   +-----------*/
   iGrpGtCnt = 0;
   iGrpLvl = 0;
   isCheckAmbiguities = FALSE;
   isPcdataDuplicate = FALSE;
   pszNamePV = pchNamesBuffer;
   ++iIdentElem;
   int iCountUselessModelSave = nbrof.useless_model;

   /*------------------------------------------------------+
   | Get the list of names in pchNamesBuffer               |
   |                                                       |
   | [117] element_type = generic_identifier |             |
   |          | name_group | ranked_element | ranked_group |
   +------------------------------------------------------*/
   pPrsstat->skip_ps();
   switch (pPrsstat->get_curr_token()) {
   case GRPO_TK:
      pPrsstat->parse_name_group(              // Parse name group
         inst_new_declared_gi_or_stem, this
      );
      break;
   case NOT_FOUND_TK:
      new_declared_gi_or_stem();               // or get individual name
      break;
   default:
      errorh(INVNAMGRP, pEh, ERH_T_SEV);
      break;
   }
   /*--------------------------------------------+
   |Find out if these are stems, rather than GIs |
   +--------------------------------------------*/
   pEnodePV = new ElementNode();
   pEnodePV->is.any = FALSE;
   pPrsstat->skip_ps();
   {
      int iLenSuffix = 0;

      if ((pPrsstat->get_curr_token() == NOT_FOUND_TK) && (pEh->is_digit())) {
         char *pch = pszNamePV-1;   // use the extra entry (GRPCNT + 1)
         do {
            if (++iLenSuffix <= NAMELEN) *(++pch) = (char)pEh->curr_char_val();
         }while (pSyntax->is_digit(pEh->read()));
         *++pch = '\0';
         pPrsstat->skip_ps();
      }

      int iReferenceCount = 0;
      GenericIdBlock *pGi;
      RankNode *pRknStem, *pRknGroup, *pRkn;

      for (
         char *pch = pchNamesBuffer, *pszSuffix = pszNamePV;
         pch != pszSuffix;
         pch += 1+NAMELEN
      ) {
         if (iLenSuffix) {
            if (expnhashElem.search_value(pch)) {
               errorh(DUPSTMELEM, pEh, ERH_E_SEV, pch);
               continue;
            }
            pRknStem = (RankNode *)expnhashStem.gimme_value(
               pch, sizeof(RankNode)
            );
            if (!(pRknStem->pRknStem)) {
               pRknStem->pRknStem = pRknStem;
            }
            int i = strlen(pch);
            if ((iLenSuffix + i) > NAMELEN) {
               errorh(BIGSTMSUF, pEh, ERH_E_SEV, pch, pszSuffix);
               continue;
            }
            memcpy(pch+i, pszSuffix, iLenSuffix+1);
         }else {
            if (expnhashStem.search_value(pch)) {
               errorh(DUPELEMSTM, pEh, ERH_E_SEV,pch);
               continue;
            }
         }
         pGi = (GenericIdBlock *)expnhashElem.gimme_value(
            pch, sizeof(GenericIdBlock)
         );
         if (pGi->pEnode) {
            errorh(DUPELEM, pEh, ERH_E_SEV, pch);
            continue;
         }
         pGi->pEnode = pEnodePV;
         if (iLenSuffix) {
            pRkn = (RankNode *)malloc(sizeof(RankNode));
            pRkn->pRknStem = pRknStem->pRknStem;
            pRknStem->pRknStem = pRkn;
            ++pRknStem->stem_cnt;
            if (!iReferenceCount) {
               pRknGroup = pRkn->pRknGroup = pRkn;
            }else {
               pRkn->pRknGroup = pRknGroup->pRknGroup;
               pRknGroup->pRknGroup = pRkn;
            }
            pGi->pRkn = pRkn;
         }
         ++iReferenceCount;                // One more referenced
      }
      if (iLenSuffix) nbrof.rkl += iReferenceCount;
      pEnodePV->iReferenceCount = iReferenceCount;
   }
   /*-----------------------+
   |Omitted tag minimization|
   +-----------------------*/
   for (
      int iCountMini = 0;
      (iCountMini < 2) && (pPrsstat->get_curr_token()==NOT_FOUND_TK);
      ++iCountMini
   ) {
      if (pSyntax->match_rsvname(pEh, O_VRN_ix)) { // "O" found
         if (!iCountMini) {                        // first time:
            pPrsstat->skip_inferable_required_ps();          // skip'n check PS
         }else {
            pPrsstat->skip_ps();                             // skip PS if any
         }
      }else {                                      // not "O" found
         if (!pSyntax->match_string(
               pEh, pSyntax->pDlmstr->value_of(MINUS_VDL_ix)
            )
         ) {                                       // neither MINUS
            break;                                 // in error
         }else {                                   // MINUS found
            if (!iCountMini) {                     // first time:
               pEnodePV->is.reqd_open_tag = TRUE;  // for open
            }else {                                // 2nd time:
               pEnodePV->is.reqd_end_tag = TRUE;   // for close
            }
         }
         pPrsstat->skip_ps();                                // skip PS if any
      }
   }
   if ((!iCountMini) && (!pSgmldcl->features.isOmittag)) {
      pEnodePV->is.reqd_open_tag = TRUE;
      pEnodePV->is.reqd_end_tag = TRUE;
   }else if (iCountMini != 2) {
      errorh(NOOTAGMIN, pEh, ERH_T_SEV);
   }

   /*----------------------------------------------------------------+
   | decode element content                                          |
   |                                                                 |
   | element_content = (declared_content | content_model)            |
   |                                                                 |
   | [125] declared_content = "CDATA" | "RCDATA" | "EMPTY"           |
   | [126] content_model = (model_group | "ANY"), (ps+, exceptions)? |
   +----------------------------------------------------------------*/
   ModelNode mdl_head;
   boolean isCheckExceptions = FALSE;

   switch (pPrsstat->get_curr_token()) {
   case GRPO_TK:
      pPrsmode->push_mode(GRP_MODE);
      mdl_head.llink = &(mdl_head);
      mdl_head.is.conn = SEQ_CONN;             // single content => seq
      ppMnodeLink = &(mdl_head.llink);
      occurTok = NULL_OCCUR;
      isCheckExceptions = TRUE;
      new_mdl_link();
      decode_model_group();
      set_mnode();
      pEnodePV->pMnode = mdl_head.llink;      // content  model
      if (isCheckAmbiguities) {
         ModelNode *pMnode = mdl_head.llink;
         if (pMnode->is.conn) {
            CheckAmbiguities(
               pMnode, GRPGTCNT, isPcdataDuplicate, pEh,
               pRsvname->name_of(PCDATA_VRN_ix)
            );
         }
      }
      pPrsmode->pop_mode();
      break;
   case NOT_FOUND_TK:
      /*
      | get element declared content
      */
      switch (pSyntax->match_rsvname_in_list(pEh, eldc_choices)) {
      case ANY_TK:
         pEnodePV->is.any = TRUE;
         pEnodePV->is.pcdata = TRUE;  // mixed cnt, and char data allowed
         isCheckExceptions = TRUE;
         break;
      case CDATA_TK:
         pEnodePV->is.dcdata = NONR_DCDATA;
         if (!pEnodePV->is.reqd_open_tag) {    // O start tag for DCDATA
            pEnodePV->is.reqd_open_tag = TRUE; // is invalid
            errorh(INVSTGMIN, pEh, ERH_I_SEV);
         }
         break;
      case RCDATA_TK:
         pEnodePV->is.dcdata = RPLC_DCDATA;
         if (!pEnodePV->is.reqd_open_tag) {    // O start tag for DCDATA
            pEnodePV->is.reqd_open_tag = TRUE; // is invalid
            errorh(INVSTGMIN, pEh, ERH_I_SEV);
         }
         break;
      case EMPTY_TK:
         pEnodePV->is.empty = TRUE;
         if (!pEnodePV->is.reqd_open_tag) {    // O start tag for EMPTY
            pEnodePV->is.reqd_open_tag = TRUE; // is invalid
            errorh(INVSTGMIN, pEh, ERH_I_SEV);
         }
         if (pEnodePV->is.reqd_end_tag) {      // - end tag for EMPTY
            pEnodePV->is.reqd_end_tag = FALSE; // is invalid
            if (iCountMini) errorh(INVETGMIN, pEh, ERH_I_SEV);
         }
         break;
      default:                                 // NOT_FOUND
         errorh(INVKWD, pEh, ERH_T_SEV);
         break;
      }
      break;
   default:
      errorh(INVDLM, pEh, ERH_T_SEV);
      break;
   }
   /*------------------------------------------------------------+
   | Decode exceptions                                           |
   | [138] exceptions = (exclusions,(ps+,inclusions)?|inclusions |
   | [140] exclusions = MINUS, name group                        |
   | [139] inclusions = PLUS, name group                         |
   +------------------------------------------------------------*/
   pEnodePV->pChxMinus = 0;           // Minus exceptions
   pEnodePV->pChxPlus = 0;            // Plus exceptions
   pPrsstat->skip_ps();
   if (isCheckExceptions) {
      switch (pPrsstat->get_curr_token()) {
      case MINUS_GRPO_TK:             // handle exclusions first
         /*------------------+
         | handle exclusions |
         +------------------*/
         ppChx = &pEnodePV->pChxMinus;
         if (pEnodePV->iReferenceCount) ++nbrof.exceptions;
         pPrsstat->parse_name_group(inst_new_exception_gi, this);
         pPrsstat->skip_ps();
         if (pPrsstat->get_curr_token() != PLUS_GRPO_TK) break;
         /* else, check inclusions */
      case PLUS_GRPO_TK:
         /*------------------+
         | handle inclusions |
         +------------------*/
         ppChx = &pEnodePV->pChxPlus;
         if (pEnodePV->iReferenceCount) ++nbrof.exceptions;
         pPrsstat->parse_name_group(inst_new_exception_gi, this);
         pPrsstat->skip_ps();
      default:
         break;                           // not an exception
      }
   }
   if (!pEnodePV->iReferenceCount) {      // undo everything...
      if (
         (!pEnodePV->is.any) && (!pEnodePV->is.empty) &&
         (!pEnodePV->is.dcdata)
      ) {
         undo_mdl(pEnodePV->pMnode);
         nbrof.useless_model = iCountUselessModelSave;
      }
      delete pEnodePV;
   }
}

/*-----------------------------------------------------new_declared_gi_or_stem-+
| A new name is declared in the name group of the ELEMENT declaration.         |
| Enter this name in the pchNamesBuffer                                        |
+-----------------------------------------------------------------------------*/
void ElementDcl::new_declared_gi_or_stem()
{
   pSyntax->read_general_name(pEh, pszNamePV);
   pszNamePV += 1 + NAMELEN;
}

/*------------------------------------------------------------new_exception_gi-+
| A new name is declared in the exception group of the ELEMENT declaration.    |
| Enter this exception in the pChx chain.                                      |
+-----------------------------------------------------------------------------*/
void ElementDcl::new_exception_gi()
{
   char pszName[NAMELEN_MAX+1];
   GenericIdBlock *pGi;

   pSyntax->read_general_name(pEh, pszName);
   pGi = (GenericIdBlock *)expnhashElem.gimme_value(
      pszName, sizeof(GenericIdBlock)
   );
   pGi->iIdentElem = iIdentElem;    // tell "check_elem" this elem was referenced
   /*-------------------+
   |Check for duplicates|
   +-------------------*/
   for (
      ExceptionChain **ppChxCurr = ppChx;
      *ppChxCurr;
      ppChxCurr = &((*ppChxCurr)->next)
   ) {
      if ((*ppChxCurr)->pGi == pGi) {
         errorh(DUPNAM, pEh, ERH_E_SEV);   // if duplicate
         return;
      }
   }
   if (pEnodePV->iReferenceCount) {        // count only if element is valid
      ++nbrof.excptnames;
      *ppChxCurr = (ExceptionChain *)malloc(sizeof(ExceptionChain));
      (*ppChxCurr)->pGi = pGi;
      (*ppChxCurr)->next = 0;
   }
}

/*----------------------------------------------------------------new_mdl_link-+
| Add a new node and prepare it to be father.                                  |
+-----------------------------------------------------------------------------*/
void ElementDcl::new_mdl_link()
{
   ModelNode *pMnode;

   pMnode = alloc_mdl_node();
   pMnode->is.occur = NULL_OCCUR;
   ppMnodeLink = &(pMnode->llink);
   *ppMnodeLink = pMnode;             // new thread set
}

/*--------------------------------------------------------------alloc_mdl_node-+
| Add a new node, either son or sibling depending on what link is pointing at. |
+-----------------------------------------------------------------------------*/
ModelNode * ElementDcl::alloc_mdl_node()
{
   ModelNode *pMnode;

   ++nbrof.model;
   if (++iGrpGtCnt > GRPGTCNT + 1) {  // +1: model group itself not counted
      errorh(BIGGRPGTC, pEh, ERH_T_SEV);
   }
   pMnode = (ModelNode *)malloc(sizeof(ModelNode));
   memset(pMnode, 0, sizeof(ModelNode));
   pMnode->rlink = *ppMnodeLink;      // propagate thread
   *ppMnodeLink = pMnode;             // establish link
   return pMnode;
}

/*-------------------------------------------------------------------set_mnode-+
| Enter here when an element token has been decoded:                           |
| Either a "group token" as in "()" or "()|" or "()?,"                         |
| Or a "GI token" as in "backm" or "backm+" or "backm+&"                       |
|                                                                              |
| occurTok has the occurence indicator,                                        |
| connTok has the connector,                                                   |
| pszNameTok has the element name                                              |
| mnkindTok has the kind of model node                                         |
|                                                                              |
| If it's a group node,                                                        |
|   it's a father already created (a father must exist before his children)    |
|   get its address from the child-father thread                               |
|   if he has no children at all: error                                        |
|   else if he has only one child, he's a useless father: remove it:           |
|      - the resulting occurence is the logical OR of the child's and          |
|        father's occurences Ex: (a?)+ == a*                                   |
|      - the resulting connector (ltag) is the child's connector (if none:     |
|        the child is a real node, so is the father - and if one, the child    |
|        is a dummy father node, and the father inherits the connector that    |
|        links the children of the child)                                      |
|     -  the resulting llink is the child's llink (same reasons as above)      |
|     -  rtag and rlink are unchanged, rtag is overwritten later               |
|     => so:  AND `occur' with the child occurence,                            |
|             then copy the rlink to the child,                                |
|             (after doing the classic update of the rthread)                  |
|             set the child as the current node, delete old node               |
|                                                                              |
|  Note: the "ltag" corresponds to the connector (AND, OR, SEQ) binding        |
|        the children: it's known only after the first child has been          |
|        processed.  So, initially, "no ltag" does *not* mean that the         |
|        node is not a father: it can have a unique son and the bind between   |
|        children is not known.  The only way to check such a case is to       |
|        ask the left-linked element who's his father.                         |
|                                                                              |
| Else, it's a regular node (a brother):                                       |
|   create it on the fly                                                       |
|   enter his name in the "bind list"                                          |
|   turn off ltag, because real node can't have chidren                        |
|                                                                              |
| Set the occurence indic for this new node                                    |
| Link it to its predecessor                                                   |
| Find out if a connector has been declared (ask the father for elder brother) |
| If none: declare the current connector (tell the eldest is there)            |
| Else: if this is not the last child, check brother connection is consistent. |
|                                                                              |
+-----------------------------------------------------------------------------*/
void ElementDcl::set_mnode()
{
   ModelNode *pMnode;

   if (!mnkindTok) {                             // dummy node
      pMnode = *ppMnodeLink;                     // restore thread
      /*----------------------------------------------------------+
      | Apply inherency rules: (4.159, p 13)                      |
      | An optional token in an OR group makes the group optional |
      | If all tokens in any group are optional, so is the group  |
      +----------------------------------------------------------*/
      if (
         ((pMnode->is.conn == OR_CONN) && (pMnode->is.one_omit)) ||
         (!pMnode->is.one_reqd)
      ) {
         occurTok = (e_Occurence)((unsigned int)occurTok | OMISSIBLE);
      }
      if (pMnode->llink == pMnode) {             // if no child at all
         errorh(NOCNTN, pEh, ERH_T_SEV);         // it's an error
      }else if (!pMnode->is.conn) {              // no child sets connector
         ModelNode *pMnodeChild = pMnode->llink; // only one child: get it
         ModelNode *pMnodePred = pMnode->rlink;  // find the predecessor
         if (pMnodePred->llink == pMnode) {      // ... in preorder:
            pMnodePred->llink = pMnodeChild;     // a father: set new son
         }else {
            pMnodePred = pMnodePred->llink;
            while (pMnodePred->rlink != pMnode) {
               pMnodePred = pMnodePred->rlink;
            }
            pMnodePred->rlink = pMnodeChild;     // an elder: new brother
         }
         pMnodeChild->rlink = pMnode->rlink;     // copy rlink into child
         free(pMnode);                           // remove current node
         ++nbrof.useless_model;
         --nbrof.model;
         pMnode = pMnodeChild;                   // replaced by the child
         occurTok = (e_Occurence)(               // OR occurence indics
            (unsigned int)occurTok | (unsigned int)pMnode->is.occur
         );
      }
   }else {
      pMnode = alloc_mdl_node();
      switch (mnkindTok) {
      case TARGET_MNKIND:
         pMnode->is.target = TRUE;
         /* fall thru */
      case REGULAR_MNKIND:
         if (pEnodePV->iReferenceCount) {
            /*
            | The GenericIdBlock name must be available
            | to issue ambiguities messages.
            */
            char *pszNewName = pszNameTok;
            pMnode->pGi = (GenericIdBlock *)expnhashElem.add_value(
               &pszNewName, sizeof(GenericIdBlock)
            );
            pMnode->pGi->pszName = pszNewName;
            if (pMnode->pGi->iIdentElem == iIdentElem) {
               pMnode->pGi->isMoreThanOnce = TRUE;
               isCheckAmbiguities = TRUE;
            }else {
               pMnode->pGi->isMoreThanOnce = FALSE;
            }
            pMnode->pGi->iIdentElem = iIdentElem;
         }
         break;
      case PCDATA_MNKIND:
         if (pEnodePV->is.pcdata) {              // already found
            if (pEnodePV->iReferenceCount) {
               isPcdataDuplicate = TRUE;         // can be ambiguous...
               isCheckAmbiguities = TRUE;
            }
         }else {
            pEnodePV->is.pcdata = TRUE;          // see TE, 11.2.3 + 7.6
         }
         pMnode->is.pcdata = TRUE;
         pMnode->pGi = 0;                        // ...means #PCDATA
         break;
      default: /* case DATATAG_MNKIND: */
         if (pEnodePV->is.pcdata) {              // already found
            if (pEnodePV->iReferenceCount) {     // can be ambiguous...
               isPcdataDuplicate = TRUE;
               isCheckAmbiguities = TRUE;
            }
         }else {
            pEnodePV->is.pcdata = TRUE;          // see TE, 11.2.4.4
         }
         pMnode->is.datatag = TRUE;
         pMnode->pOfs = pOfsPV;
         break;
      }
   }
   pMnode->is.occur = occurTok;
   if (occurTok & OMISSIBLE) {
      pMnode->rlink->is.one_omit = TRUE;
   }else {
      pMnode->rlink->is.one_reqd = TRUE;
   }
   ppMnodeLink = &pMnode->rlink;                 // prepare for siblings

   if (!((*ppMnodeLink)->is.conn)) {             // if no left tag
      (*ppMnodeLink)->is.conn = connTok;
   }
   if (!connTok) {                               // if last child,
      pMnode->is.rtag = FALSE;                   // rlink is a thread
   }else {
      pMnode->is.rtag = TRUE;                    // else, check connection
      if ((*ppMnodeLink)->is.conn != connTok) {
         errorh(DIFCONN, pEh, ERH_T_SEV);
      }
   }
}

/*----------------------------------------------------------decode_model_group-+
| Decode a model group (ISO 8879 - p40)                                        |
|                                                                              |
| [127] model_group = GRPO, ts*,content_token,ts*,   , GRPC,occurence_indic?   |
|                           (connector, ts*,content_token,ts* )*               |
|                                                                              |
| At ENTRY: curr_char == GRPO;  (no check, start AFTER: at ts*)                |
|                                                                              |
| 1) Decode "ts*,content_token,ts*"                                            |
|    [128] ts*,token_content,ts* =                                             |
|                                                                              |
|      ts*, + prim_content_token + RNI,"PCDATA"                                |
|           |                    - elem_token = generic_id,occurence_indic?    |
|           |                    + data tag group = DTGO,...                   |
|           + model group = GRPO,...                                           |
|                                                                              |
|    [30] generic_identifier = name                                            |
|                                                                              |
| 2) check GRPC or connector: if none -> error  (but at the zero level)        |
|                                                                              |
| 3) set the "model node"                                                      |
|                                                                              |
| 4) if GRPC, get occurence indic and return, else loop.                       |
|                                                                              |
+-----------------------------------------------------------------------------*/
void ElementDcl::decode_model_group()
{
   int iGrpCnt = 0;
   int curr_floor_level = pPrsstat->floor_level;           // save current pPrsstat->floor_level
   int init_ent_level  = pEh->curr_entity_level(); // remember where we started

   if (++iGrpLvl > GRPLVL) errorh(BIGGRPLVL, pEh, ERH_T_SEV);
   for (;;) {
      pPrsstat->floor_level = pEh->curr_entity_level();    // No Ee allowed
      pPrsstat->skip_ts();                                 // ts after GRPO, connector
      pPrsstat->floor_level = init_ent_level;             // allow EE to occur
      switch (pPrsstat->get_curr_token()) {
      case GRPO_TK:
         new_mdl_link();                         // get a new branch
         decode_model_group();
         break;
      case RNI_TK:
         pSyntax->match_rsvname_or_die(pEh, PCDATA_VRN_ix);
         mnkindTok = PCDATA_MNKIND;              // a PCDATA mnode
         occurTok = REP_OCCUR;                   // see 11.2.4.2 p.41
         break;
      case NOT_FOUND_TK:
         pSyntax->read_general_name(pEh, pszNameTok);
         occurTok = get_occurence_indic();
         mnkindTok = REGULAR_MNKIND;             // a REGULAR mnode
         break;
      case DTGO_TK:
         new_mdl_link();                         // get a new branch
         decode_datatag_group();
         break;
      default:
         errorh(INVDLM, pEh, ERH_T_SEV);
         break;
      }
      if (++iGrpCnt > GRPCNT) errorh(BIGGRPCNT, pEh, ERH_T_SEV);
      int curr_ent_level = pEh->curr_entity_level();
      pPrsstat->skip_ts();                                 // opt ts after content
      switch (pPrsstat->get_curr_token()) {
      case GRPC_TK:
         connTok = NULL_CONN;
         break;
      case AND_TK:
         connTok = AND_CONN;
         break;
      case SEQ_TK:
         connTok = SEQ_CONN;
         break;
      case OR_TK:
         connTok = OR_CONN;
         break;
      default:
         errorh(INVCONN, pEh, ERH_T_SEV);
         break;
      }
      set_mnode();                               // set a new node
      if (!connTok) {                            // GRPC (no more connector)
         occurTok = get_occurence_indic();
         mnkindTok = NULL_MNKIND;                // to end current structure
         --iGrpLvl;                              // bump count level
         if (pEh->curr_entity_level() != init_ent_level) {
            errorh(INVHNEST, pEh, ERH_E_SEV);
         }
         pPrsstat->floor_level = curr_floor_level; // restore current floor
         return;                                 // and return
      }else {
         if (pEh->curr_entity_level() > curr_ent_level) {
            errorh( INVDLM,  pEh, ERH_E_SEV);
         }
      }
   }
}

/*--------------------------------------------------------decode_datatag_group-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void ElementDcl::decode_datatag_group()
{
   boolean isDataTemplateGroup;
   boolean isAtPadding;
   unsigned int *aOfs, *pOfs;
   unsigned int len;
   unsigned int iGrpCnt = 0;

   if (iGrpLvl >= GRPLVL) errorh(BIGGRPLVL, pEh, ERH_T_SEV);
   ++nbrof.datatag;
   /*-----------------------+
   |Decode gi of the target |
   +-----------------------*/
   pPrsstat->skip_ts();
   pSyntax->read_general_name(pEh, pszNameTok);  // Will be used later
   pPrsstat->skip_ts();
   if (pPrsstat->get_curr_token() != SEQ_TK) errorh(INVCONN, pEh, ERH_T_SEV);

   /*------------------------------------------+
   | Decode the pattern(s) and template if any |
   +------------------------------------------*/
   pPrsstat->skip_ts();
   if (pPrsstat->get_curr_token() == GRPO_TK) {
      isDataTemplateGroup = TRUE;
      pPrsstat->skip_ts();
   }else {
      isDataTemplateGroup = FALSE;
   }
   isAtPadding = FALSE;
   pOfs = aOfs = (unsigned int *)malloc((GRPCNT+2) * sizeof(unsigned int));
   for (;;) {
      if (!(len = pPrsstat->read_parameter_literal(pchLitBuffer, TRUE))) {
         errorh(INVEMPTY, pEh, ERH_E_SEV);
      }else if ((++iGrpCnt > GRPCNT) && (!isAtPadding)) {
         errorh(BIGGRPCNT, pEh, ERH_T_SEV);
         /* --iGrpCnt; */
      }else {
         *(pchLitBuffer + len++) = '\0';   /* see last paragraph of 11.2.4.4 */
         *++pOfs = sk.park(pchLitBuffer, len);
         nbrof.datatag_text += len + NORMSEP;  // see odtattl
      }
      pPrsstat->skip_ts();
      if (!isAtPadding) {
         if (isDataTemplateGroup) {
            switch (pPrsstat->get_curr_token()) {
            case OR_TK:
               pPrsstat->skip_ts();
               continue;                 // loop for next template
            case GRPC_TK:
               pPrsstat->skip_ts();
               break;
            default:
               errorh(INVCONN, pEh, ERH_T_SEV);
               break;
            }
         }
         if (pPrsstat->get_curr_token() == SEQ_TK) { // data tag padding template
            isAtPadding = TRUE;
            pPrsstat->skip_ts();
            continue;                     // loop for padding
         }
         *++pOfs = 0;                     // No data tag padding
         ++iGrpCnt;
      }
      if (pPrsstat->get_curr_token() != DTGC_TK) errorh(NODTGC, pEh, ERH_T_SEV);
      break;
   }
   /*------------------------------------+
   | Build the model                     |
   | pszNameTok has the gi of the target |
   +------------------------------------*/
   occurTok = NULL_OCCUR;
   connTok = SEQ_CONN;
   if ((iGrpCnt <= 1) ||                   // nothing usefull was set up
      (!pSgmldcl->features.isDatatag)) {   // or DATATAG feature not enabled
      mnkindTok = REGULAR_MNKIND;          // Regular tag
      set_mnode();                         // set its node
      mnkindTok = PCDATA_MNKIND;           // default to PCDATA
      free(aOfs);                          // release area
   }else {
      mnkindTok = TARGET_MNKIND;           // TARGET tag
      set_mnode();                         // set target node
      nbrof.dtp += iGrpCnt+1;              // create DataTagPattern
      mnkindTok = DATATAG_MNKIND;
      aOfs = (unsigned int *)realloc(      // shrink
         aOfs, (iGrpCnt+1) * sizeof(unsigned int)
      );
      aOfs[0] = *pOfs;                     // padding in first entry
      *pOfs = 0;                           // zero last entry
      pOfsPV = aOfs + 1;                   // to sort only the template
      Sort::quick(iGrpCnt-1, compare_template, exchange_template, this);
      pOfsPV = aOfs;                       // point the padding
   }
   occurTok = REP_OCCUR;                   // see 11.2.4.2 p 41
   connTok = NULL_CONN;
   set_mnode();                            // set a new node

   occurTok = get_occurence_indic();
   mnkindTok = NULL_MNKIND;                // to end current structure
}

/*---------------------------------------------------------get_occurence_indic-+
| [132] occurence indicator = opt | plus | rep                                 |
+-----------------------------------------------------------------------------*/
e_Occurence ElementDcl::get_occurence_indic()
{
   switch (pPrsstat->recognize_delimiter()) {
   case OPT_TK:
      return OPT_OCCUR;     // 0 or 1
   case REP_TK:
      return REP_OCCUR;     // 0 or more
   case PLUS_TK:
      return PLUS_OCCUR;    // 1 or more
   default:
      pEh->unread(pPrsstat->get_curr_count());
      return NULL_OCCUR;    // 1
   }
}

/*------------------------------------------------------------template_compare-+
| Compare function to sort datatag template                                    |
| The longest template must be the first one in the list.                      |
+-----------------------------------------------------------------------------*/
int ElementDcl::template_compare(int i1, int i2)
{
   return strcmp(
      sk.point(*(pOfsPV + i2)), sk.point(*(pOfsPV + i1))
   );
}
/*-----------------------------------------------------------template_exchange-+
| Exchange function to sort datatag template                                   |
+-----------------------------------------------------------------------------*/
void ElementDcl::template_exchange(int i1, int i2)
{
   unsigned int ofs = *(pOfsPV + i1);

   *(pOfsPV + i1) = *(pOfsPV + i2);
   *(pOfsPV + i2) = ofs;
}

/*STATIC--------------------------------------------------------------undo_mdl-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void ElementDcl::undo_mdl(ModelNode *pMnode)
{
   --nbrof.model;
   if (pMnode->is.rtag) undo_mdl(pMnode->rlink);
   if (pMnode->is.conn) undo_mdl(pMnode->llink);
   free(pMnode);
}


/*-----------------------------------------------------------------ElementNode-+
| Default Constructor: O O ANY                                                 |
+-----------------------------------------------------------------------------*/
ElementNode::ElementNode()
{
   pMnode = 0;
   pChxMinus = 0;
   pChxPlus = 0;
   iReferenceCount = 1;
}

/*---------------------------------------------------------------------pile_up-+
| Build "stringized" model structure (ContentToken, DataTagPattern)            |
+-----------------------------------------------------------------------------*/
void ModelNode::pile_up(
   ContentToken*& pCnttokMP,
   DataTagPattern*& pDtpMP, const DataTagPattern* const aDtpTable,
   unsigned short& usRx
) const
{
   ContentToken *pCnttok = pCnttokMP++;
   pCnttok->conn = is.conn;
   pCnttok->occur = is.occur;
   pCnttok->isPcdata = is.pcdata;
   pCnttok->isDatatag = is.datatag;
   pCnttok->isTarget = is.target;
   if (is.rtag) {
      rlink->pile_up(pCnttokMP, pDtpMP, aDtpTable, usRx);
      free(rlink);
   }else {
      usRx = 0;
   }
   if (is.conn) {                          // if a child exists
      pCnttok->mdl.rx = (unsigned short)(pCnttokMP - pCnttok);
      llink->pile_up(pCnttokMP, pDtpMP, aDtpTable, pCnttok->mdl.num);
      free(llink);
   }else {
      if (is.pcdata) {                     // #PCDATA
         pCnttok->ixElem = 0;              // to keep ODT consistent
      }else if (is.datatag) {              // DATATAG
         unsigned int *pOfsCurr = pOfs;
         pCnttok->ixDtp = diffptr(pDtpMP, aDtpTable);
         do (pDtpMP++)->ofs = *pOfsCurr; while (*++pOfsCurr);
         (pDtpMP++)->ofs = 0;
         free(pOfs);
      }else {                              // TARGET or REGULAR
         pCnttok->ixElem = pGi->ixElem;
      }
   }
   ++usRx;
}

/*---------------------------------------------------------------------pile_up-+
|Build Exception array for an element type.  Last entry is -1.                 |
|The ExceptionChain structure is freed.                                        |
+-----------------------------------------------------------------------------*/
//***>NOTE: will do in a more OO fashion when times permits
int ExceptionChain::pile_up(Exception*& pExMP, const Exception* const aExTable)
{
   int ixEx = diffptr(pExMP, aExTable);
   ExceptionChain *pChx = this;

   do {
      (pExMP++)->ixElem = pChx->pGi->ixElem;
      ExceptionChain* pChxSave = pChx; // avoid access to a freed area
      pChx = pChx->next;
      free(pChxSave);
   }while (pChx);
   (pExMP++)->ixElem = -1;
   return ixEx;
}

/******************************************************************************/
