/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// elemdcl.h
///  Declarations for elemdcl
//   Pierre G. Richard 87/08/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _ELEMDCL_H
#define _ELEMDCL_H 1

enum e_ModelNodeKind {          // MNKIND, mnode_kind
   NULL_MNKIND = 0,
   TARGET_MNKIND,
   DATATAG_MNKIND,
   PCDATA_MNKIND,
   REGULAR_MNKIND
};

/*-----------------------------------------------------------------------------+
| Element type declaration                                                     |
+-----------------------------------------------------------------------------*/
class ElementDcl :
   private virtual PrologDcl           // IS_A?
{
public:
   void parse();
   void new_declared_gi_or_stem();
   void new_exception_gi();
   int  template_compare(int i1, int i2);
   void template_exchange(int i1, int i2);

private:
   ModelNode **ppMnodeLink;
   ExceptionChain **ppChx;
   ElementNode *pEnodePV;
   int iGrpGtCnt;
   int iGrpLvl;
   boolean isCheckAmbiguities;
   boolean  isPcdataDuplicate;
   char *pszNamePV;

   char pszNameTok[NAMELEN_MAX+1];
   e_Occurence occurTok;
   e_Connector connTok;
   e_ModelNodeKind mnkindTok;

   unsigned int  *pOfsPV;    // for datatag pattern / template(s)

   void new_mdl_link();
   ModelNode *alloc_mdl_node();
   void set_mnode();
   void decode_model_group();
   void decode_datatag_group();
   e_Occurence get_occurence_indic();
   void undo_mdl(ModelNode *pMnode);
};

/*-----------------------------------------------------------------------------+
| ModelNode (specific class during compilation)                                |
+-----------------------------------------------------------------------------*/
struct ModelNode {              // MNODE: Model Node
   struct Flags {
      boolean rtag        :1;   // tell rlink is right_link / thread
      e_Connector conn    :2;   // also tell llink is left_link / elem
      e_Occurence occur   :2;   // is_omittable, is_repeatable
      boolean target      :1;   // is a target tag
      boolean pcdata      :1;   // is pcdata
      boolean datatag     :1;   // is a datatag
      boolean one_reqd    :1;   // if one token in the subgroup is required
      boolean one_omit    :1;   // if one token in the subgroup is omittable
      boolean ambig_found :1;   // a GenericIdBlock mnode and ambiguity found
   };
   Flags is;
   ModelNode *rlink;            // right link if rtag, or thread
   union {
      ModelNode *llink;         // left link if ltag, or
      GenericIdBlock *pGi;      // *NOT* a thread / 0 for #PCDATA
      unsigned int *pOfs;       // Offset of datatag pattern / template(s)
   };
   void pile_up(
      ContentToken*& pCnttokMP,
      DataTagPattern*& pDtpMP, const DataTagPattern* const aDtpTable,
      unsigned short& usRx
   ) const;
};

/*-----------------------------------------------------------------------------+
| ElementNode (specific class during compilation)                              |
+-----------------------------------------------------------------------------*/
struct ElementNode {            // ENODE: Element Node
   ElementType::Flags is;
   int iReferenceCount;         // count how many times this node referenced
   union {
      ModelNode *pMnode;        // model node, and later (if solved) ....
      int ixCnttok;             // ...index in ContentToken array of the model
   };
   union {                      // Minus exceptions:
      ExceptionChain *pChxMinus; // pointer, and later (when solved)...
      int ixExMinus;            // ...index in the Exception array (-1 if none)
   };
   union {                      // Plus exceptions:
      ExceptionChain *pChxPlus; // pointer, and later (when solved)...
      int ixExPlus;             // ...index in the Exception array (-1 if none)
   };

   ElementNode();               // Default constructor (O O ANY)
};

/*-----------------------------------------------------------------------------+
| ExceptionChain (Specific class during compilation)                           |
+-----------------------------------------------------------------------------*/
struct ExceptionChain {         // CHX: exception chain
   GenericIdBlock *pGi;
   ExceptionChain *next;

   int pile_up(Exception*& pExMP, const Exception* const aExTable);
};

#endif /* _ELEMDCL_H **********************************************************/
