/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// element.c
///  Element class
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "skmgr.h"
#include "element.h"
#include "attrib.h"
#include "errorh.h"

/*----------------------------------------------------------ElementType::Flags-+
| Default Constructor for Flags                                                |
+-----------------------------------------------------------------------------*/
ElementType::Flags::Flags()
{
   memset(this, 0, sizeof (Flags));
   any = TRUE;
}

/*-----------------------------------------------------------------ElementType-+
| Constructor for Element Zero (in an OpenElementStack)                        |
+-----------------------------------------------------------------------------*/
ElementType::ElementType(ContentToken *pCnttokZero)
{
   memset(this, 0, sizeof *this);
   pCnttok = pCnttokZero;
}

/*--------------------------------------------------------------attribute_list-+
| Build the attribute definition list                                          |
+-----------------------------------------------------------------------------*/
void ElementType::attribute_list(AttributeList &attlist) const
{
   attlist.build(aAttrepDefList, iAttCount);
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries ElementType objects piled_up in an array                |
+-----------------------------------------------------------------------------*/
void ElementType::relocate_pile(
   unsigned int iMaxEntries,
   ContentToken* aCnttokTable,
   Exception* aExTable,
   AttributeRep* aAttrepTable,
   ShortrefMap* aSrmTable,
   RankList* aRklTable,
   char* strbuf
)
{
   ElementType *pElem = this;

   for (int i=0; i < iMaxEntries; ++i, ++pElem) {
      pElem->pszName = strbuf + pElem->ofsName;
      if (pElem->ixCnttok == -1) {
         pElem->pCnttok = 0;
      }else {
         pElem->pCnttok = aCnttokTable + pElem->ixCnttok;
      }
      if (pElem->ixExMinus == -1) {
         pElem->pExMinus = 0;
      }else {
         pElem->pExMinus = aExTable + pElem->ixExMinus;
      }
      if (pElem->ixExPlus == -1) {
         pElem->pExPlus = 0;
      }else {
         pElem->pExPlus = aExTable + pElem->ixExPlus;
      }
      if (pElem->ixAttrep == -1) {
         pElem->aAttrepDefList = 0;
      }else {
         pElem->aAttrepDefList = aAttrepTable + pElem->ixAttrep;
      }
      if (pElem->ixSrm == -1) {
         pElem->pSrm = 0;
      }else {
         pElem->pSrm = aSrmTable + pElem->ixSrm;
      }
      if (pElem->ixRkl == -1) {
         pElem->pRkl = 0;
      }else {
         pElem->pRkl = aRklTable + pElem->ixRkl;
      }
   }
}

/*---------------------------------------------------------------------pile_up-+
| This function is called while traversing the hash table of stems.            |
| At each RKN describing a stem, fill up the corresponding STS entry -- the    |
| next one in the STS array.  Remember in sts_ix the position of the stem in   |
| the STS array.  Then, traverse the RKN stem link chain of ranked element     |
| with the same stem and store sts_ix in the corresponding RKL entry -- the    |
| next one of the RKL array.                                                   |
| Reuse the room left by the now useless stem-chain pointer in the RKN         |
| to store the position of the corresponding RKL of the RKL array.             |
+-----------------------------------------------------------------------------*/
boolean RankNode::pile_up(
   StringKeeper &sk, const char * pszName,
   StemValue*& pStsMP, const StemValue* const aStsTable,
   RankList*& pRklMP, const RankList* const aRklTable
)
{
   int iCount = stem_cnt;
   unsigned int ixSts = diffptr(pStsMP, aStsTable);

   pStsMP->iCount = iCount;
   pStsMP->iSuffix = -1;
   pStsMP->ofsName = sk.park(pszName, 1+strlen(pszName));
   ++pStsMP;
   {
      unsigned int ixRkl = diffptr(pRklMP, aRklTable);
      RankNode *pRknStemSave = pRknStem;
      RankNode *pRknStemCurr;

      for (int i=0; i < iCount; ++i) {
         (pRklMP++)->ixSts = ixSts;
         pRknStemCurr = pRknStemSave;
         pRknStemSave = pRknStemCurr->pRknStem;   // save, because...
         pRknStemCurr->ixRkl = ixRkl++;           // this field is overwritten
      }
   }
   return TRUE;
}

/*-----------------------------------------------------------------solve_index-+
| solve ixRklGrp for a same enode                                              |
+-----------------------------------------------------------------------------*/
void RankList::solve_index(const RankNode *pRkn)
{
   unsigned int ix = pRkn->ixRkl;
   unsigned int ix0 = ix;
   do {
      pRkn = pRkn->pRknGroup;
      this[ix].ixRklGrp = pRkn->ixRkl;
      ix = pRkn->ixRkl;
   }while (ix != ix0);
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries RankList objects piled_up in an array                   |
+-----------------------------------------------------------------------------*/
void RankList::relocate_pile(unsigned int iMaxEntries, StemValue* aStsTable)
{
   RankList *pRkl = this;

   for (int i=0; i < iMaxEntries; ++i, ++pRkl) {
      pRkl->pSts = aStsTable + pRkl->ixSts;
      pRkl->pRklGrp = this + pRkl->ixRklGrp;
   }
}

/*----------------------------------------------------------------set_suffices-+
| Set current rank suffix after a full generic identifier has been validated.  |
| Set the suffix in the RankList of the ElementType, and broadcast it in       |
| all stems concerned by 7.8.1.1                                               |
|                                                                              |
| Note: before to call this function, caller must make sure the rank feature   |
|       is enabled.                                                            |
+-----------------------------------------------------------------------------*/
//***>NOTE: this violates constness of DTDdcl by changing the STEMs!!!
//***>NOTE: In fact, this structure has to be in ParseState (or copied to?)
void RankList::set_suffices(const char * pszName)
{
   int iSuffix = atoi(pszName + strlen(pSts->pszName));
   pSts->iSuffix = iSuffix;                // set the suffix

   RankList *pRklCurr = this;
   for (int i=0; i < pSts->iCount; ++i, ++pRklCurr) {
      for (
         RankList* pRklNext = pRklCurr->pRklGrp;
         pRklNext != pRklCurr;
         pRklNext = pRklNext->pRklGrp
      ) {
         pRklNext->pSts->iSuffix = iSuffix;
      }
   }
}

/*------------------------------------------------------------------first_stem-+
| Return an index to the first stem                                            |
+-----------------------------------------------------------------------------*/
int RankList::first_stem(int ixRklCurr)
{
   RankList *pRklCurr = this + ixRklCurr;
   const int ixStsCurr = pRklCurr->ixSts;

   while (ixRklCurr && (ixStsCurr == (--pRklCurr)->ixSts)) --ixRklCurr;
   return ixRklCurr;
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries DataTagPattern objects piled_up in an array             |
+-----------------------------------------------------------------------------*/
void DataTagPattern::relocate_pile(
      unsigned int iMaxEntries, char * strbuf
)
{
   DataTagPattern *pDtp = this;

   for (int i=0; i < iMaxEntries; ++i, ++pDtp) {
      pDtp->pszTemplate = strbuf + pDtp->ofs;
   }
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries StemValue objects piled_up in an array                  |
+-----------------------------------------------------------------------------*/
void StemValue::relocate_pile(
   unsigned int iMaxEntries, char * strbuf
)
{
   StemValue *pSts = this;

   for (int i=0; i < iMaxEntries; ++i, ++pSts) {
      pSts->pszName = strbuf + pSts->ofsName;
   }
}

/*----------------------------------------------------------------ContentToken-+
| Null constructor: dummy cnttokZero, in ELementZero of an OpenElementStack    |
+-----------------------------------------------------------------------------*/
ContentToken::ContentToken()
{
   memset(this, 0, sizeof *this);
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries ContentToken objects piled_up in an array               |
+-----------------------------------------------------------------------------*/
void ContentToken::relocate_pile(
   unsigned int iMaxEntries,
   ElementType *aElemTable,
   DataTagPattern *aDtpTable
)
{
   ContentToken *pCnttok = this;

   for (int i=0; i < iMaxEntries; ++i, ++pCnttok) {
      if (pCnttok->conn == NULL_CONN) {
         if (pCnttok->isPcdata) {            // PCDATA
            pCnttok->pElem = 0;
         }else if (pCnttok->isDatatag) {     // DATATAG
            pCnttok->pDtp = aDtpTable + pCnttok->ixDtp;
         }else {                             // TARGET or REGULAR
            pCnttok->pElem = aElemTable + pCnttok->ixElem;
         }
      }
   }
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries Exception objects piled_up in an array                  |
+-----------------------------------------------------------------------------*/
void Exception::relocate_pile(
   unsigned int iMaxEntries,
   ElementType *aElemTable
)
{
   Exception *pEx = this;

   for (int i=0; i < iMaxEntries; ++i, ++pEx) {
      if (pEx->ixElem == -1) {
         pEx->pElem = 0;
      }else {
         pEx->pElem = aElemTable + pEx->ixElem;
      }
   }
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries ShortrefMaps objects piled_up in an array               |
+-----------------------------------------------------------------------------*/
void ShortrefMap::relocate_pile(
   unsigned int iMaxEntries,
   unsigned int iMaxShortrefs,
   EntityDef *aEntGenTable,
   char *strbuf
)
{
   ShortrefMap *pSrm = this;

   for (int i=iMaxEntries; i-- > 0; i -= iMaxShortrefs) {
      pSrm->pszName = strbuf + pSrm->ofsName;
      ++pSrm;
      for (int j=iMaxShortrefs; j--; ++pSrm) {
         if (pSrm->ixEnt == -1) {
            pSrm->pEnt = 0;
         }else {
            pSrm->pEnt = aEntGenTable + pSrm->ixEnt;
         }
      }
   }
}
/*============================================================================*/
