/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// element.h
///  Declarations for element
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _ELEMENT_H
#define _ELEMENT_H 1
/*
|  This H file defines the structures associated to all Markup Declaration
|  directly or indirectly related to an element definition:
|  i.e <!ELEMENT, but also <!ATTLIST, <!SHORTREF, ...
*/
//******************************************************************************

#include "element.eh"   // enums

class ExpandedHash;
class StringKeeper;
class ElementType;
class ContentToken;
class AttributeRep;
class AttributeList;
class EntityDef;
class NotationType;
class Exception;
class StemValue;
class RankList;
class DataTagPattern;
class ShortrefMap;
class Component;

/*-----------------------------------------------------------------------------+
| ElementType                                                                  |
|                                                                              |
| Note: An element that has a CONREF attribute can *not* have a dcl-content    |
|       of EMPTY.  However, when there is an explicit content reference,       |
|       the "empty" bit is turned ON for the time of the notification.  This   |
|       serves as a reminder the element's content is EMPTY.                   |
+-----------------------------------------------------------------------------*/
class ElementType {             // ELEM: Element Type Definition
   friend struct GenericIdBlock; // pile_up()
   friend class InclusionListIterator;
   friend class ExclusionListIterator;
   friend class OpenElementStack;
public:
   struct Flags {               // IS:
      Flags();
      boolean reqd_open_tag :1; // Minimization
      boolean reqd_end_tag  :1; // Minimization
      boolean any           :1; // No ContentToken: model content: ANY
      boolean empty         :1; // No ContentToken: dcl content EMPTY
      e_Dcdata dcdata       :2; // No ContentToken: dcl content DCDATA
      boolean pcdata        :1; // Mixed content and data-char allowed
      boolean required      :1; // An attribute in the list is REQUIRED
      boolean current       :1; // An attribute in the list is CURRENT
      boolean conref        :1; // An attribute in the list is CONREF
      boolean notation      :1; // An attribute in the list is ELEMENT
      boolean entity        :1; // An attribute in the list is ENTITY(IES)
      boolean solved        :1; // Internal use
   };
   void relocate_pile(          // real constructor
      unsigned int iMaxEntries,
      ContentToken* aCnttokTable,
      Exception* aExTable,
      AttributeRep* aAttrepTable,
      ShortrefMap* aSrmTable,
      RankList* aRklTable,
      char* strbuf
   );
   void set_rank_suffices() const;

   /*
   |  API functions
   */
   const char * name() const;
   void attribute_list(AttributeList &attlist) const;
   boolean is_reqd_open_tag() const;
   boolean is_reqd_end_tag() const;
   boolean is_any() const;
   boolean is_empty() const;
   boolean is_pcdata() const;
   e_Dcdata declared_data_content() const;
   ContentToken * content_token() const;
   ShortrefMap * shortref_map() const;

private:
   union {                      // Start with element name (hashing)
      char *pszName;            // Element name
      unsigned int ofsName;     // (offset, when unloaded)
   };
   Flags is;
   union {                      // Null if: ANY, EMPTY, DCDATA (RCDATA | CDATA)
      ContentToken *pCnttok;    // Pointer to the model header
      int ixCnttok;             // (MDL array index of the model, when unloaded)
   };
   union {
      AttributeRep *aAttrepDefList; // Attribute definition list or 0
      int ixAttrep;             // (ATTREP array index when unloaded)
   };
   unsigned int iAttCount;      // number of defined attributes, including LINK
   union {
      Exception  *pExMinus;     // Pointer to an EX array of Minus Exceptions
      int  ixExMinus;           // (EX array index when unloaded)
   };
   union {
      Exception  *pExPlus;      // Pointer to an EX array of Plus Exceptions
      int  ixExPlus;            // (EX array index when unloaded)
   };
   union {
      ShortrefMap *pSrm;        // Pointer to a map to activate
      int  ixSrm;               // (SRM array index when unloaded)
   };
   union {
      RankList *pRkl;           // when loaded, points an RKL
      int  ixRkl;               // (RKL array index when unloaded)
   };
   ElementType(ContentToken *pCnttokZero); // Deflt constructor for ElementZero
};

/*-----------------------------------------------------------------------------+
| Short Reference Mapping - Shortref maps are described by an array of         |
| (total count of map) * (count of shortrefs + 1) `ShortRefMap' structures     |
| - the first structure is the stringized hash header to find a mapname,       |
| - the count_of_shortref others point to the ENT related (or 0 if none)       |
|   (pSrm + n points ENT related to the SR #n, 0 < n < count_of_shortref + 1)  |
+-----------------------------------------------------------------------------*/
class ShortrefMap {            // SRM  element of a (sr_max+1) array of:
   friend class ShortrefNode;
public:
   void relocate_pile(         // real constructor
      unsigned int iMaxEntries,
      unsigned int iMaxSubEntries,
      EntityDef *aEntGenTable,
      char *strbuf
   );
   /*
   |  API functions
   */
   // Get the entity mapped to SR # `iNoSr'  ** no check **
   EntityDef * entity_def(int iNoSr) const;

private:
   union {                     // element of a (sr_max+1) array of:
                               // First entry:
      char *pszName;           // Map name (hashing)
      unsigned int ofsName;    // (offset, when unloaded)
                               // Succeeding count_of_shortref entries:
      EntityDef *pEnt;         // Points each associated entity
      int  ixEnt;              // (ENT array index, when unloaded)
      char *pszNameEnt;        // (while building it: entity name pointer)
   };
   ShortrefMap();              // undefined
   ~ShortrefMap();             // undefined
};

/*-- INLINES --*/
inline EntityDef * ShortrefMap::entity_def(int iNoSr) const {
   return (this+iNoSr)->pEnt;
}

/*-----------------------------------------------------------------------------+
| Rank double link list                                                        |
+-----------------------------------------------------------------------------*/
struct RankNode {               // RKN: Rank Nodes
   union {
      RankNode *pRknStem;       // Same stem link
      unsigned int ixRkl;       // temporary: index in RankList array
   };
   union {
      unsigned int stem_cnt;    // For the STEM, count RankList with same stem
      RankNode *pRknGroup;      // Same group link
   };

   boolean pile_up(
      StringKeeper &sk, const char * pszName,
      StemValue*& pStsMP, const StemValue* const aStsTable,
      RankList*& pRklMP, const RankList* const aRklTable
   );
};

class StemValue {            // STS: Stem value
   friend RankList;
   friend boolean RankNode::pile_up(
      StringKeeper &sk, const char * pszName,
      StemValue*& pStsMP, const StemValue* const aStsTable,
      RankList*& pRklMP, const RankList* const aRklTable
   );
public:
   void relocate_pile(       // real constructor
      unsigned int iMaxEntries, char * strbuf
   );
   int suffix();
private:
   union {                   // Start with rank name (hashing)
      char *pszName;         // Rank name
      unsigned int ofsName;  // (offset, when unloaded)
   };
   int iSuffix;              // current suffix value (-1 if not set)
   int iCount;               // count of elements having this stem
   StemValue();              // undefined
   ~StemValue();             // undefined
};
/*-- INLINES --*/
inline int StemValue::suffix() { return iSuffix; }

class RankList {             // RKL: Rank List
   friend boolean RankNode::pile_up(
      StringKeeper &sk, const char * pszName,
      StemValue*& pStsMP, const StemValue* const aStsTable,
      RankList*& pRklMP, const RankList* const aRklTable
   );
public:
   void solve_index(const RankNode *pRkn);
   void relocate_pile(       // real constructor
      unsigned int iMaxEntries, StemValue* aStsTable
   );
   int first_stem(int ixRklCurr);
   void set_suffices(const char * pszName);
private:
   union {
      StemValue *pSts;       // pointer to StemValue
      int ixSts;             // (STS array index, when unloaded)
   };
   union {
      RankList *pRklGrp;     // points other ranked of which the elem is a mbr
      int ixRklGrp;          // (RKL array index, when unloaded)
   };
   RankList();               // undefined
   ~RankList();              // undefined
};

/*-----------------------------------------------------------------------------+
| Data Tag Pattern                                                             |
+-----------------------------------------------------------------------------*/
class DataTagPattern {       // DTP: array is made of 3 entries at least:
   friend struct ModelNode;  // pile_up(), but ModelNode might be not included
public:
   void relocate_pile(       // real constructor
      unsigned int iMaxEntries, char * strbuf
   );
private:
   union {
      unsigned int ofs;      // 1st is the padding (or Null Pointer if none)
      char *pszTemplate;     // 2nd is the first template
      void *v_p;             // last is a Null Pointer
   };
   DataTagPattern();         // undefined
   ~DataTagPattern();        // undefined
};

/*-----------------------------------------------------------------------------+
| Exception                                                                    |
+-----------------------------------------------------------------------------*/
class Exception {              // EX:
   friend class ExceptionListIterator;
   friend class ExceptionChain;  // pile_up()
public:
   void relocate_pile(         // real constructor
      unsigned int iMaxEntries,
      ElementType *aElemTable
   );
private:
   Exception();                // Default constructor: undefined
   ~Exception();               // Destructor: undefined
   union {
      ElementType *pElem;      // pointer to exception elem
      int ixElem;              // (ELEM array index when unloaded)
   };
};

class ExceptionListIterator {
public:
   const ElementType * next();
protected:
   Exception *pExCurr;
   enum { TOP, MIDDLE, BOTTOM } where;
   ExceptionListIterator();
};

// To list the inclusions/exclusions of a given element, one must
// construct an InclusionListIterator/ExclusionList object.
// The next() method applied to this object cycles in the list,
// returning each element -- one at a time -- until a return of
// (ElementType *)0 indicates "no more!".
class InclusionListIterator : public ExceptionListIterator {
public:
   InclusionListIterator(const ElementType *pElem);
};
class ExclusionListIterator : public ExceptionListIterator {
public:
   ExclusionListIterator(const ElementType *pElem);
};

/*-- INLINES --*/
inline ExceptionListIterator::ExceptionListIterator() { where = TOP; }
inline const ElementType * ExceptionListIterator::next() {
   ElementType *pElem;
   switch (where) {
   case TOP:
      if (pExCurr == 0) return 0;
      where = MIDDLE;
      return pExCurr->pElem;
   case MIDDLE:
      if ((pElem = (++pExCurr)->pElem) == 0) where = BOTTOM;
      return pElem;
   default: /* case BOTTOM */
      return 0;
   }
}
inline InclusionListIterator::InclusionListIterator(const ElementType *pElem) {
   pExCurr = pElem->pExPlus;
}
inline ExclusionListIterator::ExclusionListIterator(const ElementType *pElem) {
   pExCurr = pElem->pExMinus;
}

/*-----------------------------------------------------------------------------+
| Content Token (compiled form)                                                |
|                                                                              |
| If conn == NULL_CONN                                                         |
|    this content token is a primitive content token (1)                       |
|    If isDatatag is FALSE (1-a)                                               |
|       If pElem is 0                                                          |
|          this primitive content token is #PCDATA                             |
|       Else                                                                   |
|          this primitive content token is an element token                    |
|    Else (isDatatag TRUE) (1-b)                                               |
|       this primitive content token is a data tag group                       |
| Else (conn is AND_CONN, SEQ_CONN, or OR_CONN)                                |
|    this content token is a model group (2)                                   |
+-----------------------------------------------------------------------------*/
class ContentToken {           // CNTTOK:
   friend class OpenElementStack;
   friend struct ModelNode;    // pile_up()
   friend class Modele;
public:
   void relocate_pile(         // real constructor
      unsigned int iMaxEntries,
      ElementType *aElemTable,
      DataTagPattern *aDtpTable
   );
private:
   e_Connector conn     :2;    // if NULL_CONN, union contains 1); else 2)
   e_Occurence occur    :2;    // omissible,  repeatable
   boolean isTarget     :1;    // Elem token is target (next brother is Datatag)
   boolean isDatatag    :1;    // Datatag (1-b, pDtp)
   boolean isPcdata     :1;    // TRUE for #PCDATA, as well as for data_tag
   unsigned int unused  :9;    // (Future use)
   union {                     // 1) primitive content token
                               // 1-a) NULL_CONN: elem token or PCDATA
      ElementType *pElem;      // pointer to element, or PCDATA if 0
      int ixElem;              // (ELEM array index when unloaded, -1: PCDATA)

                               // 1-b) NULL_CONN + isDatatag
      DataTagPattern *pDtp;    // when loaded: pointer to the data-tag pattern
      int ixDtp;               // (DTP array index when unloaded)

      struct {                 // 2) model group
         unsigned short num;   // number of tokens in this group (*)
         unsigned short rx;    // relative index of 1st token of subgroup (**)
      }mdl;
   };
   ContentToken();             // Default constructor (for cnttokZero)
};
/*
|  (*)  - At most, GRPCNT
|  (**) - i.e. how many MDL to jump from here to the 1st subgroup token.
|         At most, GRPGTCNT
*/

/*-- INLINES --*/
inline const char * ElementType::name() const {
   return pszName;
}
inline boolean ElementType::is_reqd_open_tag() const {
   return is.reqd_open_tag? TRUE : FALSE;
}
inline boolean ElementType::is_reqd_end_tag() const {
   return is.reqd_end_tag? TRUE : FALSE;
}
inline boolean ElementType::is_any() const {
   return is.any? TRUE : FALSE;
}
inline boolean ElementType::is_empty() const {
   return is.empty? TRUE : FALSE;
}
inline boolean ElementType::is_pcdata() const {
   return is.pcdata? TRUE : FALSE;
}
inline e_Dcdata ElementType::declared_data_content() const {
   return is.dcdata;
}
inline ContentToken * ElementType::content_token() const {
   return pCnttok;
}
inline ShortrefMap * ElementType::shortref_map() const {
   return pSrm;
}
inline void ElementType::set_rank_suffices() const {
   if (pRkl) pRkl->set_suffices(pszName);
}
#endif /* _ELEMENT_H **********************************************************/
