/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// enthdlr.h
///  Declarations for enthdlr
//   Pierre G. Richard 87/04/06 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _EM_H
#define _EM_H 1
//******************************************************************************

#include "server.h"     // Server is a private base
#include "syntax.h"     // this class is used by inlines

class SGMLdcl;
class ParseMode;
class Quantity;
class DelimiterString;
class Syntax;
class EntityDef;
class ExternId;

/*-----------------------------------------------------------------------------+
| EntityControlBlock - This structure is maintained by EM, to keep a chain     |
| of opened nested entities                                                    |
+-----------------------------------------------------------------------------*/
enum e_CharKind {                 // koc
   REGULAR_KOC = 0,
   EE_KOC,                        // end of entity character
   RS_KOC,                        // true record start (new line)
   NUMCHAR_KOC                    // data character
};

struct EntityControlBlock {       // YB: entitY Block
   EntityDef *pEnt;               // entity structure
   ApplicationHandle ahlAccess;   // Access AHL
   long lOffset;                  // offset within entity
   int  iRow;                     // RS counts
   char *pszApplInfo;             // Appl. provides for msgs ER_LOC0 and ER_LOC1
   EntityControlBlock *pYbPrev;   // Previous opened entity
   long lVdiff;                   // offset-virtual offset (after replacements)

   // The following is a temporary storage used when an entity is stacked:
   int iForeItemCount;            // number of Fore (== to read) Items stacked
   union {                        // most of the time, only one
      struct {                    // hence, just this simple structure
         char ch;                 // just one char
         char koc;                // associated kind_of_char
      }oneFore;
      struct {                    // in case more than one were to unread,
         char *ach;               // the actual not-yet-read part is stacked
         char *akoc;              // with its associated kind_of_chars
      }manyFore;
   };

   // the following is irrelevant for a memory entity
   int iBufPos;                   // Current active position in buffer
   int iBufMax;                   // Last valid position in buffer
   char *achBuffer;               // Data transfer buffer
};

/*-----------------------------------------------------------------------------+
| EM (EntityDef Manager) class definition                                         |
+-----------------------------------------------------------------------------*/
class EntityHandler :
   private Server                   // Private inheritance for HAS-A
{
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   EntityHandler(const Document& doc, ParseMode *pPrsmode);   // Constructor

   void shadow_syntax(Syntax *pSyntax); //***> This function has to be protected

   int curr_char_val();
   e_CharKind curr_char_kind();
   int curr_entity_level();

   int find(const ExternId* pExid, ApplicationHandle& ahlIdent);

   //***>Note: need ParseMode, Quantity, DelimiterString
   boolean open(EntityDef *pEnt);

   //***>Note: need ParseMode, Syntax
   int read();

   int unread(unsigned int i);
   boolean close();
   void *get_pos();
   void set_pos(void *pXposblk);
   void replace(char new_char, unsigned int count, e_CharKind koc);
   void locate(
      int  iRequestedEntLevel,
      char **ppszName,
      char **ppszApplInfo,
      int  *piRow,
      long *plOffset
   );
   long offset();
   long virtual_offset();

   boolean is_nonsgml();   // is current char a NONSGML?
   boolean is_digit();     // is current char a digit?
   boolean is_funchar();   // is current char a FUNCHAR?
   boolean is_sepchar();   // is current char a SEPCHAR?
   boolean is_REchar();    // is current char a RE?
   boolean is_RSchar();    // is current char a RS?
   boolean is_SPACEchar(); // is current char a SPACE?
   boolean is_blank();     // is current char a Blank (SPACE+SEPCHAR)?
   boolean is_separator(); // is current char an S Separator? ([5])
   boolean is_sspchar();   // is current char a RE, RS, or SPACE?
   boolean is_namestrt();  // can current char starts a name? ([53])
   boolean is_name();      // is current char a name character? ([52])
   boolean is_mindata();   // is current char minimum data? [77,78]
   boolean is_minldat();   // is current char LC/UC Letter+Digit+Special?
   boolean is_fctchar();   // is current char a function character? ([54])
   boolean is_ms_char();   // is current char one of the MSO, MSI, MSS?
   boolean is_msrchar();   // is current char a MSO, or MSS?
   boolean is_msichar();   // is current char a MSI?
   boolean is_msochar();   // is current char a MSO?
   boolean is_msschar();   // is current char a MSS?
   boolean is_cref();      // does current char match the CREF constraint?
   boolean is_Ee();        // did we get the Ee signal?
   boolean is_numchar();   // is it a data character (numeric character ref)
   int     uppercase();    // uppercase current char character

   Server::is_sort_attributes;  // Att list to be sorted (access declaration)

   //***>Note: This reflects the entity manager is also used for Utility Files.
   //***>Note: The implementation is not finished. I plan to use more of the
   //***>Note: standard em_functions.  No time today.

   boolean open(
      ExternId *pExid, ApplicationHandle *pahlAccess,
      boolean isWriteMode
   );
   static boolean read(
      ApplicationHandle ahlAccess, void *pchBuffer, int iLength
   );
   static int read_record(ApplicationHandle ahlAccess, void **ppchBuffer);
   static boolean write(
      ApplicationHandle ahlAccess, const void *pchBuffer, unsigned int iLength
   );
   static boolean write_record(
      ApplicationHandle ahlAccess, const void *pchBuffer, unsigned int iLength
   );
   static boolean close(
      ApplicationHandle ahlAccess
   );

private:
   /*--------------------+
   | Private definitions |
   +--------------------*/
   enum { EE_VALUE = 0 };         // placeholder: curr_char_kind is to use
                                  // also works properly with is_xxx
   enum { BUFSIZE = 0x1000 };     // must be a power of 2 (4096 defined here)
                                  // LITLEN and BSEQLEN must be less or equal

   struct XPOSBLK {               // to set and restore positions
     EntityControlBlock yb;
     // followed by the System position block
   };

   /*------------------+
   | Private variables |
   +------------------*/
   // Shadow pointers
   Syntax          *pSyntax;   // Shadow SGMLdcl mbr, for read
   ParseMode       *pPrsmode;  // Shadow ParseState mbr, for open, read
   DelimiterString *pDlmstr;   // Shadow Syntax mbr, for open
   Quantity        *pQuantity; // Shadow Syntax mbr, for open

   int       iBufPos;                // Current position in circular buffer
   int       iBufHead;               // Head of circular buffer
   char      achBuffer[BUFSIZE];     // Circular Buffer for characters read
   char      akocBuffer[BUFSIZE];    // Circular Buffer for kind_of_chars
   EntityControlBlock *pYbCurr;      // Current EntityDef Control Block
   int       iEntLvl;                // entity level (1==PRIMARY)
   /*------------------+
   | Private functions |
   +------------------*/
   int     diff_head_pos();
   void    store_new_char(int iNewChar, e_CharKind koc);
   int     retrieve_next_char();
};


/*INLINE---------------------------------------------------------curr_char_val-+
| Give the value of the character currently read                               |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::curr_char_val()
{
   return achBuffer[iBufPos];
}

/*INLINE--------------------------------------------------------curr_char_kind-+
| Give the kind of the character currently read                                |
+-----------------------------------------------------------------------------*/
inline e_CharKind EntityHandler::curr_char_kind()
{
   return (e_CharKind)akocBuffer[iBufPos];
}

/*INLINE-----------------------------------------------------curr_entity_level-+
| Give the current entity level reached                                        |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::curr_entity_level()
{
   return iEntLvl;
}

/*INLINE--------------------------------------------------------------is_xxxxx-+
| Character classification functions for the current character read            |
+-----------------------------------------------------------------------------*/
inline boolean EntityHandler::is_nonsgml() {
   return pSyntax->is_nonsgml(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_digit() {
   return pSyntax->is_digit(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_funchar() {
   return pSyntax->is_funchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_sepchar() {
   return pSyntax->is_sepchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_REchar() {
   return pSyntax->is_REchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_RSchar() {
   return pSyntax->is_RSchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_SPACEchar() {
   return pSyntax->is_SPACEchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_blank() {
   return pSyntax->is_blank(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_separator() {
   return pSyntax->is_separator(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_sspchar() {
   return pSyntax->is_sspchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_namestrt() {
   return pSyntax->is_namestrt(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_name() {
   return pSyntax->is_name(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_mindata() {
   return pSyntax->is_mindata(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_minldat() {
   return pSyntax->is_minldat(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_fctchar() {
   return pSyntax->is_fctchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_ms_char() {
   return pSyntax->is_ms_char(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_msrchar() {
   return pSyntax->is_msrchar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_msichar() {
   return pSyntax->is_msichar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_msochar() {
   return pSyntax->is_msochar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_msschar() {
   return pSyntax->is_msschar(achBuffer[iBufPos]);
}
inline boolean EntityHandler::is_cref() {
   return pSyntax->is_cref(achBuffer[iBufPos]);
}

/*INLINE-----------------------------------------------------------------is_Ee-+
| Did we get the Ee signal?                                                    |
+-----------------------------------------------------------------------------*/
inline boolean EntityHandler::is_Ee()
{
   return (akocBuffer[iBufPos] == EE_KOC)? TRUE : FALSE;
}

/*INLINE------------------------------------------------------------is_numchar-+
| Is the current char data, because made by a numeric character reference?     |
+-----------------------------------------------------------------------------*/
inline boolean EntityHandler::is_numchar()
{
   return (akocBuffer[iBufPos] == NUMCHAR_KOC)? TRUE : FALSE;
}

/*INLINE-------------------------------------------------------------uppercase-+
| Uppercase the current character read                                         |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::uppercase()
{
   return pSyntax->uppercase(achBuffer[iBufPos]);
}

/*INLINE------------------------------------------------------------------find-+
| Find an external identifier                                                  |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::find(
   const ExternId* pExid, ApplicationHandle& ahlIdent
)
{
   return sr_find(pExid, ahlIdent);
}
#endif /* _EM_H ***************************************************************/
