/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// entity.c
///  Entity class
//   Pierre G. Richard 94/01/04 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <stdio.h>
#include <new.h>
#include "gendef.h"
#include "entity.h"
#include "enthdlr.h"
#include "externid.h"
#include "attrib.h"
#include "notation.h"
#include "rsvname.h"
#include "skmgr.h"
#include "procpil.h"

/*------------------------------------------------------------EntityDef::Flags-+
| Default Constructor for Flags                                                |
+-----------------------------------------------------------------------------*/
EntityDef::Flags::Flags()
{
   notation  = FALSE;
   bracketed = FALSE;
   rcdata = FALSE;
   external = FALSE;
   deflt = FALSE;
   parameter = FALSE;
   user = FALSE;
}

/*-------------------------------------------------------------------EntityDef-+
| Regular Constructor                                                          |
+-----------------------------------------------------------------------------*/
EntityDef::EntityDef(
   char *pszNameArg,
   const Flags &isArg,
   e_EntityKind kindArg,
   char *pszLitArg,
   int iLitLengthArg,
   ExternId *pExidArg,
   EntityNotation *pNadArg
) : is(isArg)
{
   pszName = pszNameArg;
   kind = kindArg;
   if (is.external) {
      pExid = pExidArg;
      pNad = pNadArg;
   }else {
      pszLit = pszLitArg;
      iLitLength = iLitLengthArg;
   }
}

/*-------------------------------------------------------------------EntityDef-+
| Copy Constructor                                                             |
+-----------------------------------------------------------------------------*/
EntityDef::EntityDef(
   const EntityDef& entSource, ExternId *pExidArg
) : is(entSource.is)
{
   kind = entSource.kind;
   if (is.external) {
      pExid = pExidArg;
      pNad = entSource.pNad;
   }else {
      pszLit = entSource.pszLit;
      iLitLength = entSource.iLitLength;
   }
}

/*-------------------------------------------------------------------EntityDef-+
| Copy Constructor to create a default entity                                  |
+-----------------------------------------------------------------------------*/
EntityDef::EntityDef(
   const EntityDef& entSource, char *pszNameArg, ExternId *pExidArg
) : is(entSource.is)
{
   pszName = pszNameArg;
   kind = entSource.kind;
   if (is.external) {
      pExid = pExidArg;
      pNad = entSource.pNad;
   }else {
      pszLit = entSource.pszLit;
      iLitLength = entSource.iLitLength;
   }
   is.deflt = TRUE;                      // tell its a defaulted entity
}

/*-------------------------------------------------------------------EntityDef-+
| Internal: pseudo-literal                                                     |
+-----------------------------------------------------------------------------*/
EntityDef::EntityDef(char *pszLitArg, int iLitLengthArg)
{
   pszName = 0;               // NULL for a pseudo entity
   is.rcdata = TRUE;          // that's all what it is!
   kind = TEXT_EFK;           // kind is just TEXT
   pszLit = pszLitArg;
   iLitLength = iLitLengthArg;
}

/*-------------------------------------------------------------------EntityDef-+
| Internal: SGML-specific entities                                             |
| ixName can be one of:                                                        |
|   (none) --> PRIMARY entity                                                  |
|   DOCTYPE_RRN_ix                                                             |
|   LINKTYPE_RRN_ix                                                            |
|   SYNTAX_RRN_ix                                                              |
|   CAPACITY_RRN_ix                                                            |
+-----------------------------------------------------------------------------*/
EntityDef::EntityDef(ExternId *pExidArg, e_IxRsvnameRefer ixName)
{
   if (ixName == NOT_FOUND_RRN_ix) {     // Then it's for the PRIMARY entity
      pszName = 0;
   }else {
      //***>NOTE: following cast to violate constness
      pszName = (char *)ReservedName::name_of(ixName);
   }
   is.external = TRUE;
   is.rcdata = TRUE;
   kind = SGML_EFK;
   pExid = pExidArg;
   pNad = 0;
}

/*---------------------------------------------------------------------pile_up-+
| Build entities table from the entity structure in core.                      |
|                                                                              |
| IMPLEMENTATION NOTE: it's assumed the default entity (if any) will come      |
| first -- see implementation note in EntityDcl::parse()                       |
+-----------------------------------------------------------------------------*/
boolean EntityDef::pile_up(
   StringKeeper& sk, const char *pszName,
   ExpandedHash& expnhashNotation,
   unsigned int& ofsBufferExidEntDefault,
   unsigned int& ofsLitEntDefault,
   int& ixNadEntDefault,
   EntityDef*& pEntMP, const EntityDef* const aEntTable,
   EntityNotation*& pNadMP, const EntityNotation* const aNadTable,
   Component*& pCompoMP, const Component * const aCompoTable,
   AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable,
   ExternId*& pExidMP, const ExternId* const aExidTable
)
{
   *pEntMP = *this;  // copy entity
   if (*pszName) {
      pEntMP->ofsName = sk.park(pszName, 1+strlen(pszName));
   }else {
      pEntMP->ofsName = 0;  // Unnamed: default entity
   }
   if (is.external) {
      pEntMP->ixExid = pExid->pile_up(
         sk, pEntMP->ofsName, pExidMP, aExidTable, ofsBufferExidEntDefault
      );
      if (is.deflt) {                // not TRUE for the default entity itself!
         pEntMP->ixNad = ixNadEntDefault;
      }else {
         if (is.notation) {
            pEntMP->ixNad = pNad->pile_up(
               expnhashNotation, pNadMP, aNadTable,
               pCompoMP, aCompoTable, pAttrepMP, aAttrepTable
            );
            delete pNad;          // delete EntityNotation + notation name
         }
         if (!*pszName) {         // Default Entity itself
            ixNadEntDefault = pEntMP->ixNad;
         }
      }
      delete pExid;
   }else {
      if (is.deflt) {
         pEntMP->ofsLit = ofsLitEntDefault;
      }else {
         pEntMP->iLitLength = iLitLength;
         pEntMP->ofsLit = sk.park(pszLit, 1 + iLitLength);
         free(pszLit);
      }
      if (!*pszName) {            // Default Entity
         ofsLitEntDefault = pEntMP->ofsLit;
      }
   }
   /*--------------------------------------------------------------+
   | As the entity structure in core is still valid, use the room  |
   | left by this->ofsLit to store the value of its matching       |
   | index in the EntityDef table.                                 |
   | This is a trick to transform entities pointer by their index  |
   | in the ODT structures (since pointer have no more sense in    |
   | the unloaded ODT file).                                       |
   | For instance, the name of entities in SHORTREF map is changed |
   | into their related index, avoiding (when reloading the ODT)   |
   | to do a search based on the name of the entity.  Also, the    |
   | pEnt within the component of CDATA attribute spec list is     |
   | changed into an ixEnt.                                        |
   +--------------------------------------------------------------*/
   ixEnt = diffptr(pEntMP, aEntTable);
   ++pEntMP;
   return TRUE;
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries EntityDef objects piled_up in an array                  |
+-----------------------------------------------------------------------------*/
void EntityDef::relocate_pile(
   int iMaxEntries,
   EntityNotation * aNadTable,
   ExternId * aExidTable,
   char * strbuf
)
{
   EntityDef *pEnt = this;

   for (int i=0; i < iMaxEntries; ++i, ++pEnt) {
      pEnt->pszName = strbuf + pEnt->ofsName;
      if (pEnt->is.external) {
         pEnt->pExid = aExidTable + pEnt->ixExid;
         if (pEnt->ixNad == -1) {
            pEnt->pNad = 0;
         }else {
            pEnt->pNad = aNadTable + pEnt->ixNad;
         }
      }else {
         pEnt->pszLit = strbuf + pEnt->ofsLit;
      }
   }
}

/*----------------------------------------------------------check_ref_validity-+
| Check if a reference to an entity is allowed in a given context.             |
| "check_code" is an internal code that can take the following values:         |
|                                                                              |
| Returns:                                                                     |
|  INVALID_ERC if the entity is not valid, that is:                            |
|      - an external entity, and the system was unable to locate it            |
|      - a SUBDOC (external) entity, not in CON mode                           |
|      - a NDATA (external) entity, not in CON mode                            |
|      - a PI (internal) entity, not in CON nor DS mode                        |
|     A suitable error message is issued.                                      |
|     FALSE is returned.                                                       |
|                                                                              |
|  OPENABLE_ERC if the entity is valid, and can be opened, that is:            |
|      - a parseable entity, succesfully located if external (not SUBDOC)      |
|      - a SUBDOC entity in CON mode                                           |
|      - a C/S Data (external or internal) in mode other than CON, RCD, or     |
|        RCDMS: as LITC mode (see 7.9.3)                                       |
|                                                                              |
|  UNPARSEABLE_ERC if the entity is valid, but must not be opened because it   |
|                  is not parseable, or doesn't need to be opened, that is:    |
|      - a C/S DATA (external or internal) entity in CON, RCD, RCDMS mode      |
|      - an NDATA (external) in CON mode                          [ (see 9.1)  |
|      - a PI (internal) entity in CON or DS mode                              |
|                                                                              |
+-----------------------------------------------------------------------------*/
EntityDef::e_CheckRefValidity EntityDef::check_ref_validity(
   EntityHandler *pEh, const e_RecognitionMode rcgnmode
) const
{
   if (is.external) {
      if (!pExid->find(pEh)) return INVALID_ERC;  // not valid and not openable
      if (!is.rcdata) {
         /*---------------------------------+
         | Non RCDATA (N/S/C Data or SUBDOC)|
         +---------------------------------*/
         if (
            ((kind == NDATA_EFK) || (kind == SUBDOC_EFK)) &&
            (rcgnmode != CON_MODE)        // see 9.1
         ) {
            errorh(INVENTPE, pEh, ERH_E_SEV);
            return INVALID_ERC;           // not valid and not openable
         }
         if (
            (kind != SUBDOC_EFK) && (
               (rcgnmode == CON_MODE) ||
               (rcgnmode == RCD_MODE) ||
               (rcgnmode == RCDMS_MODE)
            )
         ) {
            /*-----------------------------------------------------------+
            | Ext C/S DATA in CON, RCD, RCDMS mode, or NDATA in CON Mode |
            |                                                            |
            | For notations, find() must be called only when parsing     |
            | the document instance -- not in prolog, at compile time,   |
            | since they are not solved.  This cannot be the case:       |
            | none of the above modes can be activated while in prolog.  |
            +-----------------------------------------------------------*/
            if (pNad) {                          // find the notation, if any
               // cast: find violates the constness of ExternId
               ((ExternId *)pNad->external_id())->find(pEh);
            }
            return UNPARSEABLE_ERC;              // valid, but not openable
         }
      }
   }else {
      if (!is.rcdata) {
         /*--------------------------------+
         | Non parseable internal entities |
         +--------------------------------*/
         if (kind == PI_EFK) {                   // Internal PI
            if ((rcgnmode == CON_MODE) || (rcgnmode == DS_MODE)) {
               return UNPARSEABLE_ERC;           // valid, but not openable
            }
            errorh(INVENTPE, pEh, ERH_E_SEV);
            return INVALID_ERC;                  // not valid, not openable
         }
         /*---------------------------------------+
         | Internal C-SDATA (cannot be in prolog) |
         +---------------------------------------*/
         if (
            (rcgnmode == CON_MODE) ||
            (rcgnmode == RCD_MODE) ||
            (rcgnmode == RCDMS_MODE)
         ) {
            return UNPARSEABLE_ERC;              // valid, but not openable
         }
      }
   }
   return OPENABLE_ERC;                          // valid and openable
}

/*---------------------------------------------------------------------display-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void EntityDef::display(ParseState *pPrsstat, int iIndent) const
{
   char *pszKind;

   switch(kind) {
   case TEXT_EFK:     pszKind = "TEXT";         break;
   case CDATA_EFK:    pszKind = "CDATA";        break;
   case SDATA_EFK:    pszKind = "SDATA";        break;
   case PI_EFK:       pszKind = "PI";           break;
   case STARTTAG_EFK: pszKind = "STARTTAG";     break;
   case ENDTAG_EFK:   pszKind = "ENDTAG";       break;
   case MS_EFK:       pszKind = "MS";           break;
   case MD_EFK:       pszKind = "MD";           break;
   case NDATA_EFK:    pszKind = "NDATA";        break;
   case SUBDOC_EFK:   pszKind = "SUBDOC";       break;
   case SGML_EFK:     pszKind = "SGML";         break;
   default:           pszKind = "?????";        break;
   }
   fprintf(
      stderr, "%*.s\"%s\" type: %s, %sternal %s\n%*.sValue ",
      iIndent, " ",
      pszName? pszName : "#PRIMARY",
      is.parameter? "parameter" : "general",
      is.external? "ex" : "in",
      pszKind, iIndent+6, " "
   );
   if (!is.external) {
      fprintf(stderr, ": \"%s\"\n", pszLit);
   }else {
      fprintf(
         stderr, "in \"%s\"\n%*.sInterpreter: ",
         pExid->poem_id_handle().p, iIndent+6, " "
      );
      if (!pNad) {
         fprintf(stderr, "**unknown**\n");
      }else {
         pNad->display(pPrsstat, iIndent);
      }
   }
}

/*-------------------------------------------------------------------------new-+
| Create an EntityNotation, with space to hold the notation name (catenated)   |
| Note: this form is internal only (extended nad)                              |
+-----------------------------------------------------------------------------*/
void * EntityNotation::operator new(size_t size, const char *pszName)
{
   unsigned int iSizeName = 1 + strlen(pszName);
   char *pch = (char *)malloc(size +iSizeName);

   memset(pch, 0, size);
   memcpy(pch + size, pszName, iSizeName);
   return pch;
}

/*----------------------------------------------------------------------delete-+
| Insure consistency of the alloc/free library routines.                       |
+-----------------------------------------------------------------------------*/
void EntityNotation::operator delete(void *pVoid)
{
   free(pVoid);
}

/*-----------------------------------------------------------------------------+
| Get the notation name.  If no NCB has been created, then this is an          |
| extended NCB: get the name from the end of the NCB.                          |
+-----------------------------------------------------------------------------*/
const char * EntityNotation::name() const
{
   if (pNcb) return pNcb->external_id()->name();
   return (char *)this + sizeof(EntityNotation);
}

/*---------------------------------------------------------------------pile_up-+
| Pile-up the (expanded) EntityNotation.                                       |
|                                                                              |
| The EntityNotation (NAD) is a block that is associated to external data      |
| entities: it defines what NOTATION, and what DATA attributes are associated. |
| It is pointed from the EntityDef block of external entities.                 |
|                                                                              |
| The NotationType of an EntityNotation stays unsolved until the end of the    |
| compilation.  Also, if no attributes were defined, the Attribute pointer of  |
| the EntityNotation is not pointing the default attributes until the end of   |
| the compilation.  Now is the end of the compilation!                         |
| We need to allow `relocate_pile' to dirctly point at the right NotationType  |
| and possibily its DEFAULT attributes.  Hence, the ODT file will have all     |
| EntityDef block with Notation fully solved:                                  |
|       - Address/length of the SID of the notation (unloaded: offset/length)  |
|       - Attribute list pointer (unloaded: offset in the Attribute Table)     |
|                                                                              |
| NOTE: NotationBlock::pile_up() must have been issued!                        |
|                                                                              |
| Even if no notation, an ATTLIST can still have been defined.                 |
| Too much code not to build the ATT; however: discard it.                     |
+-----------------------------------------------------------------------------*/
int EntityNotation::pile_up(
   ExpandedHash& expnhashNotation,
   EntityNotation*& pNadMP, const EntityNotation* const aNadTable,
   Component*& pCompoMP, const Component * const aCompoTable,
   AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable
) const
{
   *pNadMP = *this;              // copy, whatever the result is...

   /*
   | Get in pNb the pointer to the NotationBlock,
   | or 0 if no NOTATION MD was declared
   */
   NotationBlock *pNb = (NotationBlock *)(
      expnhashNotation.search_value(name())
   );
   if (pAttlistSpec) {           // data attributes were associated: save them
      pNadMP->ixAttrep = pAttlistSpec->pile_up(
         pAttrepMP, aAttrepTable, pCompoMP, aCompoTable, pNb->ixAttrep
      );
      delete pAttlistSpec;
   }else if (pNb) {
      pNadMP->ixAttrep = pNb->ixAttrep;  // take default attributes
   }else {
      pNadMP->ixAttrep = -1;
   }
   int ix;
   if (!pNb) {                   // No notation associated
      pNadMP->ixNcb = -1;
      pNadMP->ixAttrep = -1;     // No notation: no attributes
      errorh(NONOTENT, ERH_E_SEV, name());
      ix = -1;
   }else {
      pNadMP->ixNcb = pNb->ixNcb;
      ix = diffptr(pNadMP, aNadTable);
   }
   ++pNadMP;
   return ix;
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries EntityNootation objects piled_up in an array            |
+-----------------------------------------------------------------------------*/
void EntityNotation::relocate_pile(
   int iMaxEntries,
   AttributeRep* aAttrepTable,
   NotationType* aNcbTable
)
{
   EntityNotation *pNad = this;

   for (int i=0; i < iMaxEntries; ++i, ++pNad) {
     if (pNad->ixAttrep == -1) {
        pNad->aAttrepSpecList = 0;
     }else {
        pNad->aAttrepSpecList = aAttrepTable + pNad->ixAttrep;
     }
     if (pNad->ixNcb == -1) {
        pNad->pNcb = 0;
     }else {
        pNad->pNcb = aNcbTable + pNad->ixNcb;
     }
   }
}

/*--------------------------------------------------------------attribute_list-+
| Build the attribute definition list                                          |
+-----------------------------------------------------------------------------*/
void EntityNotation::attribute_list(AttributeList& attlist) const {
   attlist.build(aAttrepSpecList);
}

/*---------------------------------------------------------------------display-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void EntityNotation::display(ParseState *pPrsstat, int iIndent) const
{
   fprintf(
      stderr, "(%s) \"%s\"\n",
      pNcb->external_id()->name(),
      pNcb->external_id()->poem_id_handle().p
   );
   AttributeList attlist(this);
   attlist.display(pPrsstat, iIndent+6+10);
}
/*============================================================================*/
