/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// entity.h
///  Declarations for entity
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _ENTITY_H
#define _ENTITY_H 1
/*
| This H file defines the structure associated to the Entity Declaration
*/
//******************************************************************************
#include "notation.h"

#include "rsvname.eh"  // enums
#include "prsmode.eh"  // enums

class AttributeRep;
class AttributeList;
class ParseState;
class PrologCompiler;
class EntityNotation;
class EntityHandler;
class ExternId;
class StringKeeper;
class ExpandedHash;
class Component;

/*-----------------------------------------------------------------------------+
| EntityDef                                                                    |
+-----------------------------------------------------------------------------*/
enum e_EntityKind {            // EFK: Kind
   INVALID_EFK = 0,            // Invalid
   CDATA_EFK,                  // not_rcdata | Ext/Int | notation iff Ext
   SDATA_EFK,                  // not_rcdata | Ext/Int | notation iff Ext
   MD_EFK,                     //   rcdata   |   Int   |    bracketed
   PI_EFK,                     // not_rcdata |   Int   |    bracketed
   STARTTAG_EFK,               //   rcdata   |   Int   |    bracketed
   ENDTAG_EFK,                 //   rcdata   |   Int   |    bracketed
   MS_EFK,                     //   rcdata   |   Int   |    bracketed
   NDATA_EFK,                  // not_rcdata |   Ext   |    notation
   TEXT_EFK,                   //   rcdata   |   Int   |
   SGML_EFK,                   //   rcdata   |   Ext   |
   SUBDOC_EFK                  // not_rcdata |   Ext   |
};

class EntityDef {              // ENT:
public:
   struct Flags {
      Flags();                 // Default Constructor
      boolean notation  :1;    // a notation block is associated
      boolean bracketed :1;    // must put stuff around it
      boolean rcdata    :1;    // RCDATA vs. CDATA/SDATA/NDATA/PI or SUBDOC
      boolean external  :1;    // external entity vs. memory entity
      boolean deflt     :1;    // it's the default entity
      boolean parameter :1;    // it's a parameter entity
      boolean user      :1;    // User flag - Init value is "off"
   };

   EntityDef(                  // regular constructor
      char *pszName,
      const Flags& is,
      e_EntityKind kind,
      char *pszLit,
      int iLitLength,
      ExternId *pExid,
      EntityNotation *pNad
   );
   EntityDef(                  // copy constructor
      const EntityDef& entSource, ExternId *pExidArg
   );
   EntityDef(                  // copy for a default entity
      const EntityDef& entSource, char *pszNameArg, ExternId *pExidArg
   );
   EntityDef(                  // pseudo-literal entity (internal)
      char *pszLit, int iLitLength
   );
   EntityDef(                  // SGML-specific entities (internal)
      ExternId *pExidArg,
      e_IxRsvnameRefer ixName = NOT_FOUND_RRN_ix
   );
   int ent_index() const;      // index in ENT array -- internal use
   boolean pile_up(
      StringKeeper& sk, const char *pszName,
      ExpandedHash& expnhashNotation,
      unsigned int& ofsBufferExidEntDefault,
      unsigned int& ofsLitEntDefault,
      int& ixNadEntDefault,
      EntityDef*& pEntMP, const EntityDef* const aEntTable,
      EntityNotation*& pNadMP, const EntityNotation* const aNadTable,
      Component*& pCompoMP, const Component * const aCompoTable,
      AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable,
      ExternId*& pExidMP, const ExternId* const aExidTable
   );
   void relocate_pile(
      int iMaxEntries,
      EntityNotation * aNadTable,
      ExternId * aExidTable,
      char * strbuf
   );
   enum e_CheckRefValidity { INVALID_ERC, OPENABLE_ERC, UNPARSEABLE_ERC };
   e_CheckRefValidity check_ref_validity(
      EntityHandler *pEh, const e_RecognitionMode rcgnmode
   ) const;

   /*
   |  API functions
   */
   const char * name() const;  // Name of the entity
   const char * value() const; // Value of the entity: only if Internals Ent!
   int value_length() const;   // faster strlen(value()), only if Internals Ent!
   const ExternId * external_id() const; // External id (iff is_external())
   const EntityNotation * entity_notation() const; // Notation + atributes

   // returns the "text_type" of entity, as described by the e_EntityKind enum.
   // This describes the entity types, data text and bracketed text keywords.
   e_EntityKind text_type() const;

   // are TRUE or FALSE
   boolean is_external() const;  // TRUE if external entity, FALSE if internal
   boolean is_internal() const;  // TRUE if internal entity, FALSE if external
   boolean is_bracketed() const; // TRUE if brackets, FALSE if not
   boolean is_parameter() const; // TRUE if parameter entity, FALSE if general
   boolean is_general() const;   // TRUE if general entity, FALSE if parameter
   boolean is_default() const;   // TRUE if this entity instanciates #DEFAULT
   boolean is_rcdata() const;    // FALSE if CDATA/SDATA/NDATA/PI or SUBDOC

   // Debugging...
   void display(ParseState *pPrsstat, int iIndent) const;


private:
   union {                     // Start with its name (hashing)
      char *pszName;           // EntityDef name
      unsigned int ofsName;    // (offset, when unloaded)
   };
   union {
                               // iff Internal Entity:
      char *pszLit;            // address of the literal
      unsigned int ofsLit;     // (offset, when unloaded)
                               // iff External entity:
      ExternId *pExid;         // External Identifier
      int ixExid;              // (EXID array index, when unloaded)
      int ixEnt;               // (ENT array index, temporary use)
   };
   union {                     // iff Internal Entity:
      int  iLitLength;         // length of the literal
                               // iff N/S/C Data external entity:
      EntityNotation *pNad;    // Notation + Attrib
      int  ixNad;              // (NAD array index, when unloaded)
   };
   Flags is;                   // booleans
   e_EntityKind kind;          // value field
   EntityDef(const EntityDef& entSource);  // reserved (undefined)
};

/*-- INLINES --*/
inline int EntityDef::ent_index() const { return ixEnt; }
inline const char * EntityDef::name() const { return pszName; }
inline const char * EntityDef::value() const {
   return is.external? "" : pszLit;
}
inline int EntityDef::value_length() const {
   return is.external? 0 : iLitLength;
}
inline const ExternId * EntityDef::external_id() const {
   return is.external? pExid : 0;
}
inline const EntityNotation * EntityDef::entity_notation() const {
   return is.notation? pNad : 0;
}
inline e_EntityKind EntityDef::text_type() const {
   return kind;
}
inline boolean EntityDef::is_external() const {
   return is.external? TRUE : FALSE;
}
inline boolean EntityDef::is_internal() const {
   return is.external? FALSE : TRUE;
}
inline boolean EntityDef::is_bracketed() const {
   return is.bracketed? TRUE : FALSE;
}
inline boolean EntityDef::is_parameter() const {
   return is.parameter? TRUE : FALSE;
}
inline boolean EntityDef::is_general() const {
   return is.parameter? FALSE : TRUE;
}
inline boolean EntityDef::is_default() const {
   return is.deflt? TRUE : FALSE;
}
inline boolean EntityDef::is_rcdata() const {
   return is.rcdata? TRUE : FALSE;
}

/*-----------------------------------------------------------------------------+
| Entity's notation (NAD)                                                      |
|                                                                              |
| A EntityDef Notation structure (NAD) describes the notation and              |
| data attributes associated to external C/N/S DATA entities.                  |
+-----------------------------------------------------------------------------*/
class EntityNotation {         // NAD: Entity Notation Structure
public:
   void * operator new(size_t size, EntityNotation *p);
   void * operator new(size_t size, const char *pszName);
   void operator delete(void *pVoid);
   void set_attspec(AttributeList *pAttlistSpec);
   const ExternId * external_id() const;
   int pile_up(
      ExpandedHash& expnhashNotation,
      EntityNotation*& pNadMP, const EntityNotation* const aNadTable,
      Component*& pCompoMP, const Component * const aCompoTable,
      AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable
   ) const;
   void relocate_pile(
      int iMaxEntries,
      AttributeRep* aAttrepTable,
      NotationType* aNcbTable
   );

   void attribute_list(AttributeList &attlist) const;
   const char * name() const;

   // Debugging...
   void display(ParseState *pPrsstat, int iIndent) const;

private:
   void * operator new(size_t size); // not defined
   union {
      AttributeRep *aAttrepSpecList; // Attribute Specification List
      int ixAttrep;                  // (ATTREP array index, when unloaded)
      AttributeList *pAttlistSpec;   // (attribute spec list while compiling)
   };
   union {
      NotationType *pNcb;            // Pointer to Notation Identifier
      int ixNcb;                     // (NCB array index when unloaded)
   };
};

/*-- INLINES --*/
inline const ExternId * EntityNotation::external_id() const {
   return pNcb->external_id();
}
inline void * EntityNotation::operator new(size_t, EntityNotation *p)  {
   return p;  // Specific Placement
}
inline void EntityNotation::set_attspec(AttributeList *pAttlistSpecArg) {
   pAttlistSpec = pAttlistSpecArg;
}
#endif /* _ENTITY_H ***********************************************************/
