/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// errorh.c
///  Error handler
//   Pierre G. Richard 87/04/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| Format and print an error.  Returns or exit (depending on severity).
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include "gendef.h"
#include "errorh.h"
#include "syntax.h"
#include "enthdlr.h"

/*-----------------+
| Static functions |
+-----------------*/
static void print_error(e_ErrorCode, e_ErrorSev, va_list arg_p);
static void locate_error(EntityHandler *pEh, e_ErrorCode msgcod);

/*-----------------------------------------------------------------print_error-+
| Error handler                                                                |
|                                                                              |
| This is a simplified version (*not* National Language Enabled) of the        |
| usual error handler.                                                         |
+-----------------------------------------------------------------------------*/
static void print_error(e_ErrorCode msgcod, e_ErrorSev sev, va_list arg_p)
{
   static char *messages[] = {
      // Error location messages
      /* 000 ER_NULL   */ "\n",
      /* 001 ER_LINE   */ "%s\n",
      /* 002 ER_INDIC  */ "%s\n",
      /* 003 ER_LOC0   */ "In primary entity (\"%s\"), at line %d:\n",
      /* 004 ER_LOC1   */ "In entity: %s (\"%s\"), at line %d:\n",
      /* 005 ER_LOC2   */ "In entity: %s, at offset %ld:\n",
      /* 006 ER_LOC3   */ "In literal, at offset %ld:\n",

      // Other messages
      /* 007 ABORTED   */ "Terminated.\n",
      /* 008 AMBMDL    */ "Ambiguous model on token \"%s\"\n",
      /* 009 BADRDPRIM */ "I/O error reading primary entity\n",
      /* 010 BADCLPRIM */ "Cannot close primary entity\n",
      /* 011 BADCLOSE  */ "Cannot close external entity: \"%s\"\n",
      /* 012 BADCLS    */ "Incompatible Public Text Class; ignored\n",
      /* 013 BADNOTA   */ "Notation: \"%s\" has not been defined\n",
      /* 014 BADOPDRD  */ "ODT file invalid or unreadable (code %d)\n",
      /* 015 BADOPDWR  */ "Cannot write ODT file\n",
      /* 016 BADOPEN   */ "Cannot open external entity: \"%s\"\n",
      /* 017 BADOPSRD  */ "Syntax file invalid or unreadable (code %d)\n",
      /* 018 BADOPSWR  */ "Cannot write syntax file\n",
      /* 019 BADREAD   */ "I/O error reading external entity: \"%s\"\n",
      /* 020 BIGASENT  */ "Number of entity names (%d) exceeds GRPCNT\n",
      /* 021 BIGASIDR  */ "Number of id references names (%d) exceeds GRPCNT\n",
      /* 022 BIGATTCNT */ "ATTCNT quantity exceeded; value \"%s\" ignored\n",
      /* 023 BIGAVL    */ "Normalized length of attribute value (%d) exceeds LITLEN\n",
      /* 024 BIGAVS    */ "Normalized length of attribute specification (%d) exceeds ATTSPLEN\n",
      /* 025 BIGCAP    */ "\"%s\" capacity exceeded (%lu vs. %lu allowed)\n",
      /* 026 BIGCHDESC */ "Described character %d outside of the character set\n",
      /* 027 BIGCHNB   */ "Character number too big; reference ignored\n",
      /* 028 BIGDLIT   */ "Data tag literal too long\n",
      /* 029 BIGDSCLIT */ "Too many literals in DESCSETs\n",
      /* 030 BIGDTD    */ "Too many DOCTYPE declarations, or CONCUR not enabled\n",
      /* 031 BIGENTLVL */ "EntityDef nesting exceeds ENTLVL quantity\n",
      /* 032 BIGFEAT   */ "Number exceeds the currently supported value; limited to %d\n",
      /* 033 BIGGRPCNT */ "Token exceeds GRPCNT quantity\n",
      /* 034 BIGGRPGTC */ "Token exceeds GRPGTCNT quantity\n",
      /* 035 BIGGRPLVL */ "Nested group exceeds GRPLVL quantity\n",
      /* 036 BIGLINK   */ "Too many LINK declarations, or this kind of LINK not enabled\n",
      /* 037 BIGLIT    */ "Literal too long\n",
      /* 038 BIGLITBK  */ "Bracketed literal too long\n",
      /* 039 BIGLITRCS */ "Minimum literal (%d) exceeds the reference LITLEN quantity\n",
      /* 040 BIGMODLVL */ "Mode stack overflow\n",
      /* 041 BIGMSLVL  */ "Nested MS exceeds TAGLVL quantity; MS ignored\n",
      /* 042 BIGNAME   */ "The length exceeds the NAMELEN quantity\n",
      /* 043 BIGNAMEP  */ "Parameter entity name is too long\n",
      /* 044 BIGNAMET  */ "The length exceeds the NAMELEN quantity; truncated\n",
      /* 045 BIGPI     */ "Processing Instruction too long\n",
      /* 046 BIGQTY    */ "TOTALCAP set to the largest individual value %u\n",
      /* 047 BIGSUBDOC */ "Too many SUBDOC opened, or SUBDOC not enabled\n",
      /* 048 BIGTAGLEN */ "Start-tag length (%d) exceeded TAGLEN quantity\n",
      /* 049 BIGTAGLVL */ "TAGLVL quantity exceeded at start-tag %s\n",
      /* 050 BIGTKLIST */ "Only one token allowed for this attribute\n",
      /* 051 DIFCONN   */ "Inconsistent connector\n",
      /* 052 DIFUCLC   */ "UC and LC must have same number of characters\n",
      /* 053 DUPATT    */ "Attribute %s already defined; ignored\n",
      /* 054 DUPCHDESC */ "Character number %d already described; ignored\n",
      /* 055 DUPDCLV   */ "%s already defined for attribute %s; ignored\n",
      /* 056 DUPDEF    */ "%s already defined\n",
      /* 057 DUPDLM    */ "Two identical delimiters found: \"%s\"\n",
      /* 058 DUPDRDOC  */ "CHARSET describes character \"%c\" by more than one number\n",
      /* 059 DUPDRSYN  */ "SYNTAX describes character \"%c\" by more than one number\n",
      /* 060 DUPFCTCNB */ "Duplicate Function Character number: %d\n",
      /* 061 DUPFCTREF */ "Duplicate Function Name found: \"%s\"\n",
      /* 062 DUPIDDEF  */ "Only one ID allowed; NAME assumed\n",
      /* 063 DUPIDNAME */ "Should use \"%s\" for the name of this ID attribute\n",
      /* 064 DUPNAM    */ "Duplicate specification; ignored\n",
      /* 065 DUPNAMREP */ "%s already replaces %s; replacement ignored\n",
      /* 066 DUPNOTDEF */ "Only one NOTATION allowed; token group assumed\n",
      /* 067 DUPSREF   */ "Duplicate short reference delimiter; ignored\n",
      /* 068 DUPSRMD   */ "Duplicate SHORTREF mapping declaration; ignored\n",
      /* 069 DUPSRU    */ "Element already associated with a map; ignored\n",
      /* 070 DUPVANAM  */ "%s already replaced by %s; replacement ignored\n",
      /* 071 EXITED    */ "Parser exited - RC=%d\n",
      /* 072 FAKEEND   */ "Bad end-tag; treated as empty end-tag\n",
      /* 073 IGNOREEND */ "Bad end-tag; ignored\n",
      /* 074 INVALS    */ "Extraneous attribute; ignored\n",
      /* 075 INVATTDCL */ "Invalid declared value in ATTLIST declaration\n",
      /* 076 INVATTDEF */ "Attribute %s not defined for this tag; ignored\n",
      /* 077 INVATTSPC */ "Invalid attribute value specified\n",
      /* 078 INVBASET  */ "Unknown BASESET\n",
      /* 079 INVCAP    */ "Value too big.  Capacity set to maximum (%lu)\n",
      /* 080 INVCHAR   */ "Non-SGML character found (\'%02x\'x)\n",
      /* 081 INVCHRF   */ "Unknown function name %s; reference ignored\n",
      /* 082 INVCONN   */ "A valid connector or a GRPC was expected\n",
      /* 083 INVCONREF */ "Empty element %s cannot have CONREF; ANY assumed\n",
      /* 084 INVDCLVDA */ "Invalid for Data Attributes; NAME(S) assumed\n",
      /* 085 INVDEFV   */ "Invalid for Data or Link Attributes; IMPLIED assumed\n",
      /* 086 INVDLM    */ "Invalid or misplaced delimiter\n",
      /* 087 INVDTPLIT */ "Invalid reference in data tag pattern; ignored\n",
      /* 088 INVELEMGI  */ "Element \"%s\" not declared in DTD; ignored\n",
      /* 089 INVELEMGID */ "Element not declared in DTD; \"%s O O ANY\" assumed\n",
      /* 090 INVEMPCON */ "Start-tag %s omitted with empty content\n",
      /* 091 INVEMPTY  */ "Empty string not allowed\n",
      /* 092 INVENDTAG */ "Implied end-tag %s was not declared omissible\n",
      /* 093 INVEOF    */ "Premature end of file\n",
      /* 094 INVETGMIN */ "End-tag minimization should be \"omissible\" for an EMPTY element\n",
      /* 095 INVENTATT */ "General entity \"%s\" specification invalid\n",
      /* 096 INVENTPE  */ "EntityDef type prohibited in replaceable character data; ignored\n",
      /* 097 INVEXID   */ "System identifier for \"%s\" cannot be generated by the system\n",
      /* 098 INVFCTCL  */ "Invalid Function Class\n",
      /* 099 INVFEAT   */ "The feature %s is not implemented; ignored\n",
      /* 100 INVSGFPI  */ "FORMAL YES and non-formal public identifier(s) found\n",
      /* 101 INVFXATT  */ "FIXED attribute must equal default; assigned\n",
      /* 102 INVHNEST  */ "Closing delimiter prematurely found in a nested entity\n",
      /* 103 INVIDDEF  */ "ID must be REQUIRED or IMPLIED; IMPLIED assumed\n",
      /* 104 INVKWD    */ "Unrecognizable keyword\n",
      /* 105 INVLNEST  */ "End of entity prematurely found\n",
      /* 106 INVMD     */ "Misplaced Markup Declaration\n",
      /* 107 INVMDIGN  */ "Invalid Markup Declaration; ignored\n",
      /* 108 INVMEX    */ "Invalid exclusion of element %s\n",
      /* 109 INVMLCHAR */ "Invalid character in minimum literal (\'%02x\'x); ignored\n",
      /* 110 INVMSE    */ "Marked Section End when no MS open; ignored\n",
      /* 111 INVMTYNET */ "NET delimiter is invalid for EMPTY elements; TAGC assumed\n",
      /* 112 INVNAMGRP */ "Name or name group invalid\n",
      /* 113 INVNASGN  */ "Invalid naming assignement: %c\n",
      /* 114 INVNOTA   */ "Empty element %s cannot have NOTATION; ANY assumed\n",
      /* 115 INVNOTASP */ "Notation specified with an explicit content reference\n",
      /* 116 INVNUM    */ "Invalid number\n",
      /* 117 INVPETKW  */ "Keyword ignored in parameter entity\n",
      /* 118 INVQNTY   */ "Invalid value.  Quantity not replaced\n",
      /* 119 INVREMAP  */ "SHORTREF #%d mapped to entity \"%s\"\n",
      /* 120 INVROL    */ "Invalid role\n",
      /* 121 INVSCASGN */ "Ambiguous Markup Scan character\n",
      /* 122 INVSGMLID */ "\"ISO 8879:1986\" expected; ignored\n",
      /* 123 INVSRBS1  */ "More than one BSEQ; shortref ignored\n",
      /* 124 INVSRBS2  */ "BSEQ preceeded or followed by a sepchar; shortref ignored\n",
      /* 125 INVSREF   */ "Literal not a SHORTREF; shortref ignored\n",
      /* 126 INVSTGMIN */ "Start-tag minimization should not be \"omissible\": declared content\n",
      /* 127 INVSTGREQ */ "Start-tag minimization for %s should not be \"omissible\": REQUIRED attribute\n",
      /* 128 INVSTRTAG */ "Implied start-tag %s was not declared omissible\n",
      /* 129 INVSWITCH */ "Cannot switch #%02d and #%02d: digits or letters; ignored\n",
      /* 130 ISORDOC   */ "Either INSTANCE or DOCUMENT expected\n",
      /* 131 NIMPL     */ "Not implemented (%d)\n",
      /* 132 NOATKDEF  */ "Token %s not found; attribute ignored\n",
      /* 133 NOATTDEF  */ "No attributes defined; attribute(s) ignored\n",
      /* 134 NOATTSPC  */ "Attribute %s not specified; defaulted\n",
      /* 135 NOCNTN    */ "No content token found\n",
      /* 136 NOCURATT  */ "Current attribute %s missing\n",
      /* 137 NODIG     */ "A digit was expected\n",
      /* 138 NODRDOC   */ "Significant character \"%c\" not assigned in CHARSET\n",
      /* 139 NODRDOC2  */ "Cannot map syntax char #%02d to a document char #\n",
      /* 140 NODRSYN   */ "Significant character \"%c\" not assigned in SYNTAX\n",
      /* 141 NODRSYN2  */ "Syntax character #%02d not described\n",
      /* 142 NODSC     */ "Expecting a DSC\n",
      /* 143 NODTD     */ "Expecting a DOCTYPE declaration\n",
      /* 144 NODTGC    */ "Expecting a DTGC\n",
      /* 145 NOEOF     */ "EOF expected\n",
      /* 146 NOGIREF   */ "Declared element \"%s\" does not appear in any content model\n",
      /* 147 NOGERIGN  */ "General entity \"%s\" not declared; ignored\n",
      /* 148 NOKYWD    */ "Keyword: %s expected\n",
      /* 149 NOLIT     */ "Parameter literal expected\n",
      /* 150 NOMDC     */ "Expecting a MDC\n",
      /* 151 NOMSICHAR */ "No MSICHAR function specified\n",
      /* 152 NONAME    */ "Name expected\n",
      /* 153 NONMTK    */ "Name token expected\n",
      /* 154 NONOTA    */ "No Notation MD found for %s; ignored\n",
      /* 155 NONOTENT  */ "No Notation MD found for entity %s; ignored\n",
      /* 156 NOOTAGMIN */ "Expecting omitted tag minimization\n",
      /* 157 NOPER     */ "No declaration; parameter entity ignored\n",
      /* 158 NOPS      */ "Expecting a parameter separator\n",
      /* 159 NORQDATT  */ "Required attribute %s missing\n",
      /* 160 NOSRENT   */ "Map %s: entity %s not found; SR not mapped\n",
      /* 161 NOSRM     */ "Shortref Map: %s not found; ignored\n",
      /* 162 NOTAGC    */ "TAGC expected; tag closed\n",
      /* 163 NUMLITUNU */ "Number, parm literal or UNUSED expected\n",
      /* 164 OOCDATA   */ "Data out of context; ignored\n",
      /* 165 OOCETG    */ "%s element ended prematurely\n",
      /* 166 OOCSTG    */ "Start-tag out of context; ignored\n",
      /* 167 OUTDOCDAT */ "Data found outside document; parse ended\n",
      /* 168 OUTDOCETG */ "End-tag found outside document; parse ended\n",
      /* 169 OUTDOCSTG */ "Start-tag found outside document; parse ended\n",
      /* 170 DUPFPI    */ "Entities %s and %s have the same formal public identifier\n",
      /* 171 NOLANGUP  */ "Formal Identifier: language name must be in uppercase; changed\n",
      /* 172 OOCPTDV   */ "Formal Identifier: public text display version not allowed\n",
      /* 173 INVPTDV   */ "Formal Identifier: extraneous solidi found\n",
      /* 174 NOPTDS    */ "Formal Identifier: public text designating sequence not found\n",
      /* 175 INVOWNKN  */ "Formal Identifier: invalid owner kind\n",
      /* 176 INVOWNID  */ "Formal Identifier: invalid owner identifier\n",
      /* 177 INVTXTCL  */ "Formal Identifier: invalid text class\n",
      /* 178 NOSPTXCL  */ "Formal Identifier: missing space after text class\n",
      /* 179 NOPTXDS   */ "Formal Identifier: public text description not found\n",
      /* 180 INVSWTCH2 */ "Cannot switch #%02d and #%02d: not in the syntax; ignored\n",
      /* 181 SGMLNONE  */ "Either SGMLREF or NONE expected\n",
      /* 182 UBLNCOM   */ "Unexpected entity end in comment\n",
      /* 183 UBLNLIT   */ "\"%s\" expected;  entity end, truncated\n",
      /* 184 UBLNMSO   */ "Unexpected end of entity in opened Marked Section\n",
      /* 185 UBLNPIC   */ "Unexpected end of entity in processing instruction\n",
      /* 186 UBLNTAG   */ "Unexpected end of entity in tag\n",
      /* 187 YESNO     */ "Either YES or NO must be specified\n",
      /* 188 BADOPPRIM */ "Cannot open primary entity\n",
      /* 189 DUPELEM    */ "Element %s already defined; ignored\n",
      /* 190 BIGSTMSUF */ "\"%s%s\": stem plus suffix exceed the NAMELEN quantity\n",
      /* 191 NODTDI    */ "Expecting a DOCTYPE declaration; implied from first start-tag\n",
      /* 192 INVEMPASL */ "Empty attribute specification list\n",
      /* 193 DUPELEMSTM */ "Element %s already defined as a rank stem; ignored\n",
      /* 194 DUPSTMELEM */ "Rank stem %s already defined as an element; ignored\n",
      /* 195 DUPID     */ "ID \"%s\" also identifies element #%lu; ignored\n",
      /* 196 OOCSEP    */ "Data (separator) out of context; ignored\n",
      /* 197 NOASSELEM  */ "No associated element type found; declaration ignored\n",
      /* 198 INVEMPPER */ "No token found in a parameter entity referenced within a group\n",
      /* 199 UBLNMSOD  */ "Marked Section still open; section closed\n",
      /* 200 INVMSPRO  */ "CDATA or RCDATA Marked Section in prolog; section ignored\n",
      /* 201 INVDLSTR0 */ "Delimiter string empty; ignored\n",
      /* 202 INVDLSTR1 */ "General delimiter consisting solely of function chars; ignored\n",
      /* 203 DUPDIC    */ "Delimiter-in-context \"%s\" duplicates \"%s\"\n",
      /* 204 DUPCHASGN */ "Character number %d duplicates codeset assignment of %d\n",
      /* 205 INVSGCHAR */ "Character not allowed in SGML Declaration (\'%02x\'x)\n",
      /* 206 INVPTC    */ "Formal Identifier: \"ISO Registration Number\" outside CHARSET class\n",
      /* 207 INVPTL    */ "Formal Identifier: language not a name\n",
      /* 208 NOCAPSET  */ "Expecting a capacity name and value to follow SGMLREF\n",
      /* 209 BIGALLOC  */ "Capacity points condition %d occurred on 16-bit machine; terminated\n",
      /* 210 DUPNAMDLM */ "Duplicate delimiter specification; ignored\n",
      /* 211 DUPNAMENT */ "Duplicate entity specification; ignored\n",
      /* 212 DUPNAMNOT */ "Duplicate notation name specification; ignored\n",
      /* 213 NODTSG    */ "Expecting a DOCTYPE or an optional SGML declaration\n",
      /* 214 NOVANAM   */ "%s occurs only in the SGML declaration; replacement ignored\n",
      /* 215 INVISTQTY */ "Quantity %s (%d) must equal or exceed the value required by the DTD (%d)\n",
      /* 216 INVISTDST */ "Syntax char #%02d is described differently in the reference\n",
      /* 217 NOVA2NAM  */ "%s occurs only in the SYSTEM declaration; replacement ignored\n",
      /* 218 NOREFNAM  */ "%s is not a reference reserved name; replacement ignored\n",
      /* 219 INVRPTNAM */ "Replacement cannot be a reference reserved name; ignored\n",
      /* 220 INVISTCH  */ "Description of \"%s\" may induce misplacing the end of prolog\n",
      /* 221 NOGINAME  */ "GI name expected\n",
      /* 222 MSG_222   */ "",   /* Unused */
      /* 223 INVISTDLM */ "Changing %s delimiter may induce misplacing the end of prolog\n",
      /* 224 BADREADUT */ "I/O error reading utility file\n",
      /* 225 BADWRITUT */ "I/O error writing utility file\n",
      /* 226 MSG_226   */ "",   /* Unused */
      /* 227 INVIDREF  */ "ID \"%s\" has not been defined\n",
      /* 228 INVSTGOMI */ "Start-tag %s missing and tag omission prohibited; implied\n",
      /* 229 INVETGOMI */ "End-tag %s missing and tag omission prohibited; implied\n",
      /* 230 NOGERDEF  */ "General entity \"%s\" not declared; defaulted\n",
      /* 231 INVELEMREF */ "Undeclared element \"%s\" found in one or more content models\n",
      /* 232 DUPNAMTYP */ "\"%s\" already describes a DTD or LTD; terminated\n",
      /* 233 INVDCLVLK */ "Invalid for Link Attributes; NAME(S) assumed\n",
      /* 234 NOFIXED   */ "Simple link attributes must have FIXED default value; FIXED assumed\n",
      /* 235 INVATTSLK */ "Associated element is not the base document element; ignored\n",
      /* 236 INVALINK  */ "Cannot activate LINKTYPE \"%s\"; demand ignored\n",
      /**/
      /* 237 UNKNOWN   */ "MESSAGE #%d\n"
   };
   static const char chIdentFormat[] =  "YSP%04d: (%c)";
   boolean isAbnormalEnd = FALSE;
   char c;

   switch (sev) {
   case ERH_I_SEV: c = 'I'; break;
   case ERH_E_SEV: c = 'E'; break;
   case ERH_S_SEV: c = 'S'; break;
   default:        c = 'T'; isAbnormalEnd = TRUE; break;    // case ERH_T_SEV
   }
   if (msgcod >= UNKNOWN) {           // in case unknown errcod (tests only)
      fprintf(stderr, chIdentFormat, UNKNOWN, c);
      fprintf(stderr, messages[UNKNOWN], msgcod);
   }else {
      fprintf(stderr, chIdentFormat, msgcod, c);
      vfprintf(stderr, messages[msgcod], arg_p);
   }
   if (msgcod > ER_LOC3) fprintf(stderr, "\n");  // Skip a line
   if (isAbnormalEnd) {
      errorh(ABORTED, ERH_I_SEV);
      exit(8);
   }
}

/*STATIC----------------------------------------------------------locate_error-+
| Print location messages and part of the line in error.                       |
|                                                                              |
| Note that the location of the error (arrow) works even if the error appears  |
| outside a record-oriented context.                                           |
+-----------------------------------------------------------------------------*/
static void locate_error(EntityHandler *pEh, e_ErrorCode msgcod)
{
   enum { MSG_MAXSIZ  = 80 };  // Max size of a line
   char achBuffer[MSG_MAXSIZ + 1];

   if (pEh->curr_entity_level()) { // need at least a primary entity
      {
         /*------------------------------------------------------------------+
         | Issue messages to locate where the error was found                |
         | Do this for all nested entities, starting with the PRIMARY entity |
         +------------------------------------------------------------------*/
         int row;
         long offset;
         int iRequestedEntityLevel = 1;
         char *pszName, *appl_info_p;

         do {
            pEh->locate(
               iRequestedEntityLevel,
               &pszName, &appl_info_p, &row, &offset
            );
            if (iRequestedEntityLevel==1) {
               errorh(ER_LOC0, ERH_I_SEV, appl_info_p, row);
            }else if (pszName) {                     // entity %s
               if (appl_info_p) {
                  errorh(ER_LOC1, ERH_I_SEV, pszName, appl_info_p, row);
               }else {
                  errorh(ER_LOC2, ERH_I_SEV, pszName, offset);
               }
            }else {
               errorh(ER_LOC3, ERH_I_SEV, offset);
            }
         }while (++iRequestedEntityLevel <= pEh->curr_entity_level());
      }
      {
         /*-----------------------------------------------------------------+
         | Print the current line where the error was found,                |
         | followed by a another line with an arrow under the part in error |
         +-----------------------------------------------------------------*/
         unsigned int iArrowPos = 0;
         unsigned int iBufferPos = 0;
         int iToUnread;

         do {
            pEh->unread(1);
         }while (                             // Stop when...
            (++iArrowPos < MSG_MAXSIZ - 1) && // Buffer overflow
            (!pEh->is_Ee()) &&                // End of entity
            (!pEh->is_RSchar())               // or a Record start
         );
         iToUnread = -iArrowPos;
         iBufferPos = 0;
         do {
            pEh->read();
            if (++iToUnread >= 0) {           // At the error:
               if (                           // Don't go after if...
                  (msgcod == INVSGCHAR) ||
                  (msgcod == INVCHAR) ||      // Non SGML found
                  (pEh->is_REchar()) ||       // End of record
                  (pEh->is_Ee())              // End of entity
               ) {
                  break;
               }
            }
            /*
            |Try to give a readable image of the line!
            */
            if (isgraph(pEh->curr_char_val())) {
               achBuffer[iBufferPos++] = (char)pEh->curr_char_val();   // that's OK
            }else if (
               (pEh->curr_char_val() == '\n') || (pEh->curr_char_val() == '\r')
            ) {
               if (iToUnread < 0) --iArrowPos;
            }else {
               achBuffer[iBufferPos++] = ' ';    // non-graphic, HT, VT, FF
            }
         }while (iBufferPos < MSG_MAXSIZ - 1);
         pEh->unread(iToUnread);
         achBuffer[iBufferPos] = '\0';           // AsciiZ string
         errorh(ER_LINE, ERH_I_SEV, achBuffer);
         memset(achBuffer, ' ', iArrowPos-1);
         achBuffer[iArrowPos-1] = '^';           // Put an arrow
         achBuffer[iArrowPos] = '\0';            // AsciiZ string
         errorh(ER_INDIC, ERH_I_SEV, achBuffer);
      }
   }
}

/*----------------------------------------------------------------------errorh-+
| Issue an error message - no location                                         |
+-----------------------------------------------------------------------------*/
void errorh(e_ErrorCode msgcod, e_ErrorSev sev, ...)
{
   va_list arg_p;
   va_start(arg_p, sev);
   print_error(msgcod, sev, arg_p);
   va_end(arg_p);
}

/*----------------------------------------------------------------------errorh-+
| Issue an error message preceeded by the location                             |
+-----------------------------------------------------------------------------*/
void errorh(e_ErrorCode code, EntityHandler *pEh, e_ErrorSev sev, ...)
{
   locate_error(pEh, code);
   va_list arg_p;
   va_start(arg_p, sev);
   print_error(code, sev, arg_p);
   va_end(arg_p);
}
/******************************************************************************/
