/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// errorh.h
///  Declarations for errorh
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _ERRORH_H
#define _ERRORH_H 1
//******************************************************************************

#include <stdarg.h>

class EntityHandler;
class Syntax;

/*-----------------+
| Error Severities |
+-----------------*/
enum e_ErrorSev {      // sev
   ERH_I_SEV =  0,     // Informational message
   ERH_E_SEV =  8,     // Sgml error (but Parser can recover)
   ERH_S_SEV = 12,     // Severe error (probably can't recover)
   ERH_T_SEV = 16      // Terminating error
};

enum e_BigAllocFailedCause {   // Parameter to BIGALLOC error
   RESERVED_BGLA,
   MDL_BGAL,
   ELEM_BGAL,
   ATT_BGAL,
   GENT_BGAL,
   PENT_BGAL,
   SRM_BGAL,
   NCB_BGAL,
   EXID_BGAL,
   SKMG_BGAL,          // strbuf overflow: see skmgr.c
   ADS_BGAL,
   EXC_BGAL
};

/*-----------+
|Error Codes |
+-----------*/
enum e_ErrorCode {
   /*  All the following codes are not *real* "error codes" */
   ER_NULL,                                   /* 0000 */
   ER_LINE,                                   /* 0001 */
   ER_INDIC,                                  /* 0002 */
   ER_LOC0,                                   /* 0003 */
   ER_LOC1,                                   /* 0004 */
   ER_LOC2,                                   /* 0005 */
   ER_LOC3,                                   /* 0006 */

   /* All these are used by errorh */
   ABORTED,                                   /* 0007 */
   AMBMDL,                                    /* 0008 */
   BADRDPRIM,                                 /* 0009 */
   BADCLPRIM,                                 /* 0010 */
   BADCLOSE,                                  /* 0011 */
   BADCLS,                                    /* 0012 */
   BADNOTA,                                   /* 0013 */
   BADOPDRD,                                  /* 0014 */
   BADOPDWR,                                  /* 0015 */
   BADOPEN,                                   /* 0016 */
   BADOPSRD,                                  /* 0017 */
   BADOPSWR,                                  /* 0018 */
   BADREAD,                                   /* 0019 */
   BIGASENT,                                  /* 0020 */
   BIGASIDR,                                  /* 0021 */
   BIGATTCNT,                                 /* 0022 */
   BIGAVL,                                    /* 0023 */
   BIGAVS,                                    /* 0024 */
   BIGCAP,                                    /* 0025 */
   BIGCHDESC,                                 /* 0026 */
   BIGCHNB,                                   /* 0027 */
   BIGDLIT,                                   /* 0028 */
   BIGDSCLIT,                                 /* 0029 */
   BIGDTD,                                    /* 0030 */
   BIGENTLVL,                                 /* 0031 */
   BIGFEAT,                                   /* 0032 */
   BIGGRPCNT,                                 /* 0033 */
   BIGGRPGTC,                                 /* 0034 */
   BIGGRPLVL,                                 /* 0035 */
   BIGLINK,                                   /* 0036 */
   BIGLIT,                                    /* 0037 */
   BIGLITBK,                                  /* 0038 */
   BIGLITRCS,                                 /* 0039 */
   BIGMODLVL,                                 /* 0040 */
   BIGMSLVL,                                  /* 0041 */
   BIGNAME,                                   /* 0042 */
   BIGNAMEP,                                  /* 0043 */
   BIGNAMET,                                  /* 0044 */
   BIGPI,                                     /* 0045 */
   BIGQTY,                                    /* 0046 */
   BIGSUBDOC,                                 /* 0047 */
   BIGTAGLEN,                                 /* 0048 */
   BIGTAGLVL,                                 /* 0049 */
   BIGTKLIST,                                 /* 0050 */
   DIFCONN,                                   /* 0051 */
   DIFUCLC,                                   /* 0052 */
   DUPATT,                                    /* 0053 */
   DUPCHDESC,                                 /* 0054 */
   DUPDCLV,                                   /* 0055 */
   DUPDEF,                                    /* 0056 */
   DUPDLM,                                    /* 0057 */
   DUPDRDOC,                                  /* 0058 */
   DUPDRSYN,                                  /* 0059 */
   DUPFCTCNB,                                 /* 0060 */
   DUPFCTREF,                                 /* 0061 */
   DUPIDDEF,                                  /* 0062 */
   DUPIDNAME,                                 /* 0063 */
   DUPNAM,                                    /* 0064 */
   DUPNAMREP,                                 /* 0065 */
   DUPNOTDEF,                                 /* 0066 */
   DUPSREF,                                   /* 0067 */
   DUPSRMD,                                   /* 0068 */
   DUPSRU,                                    /* 0069 */
   DUPVANAM,                                  /* 0070 */
   EXITED,                                    /* 0071 */
   FAKEEND,                                   /* 0072 */
   IGNOREEND,                                 /* 0073 */
   INVALS,                                    /* 0074 */
   INVATTDCL,                                 /* 0075 */
   INVATTDEF,                                 /* 0076 */
   INVATTSPC,                                 /* 0077 */
   INVBASET,                                  /* 0078 */
   INVCAP,                                    /* 0079 */
   INVCHAR,                                   /* 0080 */
   INVCHRF,                                   /* 0081 */
   INVCONN,                                   /* 0082 */
   INVCONREF,                                 /* 0083 */
   INVDCLVDA,                                 /* 0084 */
   INVDEFV,                                   /* 0085 */
   INVDLM,                                    /* 0086 */
   INVDTPLIT,                                 /* 0087 */
   INVELEMGI,                                  /* 0088 */
   INVELEMGID,                                 /* 0089 */
   INVEMPCON,                                 /* 0090 */
   INVEMPTY,                                  /* 0091 */
   INVENDTAG,                                 /* 0092 */
   INVEOF,                                    /* 0093 */
   INVETGMIN,                                 /* 0094 */
   INVENTATT,                                 /* 0095 */
   INVENTPE,                                  /* 0096 */
   INVEXID,                                   /* 0097 */
   INVFCTCL,                                  /* 0098 */
   INVFEAT,                                   /* 0099 */
   INVSGFPI,                                  /* 0100 */
   INVFXATT,                                  /* 0101 */
   INVHNEST,                                  /* 0102 */
   INVIDDEF,                                  /* 0103 */
   INVKWD,                                    /* 0104 */
   INVLNEST,                                  /* 0105 */
   INVMD,                                     /* 0106 */
   INVMDIGN,                                  /* 0107 */
   INVMEX,                                    /* 0108 */
   INVMLCHAR,                                 /* 0109 */
   INVMSE,                                    /* 0110 */
   INVMTYNET,                                 /* 0111 */
   INVNAMGRP,                                 /* 0112 */
   INVNASGN,                                  /* 0113 */
   INVNOTA,                                   /* 0114 */
   INVNOTASP,                                 /* 0115 */
   INVNUM,                                    /* 0116 */
   INVPETKW,                                  /* 0117 */
   INVQNTY,                                   /* 0118 */
   INVREMAP,                                  /* 0119 */
   INVROL,                                    /* 0120 */
   INVSCASGN,                                 /* 0121 */
   INVSGMLID,                                 /* 0122 */
   INVSRBS1,                                  /* 0123 */
   INVSRBS2,                                  /* 0124 */
   INVSREF,                                   /* 0125 */
   INVSTGMIN,                                 /* 0126 */
   INVSTGREQ,                                 /* 0127 */
   INVSTRTAG,                                 /* 0128 */
   INVSWITCH,                                 /* 0129 */
   ISORDOC,                                   /* 0130 */
   NIMPL,                                     /* 0131 */
   NOATKDEF,                                  /* 0132 */
   NOATTDEF,                                  /* 0133 */
   NOATTSPC,                                  /* 0134 */
   NOCNTN,                                    /* 0135 */
   NOCURATT,                                  /* 0136 */
   NODIG,                                     /* 0137 */
   NODRDOC,                                   /* 0138 */
   NODRDOC2,                                  /* 0139 */
   NODRSYN,                                   /* 0140 */
   NODRSYN2,                                  /* 0141 */
   NODSC,                                     /* 0142 */
   NODTD,                                     /* 0143 */
   NODTGC,                                    /* 0144 */
   NOEOF,                                     /* 0145 */
   NOGIREF,                                   /* 0146 */
   NOGERIGN,                                  /* 0147 */
   NOKYWD,                                    /* 0148 */
   NOLIT,                                     /* 0149 */
   NOMDC,                                     /* 0150 */
   NOMSICHAR,                                 /* 0151 */
   NONAME,                                    /* 0152 */
   NONMTK,                                    /* 0153 */
   NONOTA,                                    /* 0154 */
   NONOTENT,                                  /* 0155 */
   NOOTAGMIN,                                 /* 0156 */
   NOPER,                                     /* 0157 */
   NOPS,                                      /* 0158 */
   NORQDATT,                                  /* 0159 */
   NOSRENT,                                   /* 0160 */
   NOSRM,                                     /* 0161 */
   NOTAGC,                                    /* 0162 */
   NUMLITUNU,                                 /* 0163 */
   OOCDATA,                                   /* 0164 */
   OOCETG,                                    /* 0165 */
   OOCSTG,                                    /* 0166 */
   OUTDOCDAT,                                 /* 0167 */
   OUTDOCETG,                                 /* 0168 */
   OUTDOCSTG,                                 /* 0169 */
   DUPFPI,                                    /* 0170 */
   NOLANGUP,                                  /* 0171 */
   OOCPTDV,                                   /* 0172 */
   INVPTDV,                                   /* 0173 */
   NOPTDS,                                    /* 0174 */
   INVOWNKN,                                  /* 0175 */
   INVOWNID,                                  /* 0176 */
   INVTXTCL,                                  /* 0177 */
   NOSPTXCL,                                  /* 0178 */
   NOPTXDS,                                   /* 0179 */
   INVSWTCH2,                                 /* 0180 */
   SGMLNONE,                                  /* 0181 */
   UBLNCOM,                                   /* 0182 */
   UBLNLIT,                                   /* 0183 */
   UBLNMSO,                                   /* 0184 */
   UBLNPIC,                                   /* 0185 */
   UBLNTAG,                                   /* 0186 */
   YESNO,                                     /* 0187 */
   BADOPPRIM,                                 /* 0188 */
   DUPELEM,                                    /* 0189 */
   BIGSTMSUF,                                 /* 0190 */
   NODTDI,                                    /* 0191 */
   INVEMPASL,                                 /* 0192 */
   DUPELEMSTM,                                 /* 0193 */
   DUPSTMELEM,                                 /* 0194 */
   DUPID,                                     /* 0195 */
   OOCSEP,                                    /* 0196 */
   NOASSELEM,                                  /* 0197 */
   INVEMPPER,                                 /* 0198 */
   UBLNMSOD,                                  /* 0199 */
   INVMSPRO,                                  /* 0200 */
   INVDLSTR0,                                 /* 0201 */
   INVDLSTR1,                                 /* 0202 */
   DUPDIC,                                    /* 0203 */
   DUPCHASGN,                                 /* 0204 */
   INVSGCHAR,                                 /* 0205 */
   INVPTC,                                    /* 0206 */
   INVPTL,                                    /* 0207 */
   NOCAPSET,                                  /* 0208 */
   BIGALLOC,                                  /* 0209 */
   DUPNAMDLM,                                 /* 0210 */
   DUPNAMENT,                                 /* 0211 */
   DUPNAMNOT,                                 /* 0212 */
   NODTSG,         /* not used in YAO */      /* 0213 */
   NOVANAM,                                   /* 0214 */
   INVISTQTY,                                 /* 0215 */
   INVISTDST,                                 /* 0216 */
   NOVA2NAM,                                  /* 0217 */
   NOREFNAM,                                  /* 0218 */
   INVRPTNAM,                                 /* 0219 */
   INVISTCH,                                  /* 0220 */
   NOGINAME,                                  /* 0221 */
   MSG_222,                                   /* 0222 */
   INVISTDLM,                                 /* 0223 */
   BADREADUT,                                 /* 0224 */
   BADWRITUT,                                 /* 0225 */
   MSG_226,                                   /* 0226 */
   INVIDREF,                                  /* 0227 */
   INVSTGOMI,                                 /* 0228 */
   INVETGOMI,                                 /* 0229 */
   NOGERDEF,                                  /* 0230 */
   INVELEMREF,                                 /* 0231 */
   DUPNAMTYP,                                 /* 0232 */
   INVDCLVLK,                                 /* 0233 */
   NOFIXED,                                   /* 0234 */
   INVATTSLK,                                 /* 0235 */
   INVALINK,                                  /* 0236 */
   /**/
   UNKNOWN   /* is always at the end */       /* 0237 */
};

extern void errorh(e_ErrorCode msgcod, e_ErrorSev sev, ...);
extern void errorh(e_ErrorCode code, EntityHandler *pEh, e_ErrorSev sev, ...);

#endif /* _ERRORH_H ***********************************************************/
