/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// externid.h
///  Declarations for externid
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _EXTERNID_H
#define _EXTERNID_H 1

//******************************************************************************

#include "externid.eh"  // enums

class ParseState;
class EntityHandler;
class Syntax;
class StringKeeper;
class EntityHandler;

/*-----------------------------------------------------------------------------+
| Offset/Length                                                                |
+-----------------------------------------------------------------------------*/
struct OffsetLength {          // OFLEN: to access to a subfield
   unsigned int rel;           // Relative offset of the subfield within adlen.p
   unsigned int len;           // Subfield length
};

/*-----------------------------------------------------------------------------+
|External Identifier                                                           |
+-----------------------------------------------------------------------------*/
class ExternId {               // external resource identifier
public:
   ExternId(const char *pszName, e_TextClass asgn_cls);
   ExternId(
      ParseState *pPrsstat,
      const char *pszName,
      e_TextClass asgn_cls=INVALID_XFC
   );
   ExternId::ExternId(                // True copy
      const ExternId& exidSource,
      const char *pszName,
      e_TextClass asgn_cls
   );
   ExternId(const ExternId& exidSource, const char *pszName);
   _export ~ExternId();
   void dislocate(
      StringKeeper& sk, const ExternId& exidSource, const char *pszName
   );
   void relocate(char *strbuf);
   int pile_up(
      StringKeeper &sk,
      int ofsName,
      ExternId*& pExidMP, const ExternId* const aExidTable,
      unsigned int &ofsBufferExidEntDefault
   );
   void formalize_public_id(
      EntityHandler *pEh, Syntax *pSyntax, boolean isTrueWarning
   );
   void assign_class(
      EntityHandler *pEh, e_TextClass asgn_clsArg, boolean isFormal
   );
   static int skip_past_solidi(EntityHandler *pEh);
   void check_duplicate(StringKeeper& sk, const int iMaxEntries);
   void relocate_pile(unsigned int iMaxEntries, char * strbuf);
   boolean find(EntityHandler *pEh);

   e_TextClass assigned_class() const;   // Logical class assigned by the parser
   const ApplicationHandle& poem_id_handle() const; // POEM Identifier handle

   /*
   |  API methods
   */
   const char *name() const;       // name assigned

   const char * public_id() const; // Public ID: ASCIIZ string or NULL
   int public_id_length() const;   // faster than `strlen()'
   const char * system_id() const; // System ID: ASCIIZ string or NULL
   int system_id_length() const;   // faster than `strlen()'

   // Are TRUE or FALSE
   boolean is_identified() const;  // An identifier has been found
   boolean is_unchecked() const;   // POEM was not asked to map this ExternId
   boolean is_unknown_on_system() const; // POEM was unable to map it
   boolean is_formal() const;      // PID has a formal structure
   boolean is_avail_public_text() const; // Available public text
   boolean is_provide_ptdv() const; // System must provide Pub.Text Disp.Vers.
   boolean is_public_text_ds() const; // Pub.Text Design. Sequence vs. Language
   boolean is_alternate() const;   // Take alternate system spec if any
   boolean is_defaulted() const;   // Defaulted entity
   boolean is_registered() const;  // Owner is registered
   boolean is_iso_registration() const; // Registration is ISO
   boolean is_public() const;      // A Public Identifier was furnished
   boolean is_system() const;      // A System Identifier was furnished

   e_TextClass defined_class() const; // class, as defined in the public id

private:
   union {
      char *pszName;           // Name
      unsigned int ofsName;    // (offset, when unloaded)
   };
   union {
      char *pchBuffer;         // Identifier buffer
      unsigned int ofsBuffer;  // (offset, when unloaded)
   };                          // Identifier Subfields related to buffer
   OffsetLength sid;           // System Identifier (also an ASCIIZ string)
   OffsetLength pid;           // Public Identifier (also an ASCIIZ string)
   /*
   | iff is.VALIDFPI, redefine FPI subfields of "pid" (related to buffer)
   */
   OffsetLength ownid;         // Owner identifier
   OffsetLength txtdesc;       // Public text description
   OffsetLength lang;          // lang, or designating seq
   OffsetLength ptdv;          // Public Text Display Version
   boolean isIdentified :1;    // An identifier has been found
   boolean isChecked    :1;    // POEM checked the XID
   boolean isInvalid    :1;    // POEM was unable to map it
   boolean isValidfpi   :1;    // PID has a formal structure
   boolean isAvailptext :1;    // available public text
   boolean isProvptdv   :1;    // System must provide Pub.Text Disp.Vers.
   boolean isPtds       :1;    // Pub.Text Design. Sequence vs. Language
   boolean isAltern     :1;    // Take alternate system spec if any
   boolean isDeflt      :1;    // Default entity (including the unnamed)
   boolean isRegistered :1;    // Owner is registered
   boolean isIsoreg     :1;    // Registration is ISO
   boolean isPublic     :1;    // A Public Identifier was furnished
   boolean isSystem     :1;    // A System Identifier was furnished
   e_TextClass asgn_cls;       // Text Class assigned by the Parser
   e_TextClass spec_cls;       // Text Class as defined
   ApplicationHandle ahlIdent; // POEM Identifier handle (when solved)
};

/*-- INLINES --*/
inline const char * ExternId::name() const { return pszName; }

inline const char * ExternId::public_id() const {
   return isPublic? pchBuffer + pid.rel : 0;
}
inline int ExternId::public_id_length() const { return pid.len; }

inline const char * ExternId::system_id() const {
   return isSystem? pchBuffer + sid.rel : 0;
}
inline int ExternId::system_id_length() const { return sid.len; }

inline const ApplicationHandle& ExternId::poem_id_handle() const {
   return ahlIdent;
}

inline boolean ExternId::is_identified() const { return isIdentified; }
inline boolean ExternId::is_unchecked() const { return isChecked? FALSE : TRUE; }
inline boolean ExternId::is_unknown_on_system() const { return isInvalid; }
inline boolean ExternId::is_formal() const { return isValidfpi; }
inline boolean ExternId::is_avail_public_text() const { return isAvailptext; }
inline boolean ExternId::is_provide_ptdv() const { return isProvptdv; }
inline boolean ExternId::is_public_text_ds() const { return isPtds; }
inline boolean ExternId::is_alternate() const { return isAltern; }
inline boolean ExternId::is_defaulted() const { return isDeflt; }
inline boolean ExternId::is_registered() const { return isRegistered; }
inline boolean ExternId::is_iso_registration() const { return isIsoreg; }
inline boolean ExternId::is_public() const { return isPublic; }
inline boolean ExternId::is_system() const { return isSystem; }
inline e_TextClass ExternId::defined_class()  const { return spec_cls; }
inline e_TextClass ExternId::assigned_class() const { return asgn_cls; }

#endif /* _EXTERNID_H *********************************************************/
