/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// hash.c
///  Hashing techniques
//   Pierre G. Richard 87/01/11 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
|
| There are 2 kinds of hash classes, derived from the same `Hash' class:
|  - the `ExpandedHash', which is used as a regular hash table,
|  - the `StringizedHash', which is a stringized version of the former.
|    Hence a StringizedHash object can only be built by reloading
|    an unloaded ExpandedHash object.  ExpandedHash tables candidates to
|    unloading are also subject to some restrictions, as explained below.
|
| Constructors
| ============
|
| ------------
| An ExpandedHash table is constructed by:
|
|  ExpandedHash(unsigned short usMod = 10);
|
| where `usMod' is a number that can be guessed from the square root of
| expected number of items in the hash table.
|
| ------------
| An ExpandedHash can be downloaded if and only if:
|   1) all items have the same size,
|   2) all items can be piled up in a contiguous array
|   3) the order of grouping in this array has been obtained from the
|      traversal of the ExpandedHash table (function `traverse()')
|
|   unsigned int unload(
|      unsigned short usItemSize,
|      boolean isExtendable,
|      void  **ppVoid
|   );
|
| unloads the hash table to a mallocated piece of memory, the size of
| it is returned by this function. `ppVoid' is the address of a pointer
| to the piece of memory, `usItemSize' indicates the common size of each
| item.  It's up to the caller to free this memory, after having, for
| example, written its contents to a file.
|
| `is_extendable' is a boolean telling if the Stringized hash table
| that will be constructed later has some chances to have other items
| added to it.
|
| Note: by design, the use of Stringized hash tables should be restricted
|       to "dead" tables: no item to be added.  Extension of Stringized
|       hash tables is provided only for the cases where few items need
|       to be added.  If the number of "post-added" items is kept low,
|       it won't impact too much the efficiency of the search.
|
| ------------
| A StringizedHash table is constructed by:
|
|  StringizedHash(void *pVoid, void *pvItemBase);
|
| where `pVoid' is a pointer to the same piece of memory built by
| ExpandedHash::unload().  Of course, this piece of memory does not
| need to be at the same address.  `pvItemBase' is the address of the
| array where all items have been piled up.
|
| Functions of the Hash abstract class:
| =====================================
|
| Note: Functions prefixed by "gimme", "add" or "search":
|       "gimme" or "add" means: if you don't find it, create it
|       "search" means: just tell me if you don't find it, but don't create
|
| ------------
| To access an existing member:
|
|  void * search_value(char *pszVarName);
|
| returns either a pointer to the value associated to the variable
| (member found) or a NULL pointer (not found).
|
| ------------
| To access an existing member, or to create a new one (if unexistent)
|
|  void * add_value(char **ppszVarName, unsigned int iSizeValue)
|
| `ppszVarName' is used for Input and Output.
| On input, it points to the address of a character array, where
| the variable name has been (temporarly?) set.
| On output, the address of the character array is updated and
| points to a non-volatile area where the variable name is kept.
|
| Always returns a pointer to the value assigned to `pszVarName'.
| If this is the expanded hash table, and if the member is new,
| this points to an area of `iSizeVarValue' bytes, all bits reset
| to binary zeroes.
|
| ------------
| To get the count of variables in the hash table:
|
|  int count()
|
| Functions of the ExpandedHash abstract class:
| =============================================
|
| ------------
| To access an existing member, or to create a new one (if unexistent)
|
|  void * gimme_value(char *pszVarName, unsigned int iSizeVarValue);
|
| Always returns a pointer to the value assigned to `pszVarName'.
| If the member is new, this points to an area of `iSizeVarValue' bytes,
| all set to binary zeroes.
|
| This is a variant of `add_value()', when the caller doesn't care
| about getting access of the safe place where the variable name is stored.
|
| ------------
| To traverse all members sequentially:
|
|   traverse(void (*tvrs_function)(void *pvBase, void *pvValue) void *pvBase)
|
| ------------
| At termination:
|
|   unsigned int unload(unsigned int usItemSize, (void **)ppVoid)
|
| "stringize" the hashtable and "unload" it in core, put the address of the
| "unloaded" area in *ppVoid, return the size.
|
| This permits the caller to write the hashtable on an external device,
| as in: io_vput(fh, *ppVoid ,size); see comments in the module.
|
|
| Functions of the StringizedHash abstract class:
| ===============================================
|
| The "unloaded" hashtable is reloaded by the StringizedHash constructor:
|
|   StringizedHash(void *pVoid, void *pvItemBase);
|
|      - pVoid is the address of the "unloaded" hash table
|        obtained for example by: io_vget(fh, &pvoid)
|      - pvItemBase is the address of the sequential array of hashed items.
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include "gendef.h"
#include "hash.h"

/*******************************************************************************
*                                                                              *
*                - These routines for the ExpnHashEntry -                      *
*                                                                              *
*******************************************************************************/

/*-------------------------------------------------------------------------new-+
| Create an ExpnHashEntry, with space for the value and the name (catenated)   |
+-----------------------------------------------------------------------------*/
void * ExpandedHash::ExpnHashEntry::operator new(
    size_t size, const char *pszVarName, unsigned int iSizeValue
)
{
   ExpnHashEntry *pEhsh;
   unsigned int iSizeName = 1 + strlen(pszVarName);

   pEhsh = (ExpnHashEntry *)malloc(size + iSizeValue + iSizeName);
   pEhsh->iSize = iSizeValue;
   memset((char *)pEhsh + size, 0, iSizeValue);
   memcpy((char *)pEhsh + size + iSizeValue, pszVarName, iSizeName);
   return pEhsh;
}

/*----------------------------------------------------------------------delete-+
| Insure consistency of the alloc/free library routines.                       |
+-----------------------------------------------------------------------------*/
void ExpandedHash::ExpnHashEntry::operator delete(void *pVoid)
{
   free(pVoid);
}

/*---------------------------------------------------------------ExpnHashEntry-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
inline ExpandedHash::ExpnHashEntry::ExpnHashEntry(ExpnHashEntry **ppEhshNext)
{
   pEhshNext = *ppEhshNext;
   *ppEhshNext = this;
}

/*******************************************************************************
*                                                                              *
*         - These routines handle the "expanded" hashtable -                   *
*                                                                              *
*******************************************************************************/


/*-----------------------------------------------------------------Constructor-+
| usMod is a number related to the average of number of members the hashtable  |
| will contain.  Roughly, it can be seen as the square root of the number      |
| of members.                                                                  |
+-----------------------------------------------------------------------------*/
ExpandedHash::ExpandedHash(unsigned short usMod)
{
   iCount = 0;
   usExpnMod = usMod;
   apEhshTable = (ExpnHashEntry **) calloc((usExpnMod), sizeof apEhshTable[0]);
}

/*------------------------------------------------------------------Destructor-+
|                                                                              |
+-----------------------------------------------------------------------------*/
ExpandedHash::~ExpandedHash()
{
   for (int i=0; i < usExpnMod; ++i) {
      if (apEhshTable[i]) {
         ExpnHashEntry *pEhsh = apEhshTable[i];
         do {
            ExpnHashEntry *pEhshSave = pEhsh->pEhshNext;
            delete pEhsh;
            pEhsh = pEhshSave;
         }while (pEhsh);
      }
   }
   free(apEhshTable);
}

/*INLINE-PRIVATE-------------------------------------------------------name_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline char * ExpandedHash::name_of(const ExpnHashEntry * pEhsh) {
   if (!pEhsh->iSize) return 0;         // Entry was deleted
   return (char *)(pEhsh + 1) + pEhsh->iSize;
}

/*INLINE-PRIVATE--------------------------------------------compute_hash_value-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline unsigned int ExpandedHash::compute_hash_value(const char *pszVarName)
{
   const char *psz = pszVarName;
   unsigned int iHashValue = 0;

   while (*psz) iHashValue += *psz++;
   return iHashValue % usExpnMod;
}

/*PRIVATE------------------------------------------------gimme_or_search_token-+
| Try to access a member of the hash table:                                    |
| - if found, return a pointer to the corresponding ExpnHashEntry              |
| - if not found:                                                              |
|    - if create_it is TRUE:                                                   |
|         create a new ExpnHashEntry,                                          |
|         zero the pVarValue entry                                             |
|         return a pointer to the new ExpnHashEntry                            |
|    - if create_it is FALSE: return a NULL ExpnHashEntry pointer              |
|                                                                              |
| The collision list is kept in alphabetic order, to speed up the process.     |
+-----------------------------------------------------------------------------*/
const ExpandedHash::ExpnHashEntry * ExpandedHash::gimme_or_search_token(
   const char *pszVarName, unsigned int iSizeValue, boolean create_it
)
{
   ExpnHashEntry **ppEhsh;

   for (                          // Look for a match
      ppEhsh = apEhshTable + compute_hash_value(pszVarName);
      *ppEhsh;
      ppEhsh = &(*ppEhsh)->pEhshNext
   ) {
     int cond;

     if (!(*ppEhsh)->iSize) continue;  // virtually deleted
     cond = strcmp(pszVarName, name_of(*ppEhsh));
     if (cond < 0) break;         // Collision list is in ascending order
     if (!cond) return *ppEhsh;   // Match found
   }
   if (create_it) {               // Not found, but caller asked to create it
      ++iCount;
      return new(pszVarName, iSizeValue) ExpnHashEntry(ppEhsh);
   }
   return 0;                      // Not found
}

/*-------------------------------------------------------------------add_value-+
| Add a new member to the expanded table.                                      |
| ppszVarName: on input, points the address of the variable name               |
|              on output, points the address where the variable name           |
|                          has been internally stored                          |
+-----------------------------------------------------------------------------*/
void * ExpandedHash::add_value(
   char **ppszVarName,
   unsigned int iSizeValue
)
{
   const ExpnHashEntry *pEhsh = gimme_or_search_token(
      *ppszVarName, iSizeValue, TRUE
   );
   *ppszVarName = name_of(pEhsh);
   return value_of(pEhsh);
}

/*--------------------------------------------------------------------traverse-+
| Call "tvrs function" for all variables installed/solved:                     |
|   - Descend sequentially the Hashtable                                       |
|   - Traverse the collision chain for each hash entry                         |
| If caller returns FALSE, then the entry is virtually removed and won't be    |
| unloaded.  However, the entry is not physically deleted: this allow the      |
| caller to keep valid pointers to the to-be-deleted structure.                |
+-----------------------------------------------------------------------------*/
void ExpandedHash::traverse(
   boolean (*tvrs_function)(void *pvBase, const char *pszName, void *pvValue),
   void *pvBase
)
{
   ExpnHashEntry *pEhsh;

   for (int i=0; i < usExpnMod; ++i) {
      for (pEhsh = apEhshTable[i]; pEhsh; pEhsh = pEhsh->pEhshNext) {
         if (!pEhsh->iSize) continue;
         if (!(*tvrs_function)(pvBase, name_of(pEhsh), value_of(pEhsh))) {
            pEhsh->iSize = 0;
            --iCount;
         }
      }
   }
}

/*----------------------------------------------------------------------unload-+
| Unload (stringize and dis-allocate) the contents of the ehsh table.          |
| Returns the size of a mallocated area that contains the unloaded table.      |
|                                                                              |
| Caller is deemed to have made a copy of all items -- pointed by non-zero     |
| pVarValue -- in a sequential array (each entry being of item_size).          |
|                                                                              |
| The "unloaded" area is mallocated: it's the responsability of the caller     |
| to free it.                                                                  |
+-----------------------------------------------------------------------------*/
unsigned int ExpandedHash::unload(
   unsigned short usItemSize,
   boolean isExtendable,
   void  **ppVoid
) {
   unsigned int iSizeTable;
   StringizedHash::SgzdHashHeader *pShhdr;
   StringizedHash::SgzdHashEntry  *aShsh;

   iSizeTable = sizeof(StringizedHash::SgzdHashCompound) + (
      (usExpnMod) * sizeof(StringizedHash::SgzdHashEntry)
   );
   *ppVoid = malloc(iSizeTable);
   pShhdr = (StringizedHash::SgzdHashHeader *)(*ppVoid);
   aShsh =  (StringizedHash::SgzdHashEntry *)(
      (StringizedHash::SgzdHashCompound *)(*ppVoid) + 1
   );

   pShhdr->usMod = usExpnMod;
   pShhdr->isExtendable = isExtendable;
   pShhdr->usItemSize = usItemSize;
   {
      unsigned int ofsItem = 0;

      for (int i=0; i < usExpnMod; ++i) {
         aShsh[i].ofsItem = ofsItem;
         aShsh[i].iCollisionCount = 0;
         if (apEhshTable[i]) {
            ExpnHashEntry *pEhsh = apEhshTable[i];
            do {
               if (pEhsh->iSize) {        // if *not* virtually deleted
                  ofsItem += usItemSize;
                  ++aShsh[i].iCollisionCount;
               }
               {
                  ExpnHashEntry *pEhshSave = pEhsh->pEhshNext;
                  delete pEhsh;
                  pEhsh = pEhshSave;
               }
            }while (pEhsh);
            apEhshTable[i] = 0;
         }
      }
   }
   //** io_vput(fh, *ppVoid, iSizeTable);
   return iSizeTable;
}

/*******************************************************************************
*                                                                              *
*         - These routines handle the "stringized" hashtable -                 *
*                                                                              *
*******************************************************************************/

/*-----------------------------------------------------------------Constructor-+
| Relocate an "ehsh_unloaded" hash table                                       |
| pvItemBase point an array containing `usSzgdMod' hashed items.               |
+-----------------------------------------------------------------------------*/
StringizedHash::StringizedHash(void *pVoid, void *pvItemBase)
{
   //** io_vget(fh, &pVoid);

   SgzdHashHeader *pShhdr;

   pShhdr = (SgzdHashHeader *)pVoid;
   aShsh = (SgzdHashEntry *)((SgzdHashCompound *)pVoid + 1);

   iCount = 0;
   usSgzdMod = pShhdr->usMod;
   if (pShhdr->isExtendable) {
     apXihdr = (ExtendedItemHeader **)calloc(
        usSgzdMod, sizeof(ExtendedItemHeader *)
     );
   }else {
     apXihdr = 0;
   }
   usItemSize = pShhdr->usItemSize;
   {
      for (int i=0; i < usSgzdMod; ++i) {
         if (aShsh[i].iCollisionCount) {
            iCount += aShsh[i].iCollisionCount;
            aShsh[i].pIhdr = (ItemHeader *)(
               (char *)pvItemBase + aShsh[i].ofsItem
            );
         }else {
            aShsh[i].pIhdr = 0;
         }
      }
   }
}

/*------------------------------------------------------------------Destructor-+
|                                                                              |
+-----------------------------------------------------------------------------*/
StringizedHash::~StringizedHash()
{
   if (apXihdr) {
      for (int i=0; i < usSgzdMod; ++i) {
         if (apXihdr[i]) {
            ExtendedItemHeader *pXihdr = apXihdr[i];
            do {
               ExtendedItemHeader *pXihdrSave = pXihdr->next;
               free(pXihdr);
               pXihdr = pXihdrSave;
            }while (pXihdr);
         }
      }
      free(apXihdr);
   }
}

/*INLINE----------------------------------------------------compute_hash_value-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline unsigned int StringizedHash::compute_hash_value(const char *pszVarName)
{
   const char *psz = pszVarName;
   unsigned int iHashValue = 0;

   while (*psz) iHashValue += *psz++;
   return iHashValue % usSgzdMod;
}

/*----------------------------------------------------------------search_value-+
| Search a member in a loaded (stringized) hashtable                           |
| The collision list is kept in alphabetic order, to speed up the process.     |
+-----------------------------------------------------------------------------*/
void * StringizedHash::search_value(const char *pszVarName)
{
   int iHashValue = compute_hash_value(pszVarName);
   int iCollisionCount = aShsh[iHashValue].iCollisionCount;

   if (iCollisionCount) {
      int cond;
      ItemHeader *pIhdr = aShsh[iHashValue].pIhdr;

      for (;;) {
         cond = strcmp(pszVarName, pIhdr->pszName);
         if (cond == 0) return (void *)pIhdr;
         if ((cond > 0) && (--iCollisionCount)) {
            pIhdr = (ItemHeader *)((char *)pIhdr + usItemSize);
         }else {
            break;
         }
      }
   }
   if (apXihdr) {                  // if the table is extendable
      int cond;
      ExtendedItemHeader *pXihdr = apXihdr[iHashValue];

      while (pXihdr) {
         cond = strcmp(pszVarName, pXihdr->ihdr.pszName);
         if (cond ==0 ) return (void *)&(pXihdr->ihdr);
         if (cond > 0) {
            pXihdr = pXihdr->next;
         }else {
            break;
         }
      }
   }
   return 0;
}

/*-------------------------------------------------------------------add_value-+
| Add a new member in the extended list.                                       |
| The collision list is kept in alphabetic order, to speed up the process.     |
|                                                                              |
|   ppszVarName: on input, points the address of the variable name             |
|                on output, points the address where the variable name         |
|                           has been internally stored                         |
+-----------------------------------------------------------------------------*/
void * StringizedHash::add_value(
   char **ppszVarName,
   unsigned int iSizeValue
)
{
   int cond = 1;
   ExtendedItemHeader **ppXihdr = apXihdr + compute_hash_value(*ppszVarName);

   while (
      (*ppXihdr) &&
      (cond = strcmp(*ppszVarName, (*ppXihdr)->ihdr.pszName), cond > 0)
   ) {
      ppXihdr = &(*ppXihdr)->next;
   }
   ExtendedItemHeader *pXihdr;
   if (cond) {            // Not found (warning: cond might still be > 0!)
      unsigned int iVarNameLength = 1 + strlen(*ppszVarName);
      pXihdr = (ExtendedItemHeader *)malloc(
         sizeof (ExtendedItemHeader) - sizeof (ItemHeader)
         + iSizeValue + iVarNameLength
      );
      pXihdr->ihdr.pszName = (char *)pXihdr + (
         sizeof(ExtendedItemHeader) - sizeof(ItemHeader) + iSizeValue
      );
      memcpy(pXihdr->ihdr.pszName, *ppszVarName, iVarNameLength);
      pXihdr->next = *ppXihdr;
      *ppXihdr = pXihdr;
      ++iCount;
   }else {
      pXihdr = *ppXihdr;
   }
   *ppszVarName = pXihdr->ihdr.pszName;  // return this to caller
   return &(pXihdr->ihdr);
}
/******************************************************************************/
