/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// hash.h
///  Declarations for hash
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _HASH_H
#define _HASH_H 1
//******************************************************************************

class Hash {                            // hsh - Hash is an abstract class
public:
   virtual void *search_value(const char *pszVarName) = 0;
   virtual void *add_value(
      char **ppszVarName, unsigned int iSizeValue
   ) = 0;
   virtual int count() = 0;
   virtual ~Hash();
};

inline Hash::~Hash() { /*...*/ }

class ExpandedHash : public Hash {      // Expnhash IS-A Hash
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   ExpandedHash(unsigned short usMod = 10);
   _export ~ExpandedHash();
   void *search_value(const char *pszVarName);
   void *add_value(char **ppszVarName, unsigned int iSizeValue);
   int count();
   void *gimme_value(const char *pszVarName, unsigned int iSizeValue);
   void traverse(
      boolean (*tvrs_function)(
         void *pvBase, const char *pszName, void *pvValue
      ),
      void *pvBase
   );
   unsigned int unload(
      unsigned short usItemSize,
      boolean isExtendable,
      void  **ppVoid
   );

private:
   /*--------------------+
   | Private definitions |
   +--------------------*/
   struct ExpnHashEntry {          // EHSH
      ExpnHashEntry *pEhshNext;    // For hashing: collision chain
      unsigned int iSize;          // Size of the variable value

      ExpnHashEntry(ExpnHashEntry **ppEhshNext); // constructor
      void * operator new(
         size_t size, const char * pszVarName, unsigned int iSizeValue
      );
      void operator delete(void *);
   };
   /*------------------+
   | Private variables |
   +------------------*/
   ExpnHashEntry **apEhshTable;   // Array of (ExpnHashEntry *)
   unsigned short usExpnMod;      // Number of (ExpnHashEntry *)
   int iCount;                    // Total count of entries

   /*------------------+
   | Private functions |
   +------------------*/
   unsigned int compute_hash_value(const char *pszVarName);
   char * name_of(const ExpnHashEntry *);
   void * value_of(const ExpnHashEntry *);
   const ExpnHashEntry * gimme_or_search_token(
      const char *pszVarName, unsigned int iSizeValue, boolean create_it
   );
};

/*-----------------------------------------------------------------------------+
|                                                                              |
+-----------------------------------------------------------------------------*/
class StringizedHash : public Hash {     // Sgzdhash IS-A Hash
   friend unsigned int ExpandedHash::unload(
      unsigned short usItemSize,
      boolean isExtendable,
      void  **ppVoid
   );
public:
   /*-------------------+
   | Public definitions |
   +-------------------*/
   struct ItemHeader {       // IHDR: This defines the header of each item
      char *pszName;         // name must come first
   };

   /*-----------------+
   | Public functions |
   +-----------------*/
   StringizedHash(void *pVoid, void *pvItemBase);
   ~StringizedHash();
   void *search_value(const char *pszVarName);
   void *add_value(char **ppszVarName, unsigned int iSizeValue);
   int   count();

private:
   /*--------------------+
   | Private definitions |
   +--------------------*/
   struct ExtendedItemHeader {    // XIHDR: This defines the extended header
      ExtendedItemHeader *next;   // next or 0 if none
      ItemHeader ihdr;            // name
   };

   union SgzdHashCompound;        // Forward declaration

   struct SgzdHashHeader {        // SHHDR:
   friend union SgzdHashCompound;
      unsigned short usMod;       // Nb of entries (modulo) is kept here
      unsigned short usItemSize;  // and the size of each item
      boolean isExtendable;       // TRUE if the shsh table can be extended
   };

   struct SgzdHashEntry {         // SHSH: Entry in the stringized hash array
   friend union SgzdHashCompound;
      union {
         ItemHeader *pIhdr;       // Address of the IHDR
         unsigned int ofsItem;    // offset (only when unloaded)
      };
      unsigned int iCollisionCount;
   };

   union SgzdHashCompound {      // This union allows a proper alignment
   public:
      SgzdHashHeader hdr;
      SgzdHashEntry  o[1];
   };

   /*------------------+
   | Private variables |
   +------------------*/
   SgzdHashEntry *aShsh;          // Array of SgzdHashEntry
   unsigned short usSgzdMod;      // Number of SgzdHashEntry
   unsigned short usItemSize;     // Size of an item
   ExtendedItemHeader **apXihdr;  // non NULL if the table is extendable
   int iCount;                    // Total count of entries

   /*------------------+
   | Private functions |
   +------------------*/
   unsigned int compute_hash_value(const char *pszVarName);
};

/*INLINE-----------------------------------------------------------------count-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline int ExpandedHash::count() { return iCount; }

/*INLINE-PRIVATE------------------------------------------------------value_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline void * ExpandedHash::value_of(const ExpnHashEntry * pEhsh) {
   return pEhsh? (void *)(pEhsh + 1) : 0;
}

/*INLINE-----------------------------------------------------------gimme_value-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline void * ExpandedHash::gimme_value(
   const char *pszVarName, unsigned int iSizeValue
)
{
   return value_of(
      gimme_or_search_token(pszVarName, iSizeValue, TRUE)
   );
}

/*INLINE----------------------------------------------------------search_value-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline void * ExpandedHash::search_value(
   const char *pszVarName
)
{
   return value_of(
      gimme_or_search_token(pszVarName, 0, FALSE)
   );
}

/*INLINE-----------------------------------------------------------------count-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline int StringizedHash::count() { return iCount; }

#endif /* _HASH_H *************************************************************/
