/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// modele.h
///  Declarations for modele
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _MODELE_H
#define _MODELE_H 1
//******************************************************************************

#include "bitarray.h"   // private to OpenModelGroup

#include "quantity.eh"  // enums

class OpenElement;
class ContentToken;
class ElementType;
class OpenElementStack;
class DataTagPattern;

struct AllowedElems {
   unsigned int iCountElem;   // count of elements
   ElementType  *pElem;       // Fence list - must have count entries
   boolean      *aisAllowed;  // TRUE == Allowed  count+1 entries! (0 is PCDATA)
};

enum e_StatusModele { // MS:
   NONE_MS    = 0, // (Undefined)
   MATCH_MS   = 1, // GI is the one expected next (saved on TCB)
   FINISH_MS  = 2, // All token have been checked (saved on TCB)
   INVALID_MS = 3, // GI invalid: not element end; no required GI (saved on TCB)
   MATCHX_MS,      // Current GI was an invalid exclusion
   REQRD_MS,       // Required GI must precede proposed GI
   REQRDX_MS,      // A required GI was an invalid exclusion
   EXCLUS_MS,      // GI is invalid: exclusion (minus exception)
   INCLUS_MS,      // GI is valid as an inclusion (plus exception)
   REPARSE_MS,     // Need to reparse current char (stow_dtcb)
   OOCDATA_MS,     // Out-of-context data
   OOCSTG_MS       // Out-of-context start_tag
};

/*-----------------------------------------------------------------------------+
| Modele class                                                                 |
+-----------------------------------------------------------------------------*/
class Modele {
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   Modele() {}                               // Default constructor
   Modele(const ElementType *pElem);          // Regular constructor

   e_StatusModele mdl_goto(
      const OpenElementStack& oelemstk,       // Stack of opened element
      const ElementType * const pElemGiven,   // Proposed (0: #PCDATA)
      const ElementType ** const ppElemReturned,
      const DataTagPattern ** const ppDtpReturned,
      int *piMexDepth = 0                    // 0, or where to output mex_level
   );
   e_StatusModele mdl_fini(
      const OpenElementStack & oelemstk,
      const ElementType ** const ppElemReturned,
      const DataTagPattern ** const ppDtpReturned
   );
   void mdl_edtg(OpenElement * const pOelemCurrent);
   void mdl_aelem(
      const OpenElementStack & oelemstk, AllowedElems *pAelem
   );

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   enum e_NextContext {
      NONE_NTK,
      AFTER_HIT_NTK,
      FLUSH_NTK
   };

   /*------------------+
   | Private Variables |
   +------------------*/
   class OpenModelGroup {        // GRP:
   public:
      OpenModelGroup() : hits(GRPCNT_MAX) {}
      BitArray hits;             // Hit bits
      const ContentToken *pCnttok; // This content token is a model group [128]
      unsigned int ix;           // Index of the current token in this group
   };
   OpenModelGroup aGrpStack[GRPLVL_MAX+1]; // Array of traversed group nodes
   int ixGrpCurrent;                       // Index of the current model group

   /*------------------+
   | Private Functions |
   +------------------*/
   static void open_content_token(OpenModelGroup * pGrp, int ixToken);
   static int next_unhit(OpenModelGroup * pGrp, unsigned int first);
   static boolean all_tokens_were_hit(
      OpenModelGroup * pGrp, boolean isOptionalExcluded
   );
   static boolean is_last_token(const OpenModelGroup * const pGrp);
   boolean is_optional_token(const boolean mhit);
   boolean is_required_token(const boolean mhit);
   static int check_exceptions(
      const ElementType * const pElem, const OpenElementStack & oelemstk
   );
   void next_token(
      OpenElement * const pOelemCurrent, const e_NextContext ntk
   );
   static void include_plus_exceptions(
      AllowedElems *pAelem, const OpenElementStack & oelemstk
   );
   static void exclude_minus_exceptions(
      AllowedElems *pAelem, const OpenElementStack & oelemstk
   );
};
#endif /* _MODELE_H **********************************************************/
