/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// notadcl.c
///  NOTATION declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "prsstate.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "notadcl.h"
#include "attldcl.h"
#include "errorh.h"

/*-----------------------------------------------------------------------parse-+
| Process NOTATION MD (Notation Markup Declaration)                            |
|                                                                              |
| Enter here when NOTATION has been decoded,                                   |
| Returns with curr_char (supposed to ) == MDC                                 |
|                                                                              |
|  [148] notation declaration = mdo, "NOTATION", ps+, notation name, ps+,      |
|                               notation identifier, ps*, mdc)                 |
|                                                                              |
|  [149] notation identifier = external identifier                             |
+-----------------------------------------------------------------------------*/
void NotationDcl::parse()
{
   pPrsstat->skip_inferable_required_ps();

   /*---------------------+
   | Decode notation name |
   +---------------------*/
   NotationBlock *pNb;
   char *pszName;                // safe place for the name
   {
      char name[NAMELEN_MAX+1];
      pSyntax->read_general_name(pEh, name);
      pszName = name;
      pNb = (NotationBlock *)expnhashNotation.add_value(
         &pszName, sizeof(NotationBlock)
      );
   }
   /*------------------------+
   | Get external identifier |
   +------------------------*/
   pPrsstat->skip_ps();
   ExternId *pExid = new ExternId(pPrsstat, pszName, NOTATION_XFC);

   /*----------------------------------------+
   | Install the notation in the hash tables |
   +----------------------------------------*/
   if (pNb->pExid) {                         // if duplicate
      errorh(DUPNAMNOT, pEh, ERH_E_SEV);     // tell user error
      delete pExid;                          // release exid
   }else {
      ++nbrof.defined_notation;              // count defined <!NOTATION
      nbrof.nxid_text += (pExid->system_id_length()+pExid->public_id_length());
      pNb->pExid = pExid;                    // tell Nb about our new xid
   }
}

/*---------------------------------------------------------------------pile_up-+
| Build NotationType tables                                                    |
|                                                                              |
| NotationTypes (NCB's) are built from the information found in `NB', a block  |
| used to gather the information at compile time.  NotationType's must be seen |
| as an ExternId for a notation, followed by an Attribute pointing the default |
| AVS.  It's also an entry in the hash table for notations.  It is filled from |
| the <!NOTATION MD and the <!ATTLIST #NOTATION MD.                            |
| No NotationType is really taken into account until the <!NOTATION is         |
| encountered.  The <!ATTLIST related to NOTATIONS that haven't been defined   |
| are thrown away.                                                             |
+-----------------------------------------------------------------------------*/
boolean NotationBlock::pile_up(
   StringKeeper& sk, const char *pszName,
   NotationType*& pNcbMP, const NotationType* const aNcbTable,
   AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable,
   Component*& pCompoMP, const Component* const aCompoTable
)
{
   if (!pExid) {                             // ATTLIST w/o NOTATION
      errorh(NONOTA, ERH_E_SEV, pszName);    // tell user
      if (--pAnode->iReferenceCount == 0) delete pAnode;
      return FALSE;                          // do not unload this hash entry
   }

   NotationType *pNcb = pNcbMP++;

   pNcb->exid.dislocate(sk, *pExid, pszName); // dislocated-copy
   delete pExid;                             // and free

   if (pAnode) {                             // NOTATION and ATTLIST
      pNcb->ixAttrep = pAnode->pile_up(
         pAttrepMP, aAttrepTable, pCompoMP, aCompoTable
      );
      if (--pAnode->iReferenceCount == 0) delete pAnode;
   }else {
      pNcb->ixAttrep = -1;                   // NOTATION w/o ATTLIST
   }
   /*--------------------------------------------------+
   | As the NotationBlock structures in core are still |
   | valid, use the room left by:                      |
   | - `pExid', to store the value of the              |
   |   NotationType index of this notation in          |
   |   the NotationType table,                         |
   | - `pAnode', to store the value of its             |
   |   matching index in the Attribute table.          |
   |                                                   |
   | This is a trick to later update the               |
   | EntityNotation associated to external             |
   | entities: NAD_FILL called by ENT_BUILD.           |
   +--------------------------------------------------*/
   // to set NotationType ptr in EntityNotation:
   ixNcb = diffptr(pNcbMP, aNcbTable) - 1;
   // to solve def data att:
   ixAttrep = pNcb->ixAttrep;
   return TRUE;
}
/******************************************************************************/
