/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// oestack.h
///  Declarations for oestack
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _OESTACK_H
#define _OESTACK_H 1
//******************************************************************************

#include "modele.h"     // part of the OpenElementStack class
#include "element.h"    // part of the OpenElementStack class

class OpenElementStackIterator;

/*-----------------------------------------------------------------------------+
| This stack must be seen more like an internal working stack,                 |
| reflecting a final (stable) state reached even before all events             |
| are reported.                                                                |
| The Open Element at the top of the stack is not always the                   |
| currently reported event -- it may be in advance if elements                 |
| were inferred.  Also some elements are not stacked                           |
| (for instance: EMPTY elements)                                               |
| However looking at the top of the stack:                                     |
|   pSrm always tells what the current active map is                           |
|   pElem tells what the new non-inferred element is/will be                    |
+-----------------------------------------------------------------------------*/
class OpenElement {               // OELEM: Entry in stack of opened elements
public:
   OpenElement(                   // Regular constructor
      const ElementType *pElem,
      const DataTagPattern *pDtp,
      const OpenElement *pOelemPrevious
   );
   Modele mdlCurrent;             // Model this element satisfies so far
   const ElementType *pElem;      // Element type for this tag
   const DataTagPattern *pDtp;    // Data tag pattern (if any)
   const ShortrefMap *pSrm;       // Shortref map (inherited, or modified)
   unsigned int msStatus    :2;   // MATCH, INVALID, FINISH (temp: NONE)
   boolean isNetEnabling    :1;   // Tag enables NET
   boolean isNetEnabled     :1;   // NET enabled now and/or before
   boolean isPex            :1;   // Tag was plus exception
   boolean isMhit           :1;   // Header model group was hit
   boolean isDataTagEnabled :1;   // DATATAG enabled now and/or before
   boolean isEmptyOrBad     :1;   // Tag has been omitted and empty content
   boolean isZero           :1;   // Recognize the "zero" (header) element
};

class OpenElementStack {          // OESTK
   friend OpenElementStackIterator;
public:

   /*-----------------+
   | Public Functions |
   +-----------------*/
   OpenElementStack();
   OpenElementStack(const OpenElementStack& source);
   _export ~OpenElementStack();
   void activate(ElementType *pElemBase, int iMaxEntries);
   OpenElementStack& operator=(const OpenElementStack&);
   OpenElement * operator[](int ix) const;  // get ix-th tcb (no lvalue)
   OpenElement * first() const;             // get first value
   OpenElement * push(const ElementType *pElem, const DataTagPattern *pDtp);
   OpenElement * push(const OpenElement *pSource);
   boolean pop();
private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   struct StackEntry {           // stack entry
      StackEntry(
          const ElementType *pElem,
          const DataTagPattern *pDtp,
          const OpenElement *pOelemPrevious
      ) : oelem(pElem, pDtp, pOelemPrevious) {}
      StackEntry(const OpenElement * pSource) :
          oelem(*pSource) {}     // use the default bitwise copy
      StackEntry *previous;
      OpenElement oelem;
   };

   /*------------------+
   | Private Variables |
   +------------------*/
   int iStkentMax;                     // maximum stack entries
   int ixStkentTop;                    // top entry
   StackEntry *pStkentTop;             // top entry
   ElementType elemZero;
   ContentToken cnttokZero;
};

class OpenElementStackIterator {
public:
   OpenElementStackIterator(const OpenElementStack& oestk);
   const OpenElement * prev();
private:
   const OpenElementStack::StackEntry * pStkent;
};
#endif /* _OESTACK_H **********************************************************/
