/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// parser.h
///  Declarations for parser and parserm
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _PARSER_H
#define _PARSER_H 1
/*__automake__
  Source "parserm.cpp";
*/
//******************************************************************************

#include "prsstate.h"   // private base (HAS-A relationship)
#include "event.h"      // part of SGMLparser

#include "element.eh"   // enums
#include "event.eh"

class Document;

class SGMLparser :
   private virtual ParseState  // USES-A (based on) relationship
{
   friend ParseState::ParseState(const SGMLparser *pParserSource);
public:
   // Create a new parser session.
   _export SGMLparser(const Document& doc);

   // Create another parser session from a given ParseState (to resume from)
   SGMLparser(const ParseState& prsctxt);

   // identify returns a string telling who we are
   static const char * const _export identify();

   // Report events once at a time
   e_Event _export event();

   // zstring() returns an ASCIIZ string, the contents of which is specific
   // to the event lastly reported
   const char * _export zstring();

   // display() shows the content of an event.
   // (this is effective only on primary events)
   void _export display();

   // enable or disable events
   SGMLparser& _export operator<<(e_Event);
   SGMLparser& _export operator>>(e_Event);

   // capacity() gets the value reached for each capacity name
   unsigned long _export capacity(e_IxCapacity ix);

#ifdef _PKHACKED 
   Event const *currentEvent () { return evlst.operator->() ; } ;
   EventList const &eventList () { return evlst ; } ;
#endif

private:
   EventList evlst;            // Primary event list

   boolean parse_prolog();
   boolean activate_base_document();

   boolean parse_instance();
   boolean process_data();
   void process_entity(EntityDef *pEnt);
   boolean process_Ee();
   void process_usedcl();

   void copy_context_into(ParseState& prsctxt) const;

   void start_base_element();
   boolean start_element();
   boolean prepare_for_data();
   void end_element(
      unsigned int iCountElementToEnd,
      e_InferBy inferby,
      e_Minimize minias
   );

   void check_events(const int iCountStart);
   void record_elem_start(
      const ElementType *pElem,
      const AttributeList &attlist,
      const DataTagPattern *pDtp,
      const e_InferBy inferby,
      const e_Minimize minias = NO_MNF,
      boolean isInclusion = FALSE,
      boolean isExplConref = FALSE
   );
   boolean record_elem_end(
      const ElementType **ppElemReturned,
      const DataTagPattern **ppDtpReturned,
      const e_InferBy inferby,
      const e_Minimize minias = NO_MNF
   );
   e_StatusModele validate_element_context(
      const ElementType *pElemFound,
      const DataTagPattern **ppDtpReturned,
      const e_InferBy inferby
   );
   int how_many_NET_end();
   int how_many_element_end();
};
#endif /* _PARSER_H ***********************************************************/
