/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// parserm.c
///  Parse an SGML document and report events (part 1 of 2)
//   Pierre G. Richard 93/10/05 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

//***>NOTE: Splitting the parser in two files is temporary (?)
//***>NOTE: Main purpose is to make my life easier while finishing testing.

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gendef.h"
#include "rsvname.h"
#include "parser.h"
#include "event.h"
#include "element.h"
#include "quantity.h"
#include "capacity.h"
#include "hash.h"
#include "enthdlr.h"
#include "prsmode.h"
#include "syntax.h"
#include "sgmldcl.h"
#include "dtddcl.h"
#include "procpil.h"
#include "utilctl.h"

static boolean tvrs_check_activation(void *, char *pszName);

/*--------------------------------------------------------------------identify-+
| Identify myself                                                              |
+-----------------------------------------------------------------------------*/
const char * const SGMLparser::identify()
{
   return
      "\t\tProject YAO Object SGML Parser (OSGML)\n"
      "\t\tVersion 1.0 Alpha level 4.0"
   ;
}

/*-----------------------------------------------------------------------event-+
| Report the next relevant event                                               |
+-----------------------------------------------------------------------------*/
e_Event SGMLparser::event()
{
   for (;;) {
      /*--------------------------------------------+
      | if the primary event list is empty, fill it |
      +--------------------------------------------*/
      while (!evlst.count()) {                   // until events to report
         switch (pst0) {
         case INITIALIZED_PST0:
            /*
            | About to parse the SGML declaration
            */
            if (!pEh->open(&entPrimary)) {
               errorh(ABORTED, ERH_T_SEV);
            }
            floor_level = pEh->curr_entity_level();    // No Ee allowed
            evlst.append(EntityStart(&entPrimary));
            pPrsmode->set_mode(INI_MODE);        // STAGO also recognized
            skip_S();
            if (
               (get_curr_token() == MDO_TK) &&   // MDO found
               (pSyntax->match_rsvname(pEh, SGML_RRN_ix)) // followed by "SGML"
            ) {                                  // variant (USER ?) OCS
               evlst.append(SgmlStart());        // Tell SGML found
               pst0 = SGMLDCL_STARTED_PST0;
            }else {                              // no <!SGML found
               activate_SGMLdcl(UTSYNL_XFC);     // Load the Local OCS
               pEh->unread(get_curr_count());    // unread the MDO or STAGO
               pst0 = SGMLDCL_DONE_PST0;
            }
            break;

         case SGMLDCL_STARTED_PST0:
            {
               pNewSgmldcl = new NewSGMLdcl(     // Construct a CS compiler
                  this, evlst
               );
               pNewSgmldcl->compile();
               delete pNewSgmldcl;
               evlst.append(SgmlEnd());
               activate_SGMLdcl(UTSYNU_XFC);     // Load the Variant OCS
               if (pSgmldcl->isScopeInstance) {  // if scope instance
                  activate_syntax(UTSYNR_XFC);   // get the reference syntax
               }
               pst0 = SGMLDCL_DONE_PST0;
               pEh->read();                      // change #5 in ocscomp.c
            }
            break;

         case SGMLDCL_DONE_PST0:
            /*
            | About to parse the prolog
            */
            pPrsmode->set_mode(INI_MODE);        // also to point the right dv
            pst0 = PROLOG_STARTED_PST0;
            pst1 = OTHER_PROLOG_PST1;
            pst2 = NOTHING_FOUND_PST2;
            break;

         case PROLOG_STARTED_PST0:
            if (!parse_prolog()) pst0 = PROLOG_DONE_PST0;
            break;

         case PROLOG_DONE_PST0:
            sstrlActiveType.traverse(tvrs_check_activation, 0); // final checks
            if (pSgmldcl->isScopeInstance) {     // if scope instance
               activate_syntax(UTSYNU_XFC);      // Activate variant syntax
            }
            pPrsmode->set_mode(CON_MODE);
            recognize_delimiter();
            if (!activate_base_document()) {
               errorh(NODTD, pEh, ERH_T_SEV);
               pst0 = INSTANCE_DONE_PST0;
               break;
            }
            if (
               (get_curr_token() != STAGO_TK) &&
               (get_curr_token() != STAGO_TAGC_TK)
            ) {
               start_base_element();
            }
            pst0 = INSTANCE_STARTED_PST0;
            pst1 = TAG_PST1;
            pst2 = EE_BEFORE_CLOSE_PST2;
            pEh->unread(get_curr_count()); //parse_instance recognizes delimiter
            break;

         case INSTANCE_STARTED_PST0:
            if (!parse_instance()) pst0 = INSTANCE_DONE_PST0;
            break;

         case INSTANCE_DONE_PST0:
            pEh->close();                  // This has to be done somewhere...
            pst0 = ALL_DONE_PST0;
            break;

         default:   // ALL_DONE_PST0
            return EOP_EV;
         }
         evlst.position(0);                // position at first event
      }
      /*---------------------------------------+
      | Empty the primary event list while     |
      | decomposing it into elementary events  |
      +---------------------------------------*/
      do {
         // Use downcasting... until "template" work better
         PrimEvent *pPrmev = (PrimEvent *)((Event *)evlst);
         EventList& evlstDcmp = pPrmev->evlst;

         if (!pPrmev->isDecomposed) {
            pPrmev->isDecomposed = TRUE;
            pPrmev->init_decompose();
            if (isEnabledEvent[(int)pPrmev->id]) {
               return pPrmev->id;
            }
         }
         while (evlstDcmp.count()) {
            DecomposedEvent *pDcmpev = (DecomposedEvent *)((Event *)evlstDcmp);
            e_Event id = pDcmpev->decompose(this, evlstDcmp);
            if (
               (id != EOP_EV) && (isEnabledEvent[(int)id])
            ) {
               return id;
            }
         }
         evlst.remove();         // remove this one,
         ++evlst;                // take next
      }while (evlst.count());
   }
}

/*---------------------------------------------------------------------zstring-+
| Return an ASCIIZ string specific to the current event                        |
+-----------------------------------------------------------------------------*/
const char * SGMLparser::zstring()
{
   Event *pEv = (Event *)evlst;
   if (pEv) {
      Event *pDcmpev = (((PrimEvent *)pEv)->evlst)[-1];
      if (pDcmpev) pEv = pDcmpev;
      return pEv->zstring();
   }
   return "???";
}

/*---------------------------------------------------------------------display-+
| Display a primary, not yet decomposed event                                  |
+-----------------------------------------------------------------------------*/
void SGMLparser::display()
{
   PrimEvent *pPrmev = (PrimEvent *)((Event *)evlst);
   if ((pPrmev) && (pPrmev->evlst.count() <= 1)) {  // Prim event not decomposed
      pPrmev->display(this);
   }
}

/*------------------------------------------------------------------operator<<-+
| Enable events                                                                |
+-----------------------------------------------------------------------------*/
SGMLparser& SGMLparser::operator<<(e_Event evid)
{
   if (evid == ALL_EV) {
      isEnabledEvent = TRUE;
   }else {
      isEnabledEvent += (int)evid;
   }
   return *this;
}

/*------------------------------------------------------------------operator>>-+
| Disable events                                                               |
+-----------------------------------------------------------------------------*/
SGMLparser& SGMLparser::operator>>(e_Event evid)
{
   if (evid == ALL_EV) {
      isEnabledEvent = FALSE;
   }else {
      isEnabledEvent -= (int)evid;
   }
   return *this;
}

/*--------------------------------------------------------------------capacity-+
| Return the value currently reached for a given capacity                      |
+-----------------------------------------------------------------------------*/
unsigned long SGMLparser::capacity(e_IxCapacity ix)
{
   return pSgmldcl->pCapacity->compute(ix, pDtddcl->aulCapacityCount);
}

/*------------------------------------------------------activate_base_document-+
|                                                                              |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::activate_base_document()
{
   ExternId *pExidOdt;
   UtilityIdentifier *pUtidOdt = utilctl.find_utid_base();

   if (pUtidOdt) {
      pExidOdt = &(pUtidOdt->exid);
   }else {
      /*----------------------------------------------------------+
      | This is a recovery procedure to guess a doctype when no   |
      | DOCTYPE declaration has been found.  If the document      |
      | instance starts with a start-tag, then assume the doctype |
      | is the generic identifier of this start-tag.              |
      +----------------------------------------------------------*/
      if (get_curr_token() != STAGO_TK) return FALSE;

      char pszName[NAMELEN_MAX + 1];

      pEh->unread(pSyntax->read_general_name(pEh, pszName)); // Get GI
      utilctl.new_utid(                 // create a minimal utid
         pEh, pSgmldcl, sstrlActiveType, TRUE, pszName, NOT_A_LINK_LKKIND
      );
      ExternId *pExidSource = new ExternId(pszName, DTD_XFC);
      pExidOdt = utilctl.build_compiled_exid(UTODTR_XFC, pExidSource);
      delete pExidSource;
      if (!pExidOdt->find(pEh)) return FALSE;
      errorh(NODTDI, pEh, ERH_E_SEV);   // Recovery OK, but not in ISO
   }
   {
      ApplicationHandle ahlAccess = open_ODT(pExidOdt, FALSE);
      pDtddcl = new DTDdcl(ahlAccess, pSyntax);   // load the ODT
      pEh->close(ahlAccess);
   }
   pHshGenEnt = pDtddcl->pHshGenEnt;              // Shadow
   pHshPrmEnt = pDtddcl->pHshPrmEnt;              // Shadow

   /*--------------------------+
   | Add the capacities, check |
   +--------------------------*/
   for (
      int i=0;
      i < sizeof aulCapacityCount / sizeof aulCapacityCount[0];
      ++i
   ) {
      aulCapacityCount[i] += (pDtddcl->aulCapacityCount)[i];
   }
   pSgmldcl->pCapacity->capa_chk(aulCapacityCount);

   /*-------------------------------+
   | init the stack of open element |
   +-------------------------------*/
   oelemstk.activate(
       pDtddcl->pElemBase,
       pSyntax->pQuantity->value_of(TAGLVL_QTY_ix)
   );
   return TRUE;    // done.
}

/*----------------------------------------------------------------parse_prolog-+
| [7] prolog = other prolog*, base document type declaration,                  |
|              (document type declaration | other prolog)*,                    |
|              (link type declaration | other prolog)*                         |
|                                                                              |
| [8] other prolog = comment declaration | processing instruction | s          |
|                                                                              |
| Returns FALSE when prolog has ended.                                         |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::parse_prolog()
{
   static const VarntChoiceEntry prolog_choices[] = {
      { DOCTYPE_VRN_ix,   DOCTYPE_TK   },
      { LINKTYPE_VRN_ix,  LINKTYPE_TK  },
      { NOT_FOUND_VRN_ix, NOT_FOUND_TK }
   };

   switch (pst1) {

   case DOCTYPE_STARTED_PST1:
      /*
      | if NOTHING_FOUND, this is the first DTD, i.e the base dtd
      */
      pPrsmode->push_mode(MD_MODE);
      pProcpil = new PrologCompiler(        // Construct a DTD compiler
         this, evlst, (pst2 == NOTHING_FOUND_PST2)? TRUE : FALSE
      );
      pst1 = DOCTYPE_PROCESSING_PST1;
      pst2 = DOCTYPE_FOUND_PST2;
      return TRUE;

   case DOCTYPE_PROCESSING_PST1:
      if (!pProcpil->compile()) pst1 = DOCTYPE_DONE_PST1;
      return TRUE;

   case DOCTYPE_DONE_PST1:
      delete pProcpil;
      pPrsmode->pop_mode();
      evlst.append(DoctypeEnd());
      pst1 = OTHER_PROLOG_PST1;
      return TRUE;

   case LINKTYPE_STARTED_PST1:
      pPrsmode->push_mode(MD_MODE);
      pProcpil = new PrologCompiler(this, evlst); // Construct a LTD compiler
      pst1 = LINKTYPE_PROCESSING_PST1;
      pst2 = LINKTYPE_FOUND_PST2;
      return TRUE;

   case LINKTYPE_PROCESSING_PST1:
      if (!pProcpil->compile()) pst1 = LINKTYPE_DONE_PST1;
      return TRUE;

   case LINKTYPE_DONE_PST1:
      delete pProcpil;
      pPrsmode->pop_mode();
      evlst.append(LinktypeEnd());
      pst1 = OTHER_PROLOG_PST1;
      return TRUE;

   default: // OTHER_PROLOG_PST1  --  Skip "other prologs"
      skip_S();
      switch (get_curr_token()) {
      case MDO_COM_TK:
         evlst.append(Comment());          // Comment skipped
         skip_cdc();                       // skip comment declaration
         return TRUE;
      case MDO_MDC_TK:                     // "short" comment
         evlst.append(Comment());
         return TRUE;
      case PIO_TK:                         // Processing Instruction
         evlst.append(ProcessingInstruction(pEh, pSyntax));
         return TRUE;
      case MDO_TK:
         switch (pSyntax->match_rsvname_in_list(pEh, prolog_choices)) {
         case DOCTYPE_TK:
            if (pst2 == LINKTYPE_FOUND_PST2) {  // DTD cannot follow a LPD
               errorh(INVMD, pEh, ERH_T_SEV);
            }
            evlst.append(DoctypeStart());
            pst1 = DOCTYPE_STARTED_PST1;
            return TRUE;
         case LINKTYPE_TK:
            if (pst2 == NOTHING_FOUND_PST2) {  // LPD: one DTD must preceed
               errorh(INVMD, pEh, ERH_T_SEV);
            }
            evlst.append(LinktypeStart());
            pst1 = LINKTYPE_STARTED_PST1;
            return TRUE;
         default:
            /*
            | Quiet: this can be a MD allowed in document instance
            */
            break;
         }
         break;
      default:                              // NOT_FOUND, STAGO, combined STAGO
         break;
      }
      break;

   }
   pEh->unread(get_curr_count());          // Unread STAGO_x, or MDO..
   return FALSE;                           // done.
}

/*STATIC-------------------------------------------------tvrs_check_activation-+
| Traversal of the `ActiveType' string list, to find out if all demand of LINK |
| activation were answered.   Free file positions.                             |
+-----------------------------------------------------------------------------*/
static boolean tvrs_check_activation(void *, char *pszName)
{
   errorh(INVALINK, ERH_E_SEV, pszName);   // unanswered
   //***NOTE: file positions, if any, must be freed! (not done)
   return TRUE;                            // remove entry, anyway
}

/******************************************************************************/
