/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// procpil.c
///  Prolog compiler
//   Pierre G. Richard 88/03/29 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <new.h>
#include <assert.h>

#include "quantity.eh"

#include "gendef.h"
#include "enthdlr.h"
#include "prsmode.h"
#include "event.h"
#include "sgmldcl.h"
#include "shortref.h"
#include "dtddcl.h"
#include "notation.h"
#include "capacity.h"
#include "quantity.h"
#include "prsstate.h"
#include "rsvname.h"
#include "procpil.h"

/*--------------------------------------------------------------PrologCompiler-+
| Constructor: prepare to compile a DTD                                        |
+-----------------------------------------------------------------------------*/
PrologCompiler::PrologCompiler(
   ParseState *pPrsstatArg, EventList& evlstArg, boolean isBaseDtdArg
) :
   PrologDcl(pPrsstatArg),
   compile_what(&PrologCompiler::compile_DTD),
   evlst(evlstArg),
   utilctl(pPrsstatArg->utilctl)
{
   pPrsstatArg->pHshGenEnt = &expnhashGenEnt;
   pPrsstatArg->pHshPrmEnt = &expnhashPrmEnt;
   isBaseDtd = isBaseDtdArg;
   cst0 = INITIALIZED_CST0;
}

/*--------------------------------------------------------------PrologCompiler-+
| Constructor: prepare to compile a LTD                                        |
+-----------------------------------------------------------------------------*/
PrologCompiler::PrologCompiler(
   ParseState *pPrsstatArg, EventList& evlstArg
) :
   PrologDcl(pPrsstatArg),
   compile_what(&PrologCompiler::compile_LTD),
   evlst(evlstArg),
   utilctl(pPrsstatArg->utilctl)
{
   pPrsstatArg->pHshGenEnt = &expnhashGenEnt;
   pPrsstatArg->pHshPrmEnt = &expnhashPrmEnt;
   isBaseDtd = FALSE;
   cst0 = INITIALIZED_CST0;
}

/*-------------------------------------------------------------~PrologCompiler-+
| Destructor                                                                   |
| Zero the hash tables in ParseState; in case no re-initialization from the    |
| ODT/OLT is done, the error will be made obvious.                             |
+-----------------------------------------------------------------------------*/
PrologCompiler::~PrologCompiler()
{
   pPrsstat->pHshGenEnt = 0;
   pPrsstat->pHshPrmEnt = 0;
}

/*-----------------------------------------------------------------------setup-+
| Setup an appropriate environment before to compile a DTD or LTD.             |
|                                                                              |
|  - "pExidSourceArg" points to a working ExternId structure created by the    |
|    caller.                                                                   |
|                                                                              |
|  - "txtclsAsgn" is the assigned class (either DTD_XFC, or LPD_XFC)           |
|                                                                              |
|  - "lkkind" is the kind of link: NOT_A_LKKIND for a DTD,                     |
|         SIMPLE_LKKIND, IMPLICIT_LKKIND, or EXPLICIT_LKKIND for a LTD         |
|                                                                              |
|  - "txtclsCompiled" is the class used to look for a compiled form            |
|    either: UTODTR_XFC, or UTOLTR_XFC, or INVALID_XFC if a compiled           |
|    form cannot be used.                                                      |
|                                                                              |
+-----------------------------------------------------------------------------*/
void PrologCompiler::setup(
   ExternId *pExidSourceArg,
   e_TextClass txtclsAsgn,
   e_LinkKind lkkindArg,
   e_TextClass txtclsCompiledArg
) {
   static const VarntChoiceEntry dtss_choices[] = {
      { ENTITY_VRN_ix,    ENTITY_TK    },
      { ELEMENT_VRN_ix,   ELEMENT_TK   },
      { ATTLIST_VRN_ix,   ATTLIST_TK   },
      { NOTATION_VRN_ix,  NOTATION_TK  },
      { SHORTREF_VRN_ix,  SHORTREF_TK  },
      { USEMAP_VRN_ix,    USEMAP_TK    },
      { NOT_FOUND_VRN_ix, NOT_FOUND_TK }
   };
   static VarntChoiceEntry const ltss_choices[] = {
      { ATTLIST_VRN_ix,   ATTLIST_TK   },
      { ENTITY_VRN_ix,    ENTITY_TK    },
      { LINK_VRN_ix,      LINK_TK      },
      { IDLINK_VRN_ix,    IDLINK_TK    },
      { NOT_FOUND_VRN_ix, NOT_FOUND_TK }
   };

   if (txtclsAsgn == DTD_XFC) {
      specific_choices = dtss_choices;
   }else {
      specific_choices = ltss_choices;
   }
   lkkind = lkkindArg;
   txtclsCompiled = txtclsCompiledArg;
   pExidSource = pExidSourceArg;

   cst1 = INITIALIZED_CST1;
   isRealDeclaration = FALSE;
}

/*-----------------------------------------------------------------compile_DTD-+
| Compile a DTD                                                                |
| Enter here when DOCTYPE has been decoded.                                    |
|                                                                              |
| [110] document type declaration =                                            |
|       mdo, "DOCTYPE", ps+, document type name,                               |
|       (ps +, external identifier)?,                                          |
|       (ps +, dso, document type declaration subset, dsc)?,                   |
|       ps*, mdc                                                               |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::compile_DTD()
{
   switch (cst0) {
   case INITIALIZED_CST0:
      pPrsstat->skip_specific_ps();
      {
         char pszNameType[NAMELEN_MAX + 1];

         pSyntax->read_general_name(pEh, pszNameType);
         pUtidOdt = utilctl.new_utid(
            pEh, pSgmldcl, pPrsstat->sstrlActiveType, isBaseDtd,
            pszNameType, NOT_A_LINK_LKKIND
         );
      }
      if (!pUtidOdt) errorh(ABORTED, ERH_T_SEV);
      pGiDocElement = (GenericIdBlock *)expnhashElem.gimme_value(
         pUtidOdt->exid.name(), sizeof(GenericIdBlock)
      );
      // Since CONCUR is not implemented, we will suppose pUtidOdt->isActive
      if (
         (pSgmldcl->features.isSimple) ||
         (pSgmldcl->features.isImplicit) ||
         (pSgmldcl->features.isExplicit)
      ) {
         pUtidOdt->position = pEh->get_pos(); // note the position of the DTD
      }
      pExidDtd = new ExternId(pUtidOdt->exid.name(), DTD_XFC);
      setup(                                   // setups to compile the DTD
         pExidDtd,
         DTD_XFC,
         pUtidOdt->lkkind,                     // NOT_A_LKKIND
         UTODTR_XFC
      );
      cst0 = COMPILING_DTD_CST0;
      /* fall thru */

   case COMPILING_DTD_CST0:
      if (compile_subset_and_external()) return TRUE;
      if (isRealDeclaration) {
         unload(pExidDtd, UTODTW_XFC);
      }
      delete pExidDtd;
      return FALSE;

   default:       // Never hit
      break;
   }
   return FALSE;  // Never hit
}

/*-----------------------------------------------------------------compile_LTD-+
| Compile a LTD                                                                |
| Enter here when LINKTYPE has been decoded.                                   |
|                                                                              |
| [154] link type declaration =                                                |
|                     mdo, "LINKTYPE", ps+, link type name, ps+,               |
|                     simple link specification |                              |
|                     implicit link specification |                            |
|                     explicit link specification ),                           |
|                     (ps +, external identifier)?,                            |
|                     (ps +, dso, link type declaration subset, dsc)?,         |
|                     ps*, mdc                                                 |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::compile_LTD()
{
   switch (cst0) {
   case INITIALIZED_CST0:
      pPrsstat->skip_specific_ps();
      {
         char pszNameType[NAMELEN_MAX + 1];

         pSyntax->read_general_name(pEh, pszNameType);
         pUtidOlt = utilctl.new_utid(
            pEh, pSgmldcl, pPrsstat->sstrlActiveType, FALSE,
            pszNameType, link_specification()
         );
      }
      if (!pUtidOlt) errorh(ABORTED, ERH_T_SEV);
      if (pUtidOlt->isActive) {
         pUtidOlt->position = pEh->get_pos(); // note the position of the LTD
      }
      pUtidOdt = utilctl.find_utid_base();
      pGiDocElement = (GenericIdBlock *)expnhashElem.gimme_value(
         pUtidOdt->exid.name(), sizeof(GenericIdBlock)
      );
      pExidLpd = new ExternId(pUtidOlt->exid.name(), LPD_XFC);
      setup(                                   // setups to compile a LPD
         pExidLpd,
         LPD_XFC,
         pUtidOlt->lkkind,                     // SIMPLE, IMPLICIT, EXPLICIT
         UTOLTR_XFC
      );
      cst0 = COMPILING_LTD_CST0;
      /* fall thru */

   case COMPILING_LTD_CST0:                    // compile the LTD
      if (compile_subset_and_external()) return TRUE;
      if ((!isRealDeclaration) || (!pUtidOlt->isActive)) {
         cst0 = ENDED_CST0;
         return TRUE;
      }
      position = pEh->get_pos();               // position after the LTD
      pEh->set_pos(pUtidOdt->position);        // seek to the DTD source
      pExidDtd = new ExternId(pUtidOdt->exid.name(), DTD_XFC);
      setup(                                   // setups to recompile the DTD
         pExidDtd,
         DTD_XFC,
         pUtidOdt->lkkind,                     // NOT_A_LINK
         INVALID_XFC                           // force to recompile the DTD
      );
      cst0 = COMPILING_DTD_CST0;
      /* fall thru */

   case COMPILING_DTD_CST0:                    // compile the DTD
      if (compile_subset_and_external()) return TRUE;
      if (isRealDeclaration) {
         unload(pExidLpd, UTOLTW_XFC);
//***>NOTE: Not yet done
//*   }else {
//*      errorh: CANNOT READ THE RELEVANT DTD
      }
      delete pExidDtd;
      pEh->set_pos(position);                 // seek back to previous pos
      free(position);
      /* fall thru */

   case ENDED_CST0:
      delete pExidLpd;
      return FALSE;

   default:       // Never hit
      break;
   }
   return FALSE;  // Never hit
}

/*-------------------------------------------------compile_subset_and_external-+
| This function processes the part of a declaration made of a file             |
| and a subset (DTD or LTD's):                                                 |
| PrologCompiler::setup must have been called first.                           |
|                                                                              |
|  (ps +, external identifier)?, (ps +, dso, subset, dsc)?, ps*, mdc           |
|                                                                              |
| (i.e. this is what follows:                                                  |
|       mdo, "DOCTYPE | LINKTYPE", ps+, document/link type name)               |
|                                                                              |
| The subset is parsed first, then the file.                                   |
|                                                                              |
| TRUE is returned until compile ended.                                        |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::compile_subset_and_external()
{
   switch (cst1) {

   case INITIALIZED_CST1:
      pPrsstat->skip_specific_ps();
      switch (pPrsstat->get_curr_token()) {
      case NOT_FOUND_TK:                // external ID: decode it
         new(pExidSource) ExternId(     // fill_up pExidSource; skip_ps done
            pPrsstat, pExidSource->name(), pExidSource->assigned_class()
         );
         if (pPrsstat->get_curr_token() != DSO_TK) {
            cst1 = PARSE_SUBSET_DONE_CST1;
         }else {
            cst1 = PARSE_SUBSET_STARTED_CST1;
            cst2 = STARTING_CST2;
         }
         break;
      case DSO_TK:
         cst1 = PARSE_SUBSET_STARTED_CST1;
         cst2 = STARTING_CST2;
         break;
      default:
         cst1 = PARSE_SUBSET_DONE_CST1;
         break;
      }
      return TRUE;

   case PARSE_SUBSET_STARTED_CST1:
      if (!compile_declaration_set()) {
         if (pPrsstat->get_curr_token() != DSC_TK) {
            errorh(NODSC, pEh, ERH_E_SEV);
         }else {
            pPrsstat->skip_specific_ps();
         }
         cst1 = PARSE_SUBSET_DONE_CST1;
      }
      return TRUE;

   case PARSE_SUBSET_DONE_CST1:
      if (pPrsstat->get_curr_token() != MDC_TK) errorh(NOMDC, pEh, ERH_T_SEV);

      if (!isRealDeclaration) {                // no subset found
         if (txtclsCompiled != INVALID_XFC) {  // not forced compilation
            if (utilctl.build_compiled_exid(
                  txtclsCompiled, pExidSource
               )->find(pEh)
            ) {
               return FALSE;     // Not processed, but a compiled form exists
            }
         }
      }else {
         if (!pExidSource->is_identified()) {
            return FALSE; // subset was processed and there is no external id
         }
      }
      /*---------------------------------------------+
      | IF                                           |
      |    no subset has been found                  |
      |      AND                                     |
      |        cannot use compiled form  (forced)    |
      |        OR                                    |
      |        no compiled form exists,              |
      |    OR                                        |
      |    subset to be combined to an external file |
      | THEN                                         |
      |    Parse the source DTD                      |
      +---------------------------------------------*/
      pEntSource = new EntityDef(
         pExidSource,
         pExidSource->assigned_class() == DTD_XFC?
            DOCTYPE_RRN_ix : LINKTYPE_RRN_ix
      );
      if (!pEh->open(pEntSource)) {
         return FALSE;
      }
      cst1 = PARSE_EXTERNAL_STARTED_CST1;
      cst2 = STARTING_CST2;
      return TRUE;

   case PARSE_EXTERNAL_STARTED_CST1:
      if (!compile_declaration_set()) {
         if (!pEh->is_Ee()) errorh(NOEOF, pEh, ERH_E_SEV);
         pEh->close();
         delete pEntSource;
         return FALSE;
      }
      return TRUE;

   default:       // Never hit
      break;
   }
   return FALSE;  // Never hit
}

/*-----------------------------------------------------compile_declaration_set-+
| Process each individual declaration found within a DTD or LTD                |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::compile_declaration_set()
{
   switch (cst2) {

   case STARTING_CST2:
      init_ent_level  = pEh->curr_entity_level(); // remember where we started
      init_floor_level = pPrsstat->floor_level;    // save current floor_level
      pPrsstat->floor_level = init_ent_level;     // set new floor level
      pPrsmode->push_mode(DS_MODE);
      cst2 = STARTED_CST2;
      /* fall thru */

   case STARTED_CST2:
      /*----------------------------------------------------+
      | Skip declaration separator (was skip_ds)            |
      |                                                     |
      | [70] ts = s | Ee | parameter entity reference       |
      | [71] ds = s | Ee | parameter entity reference       |
      |      | comment declaration | processing instruction |
      |      | marked section declaration                   |
      +----------------------------------------------------*/
      {
         EntityDef *pEntPI = 0;

         pPrsstat->isQuietAtFloor = TRUE;  // allow an Ee to occur
         pPrsstat->skip_ts(&pEntPI);
         pPrsstat->isQuietAtFloor = FALSE;
         if (pEntPI) {
            evlst.append(EntityStart(pEntPI));
            evlst.append(EntityEnd());
            return TRUE;
         }
      }
      switch (pPrsstat->get_curr_token()) {
      case MDO_COM_TK:
         evlst.append(Comment());          // Comment skipped
         pPrsstat->skip_cdc();             // skip comment declaration
         return TRUE;
      case MDO_MDC_TK:                     // "short" comment
         evlst.append(Comment());
         return TRUE;
      case PIO_TK:                         // Processing Instruction
         evlst.append(ProcessingInstruction(pEh, pSyntax));
         return TRUE;
      case MDO_DSO_TK:                     // marked section declaration
         if (!pPrsstat->marked_section_open()) {
            evlst.append(Error());
            pPrsstat->marked_section_ignore(); // MSC_MDC reached
         }else switch (pPrsmode->get_mode()) {
         case IGNMS_MODE:
            evlst.append(SectionStart(SectionStart::IGNORE));
            pPrsstat->marked_section_ignore(); // MSC_MDC reached
            evlst.append(SectionEnd());
            pPrsstat->marked_section_close();
         default:                          // INCLUDE or TEMP (no FCDMS, RCDMS)
            evlst.append(SectionStart(SectionStart::INCLUDE));
         }
         return TRUE;
      case MSC_MDC_TK:                     // ]]>
         pPrsstat->marked_section_close();
         return TRUE;
      case MDO_TK:                         // <!
         isRealDeclaration = TRUE;
         cst2 = PROCESSING_CST2;
         break;                            // no needs to return: can fall thru!
      default:                             // DSC, NOT_FOUND
         cst2 = DONE_CST2;
         return TRUE;                      // sdsrc is set
      }
      /* fall thru */

   case PROCESSING_CST2:
      pPrsstat->floor_level = pEh->curr_entity_level(); // set floor_level
      pPrsmode->push_mode(MD_MODE);
      switch (pSyntax->match_rsvname_in_list(pEh, specific_choices)) {
      case ENTITY_TK:                      // DOCTYPE and LINKTYPE
         EntityDcl::parse();
         break;
      case ELEMENT_TK:                     // DOCTYPE only
         ElementDcl::parse();
         break;
      case ATTLIST_TK:                     // DOCTYPE and LINKTYPE
         AttlistDcl::parse(lkkind);
         break;
      case NOTATION_TK:                    // DOCTYPE only
         NotationDcl::parse();
         break;
      case SHORTREF_TK:                    // DOCTYPE only
         ShortrefDcl::parse();
         break;
      case USEMAP_TK:                      // DOCTYPE only
         UsemapDcl::parse();
         break;
      case LINK_TK:                        // LINKTYPE only
         if (lkkind == SIMPLE_LKKIND) {
            errorh(INVKWD, pEh, ERH_T_SEV);
         }
         LinkDcl::parse();
         break;
      case IDLINK_TK:                      // LINKTYPE only
         if (lkkind == SIMPLE_LKKIND) {
            errorh(INVKWD, pEh, ERH_T_SEV);
         }
         IdLinkDcl::parse();
         break;
      default:
         errorh(INVKWD, pEh, ERH_T_SEV);
         break;
      }
      if (pPrsstat->get_curr_token() != MDC_TK) {
         errorh(NOMDC, pEh, ERH_T_SEV);
      }
      pPrsmode->pop_mode();
      if (pEh->curr_entity_level() != pPrsstat->floor_level) {
         errorh(INVHNEST, pEh, ERH_T_SEV);
      }
      pPrsstat->floor_level = init_ent_level;   // restore floor_level
      cst2 = STARTED_CST2;
      return TRUE;

   case DONE_CST2:
      pPrsmode->pop_mode();
      if (pEh->curr_entity_level() != init_ent_level) {
         errorh(INVHNEST, pEh, ERH_T_SEV);
      }
      if (pPrsstat->ms_level) {
         errorh(UBLNMSOD, pEh, ERH_E_SEV);
         pPrsstat->ms_level = 0;
      }
      pPrsstat->floor_level = init_floor_level;  // restore floor_level;
      return FALSE;
   default:       // Never hit
      break;
   }
   return FALSE;  // Never hit
}

/*----------------------------------------------------------link_specification-+
| [156] simple link specification   = rni, "SIMPLE", ps+, rni, "IMPLIED"       |
|                                                                              |
| [157] implicit link specification =                                          |
|                     source document type name, ps+, rni, "IMPLIED"           |
|                                                                              |
| [158] explicit link specification =                                          |
|                     source document type name, ps+,                          |
|                     result document type name                                |
+-----------------------------------------------------------------------------*/
e_LinkKind PrologCompiler::link_specification()
{
   e_LinkKind lkkind;

   pPrsstat->skip_specific_ps();
   switch (pPrsstat->get_curr_token()) {
   case RNI_TK:                         // must be a simple link...
      pSyntax->match_rsvname_or_die(pEh, SIMPLE_VRN_ix);
      pPrsstat->skip_specific_ps();
      pSyntax->match_rsvname_or_die(pEh, IMPLIED_VRN_ix);
      lkkind = SIMPLE_LKKIND;
      break;
   case NOT_FOUND_TK:                   // Explicit or Implicit
      {
         char  pszNameSourceDT[NAMELEN_MAX+1];
         char  pszNameResultDT[NAMELEN_MAX+1];

         pSyntax->read_general_name(pEh, pszNameSourceDT);
         pPrsstat->skip_specific_ps();
         switch (pPrsstat->get_curr_token()) {
         case RNI_TK:                   // must be an implicit link...
            pSyntax->match_rsvname_or_die(pEh, IMPLIED_VRN_ix);
            lkkind = IMPLICIT_LKKIND;
            break;
         case NOT_FOUND_TK:             // must be an explicit link
            pSyntax->read_general_name(pEh, pszNameResultDT);
            lkkind = EXPLICIT_LKKIND;
            break;
         default:
            errorh(INVDLM, pEh, ERH_T_SEV);
            break;
         }
      }
      break;
   default:
      errorh(INVDLM, pEh, ERH_T_SEV);
      break;
   }
   return lkkind;
}

/*-------------------------------------------------tvrs_check_declared_element-+
| This procedure is called while traversing the hash table of elements         |
| Check if elements have all been declared.                                    |
| This is done before creating and filling the ElementType array, so we can    |
| add/delete(?) ElementType structures with minimized troubles...              |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_check_declared_element(
   void *pvBase, const char *pszName, void *pvValue
)
{
   GenericIdBlock * pGi = (GenericIdBlock *)pvValue;

   if (!pGi->pEnode) {
      errorh(INVELEMREF, ERH_I_SEV, pszName);
      pGi->pEnode = new ElementNode();       // Default is O O ANY
      /*
      | To denote it is generated, pEnode->is.pcdata is not set to TRUE!
      */
   }else {
      if (
         (pGi->iIdentElem == 0) &&
         (pGi != ((PrologCompiler *)pvBase)->pGiDocElement)
      ) {
         errorh(NOGIREF, ERH_I_SEV, pszName);
      }
   }
   pGi->ixElem = (((PrologCompiler *)pvBase)->ixElemMP)++;
   return TRUE;
}

/*----------------------------------------------------------tvrs_pile_up_ranks-+
| This procedure is called while traversing the hash table of stems            |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_pile_up_ranks(
   void *pvBase, const char *pszName, void *pvValue
)
{
   PrologCompiler * pProcpil = (PrologCompiler *)pvBase;

   return ((RankNode *)pvValue)->pile_up(
      pProcpil->sk, pszName,
      pProcpil->pStsMP, pProcpil->aStsTable,
      pProcpil->pRklMP, pProcpil->aRklTable
   );
}

/*------------------------------------------------------tvrs_pile_up_notations-+
| This procedure is called while traversing the hash table of NotationBlocks   |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_pile_up_notations(
   void *pvBase, const char *pszName, void *pvValue
)
{
   PrologCompiler * pProcpil = (PrologCompiler *)pvBase;

   return ((NotationBlock *)pvValue)->pile_up(
      pProcpil->sk, pszName,
      pProcpil->pNcbMP, pProcpil->aNcbTable,
      pProcpil->pAttrepMP, pProcpil->aAttrepTable,
      pProcpil->pCompoMP, pProcpil->aCompoTable
   );
}

/*-------------------------------------------------------tvrs_pile_up_entities-+
| This procedure is called while traversing the hash table of Entities         |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_pile_up_entities(
   void *pvBase, const char *pszName, void *pvValue
)
{
   PrologCompiler * pProcpil = (PrologCompiler *)pvBase;

   return ((EntityDef *)pvValue)->pile_up(
      pProcpil->sk, pszName,
      pProcpil->expnhashNotation,
      pProcpil->ofsBufferExidEntDefault,
      pProcpil->ofsLitEntDefault,
      pProcpil->ixNadEntDefault,
      pProcpil->pEntMP, pProcpil->aEntTable,
      pProcpil->pNadMP, pProcpil->aNadTable,
      pProcpil->pCompoMP, pProcpil->aCompoTable,
      pProcpil->pAttrepMP, pProcpil->aAttrepTable,
      pProcpil->pExidMP, pProcpil->aExidTable
   );
}

/*------------------------------------------------------tvrs_pile_up_shortrefs-+
| This procedure is called while traversing the hash table of shortref maps    |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_pile_up_shortrefs(
   void *pvBase, const char *pszName, void *pvValue
)
{
   PrologCompiler * pProcpil = (PrologCompiler *)pvBase;

   return ((ShortrefNode *)pvValue)->pile_up(
      pProcpil->sk, pszName, pProcpil->pShortref->count(),
      pProcpil->expnhashGenEnt, pProcpil->pSrmMP, pProcpil->aSrmTable
   );
}

/*------------------------------------------------------------tvrs_pile_up_gis-+
| This procedure is called while traversing the hash table of GenericIdBlock   |
+-----------------------------------------------------------------------------*/
boolean PrologCompiler::tvrs_pile_up_gis(
   void *pvBase, const char *pszName, void *pvValue
)
{
   PrologCompiler * pProcpil = (PrologCompiler *)pvBase;

   return ((GenericIdBlock *)pvValue)->pile_up(
      pProcpil->sk, pszName, pProcpil->aElemTable,
      pProcpil->pCnttokMP, pProcpil->aCnttokTable,
      pProcpil->pAttrepMP, pProcpil->aAttrepTable,
      pProcpil->pCompoMP, pProcpil->aCompoTable,
      pProcpil->pExMP, pProcpil->aExTable,
      pProcpil->pDtpMP, pProcpil->aDtpTable,
      pProcpil->aRklTable
   );
}

/*INLINE--------------------------------------------unload_and_write_hashtable-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline void PrologCompiler::unload_and_write_hashtable(
   ApplicationHandle ahlAccess,
   ExpandedHash& expnhash,
   int iSizeItem,
   boolean isExtendable
)
{
   void *pVoid;
   int iSizeTotal;

   iSizeTotal = expnhash.unload(
      (unsigned short)iSizeItem, isExtendable, &pVoid
   );
   pEh->write_record(ahlAccess, pVoid, iSizeTotal);
   free(pVoid);
}

/*----------------------------------------------------------------------unload-+
|Stringize and unload the ODT or OLT structure to an utility file              |
+-----------------------------------------------------------------------------*/
void PrologCompiler::unload(ExternId *pExid, e_TextClass clsOxt)
{
   int ixDocElement;           // index in ElementType array
   char *achCnttokTable;
   char *achElemTable;
   char *achExTable;
   char *achAttrepTable;
   char *achCompoTable;
   char *achEntGenTable;
   char *achEntPrmTable;
   char *achSrmTable;
   char *achDtpTable;
   char *achNadTable;
   char *achNcbTable;
   char *achExidTable;
   char *achRklTable;
   char *achStsTable;
   unsigned long cap_count[CAP_MAX_SIZE];
   const int iCountGenEnt = expnhashGenEnt.count();
   const int iCountPrmEnt = expnhashPrmEnt.count();
   const int iCountElem = expnhashElem.count();
   const int iCountStem = expnhashStem.count();
   const int iCountShortref = pShortref->count();
   const int iCountSrMap = expnhashSrMap.count();

   ixElemMP = 0;
   expnhashElem.traverse(tvrs_check_declared_element, this);
   ixDocElement = pGiDocElement->ixElem;

   /*-----------------------+
   |Save the capacity count |
   +-----------------------*/
   cap_count[ENTCAP_ix]   = iCountGenEnt + iCountPrmEnt;
   cap_count[ENTCHCAP_ix] = nbrof.g_text + nbrof.p_text + nbrof.exid_text;
   cap_count[ELEMCAP_ix]  = iCountElem;
   cap_count[GRPCAP_ix]   = nbrof.model + nbrof.useless_model;
   cap_count[MAPCAP_ix]   = (iCountSrMap-1) * (1+iCountShortref);
   cap_count[NOTCAP_ix]   = nbrof.defined_notation;
   cap_count[NOTCHCAP_ix] = nbrof.nxid_text;
   cap_count[ATTCAP_ix]   = nbrof.attributes;
   cap_count[AVGRPCAP_ix] = nbrof.att_dclv;
   cap_count[ATTCHCAP_ix] = nbrof.dclv_text;
   cap_count[EXGRPCAP_ix] = nbrof.exceptions;
   cap_count[EXNMCAP_ix]  = nbrof.excptnames;
   cap_count[DTACAP_ix]   = nbrof.datatag;       // doesn't exist in ISO
   cap_count[DTACHCAP_ix] = nbrof.datatag_text;  // doesn't exist in ISO
   cap_count[IDCAP_ix]    = 0;                   // irrelevant for DTD compile
   cap_count[IDREFCAP_ix] = 0;                   // irrelevant for DTD compile
   cap_count[LKNMCAP_ix]  = 0;                   // NOT YET IMPLEMENTED
   cap_count[LKSETCAP_ix] = 0;                   // NOT YET IMPLEMENTED
   cap_count[TOTALCAP_ix] = 0;                   // irrelevant
   pSgmldcl->pCapacity->capa_chk(cap_count);

   achCnttokTable = new char[nbrof.model * sizeof (ContentToken)];
   pCnttokMP = aCnttokTable = (ContentToken *)achCnttokTable;
   if (!achCnttokTable) errorh(BIGALLOC, ERH_T_SEV, MDL_BGAL);

   achElemTable = new char[iCountElem * sizeof(ElementType)];
   aElemTable = (ElementType *)achElemTable;
   if (!achElemTable) errorh(BIGALLOC, ERH_T_SEV, ELEM_BGAL);

   achExTable = new char[
      (nbrof.exceptions + nbrof.excptnames) * sizeof(Exception)
   ];
   pExMP = aExTable = (Exception *)achExTable;
   if (!achExTable) errorh(BIGALLOC, ERH_T_SEV, EXC_BGAL);

   achAttrepTable = new char[nbrof.attributes * sizeof(AttributeRep)];
   pAttrepMP = aAttrepTable = (AttributeRep *)achAttrepTable;
   if (!achAttrepTable) errorh(BIGALLOC, ERH_T_SEV, ATT_BGAL);

   achCompoTable = new char[nbrof.components * sizeof(Component)];
   pCompoMP = aCompoTable = (Component *)achCompoTable;
   if (!achCompoTable) errorh(BIGALLOC, ERH_T_SEV, ADS_BGAL);    // documented?

   achEntGenTable = new char[iCountGenEnt * sizeof(EntityDef)];
   if (!achEntGenTable) errorh(BIGALLOC, ERH_T_SEV, GENT_BGAL);

   achEntPrmTable = new char[iCountPrmEnt * sizeof(EntityDef)];
   if (!achEntPrmTable) errorh(BIGALLOC, ERH_T_SEV, PENT_BGAL);

   achSrmTable = new char[iCountSrMap*(1+iCountShortref) * sizeof(ShortrefMap)];
   pSrmMP = aSrmTable = (ShortrefMap *)achSrmTable;
   if (!achSrmTable) errorh(BIGALLOC, ERH_T_SEV, SRM_BGAL);

   achNcbTable = new char[nbrof.defined_notation * sizeof(NotationType)];
   pNcbMP = aNcbTable = (NotationType *) achNcbTable;
   if (!achNcbTable) errorh(BIGALLOC, ERH_T_SEV, NCB_BGAL);

   achExidTable = new char[(pPrsstat->iCountOfExid) * sizeof (ExternId)];
   pExidMP = aExidTable = (ExternId *)achExidTable;
   if (!achExidTable) errorh(BIGALLOC, ERH_T_SEV, EXID_BGAL);

   achDtpTable = new char[nbrof.dtp * sizeof (DataTagPattern)];
   pDtpMP = aDtpTable = (DataTagPattern *)achDtpTable;

   achNadTable = new char[nbrof.nad * sizeof(EntityNotation)];
   pNadMP = aNadTable = (EntityNotation *)achNadTable;

   achRklTable = new char[nbrof.rkl * sizeof (RankList)];
   pRklMP = aRklTable = (RankList *)achRklTable;

   achStsTable = new char [iCountStem * sizeof (StemValue)];
   pStsMP = aStsTable = (StemValue *)achStsTable;

   pEntMP = aEntTable = (EntityDef *)achEntGenTable;
   expnhashStem.traverse(tvrs_pile_up_ranks, this);
   expnhashNotation.traverse(tvrs_pile_up_notations, this); // 1) notations
   expnhashGenEnt.traverse(tvrs_pile_up_entities, this);    // 2) gen ent

   pEntMP = aEntTable = (EntityDef *)achEntPrmTable;
   expnhashPrmEnt.traverse(tvrs_pile_up_entities, this);  // 3) prm ent
   if (pSgmldcl->features.isFormal) {                     // FORMAL YES
      aExidTable->check_duplicate(sk, pPrsstat->iCountOfExid);
   }
   expnhashSrMap.traverse(tvrs_pile_up_shortrefs, this);  // 4) build srm
   expnhashElem.traverse(tvrs_pile_up_gis, this);         // 5) build elem, att

   aCompoTable->pile_up(nbrof.components);                // 6) unlocate compo

   /*-----------------+
   |Write the OxT file|
   +-----------------*/
   ApplicationHandle ahl = pPrsstat->open_ODT(
      utilctl.build_compiled_exid(clsOxt, pExid), TRUE
   );

   pEh->write(ahl, &ixDocElement, sizeof(int));
   pEh->write(ahl, &iCountShortref, sizeof(unsigned int)); // OCS dependency
   pQuantity->unload(ahl);                                // OCS dependency
   pSyntax->write_charset(ahl);                           // OCS dependency
   pEh->write(ahl, cap_count, sizeof(cap_count));

   pEh->write_record(ahl, achCnttokTable, nbrof.model * sizeof (ContentToken));
   delete [] achCnttokTable;

   pEh->write_record(ahl, achElemTable, iCountElem * sizeof (ElementType));
   delete [] achElemTable;

   pEh->write_record(
      ahl, achExTable,
      (nbrof.exceptions + nbrof.excptnames) * sizeof (Exception)
   );
   delete [] achExTable;
   unload_and_write_hashtable(ahl, expnhashElem, sizeof (ElementType), FALSE);

   /*
   | The Attribute array can have been shrinked if errors were found.
   | Get the final size from the pointers difference.
   */
   {
      char *pch = (char *)pAttrepMP;

      pEh->write_record(ahl, achAttrepTable, diffptr(pch, achAttrepTable));
      delete [] achAttrepTable;
   }

   pEh->write_record(ahl, achCompoTable, nbrof.components * sizeof (Component));
   delete [] achCompoTable;

   pEh->write_record(ahl, achEntGenTable, iCountGenEnt * sizeof (EntityDef));
   unload_and_write_hashtable(ahl, expnhashGenEnt, sizeof (EntityDef), TRUE);
   delete [] achEntGenTable;

   pEh->write_record(ahl, achEntPrmTable, iCountPrmEnt * sizeof (EntityDef));
   unload_and_write_hashtable(ahl, expnhashPrmEnt, sizeof (EntityDef), FALSE);
   delete [] achEntPrmTable;

   /*
   | The number of Shortref maps can have decreased
   | if USEMAP to undefined maps were found.  Recompute.
   */
   pEh->write_record(
      ahl, achSrmTable,
      (expnhashSrMap.count()) * (1+iCountShortref) * sizeof (ShortrefMap)
   );
   unload_and_write_hashtable(
      ahl, expnhashSrMap, (1+iCountShortref) * sizeof (ShortrefMap), FALSE
   );
   delete [] achSrmTable;

   pEh->write_record(ahl, achDtpTable, nbrof.dtp * sizeof (DataTagPattern));
   delete [] achDtpTable;

   pEh->write_record(ahl, achNadTable, nbrof.nad * sizeof (EntityNotation));
   delete [] achNadTable;

   pEh->write_record(
      ahl, achNcbTable, nbrof.defined_notation * sizeof (NotationType)
   );
   unload_and_write_hashtable(
      ahl, expnhashNotation, sizeof (NotationType), FALSE
   );
   delete [] achNcbTable;

   pEh->write_record(
      ahl, achExidTable, pPrsstat->iCountOfExid * sizeof (ExternId)
   );
   delete [] achExidTable;

   pEh->write_record(ahl, achStsTable, iCountStem * sizeof (StemValue));
   unload_and_write_hashtable(ahl, expnhashStem, sizeof (StemValue), FALSE);
   delete [] achStsTable;

   pEh->write_record(ahl, achRklTable, nbrof.rkl * sizeof (RankList));
   delete [] achRklTable;

   sk.output(pEh->write, ahl);
   pEh->close(ahl);
}
/******************************************************************************/
