/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// procpil.h
///  Declarations for procpil
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _PROCPIL_H
#define _PROCPIL_H 1
//******************************************************************************

#include "prodcl.h"     // base class
#include "srefdcl.h"    // base class
#include "elemdcl.h"    // base class
#include "attldcl.h"    // base class
#include "notadcl.h"    // base class
#include "entdcl.h"     // base class
#include "umapdcl.h"    // base class
#include "linkdcl.h"    // base class
#include "idlkdcl.h"    // base class

/*-----------------------------------------------------------------------------+
| Document Type and Link Type declaration compilers.                           |
| These tools create the binary form of DTD's or LTD's                         |
+-----------------------------------------------------------------------------*/
class PrologCompiler :
   private virtual PrologDcl,          // HAS-A
   private virtual ElementDcl,         // HAS-A
   private virtual AttlistDcl,         // HAS-A
   private virtual EntityDcl,          // HAS-A
   private virtual NotationDcl,        // HAS-A
   private virtual UsemapDcl,          // HAS-A
   private virtual ShortrefDcl,        // HAS-A
   private virtual LinkDcl,            // HAS-A
   private virtual IdLinkDcl           // HAS-A
{
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   PrologCompiler(        // to compile a DTD
      ParseState *pPrsstat, EventList& evlst, boolean isBaseDtd
   );
   PrologCompiler(        // to compile a LTD
      ParseState *pPrsstatArg, EventList& evlst
   );
   ~PrologCompiler();

   boolean compile();
   void unload(ExternId *pExid, e_TextClass clsOxt);

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   enum e_PrologCompilerState0 {
      INITIALIZED_CST0,
      COMPILING_LTD_CST0,
      COMPILING_DTD_CST0,
      ENDED_CST0
   };

   enum e_PrologCompilerState1 {
      INITIALIZED_CST1,
      PARSE_SUBSET_STARTED_CST1,
      PARSE_SUBSET_DONE_CST1,
      PARSE_EXTERNAL_STARTED_CST1
   };
   enum e_PrologCompilerState2 {
      STARTING_CST2,
      STARTED_CST2,
      PROCESSING_CST2,
      DONE_CST2
   };
   /*-------------+
   | Private Data |
   +-------------*/
   UtilityControl& const utilctl;
   EventList& evlst;
   EntityDef *pEntSource;

   e_PrologCompilerState0 cst0;
   e_PrologCompilerState1 cst1;
   e_PrologCompilerState2 cst2;
   boolean isBaseDtd;
   void *position;
   UtilityIdentifier *pUtidOdt;
   UtilityIdentifier *pUtidOlt;
   ExternId *pExidDtd;
   ExternId *pExidLpd;

   // These values are initialized by setup()
   ExternId  *pExidSource;      // either pExidDtd or pExidLpd
   e_LinkKind lkkind;
   const VarntChoiceEntry *specific_choices;
   e_TextClass txtclsCompiled;
   boolean isRealDeclaration;   // a `real' declaration has been found

   int init_floor_level;        // to save floor_level when starting
   int init_ent_level;          // to save nesting entity level when starting

   unsigned int ofsBufferExidEntDefault;
   unsigned int ofsLitEntDefault;
   int ixNadEntDefault;
   unsigned int ixElemMP;

   ContentToken   *aCnttokTable, *pCnttokMP;
   ElementType    *aElemTable;
   Exception      *aExTable,     *pExMP;
   AttributeRep   *aAttrepTable, *pAttrepMP;
   Component      *aCompoTable,  *pCompoMP;
   ExternId       *aExidTable,   *pExidMP;
   EntityDef      *aEntTable,    *pEntMP;
   ShortrefMap    *aSrmTable,    *pSrmMP;
   DataTagPattern *aDtpTable,    *pDtpMP;
   EntityNotation *aNadTable,    *pNadMP;
   NotationType   *aNcbTable,    *pNcbMP;
   RankList       *aRklTable,    *pRklMP;
   StemValue      *aStsTable,    *pStsMP;

   /*------------------+
   | Private Functions |
   +------------------*/
   void setup(
      ExternId *pExidSource,
      e_TextClass txtclsAsgn,
      e_LinkKind lkkind,
      e_TextClass txtclsCompiled
   );

   boolean (PrologCompiler::*compile_what)();
   boolean compile_DTD();
   boolean compile_LTD();
   boolean compile_subset_and_external();
   boolean compile_declaration_set();

   void unload_and_write_hashtable(
      ApplicationHandle ahlAccess,
      ExpandedHash& expnhash,
      int iSizeItem,
      boolean isExtendable
   );
   e_LinkKind link_specification();

   static boolean tvrs_check_declared_element(
      void *pvBase, const char *pszName, void *pvValue
   );
   static boolean tvrs_pile_up_ranks(
      void *pvBase, const char *pszName, void *pvValue
   );
   static boolean tvrs_pile_up_notations(
      void *pvBase, const char *pszName, void *pvValue
   );
   static boolean tvrs_pile_up_entities(
      void *pvBase, const char *pszName, void *pvValue
   );
   static boolean tvrs_pile_up_shortrefs(
      void *pvBase, const char *pszName, void *pvValue
   );
   static boolean tvrs_pile_up_gis(
      void *pvBase, const char *pszName, void *pvValue
   );
};

/*-- INLINES --*/
inline boolean PrologCompiler::compile() {  // homemade virtual
   return (this->*(this->compile_what))();
}
#endif /* _PROCPIL_H **********************************************************/
