/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prodcl.c
///  Base class for elementary declaration compilers
//   Pierre G. Richard 94/02/02 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| This class is a base for all elementary declaration compilers.
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "prodcl.h"
#include "prsstate.h"
#include "syntax.h"
#include "quantity.h"
#include "shortref.h"
#include "srefdcl.h"    // to create the EMPTY shortref map
#include "elemdcl.h"
#include "attldcl.h"
#include "errorh.h"

/*-------------------------------------------------------------------PrologDcl-+
| Initialize data used by the elementary compilers.                            |
| Create an empty shortref map.                                                |
| The empty map is treated like any other, except that there is no declaration |
| entered -- no <!SHORTREF #EMPTY>.                                            |
| Do not confuse an EMPTY map (that can be activated: <!USEMAP #EMPTY...)      |
| with no map activated:                                                       |
|   - there is a "real" address for the empty map                              |
|     (vs. no map activated == (SRM *)0)                                       |
|   - there is also a name (an empty string, but it exists!)                   |
+-----------------------------------------------------------------------------*/
PrologDcl::PrologDcl(
   ParseState *pPrsstatArg
) :
   pPrsstat(pPrsstatArg),
   pEh(pPrsstatArg->pEh),
   pSgmldcl(pPrsstatArg->pSgmldcl),
   pPrsmode(pPrsstatArg->pPrsmode),
   pSyntax(pPrsstatArg->pSyntax),
   pShortref(pPrsstatArg->pSyntax->pShortref),
   pRsvname(pPrsstatArg->pSyntax->pRsvname),
   pQuantity(pPrsstatArg->pSyntax->pQuantity),
   expnhashGenEnt(ENT_HASHSIZE),
   expnhashPrmEnt(ENT_HASHSIZE),
   expnhashElem(ELEM_HASHSIZE),
   expnhashNotation(NOTA_HASHSIZE),
   expnhashSrMap(SRM_HASHSIZE),
   expnhashStem(STEM_HASHSIZE),
   iSrMax(pPrsstatArg->pSyntax->pShortref->count()),
   ATTCNT(pPrsstatArg->pSyntax->pQuantity->value_of(ATTCNT_QTY_ix)),
   GRPCNT(pPrsstatArg->pSyntax->pQuantity->value_of(GRPCNT_QTY_ix)),
   GRPGTCNT(pPrsstatArg->pSyntax->pQuantity->value_of(GRPGTCNT_QTY_ix)),
   GRPLVL(pPrsstatArg->pSyntax->pQuantity->value_of(GRPLVL_QTY_ix)),
   LITLEN(pPrsstatArg->pSyntax->pQuantity->value_of(LITLEN_QTY_ix)),
   NAMELEN(pPrsstatArg->pSyntax->pQuantity->value_of(NAMELEN_QTY_ix)),
   NORMSEP(pPrsstatArg->pSyntax->pQuantity->value_of(NORMSEP_QTY_ix))
{
   memset(&nbrof, 0, sizeof nbrof);

   achNameOfIDAttribute[0] = '\0';
   iIdentElem = 0;
   pchLitBuffer = new char [LITLEN + 1];
   if (GRPCNT > ATTCNT) {
      pchNamesBuffer = new char [(GRPCNT+1)*(NAMELEN+1)];
   }else {
      pchNamesBuffer = new char [(ATTCNT+1)*(NAMELEN+1)];
   }
   /*
   | create the empty SR map
   */
   new(
      expnhashSrMap.gimme_value("", sizeof(ShortrefNode))
   ) ShortrefNode(pShortref->count());
}

/*PROTECTED----------------------------------------------------------PrologDcl-+
| Default constructor                                                          |
| This constructor cannot be called.  It is required because Instance has      |
| access to the unique and already created PrologDcl.                          |
+-----------------------------------------------------------------------------*/
PrologDcl::PrologDcl() :
   pPrsstat(0), pEh(0), pSyntax(0), pSgmldcl(0), pPrsmode(0), pShortref(0),
   pRsvname(0), pQuantity(0), iSrMax(0), ATTCNT(0), GRPCNT(0), GRPGTCNT(0),
   GRPLVL(0), LITLEN(0), NAMELEN(0), NORMSEP(0)
{
   assert("PrologDcl NULL constructor cannot be called" == FALSE);
}

/*------------------------------------------------------------------~PrologDcl-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
PrologDcl::~PrologDcl()
{
   delete [] pchLitBuffer;
   delete [] pchNamesBuffer;
}

/*---------------------------------------------------------------------pile_up-+
| Build ElementType, ContentToken, and Attribute arrays from ElementNode,      |
| ModelNode and AttributeNode.  These structures (+ each AttributeItem) are    |
| then freed.                                                                  |
+-----------------------------------------------------------------------------*/
boolean GenericIdBlock::pile_up(
   StringKeeper& sk,
   const char *pszName,
   ElementType* const aElemTable,
   ContentToken*& pCnttokMP, const ContentToken* const aCnttokTable,
   AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable,
   Component*& pCompoMP, const Component* const aCompoTable,
   Exception*& pExMP, const Exception* const aExTable,
   DataTagPattern*& pDtpMP, const DataTagPattern* const aDtpTable,
   RankList * const aRklTable
)
{
   /*----------------+
   | dislocate enode |
   +----------------*/
   if (!pEnode->is.solved) {               // if not already "solved":
      if (pEnode->pChxMinus) {
         pEnode->ixExMinus = pEnode->pChxMinus->pile_up(pExMP, aExTable);
      }else {
         pEnode->ixExMinus = -1;
      }
      if (pEnode->pChxPlus) {
         pEnode->ixExPlus = pEnode->pChxPlus->pile_up(pExMP, aExTable);
      }else {
         pEnode->ixExPlus = -1;
      }
      if ((!pEnode->is.any) && (!pEnode->is.empty) && (!pEnode->is.dcdata)) {
         unsigned short rx;
         int ixCnttok = diffptr(pCnttokMP, aCnttokTable);
         pEnode->pMnode->pile_up(pCnttokMP, pDtpMP, aDtpTable, rx);
         free(pEnode->pMnode);
         pEnode->ixCnttok = ixCnttok;
      }else {
         pEnode->ixCnttok = -1;            // no model in such case
      }
      if (pRkn) aRklTable->solve_index(pRkn);
   }

   /*-------------------------------+
   | set up ixRkl to the first stem |
   +-------------------------------*/
   int ixRkl;
   if (!pRkn) {
      ixRkl = -1;
   }else {
      ixRkl = aRklTable->first_stem(pRkn->ixRkl);
      free(pRkn);
   }

   /*-------------------------------+
   | get Attribute index, and count |
   +-------------------------------*/
   unsigned int iAttCount;
   int ixAttrep;
   if (!pAnode) {                    // no attributes associated:
      iAttCount = 0;                 // number of defined attributes
      ixAttrep = -1;                 // put a dummy value
   }else {
      pEnode->is.required = pAnode->is.required;
      pEnode->is.current = pAnode->is.current;
      pEnode->is.conref = pAnode->is.conref;
      pEnode->is.notation = pAnode->is.notation;
      pEnode->is.entity = pAnode->is.entity;
      if ((pEnode->is.required)&&(!pEnode->is.reqd_open_tag)) { // see 7.3.1.1
         errorh(INVSTGREQ, ERH_I_SEV, pszName);
         pEnode->is.reqd_open_tag = TRUE;
      }
      if ((pEnode->is.empty)&&((pEnode->is.conref)||(pEnode->is.notation))) {
         pEnode->is.empty = FALSE;
         pEnode->is.any = TRUE;
         if (pEnode->is.conref) {               // last paragraph in 11.3.4
            errorh(INVCONREF, ERH_E_SEV, pszName);
         }
         if (pEnode->is.notation) {             // last paragraph in 11.3.3
            errorh(INVNOTA, ERH_E_SEV, pszName);
         }
      }
      iAttCount = pAnode->attlist.count();
      ixAttrep = pAnode->pile_up(
         pAttrepMP, aAttrepTable, pCompoMP, aCompoTable
      );
      if (--pAnode->iReferenceCount == 0) delete pAnode;
   }
   if (pAnodeLink) {
      iAttCount += pAnodeLink->attlist.count(); // bump count
      pAnodeLink->pile_up(                      // pile_up and catenate?
         pAttrepMP, aAttrepTable, pCompoMP, aCompoTable, pAnode? TRUE : FALSE
      );
      if (--pAnodeLink->iReferenceCount == 0) delete pAnodeLink;
   }

   /*-------------------------------------------------+
   | Construct the dislocated ElementType in the pile |
   +-------------------------------------------------*/
   {
      ElementType * const pElem = aElemTable + ixElem;
      pElem->ofsName = sk.park(pszName, 1+strlen(pszName));
      pElem->is = pEnode->is;
      pElem->ixCnttok = pEnode->ixCnttok;
      pElem->ixAttrep = ixAttrep;
      pElem->iAttCount = iAttCount;
      pElem->ixExMinus = pEnode->ixExMinus;
      pElem->ixExPlus = pEnode->ixExPlus;
      pElem->ixSrm = pSnode? pSnode->pile_index() : -1;
      pElem->ixRkl = ixRkl;
   }
   if (!pEnode->is.solved) {                 // "solved" now
      if (--pEnode->iReferenceCount == 0) {  // no more needed
         delete pEnode;
      }else {
         pEnode->is.solved = TRUE;           // else, don't dislocate twice!
      }
   }
   return TRUE;
}
/*============================================================================*/
