/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prodcl.h
///  Declarations for prodcl
//   Pierre G. Richard 94/02/02 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _PRODCL_H
#define _PRODCL_H 1
//******************************************************************************

#include "hash.h"       // part of one class
#include "skmgr.h"      // part of one class

class ParseState;
class EntityHandler;
class ParseMode;
class SGMLdcl;
class Syntax;
class Shortref;
class ReservedName;
class Quantity;
class ElementType;
class ContentToken;
class AttributeRep;
class Exception;
class RankList;
class DataTagPattern;
class Component;

class ShortrefNode;
struct GenericIdBlock;
struct ElementNode;
struct AttributeNode;
struct ModelNode;
struct RankNode;

/*-----------------------------------------------------------------------------+
| Prolog Declaration class                                                     |
+-----------------------------------------------------------------------------*/
class PrologDcl
{
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   PrologDcl(ParseState *pPrsstat);
   ~PrologDcl();

protected:
   /*---------------+
   | Protected Data |
   +---------------*/
   ParseState * const pPrsstat;
   EntityHandler * const pEh;
   Syntax * const pSyntax;
   SGMLdcl * const pSgmldcl;
   ParseMode * const pPrsmode;
   Shortref * const pShortref;
   ReservedName * const pRsvname;
   Quantity * const pQuantity;
   GenericIdBlock *pGiDocElement;

   const iSrMax;
   const int ATTCNT;
   const int GRPCNT;
   const int GRPGTCNT;
   const int GRPLVL;
   const int LITLEN;
   const int NAMELEN;
   const int NORMSEP;

   ExpandedHash expnhashElem;
   ExpandedHash expnhashGenEnt;      // also shadowed to ParseState
   ExpandedHash expnhashPrmEnt;      // also shadowed to ParseState
   ExpandedHash expnhashSrMap;
   ExpandedHash expnhashNotation;
   ExpandedHash expnhashStem;

   StringKeeper sk;

   struct NbrOf {
      unsigned int defined_notation;
      unsigned int model;
      unsigned int useless_model;
      unsigned int exceptions;
      unsigned int excptnames;
      unsigned int rkl;
      unsigned int attributes;
      unsigned int att_dclv;
      unsigned int dclv_text;
      unsigned int components;
      unsigned int g_text;
      unsigned int p_text;
      unsigned int dtp;
      unsigned int datatag;
      unsigned int datatag_text;
      unsigned int nad;
      unsigned int nxid_text;
      unsigned int exid_text;
   }nbrof;

   char *pchNamesBuffer;                      // General buffer for name lists
   char *pchLitBuffer;                        // General buffer to get a literal
   int iIdentElem;                            // Incremented by ElementDcl
   char achNameOfIDAttribute[NAMELEN_MAX+1];  // Set by AttlistDcl

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   PrologDcl();            // Default constructor.  Must be never called

   /*--------------------+
   | Private Definitions |
   +--------------------*/
   enum { ENT_HASHSIZE = 29 }; // For about 100 entries
   enum { ELEM_HASHSIZE = 53 }; // For about 200 entries
   enum { NOTA_HASHSIZE = 5 }; // For about  20 entries
   enum { SRM_HASHSIZE =  5 }; // For about  20 entries
   enum { STEM_HASHSIZE = 7 }; // For about  50 entries
};

/*-----------------------------------------------------------------------------+
| GenericIdBlock groups all characteristics relative to a GI                   |
+-----------------------------------------------------------------------------*/
struct GenericIdBlock {         // GI: value_of(pExpnhashElem)
   ElementNode *pEnode;         // ElementNode
   AttributeNode *pAnode;       // AttributeNode
   AttributeNode *pAnodeLink;   // AttributeNode for LINK attributes
   ShortrefNode *pSnode;        // ShortrefNode for shortref map
   RankNode *pRkn;              // Rank Node
   unsigned int ixElem;         // Elem array index (set at end of compilation)
   char *pszName;               // GenericIdBlock name (ambiguities error msg)
   unsigned int iIdentElem :15; // seq nbr, check dup GIs in a same model
   boolean isMoreThanOnce  :1;  // this GI occurs more than once in the model

   boolean pile_up(
      StringKeeper& sk,
      const char *pszName,
      ElementType* const aElemTable,
      ContentToken*& pCnttokMP, const ContentToken* const aCnttokTable,
      AttributeRep*& pAttrepMP, const AttributeRep* const aAttrepTable,
      Component*& pCompoMP, const Component* const aCompoTable,
      Exception*& pExMP, const Exception* const aExTable,
      DataTagPattern*& pDtpMP, const DataTagPattern* const aDtpTable,
      RankList * const aRklTable
   );
};

#endif /* _PRODCL_H ***********************************************************/
