/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prsmode.c
///  Handle recognition mode for SGML delimiters
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "prsmode.h"
#include "errorh.h"

/*-------------------------------------------------------------------Parsemode-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
ParseMode::ParseMode()
{
   enable_markup_recognition();
   mode_stack_index = 0;
   set_mode(INI_MODE);
}

/*-------------------------------------------------------------------push_mode-+
| Push the current recognition mode, and set a new one.                        |
|"mode" is one of xxx_MODE, defined in "mode.h"                                |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
void ParseMode::push_mode(e_RecognitionMode mode)
{
   if (mode_stack_index > MAX_MODE_LVL) {
      errorh(BIGMODLVL, ERH_T_SEV);
   }
   mode_stack[mode_stack_index++] = curr_mode;
   set_mode(mode);
}

/*--------------------------------------------------------------------pop_mode-+
|Restore previous recognition mode.                                            |
|"mode" is one of xxx_MODE, defined in "mode.h"                                |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
void ParseMode::pop_mode()
{
   set_mode(mode_stack[--mode_stack_index]);
}
