/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prsmode.h
///  Declarations for prsmode
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _PRSMODE_H
#define _PRSMODE_H 1
/*
| The ParseMode Class is designed to collect the state that can
| affect the recognition of delimiters.
*/
//******************************************************************************

#include "prsmode.eh"   // enums

class ParseMode {
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   ParseMode();

   void set_mode(e_RecognitionMode);     // set current mode
   e_RecognitionMode get_mode();         // what is the current mode?

   void enable_markup_recognition();         // enable markup
   void disable_markup_recognition();        // disable markup
   boolean is_markup_recognition_disabled(); // is markup disabled?

   void push_mode(e_RecognitionMode);    // push mode on a stack
   void pop_mode();                          // restore mode from stack

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   // The following is a "safe" assumption of what the maximum mode nesting
   // level is.  This number has no relationship with the syntax.
   enum { MAX_MODE_LVL = 16 };    // modes: maximum level of nesting

   /*------------------+
   | Private Variables |
   +------------------*/
   boolean isMarkupHidden;
   e_RecognitionMode curr_mode;
   e_RecognitionMode mode_stack[MAX_MODE_LVL];
   int mode_stack_index;
};

/*--------------------------------------------------------------------set_mode-+
| Change the recognition mode                                                  |
|                                                                              |
| -- INVARIANTS --                                                             |
| `curr_mode' contains the mode which is currently active.                     |
| -- INPUT --                                                                  |
| `mode' holds the new recognition mode                                        |
| -- PRECONDITION --                                                           |
| `mode' is one of the `e_RecognitionMode' value                               |
| -- OUTPUT --                                                                 |
| (none)                                                                       |
| -- POSTCONDITION --                                                          |
| `curr_mode' contains the new mode                                            |
+-----------------------------------------------------------------------------*/
inline void ParseMode::set_mode(e_RecognitionMode new_mode)
{
   curr_mode = new_mode;
}

/*--------------------------------------------------------------------get_mode-+
| What's the current recognition mode?                                         |
|                                                                              |
| -- INVARIANTS --                                                             |
| `curr_mode' contains the mode which is currently active.                     |
| -- INPUT --                                                                  |
| (none)                                                                       |
| -- PRECONDITION --                                                           |
| (none)                                                                       |
| -- OUTPUT --                                                                 |
| `mode' is one of the `e_RecognitionMode' value                               |
| -- POSTCONDITION --                                                          |
| (none)                                                                       |
+-----------------------------------------------------------------------------*/
inline e_RecognitionMode ParseMode::get_mode()
{
   return curr_mode;
}

/*--------------------------------------------------disable_markup_recognition-+
|Do not allow any MARKUP to be recognized.                                     |
|`enable_markup_recognition' restores to normal state.                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline void ParseMode::disable_markup_recognition()
{
   isMarkupHidden = TRUE;
}

/*----------------------------------------------is_markup_recognition_disabled-+
|Indicates if markup recognition is disabled                                   |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline boolean ParseMode::is_markup_recognition_disabled()
{
   return isMarkupHidden;
}

/*---------------------------------------------------enable_markup_recognition-+
|Allow recognition of markup.                                                  |
|It can have been previoulsy disabled by `disable_markup_recognition'.         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline void ParseMode::enable_markup_recognition()
{
   isMarkupHidden = FALSE;
}
#endif /* _PRSMODE_H **********************************************************/
