/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prsstate.h
///  Declarations for prsstate
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _PRSCTXT_H
#define _PRSCTXT_H 1
/*__automake__
 Source "tag.cpp";
*/
//******************************************************************************

#include "hash.h"       // part of the class
#include "oestack.h"    // part of the class
#include "rcdstate.h"   // part of the class
#include "utilctl.h"    // part of the class
#include "entity.h"     // part of the class
#include "externid.h"   // part of the class
#include "strlst.h"     // part of the class

#include "lextoken.eh"  // enums
#include "capacity.eh"
#include "element.eh"

class SGMLparser;
class PrologCompiler;
class ParseMode;
class SGMLdcl;
class NewSGMLdcl;
class Syntax;
class DelimiterVector;
class EntityHandler;
class ExpandedHash;
class DTDdcl;
class Document;
class ElementType;
class Attribute;
class AttributeList;
class Component;

enum e_ParseState0 {        // Order does matter
   INITIALIZED_PST0,        // At the very beginning...
   SGMLDCL_STARTED_PST0,    // Parsing the SGML dcl
   SGMLDCL_DONE_PST0,       // About to parse the Prolog
   PROLOG_STARTED_PST0,     // Parsing the Prolog
   PROLOG_DONE_PST0,        // About to parse the Instance
   INSTANCE_STARTED_PST0,   // Parsing the instance
   INSTANCE_DONE_PST0,      // At the very end
   ALL_DONE_PST0            // In case consumer insists...
};

enum e_ParseState1 {
   // ------------------------ In PROLOG_STARTED_PST0;
   OTHER_PROLOG_PST1,       // parsing other prolog ([7], [8])
   DOCTYPE_STARTED_PST1,    // parsing a doctype
   DOCTYPE_PROCESSING_PST1,
   DOCTYPE_DONE_PST1,
   LINKTYPE_STARTED_PST1,   // parsing a linktype
   LINKTYPE_PROCESSING_PST1,
   LINKTYPE_DONE_PST1,
   // ------------------------ In INSTANCE_STARTED_PST0
   TAG_PST1,                // tag, is also the floor state
   DATA_PST1,               // character data
   DATA_STOWED_PST1,
   DATA_DISCARD_CHAR_PST1,
   DATA_PENDING_CHAR_PST1,
   INVDATA_PST1             // invalid character
};

enum e_ParseState2 {
   // ------------------------ In PROLOG_STARTED_PST0;
   NOTHING_FOUND_PST2,      // before any doctype/linktype declaration
   DOCTYPE_FOUND_PST2,      // at least one doctype was found
   LINKTYPE_FOUND_PST2,     // at least, one linktype was found
   // ------------------------ In INSTANCE_STARTED_PST0
   EE_BEFORE_CLOSE_PST2,    // Ee found: about to close...
   EE_AFTER_CLOSE_PST2      // Ee found and reported
};

enum e_AttSpecKind {
   DEF_TGF,                 // Default AVS being parsed (DATA or TAG)
   TAG_TGF,                 // Tag Attribute Specification being parsed
   DATA_TGF                 // Data Attribute Specification being parsed
};

enum e_RcAttSpecList {      // Return code from att_spec_list
   NO_CHANGE_ASLRC,         // Nothing changed: can use the default ASL
   CHANGED_ASLRC,           // At least one AVS was found
   CONREF_CHANGED_ASLRC     // AVS found for a CONREF
};

/*-----------------------------------------------------------------------------+
| ParseState class                                                             |
+-----------------------------------------------------------------------------*/
class ParseState
{
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   /*
   | This constructor is used to get a ParseState at a given
   | point, so it's then possible to resume from this point.
   */
   ParseState(const SGMLparser *pParserSource);

   _export ~ParseState();

   void activate_SGMLdcl(e_TextClass clsOcs);
   void activate_syntax(e_TextClass clsOcs);

   e_LexToken get_curr_token();           // Current delimiter found
   unsigned int get_curr_count();             // Current delimiter length

   unsigned int read_processing_instr(EntityHandler *pEh, char *pchBuffer);
   unsigned int read_minimum_literal(EntityHandler *pEh, char *pchBuffer);
   unsigned int read_cdata_literal(EntityHandler *pEh, char *pchBuffer);

   boolean  may_be_delimiter();
   e_LexToken recognize_delimiter();

   unsigned int read_parameter_literal(char *pchBuffer, boolean for_datatag);

   void parse_name_group(void (*inst_function)(void *pvBase), void *pvBase);

   boolean marked_section_open();        // Enter a Marked Section
   boolean marked_section_close();       // Exit a Marked Section
   void marked_section_ignore();         // Skip over an ignored marked section

   boolean skip_S();                     // Skip S separator(s) [5]
   boolean skip_ts(EntityDef **ppEntPI=0);  // Skip token separator(s) [70]
   boolean skip_ps();                    // Skip parameter separator [65]
   void    skip_specific_ps();           // Skip specific parameter separator
   void    skip_inferable_required_ps(); // Skip required ps, but recover
   void    skip_cdc();                   // Skip a comment declaration [91]

   void adjust_recognition_mode(e_Dcdata dcdata);

   int att_val_spec(
      char * const pchBuf,               // Where to put the AVS
      const Attribute *pAtt,             // Attribute or 0
      e_AttSpecKind tgf,                 //
      Component*& pCompo,                // Output: value components
      unsigned int& uiNormedLenVal       // Output: normalized length
   );
   e_RcAttSpecList att_spec_list(
      const AttributeList *pAttlist,     // 0 if must get rid of ASL
      e_AttSpecKind tgf,                 // TAG_TGF or DATA_TGF
      unsigned int& uiNormedLenVal       // Output: normalized length
   );
   boolean finish_parse_tag(
      const AttributeList *pAttlist = 0  // 0 if must get rid of the ASL
   );
   void validate_start_element(
      const ElementType *pElem,
      const AttributeList&,
      boolean isInferred, boolean isExplConref
   );
   void validate_end_element(
      const ElementType *pElem,
      boolean isInferred, boolean isPremature, boolean isBadEmpty
   );
   ApplicationHandle open_OCS(
      e_TextClass clsOcs, boolean isWriteMode
   );
   ApplicationHandle open_ODT(
      ExternId *pExid, boolean isWriteMode
   );

//***>NOTE: protected removed - design must be slightly reworked (compilers)
// protected:
   /*--------------------+
   | Protected Variables |
   +--------------------*/
   BitArray isEnabledEvent;
   SGMLdcl *pSgmldcl;
   DTDdcl *pDtddcl;
   EntityHandler *pEh;         // Current entity manager
   Syntax *pSyntax;            // Shadow SGMLdcl mbr
   UtilityControl utilctl;
   ParseMode *pPrsmode;
   NewSGMLdcl *pNewSgmldcl;
   PrologCompiler *pProcpil;
   SortedStringList sstrlActiveType;
   int iCountOfExid;           // count of external identifiers
   e_ParseState0 pst0;
   e_ParseState1 pst1;
   e_ParseState2 pst2;

   Hash *pHshGenEnt;           // for general entities   (Expn, or DTD Shadow)
   Hash *pHshPrmEnt;           // for parameter entities (Expn, or DTD Shadow)

   /*
   | The capacity counters:
   |  - accumulate the values of each DTDdcl instanciated,
   |  - might increase during the instance parse, namely for:
   |    ENTCAP (10.5.1.2), ATTCHCAP, IDCAP, IDREFCAP
   */
   unsigned long aulCapacityCount[CAP_MAX_SIZE]; // capacity reached

   // Used when parsing the instance
   ExpandedHash ExpnhashId;      // Hash table for ID attributes
   const ElementType *pElemLastEnded;
   OpenElementStack oelemstk;
   RecordState rcdstate;
   ExternId exidPrimary; EntityDef entPrimary;  // in this order!
   boolean isNetEnabled;         // TRUE if within a NET started element
   boolean isQuietAtFloor;       // quiet if Ee reached at floor_level
   int iCountForcedData;         // suppress recognition of char in unmapped SR

   int curr_elem_id;             // last list preorder sequence number

                                 // nesting levels:
   int     ms_ent_level;         // for an RCDATA marked section
   int     cn_ent_level;         // for RCDATA in content
   int     floor_level;          // can't close under this level
   int     ms_level;             // Marked section depth, signed!

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   ParseState();                 // Default constructor.  Must be never called

   /*
   | Regular constructor, called by SGMLparser(Document *)
   | to enter a new parser session.
   */
   ParseState(const Document& doc);

   /*
   | Copy constructor
   | This constructor is used to resume a Parser session form a given
   | ParseState.  Only called by SGMLparser(const ParseState& prsctxt),
   | it is therefore protected.
   */
   //***>Note: not yet implemented
   ParseState(const ParseState& prsctxtSource);

   EntityDef *find_general_entity(char *pszName);
   EntityDef *open_general_entity_reference();

private:
   /*------------------+
   | Private Variables |
   +------------------*/

   DelimiterVector *pDv;       // Shadow Syntax mbr for recognize delimiter

   e_LexToken curr_token;      // current delimiter token found
   int  curr_count;            // length of current delimiter

   /*------------------+
   | Private Functions |
   +------------------*/
   void shadow_syntax(Syntax *pSyntaxArg);

   //***>Note: needs Em, Syntax, ParseMode
   boolean check_contextual_constraint(e_LexToken);

   //***>Note: needs pDtddcl
   void open_parameter_entity_reference(EntityDef **ppEnt = 0);
};

/*INLINE--------------------------------------------------------get_curr_token-+
| What's the current delimiter found?                                          |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline e_LexToken ParseState::get_curr_token() {
   return curr_token;
}

/*INLINE--------------------------------------------------------get_curr_count-+
| What's the current delimiter length?                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline unsigned int ParseState::get_curr_count() {
   return curr_count;
}
#endif /* _PRSCTXT_H **********************************************************/
