/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// quantity.c
///  Class for SGML quantities
//   Pierre G. Richard 89/02/14 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "enthdlr.h"
#include "quantity.h"
#include "rsvname.h"
#include "errorh.h"

/*--------------------------------------------------------------------Quantity-+
| Constructor from scratch (i.e Ref Concrete Syntax)                           |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Quantity::Quantity()
{
   qntytab[ATTCNT_QTY_ix]    =  ATTCNT_RCS;
   qntytab[ATTSPLEN_QTY_ix]  =  ATTSPLEN_RCS;
   qntytab[BSEQLEN_QTY_ix]   =  BSEQLEN_RCS;
   qntytab[DTAGLEN_QTY_ix]   =  DTAGLEN_RCS;
   qntytab[DTEMPLEN_QTY_ix]  =  DTEMPLEN_RCS;
   qntytab[ENTLVL_QTY_ix]    =  ENTLVL_RCS;
   qntytab[GRPCNT_QTY_ix]    =  GRPCNT_RCS;
   qntytab[GRPGTCNT_QTY_ix]  =  GRPGTCNT_RCS;
   qntytab[GRPLVL_QTY_ix]    =  GRPLVL_RCS;
   qntytab[LITLEN_QTY_ix]    =  LITLEN_RCS;
   qntytab[NAMELEN_QTY_ix]   =  NAMELEN_RCS;
   qntytab[NORMSEP_QTY_ix]   =  NORMSEP_RCS;
   qntytab[PILEN_QTY_ix]     =  PILEN_RCS;
   qntytab[TAGLEN_QTY_ix]    =  TAGLEN_RCS;
   qntytab[TAGLVL_QTY_ix]    =  TAGLVL_RCS;
}

/*--------------------------------------------------------------------Quantity-+
| Constructor from OCS                                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Quantity::Quantity(ApplicationHandle ahlAccess)
{
   EntityHandler::read(ahlAccess, qntytab, sizeof qntytab);
}

/*---------------------------------------------------------------------name_of-+
|Get the reserved name associated to an index in the quantity table.           |
+-----------------------------------------------------------------------------*/
const char * const Quantity::name_of(e_IxQuantity ix)
{
   const char * pszNameQnty;

   switch (ix) {
   case ATTCNT_QTY_ix:
      pszNameQnty = ReservedName::name_of(ATTCNT_RRN_ix);
      break;
   case ATTSPLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(ATTSPLEN_RRN_ix);
      break;
   case BSEQLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(BSEQLEN_RRN_ix);
      break;
   case DTAGLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(DTAGLEN_RRN_ix);
      break;
   case DTEMPLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(DTEMPLEN_RRN_ix);
      break;
   case ENTLVL_QTY_ix:
      pszNameQnty = ReservedName::name_of(ENTLVL_RRN_ix);
      break;
   case GRPCNT_QTY_ix:
      pszNameQnty = ReservedName::name_of(GRPCNT_RRN_ix);
      break;
   case GRPGTCNT_QTY_ix:
      pszNameQnty = ReservedName::name_of(GRPGTCNT_RRN_ix);
      break;
   case GRPLVL_QTY_ix:
      pszNameQnty = ReservedName::name_of(GRPLVL_RRN_ix);
      break;
   case LITLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(LITLEN_RRN_ix);
      break;
   case NAMELEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(NAMELEN_RRN_ix);
      break;
   case NORMSEP_QTY_ix:
      pszNameQnty = ReservedName::name_of(NORMSEP_RRN_ix);
      break;
   case PILEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(PILEN_RRN_ix);
      break;
   case TAGLEN_QTY_ix:
      pszNameQnty = ReservedName::name_of(TAGLEN_RRN_ix);
      break;
   default:  /* case TAGLVL_QTY_ix: */
      pszNameQnty = ReservedName::name_of(TAGLVL_RRN_ix);
      break;
   }
   return pszNameQnty;
}

/*-----------------------------------------------------read_and_check_quantity-+
| Check the Quantity used to create the ODT is still compatible with the       |
| one currently active.                                                        |
+-----------------------------------------------------------------------------*/
boolean Quantity::read_and_check_quantity(ApplicationHandle ahlAccess)
{
   boolean rc = TRUE;
   unsigned int aiDtdQnty[sizeof qntytab / sizeof qntytab[0]];

   EntityHandler::read(ahlAccess, aiDtdQnty, sizeof aiDtdQnty);
   for (int i=0; i < sizeof qntytab / sizeof qntytab[0]; ++i) {
      if (qntytab[i] < aiDtdQnty[i]) {
         errorh(
            INVISTQTY, ERH_E_SEV, name_of((e_IxQuantity)i),
            qntytab[i], aiDtdQnty[i]
         );
         rc = FALSE;
      }
   }
   return rc;
}

/*----------------------------------------------------------------------unload-+
| Write the current quantity to disk.                                          |
| This function is also used to save the quantity the DTD was compiled with    |
| inside the ODT.  This permits to check the compatibility of an ODT with any  |
| active OCS (see read_and_check_quantity).                                    |
+-----------------------------------------------------------------------------*/
void Quantity::unload(ApplicationHandle ahlAccess)
{
   EntityHandler::write(ahlAccess, qntytab, sizeof qntytab);
}

/*-------------------------------------------------------------------set_value-+
| Set new quantity values                                                      |
+-----------------------------------------------------------------------------*/
boolean NewQuantity::set_value(e_IxQuantity ix, unsigned long value)
{
   unsigned long min = 1;
   unsigned long max = 1;

   /*-----------------------+
   | limit each quantity to |
   | a specific min/max     |
   +-----------------------*/
   min = 1;
   switch (ix) {
   case ATTSPLEN_QTY_ix:
      max = ATTSPLEN_MAX;
      break;
   case ATTCNT_QTY_ix:
      max = ATTCNT_MAX;
      break;
   case LITLEN_QTY_ix:     /* Warning: see BUFSIZE in GLOBAL.H */
      min = NAMELEN_MAX;
      max = LITLEN_MAX;
      break;
   case NAMELEN_QTY_ix:
      max = NAMELEN_MAX;
      break;
   case NORMSEP_QTY_ix:
      max = NORMSEP_MAX;
      break;
   case PILEN_QTY_ix:
      max = PILEN_MAX;
      break;
   case TAGLEN_QTY_ix:
      max = TAGLEN_MAX;
      break;
   case GRPCNT_QTY_ix:
      max = GRPCNT_MAX;
      break;
   case ENTLVL_QTY_ix:
      max = ENTLVL_MAX;
      break;
   case GRPGTCNT_QTY_ix:
      max = GRPGTCNT_MAX;
      break;
   case TAGLVL_QTY_ix:
      max = TAGLVL_MAX;
      break;
   case BSEQLEN_QTY_ix:    /* Warning: see BUFSIZE in GLOBAL.H */
   case DTAGLEN_QTY_ix:
   case DTEMPLEN_QTY_ix:
   case GRPLVL_QTY_ix:
   default:
      min = max = qntytab[ix];
      break;
   }
   if ((value < min) || (value > max)) return FALSE;

   qntytab[ix] = (unsigned int)value;
   return TRUE;
}
/******************************************************************************/
