/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// rcdstate.c
///  Record boundaries conditions to generate relevant record end
//   Pierre G. Richard 88/05/28 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "rcdstate.h"

/*-----------------------------------------------------------------------------+
| Record boundaries conditions (7.6.1) are described by a FlowGraph.           |
|                                                                              |
| Theoretically each time a new element is started, either proper or included, |
| the element's current state is stacked a a new FlowGraph is started.  This   |
| enforces the rule for each element being treated as an atom.  According      |
| to this, the graph can be reduced to 5 states and 4 events:                  |
|                                                                              |
|   - RS      : Record Start                                                   |
|   - RE      : Record End                                                     |
|   - DATA    : Data or Proper subelement                                      |
|   - OTHER   : Included subelement, Markup Declaration, PI                    |
|                                                                              |
|   Current|   New State = f(Events)   |                                       |
|    State |                           |                                       |
|          |  RS  |  RE  | DATA |OTHER |                                       |
|   -------+------+------+------+------+                                       |
|      0   |  2   |  2   |  1   |  0   |                                       |
|      1   |  2   |  3   |  1   |  1   |                                       |
|      2   |  2   |  3   |  1   |  0   |                                       |
|      3   |  3   | 3+RE | 1+RE |  4   |                                       |
|      4   |  3   |  3   | 1+RE |  4   |                                       |
|                                                                              |
| State 0 and 2 reflect "RE due to markup" conditions.                         |
| State 3 and 4 are "RE pending" conditions: the RE will be generated          |
| or not according to the last RE rule (rule b).                               |
| State 0 and 4 are to discard the RE following an included subelement.        |
| State 0 is also the initial state (graph reduction).                         |
| State 1 is when a DATA or proper subelement is included.                     |
|                                                                              |
| Practically entering a new proper element doesn't require a stack operation: |
| the RE (if any) will be generated before the proper subelement is entered    |
| so the state "popped" when the proper subelement ends is always the same.    |
| So the start and ends of a proper subelement (STG and ETG) can be added to   |
| the list of events.  However, for included subelement, stacking the          |
| element's current state is mandatory.  So the final graph is as shown below. |
+-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------RecordState-+
| Constructor.                                                                 |
+-----------------------------------------------------------------------------*/
RecordState::RecordState()
{
   pState = state;
   *pState = 0;
}
/******************************************************************************/
