/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// rcdstate.h
///  Declarations for rcdstate
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _RCDSTATE_H
#define _RCDSTATE_H 1
/*
| This class describes the record boundaries conditions of 7.6.1.
|
| In contents, the following events:     must be declared by calling:
|    Record start:                       `declare_RS'
|    Record end:                         `declare_RE'
|    data (not RE, nor RS):              `declare_data'
|    start of a proper sub-element:      `declare_start_proper_elem'
|    end of a proper sub-element:        `declare_end_proper_elem'
|    start of an included sub-element:   `declare_SITG'
|    end of an included sub-element:     `declare_EITG'
|    other markup declarations:          `declare_other_markup_dcl'
|
| Three of these functions:
|      `declare_RE',
|      `declare_start_proper_elem',
|      `declare_data'
| might return TRUE, meanning that a significant record end has been found.
|
*/
//******************************************************************************

#include "quantity.eh"  // enums

class RecordState {
public:
   RecordState();
   void declare_RS();
   boolean declare_RE();
   boolean declare_start_proper_elem();
   void declare_end_proper_elem();
   boolean declare_data();
   void declare_other_markup_dcl();
   void declare_start_included_elem();
   void declare_end_included_elem();
   void reset();
private:
   struct FlowGraphNode {
      char new_state;
      boolean re_generated;
   };
   char *pState;
   char state[TAGLVL_MAX];
};

/*-----------------------------------------------------------------------reset-+
| Reset the current state.                                                     |
+-----------------------------------------------------------------------------*/
inline void RecordState::reset()
{
   *pState = 0;
}

/*------------------------------------------------------------------declare_RS-+
| Register a Record Start                                                      |
+-----------------------------------------------------------------------------*/
inline void RecordState::declare_RS()
{
   static RecordState::FlowGraphNode fgRS[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+----------
         {2,FALSE} , {2,FALSE} , {2,FALSE} , {3,FALSE} , {3,FALSE}
   };
   *pState = fgRS[*pState].new_state;
}

/*------------------------------------------------------------------declare_RE-+
| Register a Record End                                                        |
+-----------------------------------------------------------------------------*/
inline boolean RecordState::declare_RE()
{
   static RecordState::FlowGraphNode fgRE[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+-----------
         {2,FALSE} , {3,FALSE} , {3,FALSE} , {3,TRUE } , {3,FALSE}
   };
   RecordState::FlowGraphNode old_node = fgRE[*pState];
   *pState = old_node.new_state;
   return old_node.re_generated;
}

/*---------------------------------------------------declare_start_proper_elem-+
| Register the start of a proper element                                       |
+-----------------------------------------------------------------------------*/
inline boolean RecordState::declare_start_proper_elem()
{
   static RecordState::FlowGraphNode fgSTG[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+-----------
         {0,FALSE} , {0,FALSE} , {0,FALSE} , {0,TRUE } , {0,TRUE }
   };
   FlowGraphNode old_node = fgSTG[*pState];
   *pState = old_node.new_state;
   return old_node.re_generated;
}

/*-----------------------------------------------------declare_end_proper_elem-+
| Register the end of a proper element                                         |
+-----------------------------------------------------------------------------*/
inline void RecordState::declare_end_proper_elem()
{
   static RecordState::FlowGraphNode fgETG[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+-----------
         {1,FALSE} , {1,FALSE} , {1,FALSE} , {1,FALSE} , {1,FALSE}
   };
   *pState = fgETG[*pState].new_state;
}

/*----------------------------------------------------------------declare_data-+
| Register data                                                                |
+-----------------------------------------------------------------------------*/
inline boolean RecordState::declare_data()
{
   static RecordState::FlowGraphNode fgDATA[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+-----------
         {1,FALSE} , {1,FALSE} , {1,FALSE} , {1,TRUE } , {1,TRUE }
   };
   RecordState::FlowGraphNode old_node = fgDATA[*pState];
   *pState = old_node.new_state;
   return old_node.re_generated;
}

/*----------------------------------------------------declare_other_markup_dcl-+
| Register other markp declaration                                             |
+-----------------------------------------------------------------------------*/
inline void RecordState::declare_other_markup_dcl()
{
   static RecordState::FlowGraphNode fgOTHER[] = {
      //    {New State, re_generated} = f[Events, OldState]
      //    0      ,     1     ,     2     ,     3     ,     4
      //-----------+-----------+-----------+-----------+-----------
         {0,FALSE} , {1,FALSE} , {0,FALSE} , {4,FALSE} , {4,FALSE}
   };
   *pState = fgOTHER[*pState].new_state;
}

/*-------------------------------------------------declare_start_included_elem-+
| Enter a new flowgraph at the start of an included sub-element                |
+-----------------------------------------------------------------------------*/
inline void RecordState::declare_start_included_elem()
{
   declare_other_markup_dcl();
   *(++pState) = 0;
}

/*---------------------------------------------------declare_end_included_elem-+
| Resume previous flowgraph at the end of an included sub-element              |
+-----------------------------------------------------------------------------*/
inline void RecordState::declare_end_included_elem()
{
   --pState;
}
#endif /* _RCDSTATE_H *********************************************************/
