/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// refstr.c
///  Reference strings of the concrete syntax
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| This module is created by GnName (NOT YET CONVERTED)
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "refstr.h"
#include "dlmstr.h"
#include "rsvname.h"
#include <stdlib.h>
#include <string.h>

enum e_NameIdentification {
   DLVL_CHARID = 'a',
   CAPA_CHARID = 'c',
   DLIM_CHARID = 'd',
   FCLS_CHARID = 'f',
   GEN_CHARID  = 'g',
   QNTY_CHARID = 'q',
   SGML_CHARID = 's',
   SYST_CHARID = 'y'
};

static const char iso_bnr[14]   =  "ISO 8879:1986";
static const char dAND[5]       =  "dAND"         ;
static const char gANY[5]       =  "gANY"         ;
static const char sAPPINFO[9]   =  "sAPPINFO"     ;
static const char yASN1[6]      =  "yASN1"        ;
static const char cATTCAP[8]    =  "cATTCAP"      ;
static const char cATTCHCAP[10] =  "cATTCHCAP"    ;
static const char qATTCNT[8]    =  "qATTCNT"      ;
static const char gATTLIST[9]   =  "gATTLIST"     ;
static const char qATTSPLEN[10] =  "qATTSPLEN"    ;
static const char cAVGRPCAP[10] =  "cAVGRPCAP"    ;
static const char sBASESET[9]   =  "sBASESET"     ;
static const char qBSEQLEN[9]   =  "qBSEQLEN"     ;
static const char gCAPACITY[10] =  "gCAPACITY"    ;
static const char gCDATA[7]     =  "gCDATA"       ;
static const char yCHANGES[9]   =  "yCHANGES"     ;
static const char gCHARSET[9]   =  "gCHARSET"     ;
static const char dCOM[5]       =  "dCOM"         ;
static const char sCONCUR[8]    =  "sCONCUR"      ;
static const char gCONREF[8]    =  "gCONREF"      ;
static const char sCONTROLS[10] =  "sCONTROLS"    ;
static const char dCRO[5]       =  "dCRO"         ;
static const char gCURRENT[9]   =  "gCURRENT"     ;
static const char sDATATAG[9]   =  "sDATATAG"     ;
static const char gDEFAULT[9]   =  "gDEFAULT"     ;
static const char sDELIM[7]     =  "sDELIM"       ;
static const char yDELIMLEN[10] =  "yDELIMLEN"    ;
static const char sDESCSET[9]   =  "sDESCSET"     ;
static const char gDOCTYPE[9]   =  "gDOCTYPE"     ;
static const char gDOCUMENT[10] =  "gDOCUMENT"    ;
static const char dDSC[5]       =  "dDSC"         ;
static const char dDSO[5]       =  "dDSO"         ;
static const char cDTACAP[8]    =  "cDTACAP"      ;
static const char cDTACHCAP[10] =  "cDTACHCAP"    ;
static const char qDTAGLEN[9]   =  "qDTAGLEN"     ;
static const char gDTD[5]       =  "gDTD"         ;
static const char qDTEMPLEN[10] =  "qDTEMPLEN"    ;
static const char dDTGC[6]      =  "dDTGC"        ;
static const char dDTGO[6]      =  "dDTGO"        ;
static const char cELEMCAP[9]   =  "cELEMCAP"     ;
static const char gELEMENT[9]   =  "gELEMENT"     ;
static const char gELEMENTS[10] =  "gELEMENTS"    ;
static const char gEMPTY[7]     =  "gEMPTY"       ;
static const char gENDTAG[8]    =  "gENDTAG"      ;
static const char cENTCAP[8]    =  "cENTCAP"      ;
static const char cENTCHCAP[10] =  "cENTCHCAP"    ;
static const char gENTITIES[10] =  "gENTITIES"    ;
static const char gENTITY[8]    =  "gENTITY"      ;
static const char qENTLVL[8]    =  "qENTLVL"      ;
static const char dERO[5]       =  "dERO"         ;
static const char dETAGO[7]     =  "dETAGO"       ;
static const char yEXCLUDE[9]   =  "yEXCLUDE"     ;
static const char cEXGRPCAP[10] =  "cEXGRPCAP"    ;
static const char cEXNMCAP[9]   =  "cEXNMCAP"     ;
static const char sEXPLICIT[10] =  "sEXPLICIT"    ;
static const char sFEATURES[10] =  "sFEATURES"    ;
static const char gFIXED[7]     =  "gFIXED"       ;
static const char sFORMAL[8]    =  "sFORMAL"      ;
static const char fFUNCHAR[9]   =  "fFUNCHAR"     ;
static const char sFUNCTION[10] =  "sFUNCTION"    ;
static const char sGENERAL[9]   =  "sGENERAL"     ;
static const char dGRPC[6]      =  "dGRPC"        ;
static const char cGRPCAP[8]    =  "cGRPCAP"      ;
static const char qGRPCNT[8]    =  "qGRPCNT"      ;
static const char qGRPGTCNT[10] =  "qGRPGTCNT"    ;
static const char qGRPLVL[8]    =  "qGRPLVL"      ;
static const char dGRPO[6]      =  "dGRPO"        ;
static const char gID[4]        =  "gID"          ;
static const char cIDCAP[7]     =  "cIDCAP"       ;
static const char gIDLINK[8]    =  "gIDLINK"      ;
static const char gIDREF[7]     =  "gIDREF"       ;
static const char cIDREFCAP[10] =  "cIDREFCAP"    ;
static const char gIDREFS[8]    =  "gIDREFS"      ;
static const char gIGNORE[8]    =  "gIGNORE"      ;
static const char sIMPLICIT[10] =  "sIMPLICIT"    ;
static const char gIMPLIED[9]   =  "gIMPLIED"     ;
static const char gINCLUDE[9]   =  "gINCLUDE"     ;
static const char gINITIAL[9]   =  "gINITIAL"     ;
static const char sINSTANCE[10] =  "sINSTANCE"    ;
static const char sLCNMCHAR[10] =  "sLCNMCHAR"    ;
static const char sLCNMSTRT[10] =  "sLCNMSTRT"    ;
static const char gLINK[6]      =  "gLINK"        ;
static const char gLINKTYPE[10] =  "gLINKTYPE"    ;
static const char dLIT[5]       =  "dLIT"         ;
static const char dLITA[6]      =  "dLITA"        ;
static const char qLITLEN[8]    =  "qLITLEN"      ;
static const char cLKNMCAP[9]   =  "cLKNMCAP"     ;
static const char cLKSETCAP[10] =  "cLKSETCAP"    ;
static const char gLPD[5]       =  "gLPD"         ;
static const char cMAPCAP[8]    =  "cMAPCAP"      ;
static const char gMD[4]        =  "gMD"          ;
static const char dMDC[5]       =  "dMDC"         ;
static const char dMDO[5]       =  "dMDO"         ;
static const char sMINIMIZE[10] =  "sMINIMIZE"    ;
static const char dMINUS[7]     =  "dMINUS"       ;
static const char yMODEL[7]     =  "yMODEL"       ;
static const char gMS[4]        =  "gMS"          ;
static const char dMSC[5]       =  "dMSC"         ;
static const char fMSICHAR[9]   =  "fMSICHAR"     ;
static const char fMSOCHAR[9]   =  "fMSOCHAR"     ;
static const char fMSSCHAR[9]   =  "fMSSCHAR"     ;
static const char gNAME[6]      =  "gNAME"        ;
static const char sNAMECASE[10] =  "sNAMECASE"    ;
static const char qNAMELEN[9]   =  "qNAMELEN"     ;
static const char gNAMES[7]     =  "gNAMES"       ;
static const char sNAMING[8]    =  "sNAMING"      ;
static const char gNDATA[7]     =  "gNDATA"       ;
static const char dNET[5]       =  "dNET"         ;
static const char gNMTOKEN[9]   =  "gNMTOKEN"     ;
static const char gNMTOKENS[10] =  "gNMTOKENS"    ;
static const char sNO[4]        =  "sNO"          ;
static const char sNONE[6]      =  "sNONE"        ;
static const char gNONSGML[9]   =  "gNONSGML"     ;
static const char qNORMSEP[9]   =  "qNORMSEP"     ;
static const char gNOTATION[10] =  "gNOTATION"    ;
static const char cNOTCAP[8]    =  "cNOTCAP"      ;
static const char cNOTCHCAP[10] =  "cNOTCHCAP"    ;
static const char gNUMBER[8]    =  "gNUMBER"      ;
static const char gNUMBERS[9]   =  "gNUMBERS"     ;
static const char gNUTOKEN[9]   =  "gNUTOKEN"     ;
static const char gNUTOKENS[10] =  "gNUTOKENS"    ;
static const char gO[3]         =  "gO"           ;
static const char sOMITTAG[9]   =  "sOMITTAG"     ;
static const char dOPT[5]       =  "dOPT"         ;
static const char dOR[4]        =  "dOR"          ;
static const char sOTHER[7]     =  "sOTHER"       ;
static const char yPACK[6]      =  "yPACK"        ;
static const char gPCDATA[8]    =  "gPCDATA"      ;
static const char dPERO[6]      =  "dPERO"        ;
static const char gPI[4]        =  "gPI"          ;
static const char dPIC[5]       =  "dPIC"         ;
static const char qPILEN[7]     =  "qPILEN"       ;
static const char dPIO[5]       =  "dPIO"         ;
static const char dPLUS[6]      =  "dPLUS"        ;
static const char gPOSTLINK[10] =  "gPOSTLINK"    ;
static const char gPUBLIC[8]    =  "gPUBLIC"      ;
static const char sQUANTITY[10] =  "sQUANTITY"    ;
static const char sRANK[6]      =  "sRANK"        ;
static const char gRCDATA[8]    =  "gRCDATA"      ;
static const char gRE[4]        =  "gRE"          ;
static const char dREFC[6]      =  "dREFC"        ;
static const char dREP[5]       =  "dREP"         ;
static const char gREQUIRED[10] =  "gREQUIRED"    ;
static const char gRESTORE[9]   =  "gRESTORE"     ;
static const char dRNI[5]       =  "dRNI"         ;
static const char gRS[4]        =  "gRS"          ;
static const char sSCOPE[7]     =  "sSCOPE"       ;
static const char gSDATA[7]     =  "gSDATA"       ;
static const char ySDIF[6]      =  "ySDIF"        ;
static const char fSEPCHAR[9]   =  "fSEPCHAR"     ;
static const char dSEQ[5]       =  "dSEQ"         ;
static const char ySEQUENCE[10] =  "ySEQUENCE"    ;
static const char sSGML[6]      =  "sSGML"        ;
static const char sSGMLREF[9]   =  "sSGMLREF"     ;
static const char gSHORTREF[10] =  "gSHORTREF"    ;
static const char sSHORTTAG[10] =  "sSHORTTAG"    ;
static const char sSHUNCHAR[10] =  "sSHUNCHAR"    ;
static const char gSIMPLE[8]    =  "gSIMPLE"      ;
static const char gSPACE[7]     =  "gSPACE"       ;
static const char ySRCNT[7]     =  "ySRCNT"       ;
static const char ySRLEN[7]     =  "ySRLEN"       ;
static const char dSTAGO[7]     =  "dSTAGO"       ;
static const char gSTARTTAG[10] =  "gSTARTTAG"    ;
static const char gSUBDOC[8]    =  "gSUBDOC"      ;
static const char sSWITCHES[10] =  "sSWITCHES"    ;
static const char gSYNTAX[8]    =  "gSYNTAX"      ;
static const char gSYSTEM[8]    =  "gSYSTEM"      ;
static const char sTAB[5]       =  "sTAB"         ;
static const char dTAGC[6]      =  "dTAGC"        ;
static const char qTAGLEN[8]    =  "qTAGLEN"      ;
static const char qTAGLVL[8]    =  "qTAGLVL"      ;
static const char gTEMP[6]      =  "gTEMP"        ;
static const char gTEXT[6]      =  "gTEXT"        ;
static const char cTOTALCAP[10] =  "cTOTALCAP"    ;
static const char sUCNMCHAR[10] =  "sUCNMCHAR"    ;
static const char sUCNMSTRT[10] =  "sUCNMSTRT"    ;
static const char yUNPACK[8]    =  "yUNPACK"      ;
static const char sUNUSED[8]    =  "sUNUSED"      ;
static const char gUSELINK[9]   =  "gUSELINK"     ;
static const char gUSEMAP[8]    =  "gUSEMAP"      ;
static const char yVALIDATE[10] =  "yVALIDATE"    ;
static const char dVI[4]        =  "dVI"          ;
static const char sYES[5]       =  "sYES"         ;

static const char aAND[2]       =  "&"   ;
static const char aCOM[3]       =  "--"  ;
static const char aCRO[3]       =  "&#"  ;
static const char aDSC[2]       =  "]"   ;
static const char aDSO[2]       =  "["   ;
static const char aDTGC[2]      =  "]"   ;
static const char aDTGO[2]      =  "["   ;
static const char aERO[2]       =  "&"   ;
static const char aETAGO[3]     =  "</"  ;
static const char aGRPC[2]      =  ")"   ;
static const char aGRPO[2]      =  "("   ;
static const char aLIT[3]       =  "\""  ;
static const char aLITA[2]      =  "'"   ;
static const char aMDC[2]       =  ">"   ;
static const char aMDO[3]       =  "<!"  ;
static const char aMINUS[2]     =  "-"   ;
static const char aMSC[3]       =  "]]"  ;
static const char aNET[2]       =  "/"   ;
static const char aOPT[2]       =  "?"   ;
static const char aOR[2]        =  "|"   ;
static const char aPERO[2]      =  "%"   ;
static const char aPIC[2]       =  ">"   ;
static const char aPIO[3]       =  "<?"  ;
static const char aPLUS[2]      =  "+"   ;
static const char aREFC[2]      =  ";"   ;
static const char aREP[2]       =  "*"   ;
static const char aRNI[2]       =  "#"   ;
static const char aSEQ[2]       =  ","   ;
static const char aSTAGO[2]     =  "<"   ;
static const char aTAGC[2]      =  ">"   ;
static const char aVI[2]        =  "="   ;
static const char aMDO_DSO[4]   =  "<![" ;
static const char aMSC_MDC[4]   =  "]]>" ;

/*-----------------------------------------------------------------------------+
| `achReferNameValues' is a table of all (constant) reserved names of the      |
| Reference, including the ones found in the System Declaration.  Each         |
| keyword is preceeded by a character that identify the type of the            |
| keyword, according to the CHARID values.  Indices in `aRsvnameRefer' are     |
| found in the `refer_rsvname_ix' enum.                                        |
+-----------------------------------------------------------------------------*/
static const char *achReferNameValues[RRN_MAX_SIZE] = {
   dAND+1     , gANY+1     , sAPPINFO+1 , yASN1+1    ,
   cATTCAP+1  , cATTCHCAP+1, qATTCNT+1  , gATTLIST+1 ,
   qATTSPLEN+1, cAVGRPCAP+1, sBASESET+1 , qBSEQLEN+1 ,
   gCAPACITY+1, gCDATA+1   , yCHANGES+1 , gCHARSET+1 ,
   dCOM+1     , sCONCUR+1  , gCONREF+1  , sCONTROLS+1,
   dCRO+1     , gCURRENT+1 , sDATATAG+1 , gDEFAULT+1 ,
   sDELIM+1   , yDELIMLEN+1, sDESCSET+1 , gDOCTYPE+1 ,
   gDOCUMENT+1, dDSC+1     , dDSO+1     , qDTAGLEN+1 ,
   gDTD+1     , qDTEMPLEN+1, dDTGC+1    , dDTGO+1    ,
   cELEMCAP+1 , gELEMENT+1 , gELEMENTS+1, gEMPTY+1   ,
   gENDTAG+1  , cENTCAP+1  , cENTCHCAP+1, gENTITIES+1,
   gENTITY+1  , qENTLVL+1  , dERO+1     , dETAGO+1   ,
   yEXCLUDE+1 , cEXGRPCAP+1, cEXNMCAP+1 , sEXPLICIT+1,
   sFEATURES+1, gFIXED+1   , sFORMAL+1  , fFUNCHAR+1 ,
   sFUNCTION+1, sGENERAL+1 , dGRPC+1    , cGRPCAP+1  ,
   qGRPCNT+1  , qGRPGTCNT+1, qGRPLVL+1  , dGRPO+1    ,
   gID+1      , cIDCAP+1   , gIDLINK+1  , gIDREF+1   ,
   cIDREFCAP+1, gIDREFS+1  , gIGNORE+1  , sIMPLICIT+1,
   gIMPLIED+1 , gINCLUDE+1 , gINITIAL+1 , sINSTANCE+1,
   sLCNMCHAR+1, sLCNMSTRT+1, gLINK+1    , gLINKTYPE+1,
   dLIT+1     , dLITA+1    , qLITLEN+1  , cLKNMCAP+1 ,
   cLKSETCAP+1, gLPD+1     , cMAPCAP+1  , gMD+1      ,
   dMDC+1     , dMDO+1     , sMINIMIZE+1, dMINUS+1   ,
   yMODEL+1   , gMS+1      , dMSC+1     , fMSICHAR+1 ,
   fMSOCHAR+1 , fMSSCHAR+1 , gNAME+1    , sNAMECASE+1,
   qNAMELEN+1 , gNAMES+1   , sNAMING+1  , gNDATA+1   ,
   dNET+1     , gNMTOKEN+1 , gNMTOKENS+1, sNO+1      ,
   sNONE+1    , gNONSGML+1 , qNORMSEP+1 , gNOTATION+1,
   cNOTCAP+1  , cNOTCHCAP+1, gNUMBER+1  , gNUMBERS+1 ,
   gNUTOKEN+1 , gNUTOKENS+1, gO+1       , sOMITTAG+1 ,
   dOPT+1     , dOR+1      , sOTHER+1   , yPACK+1    ,
   gPCDATA+1  , dPERO+1    , gPI+1      , dPIC+1     ,
   qPILEN+1   , dPIO+1     , dPLUS+1    , gPOSTLINK+1,
   gPUBLIC+1  , sQUANTITY+1, sRANK+1    , gRCDATA+1  ,
   gRE+1      , dREFC+1    , dREP+1     , gREQUIRED+1,
   gRESTORE+1 , dRNI+1     , gRS+1      , sSCOPE+1   ,
   gSDATA+1   , ySDIF+1    , fSEPCHAR+1 , dSEQ+1     ,
   ySEQUENCE+1, sSGML+1    , sSGMLREF+1 , gSHORTREF+1,
   sSHORTTAG+1, sSHUNCHAR+1, gSIMPLE+1  , gSPACE+1   ,
   ySRCNT+1   , ySRLEN+1   , dSTAGO+1   , gSTARTTAG+1,
   gSUBDOC+1  , sSWITCHES+1, gSYNTAX+1  , gSYSTEM+1  ,
   sTAB+1     , dTAGC+1    , qTAGLEN+1  , qTAGLVL+1  ,
   gTEMP+1    , gTEXT+1    , cTOTALCAP+1, sUCNMCHAR+1,
   sUCNMSTRT+1, yUNPACK+1  , sUNUSED+1  , gUSELINK+1 ,
   gUSEMAP+1  , yVALIDATE+1, dVI+1      , sYES+1
};

/*-----------------------------------------------------------------------------+
| `aKywdVarntNameValues' is a table of the "actual" reserved names, as a       |
| variant concrete syntax could have change any of them.  This table           |
| contains only the names that may be found in prologs and document            |
| instances.  Indices in `aRsvnameVarnt' are found in the                      |
| `varnt_rsvname_ix' enum.                                                     |
+-----------------------------------------------------------------------------*/
static KeyWord aKywdVarntNameValues[VRN_MAX_SIZE] = {
   gANY+1     , gATTLIST+1 , gCAPACITY+1, gCDATA+1   ,
   gCHARSET+1 , gCONREF+1  , gCURRENT+1 , gDEFAULT+1 ,
   gDOCTYPE+1 , gDOCUMENT+1, gDTD+1     , gELEMENT+1 ,
   gELEMENTS+1, gEMPTY+1   , gENDTAG+1  , gENTITIES+1,
   gENTITY+1  , gFIXED+1   , gID+1      , gIDLINK+1  ,
   gIDREF+1   , gIDREFS+1  , gIGNORE+1  , gIMPLIED+1 ,
   gINCLUDE+1 , gINITIAL+1 , gLINK+1    , gLINKTYPE+1,
   gLPD+1     , gMD+1      , gMS+1      , gNAME+1    ,
   gNAMES+1   , gNDATA+1   , gNMTOKEN+1 , gNMTOKENS+1,
   gNONSGML+1 , gNOTATION+1, gNUMBER+1  , gNUMBERS+1 ,
   gNUTOKEN+1 , gNUTOKENS+1, gO+1       , gPCDATA+1  ,
   gPI+1      , gPOSTLINK+1, gPUBLIC+1  , gRCDATA+1  ,
   gRE+1      , gREQUIRED+1, gRESTORE+1 , gRS+1      ,
   gSDATA+1   , gSHORTREF+1, gSIMPLE+1  , gSPACE+1   ,
   gSTARTTAG+1, gSUBDOC+1  , gSYNTAX+1  , gSYSTEM+1  ,
   gTEMP+1    , gTEXT+1    , gUSELINK+1 , gUSEMAP+1
};

/*-----------------------------------------------------+
| `ReferDlmValues' is a structure defining the string  |
| value of all (constant) delimiters of the            |
| Reference Concrete Syntax.                           |
+-----------------------------------------------------*/
static const char *achReferDlmValues[RDL_MAX_SIZE] = {
   aAND      , aCOM      , aCRO      , aDSC      , aDSO      ,
   aDTGC     , aDTGO     , aERO      , aETAGO    , aGRPC     ,
   aGRPO     , aLIT      , aLITA     , aMDC      , aMDO      ,
   aMINUS    , aMSC      , aNET      , aOPT      , aOR       ,
   aPERO     , aPIC      , aPIO      , aPLUS     , aREFC     ,
   aREP      , aRNI      , aSEQ      , aSTAGO    , aTAGC     ,
   aVI       , aMDO_DSO  , aMSC_MDC
};

/*-------------------------------------------------------+
| `ReferQuantityNames' is a structure defining the names |
| of all quantities of the Reference Concrete Syntax.    |
+-------------------------------------------------------*/
static const char *achReferQuantityNames[QTY_MAX_SIZE] = {
   qATTCNT+1  , qATTSPLEN+1, qBSEQLEN+1 , qDTAGLEN+1 , qDTEMPLEN+1,
   qENTLVL+1  , qGRPCNT+1  , qGRPGTCNT+1, qGRPLVL+1  , qLITLEN+1  ,
   qNAMELEN+1 , qNORMSEP+1 , qPILEN+1   , qTAGLEN+1  , qTAGLVL+1
};

/*-------------------------------------------------------+
| `ReferCapacityNames' is a structure defining the names |
| of all capacities of the Reference Concrete Syntax.    |
+-------------------------------------------------------*/
static const char *achReferCapacityNames[CAP_MAX_SIZE] = {
   cATTCAP+1  , cATTCHCAP+1, cAVGRPCAP+1,
   cDTACAP+1  , cDTACHCAP+1,
   cELEMCAP+1 , cENTCAP+1  ,
   cENTCHCAP+1, cEXGRPCAP+1, cEXNMCAP+1 , cGRPCAP+1  , cIDCAP+1   ,
   cIDREFCAP+1, cLKNMCAP+1 , cLKSETCAP+1, cMAPCAP+1  , cNOTCAP+1  ,
   cNOTCHCAP+1, cTOTALCAP+1
};

/*----------------------------------------------+
| This is a table for Variant Delimiter strings |
| Not *all* delimiters are included             |
+----------------------------------------------*/
static KeyWord aKywdVarntDlmValues[VDL_MAX_SIZE]  = {
   aCOM,     aCRO,     aERO,     aETAGO,
   aLIT,     aLITA,    aMDC,     aMDO,
   aMINUS,   aPERO,    aPIC,     aPIO,
   aREFC,    aRNI,     aSTAGO,   aTAGC,
   aVI,      aMDO_DSO, aMSC_MDC
};

/*-----------------------------------------------------+
| `DelimiterItem' and `DoubleDelimiterItem' structure  |
| define the characteristics of the Reference Concrete |
| Syntax delimiters as shown in figure 3 and 4 of      |
| ISO 8879.                                            |
+-----------------------------------------------------*/
struct DelimiterItem {        /* DLMV: */
   const char *pszName;
   e_LexToken token;
   unsigned int allmode;
   const char *assigned;
   e_IxDlmstrVarnt ix;
};

static DelimiterItem const dlmv_tab[DLI_MAX_SIZE] = {
   { dAND+1     , AND_TK     ,    4, aAND  , NOT_FOUND_VDL_ix },
   { dCOM+1     , COM_TK     ,   16, aCOM  , COM_VDL_ix       },
   { dCRO+1     , CRO_TK     , 1417, aCRO  , CRO_VDL_ix       },
   { dDSC+1     , DSC_TK     ,   82, aDSC  , NOT_FOUND_VDL_ix },
   { dDSO+1     , DSO_TK     ,   16, aDSO  , NOT_FOUND_VDL_ix },
   { dDTGC+1    , DTGC_TK    ,    4, aDTGC , NOT_FOUND_VDL_ix },
   { dDTGO+1    , DTGO_TK    ,    4, aDTGO , NOT_FOUND_VDL_ix },
   { dERO+1     , ERO_TK     , 1289, aERO  , ERO_VDL_ix       },
   { dETAGO+1   , ETAGO_TK   ,  801, aETAGO, ETAGO_VDL_ix     },
   { dGRPC+1    , GRPC_TK    ,    4, aGRPC , NOT_FOUND_VDL_ix },
   { dGRPO+1    , GRPO_TK    ,   20, aGRPO , NOT_FOUND_VDL_ix },
   { dLIT+1     , LIT_TK     ,  252, aLIT  , LIT_VDL_ix       },
   { dLITA+1    , LITA_TK    ,  252, aLITA , LITA_VDL_ix      },
   { dMDC+1     , MDC_TK     ,   16, aMDC  , MDC_VDL_ix       },
   { dMDO+1     , MDO_TK     , 8195, aMDO  , MDO_VDL_ix       },
   { dMINUS+1   , MINUS_TK   ,    0, aMINUS, MINUS_VDL_ix     },
   { dMSC+1     , MSC_TK     ,    0, aMSC  , NOT_FOUND_VDL_ix },
   { dNET+1     , NET_TK     ,  801, aNET  , NOT_FOUND_VDL_ix },
   { dOPT+1     , OPT_TK     ,    4, aOPT  , NOT_FOUND_VDL_ix },
   { dOR+1      , OR_TK      ,    4, aOR   , NOT_FOUND_VDL_ix },
   { dPERO+1    , PERO_TK    ,  150, aPERO , PERO_VDL_ix      },
   { dPIC+1     , PIC_TK     ,    0, aPIC  , PIC_VDL_ix       },
   { dPIO+1     , PIO_TK     , 8195, aPIO  , PIO_VDL_ix       },
   { dPLUS+1    , PLUS_TK    ,    4, aPLUS , NOT_FOUND_VDL_ix },
   { dREFC+1    , REFC_TK    ,    0, aREFC , REFC_VDL_ix      },
   { dREP+1     , REP_TK     ,    4, aREP  , NOT_FOUND_VDL_ix },
   { dRNI+1     , RNI_TK     ,   20, aRNI  , RNI_VDL_ix       },
   { dSEQ+1     , SEQ_TK     ,    4, aSEQ  , NOT_FOUND_VDL_ix },
   { gSHORTREF+1, SHORTREF_TK,    1, 0     , NOT_FOUND_VDL_ix },
   { dSTAGO+1   , STAGO_TK   , 8225, aSTAGO, STAGO_VDL_ix     },
   { dTAGC+1    , TAGC_TK    ,   32, aTAGC , TAGC_VDL_ix      },
   { dVI+1      , VI_TK      ,   96, aVI   , VI_VDL_ix        }
};

struct DoubleDelimiterItem {  /* DDLMV: */
   e_LexToken tk1;
   e_LexToken tk2;
   e_LexToken token;
   unsigned int allmode;
};

static DoubleDelimiterItem const ddlmv_tab[DDLI_MAX_SIZE] = {
   { ERO_TK     , GRPO_TK    , ERO_GRPO_TK        , 1289 },
   { ETAGO_TK   , TAGC_TK    , ETAGO_TAGC_TK      ,  801 },
   { ETAGO_TK   , GRPO_TK    , ETAGO_GRPO_TK      ,  801 },
   { MDO_TK     , COM_TK     , MDO_COM_TK         , 8195 },
   { MDO_TK     , DSO_TK     , MDO_DSO_TK         , 4099 },
   { MDO_TK     , MDC_TK     , MDO_MDC_TK         , 8195 },
   { MSC_TK     , MDC_TK     , MSC_MDC_TK         , 7171 },
   { PERO_TK    , GRPO_TK    , PERO_GRPO_TK       ,  150 },
   { STAGO_TK   , TAGC_TK    , STAGO_TAGC_TK      , 8225 },
   { STAGO_TK   , GRPO_TK    , STAGO_GRPO_TK      , 8225 },
   { MINUS_TK   , GRPO_TK    , MINUS_GRPO_TK      ,   16 },
   { PLUS_TK    , GRPO_TK    , PLUS_GRPO_TK       ,   16 }
};

/*-------------------------------------------------------------------get_index-+
| Performs a binary search in the aKywdVarntNameValues sorted array.           |
| Returns the index or NOT_FOUND_VRN_ix if not found.                          |
+-----------------------------------------------------------------------------*/
void ReferenceString::get_index(
   const char *needle, e_IxRsvnameVarnt * const pix
)
{
   const KeyWord *haystack = aKywdVarntNameValues;
   int num = (sizeof aKywdVarntNameValues / sizeof aKywdVarntNameValues[0]) - 1;

   int low = 0;                    // low bound
   while (low <= num) {
      int mid = ((unsigned int)low + (unsigned int)num) / 2;
      int cond = strcmp(needle, (haystack + mid)->p);
      if (cond < 0) {
         num = mid - 1;            // set low range
      }else if (cond > 0) {
         low = mid + 1;
      }else {                      // found
         *pix = (e_IxRsvnameVarnt)mid;
         return;
      }
   }
   *pix = NOT_FOUND_VRN_ix;
}

/*-------------------------------------------------------------------get_index-+
| Performs a binary search in the dlmv_tab sorted array.                       |
| Returns the index or NOT_FOUND_DLI_ix if not found.                          |
+-----------------------------------------------------------------------------*/
void ReferenceString::get_index(
   const char *needle, e_IxDlmitmRefer * const pix
)
{
   const DelimiterItem *haystack = dlmv_tab;
   int num = (sizeof dlmv_tab / sizeof dlmv_tab[0]) - 1;

   int low = 0;                    // low bound
   while (low <= num) {
      int mid = ((unsigned int)low + (unsigned int)num) / 2;
      int cond = strcmp(needle, (haystack + mid)->pszName);
      if (cond < 0) {
         num = mid - 1;            // set low range
      }else if (cond > 0) {
         low = mid + 1;
      }else {                      // found
         *pix = (e_IxDlmitmRefer)mid;
         return;
      }
   }
   *pix = NOT_FOUND_DLI_ix;
}

/*-------------------------------------------------------------------get_index-+
| Performs a binary search in the achReferNameValues sorted array.             |
| Returns the index or NOT_FOUND_RRN_ix if not found.                          |
+-----------------------------------------------------------------------------*/
void ReferenceString::get_index(
   const char *needle, e_IxRsvnameRefer * const pix
)
{
   const char **haystack = achReferNameValues;
   int num = (sizeof achReferNameValues / sizeof achReferNameValues[0]) - 1;

   int low = 0;                    // low bound
   while (low <= num) {
      int mid = ((unsigned int)low + (unsigned int)num) / 2;
      int cond = strcmp(needle, haystack[mid]);
      if (cond < 0) {
         num = mid - 1;            // set low range
      }else if (cond > 0) {
         low = mid + 1;
      }else {                      // found
         *pix = (e_IxRsvnameRefer)mid;
         return;
      }
   }
   *pix = NOT_FOUND_RRN_ix;
}

/*-------------------------------------------------------------------get_index-+
| Performs a binary search in the achReferQuantityNames sorted array.          |
| Returns the index or NOT_FOUND_QTY_ix if not found.                          |
+-----------------------------------------------------------------------------*/
void ReferenceString::get_index(
   const char *needle, e_IxQuantity * const pix
)
{
   const char **haystack = achReferQuantityNames;
   int num = (
      sizeof achReferQuantityNames / sizeof achReferQuantityNames[0]
   ) - 1;

   int low = 0;                    // low bound
   while (low <= num) {
      int mid = ((unsigned int)low + (unsigned int)num) / 2;
      int cond = strcmp(needle, haystack[mid]);
      if (cond < 0) {
         num = mid - 1;            // set low range
      }else if (cond > 0) {
         low = mid + 1;
      }else {                      // found
         *pix = (e_IxQuantity)mid;
         return;
      }
   }
   *pix = NOT_FOUND_QTY_ix;
}

/*-------------------------------------------------------------------get_index-+
| Performs a binary search in the achReferCapacityNames sorted array.          |
| Returns the index or NOT_FOUND_CAP_ix if not found.                          |
+-----------------------------------------------------------------------------*/
void ReferenceString::get_index(
   const char *needle, e_IxCapacity * const pix
)
{
   const char **haystack = achReferCapacityNames;
   int num = (
      sizeof achReferCapacityNames / sizeof achReferCapacityNames[0]
   ) - 1;

   int low = 0;                    // low bound
   while (low <= num) {
      int mid = ((unsigned int)low + (unsigned int)num) / 2;
      int cond = strcmp(needle, haystack[mid]);
      if (cond < 0) {
         num = mid - 1;            // set low range
      }else if (cond > 0) {
         low = mid + 1;
      }else {                      // found
         *pix = (e_IxCapacity)mid;
         return;
      }
   }
   *pix = NOT_FOUND_CAP_ix;
}

/*--------------------------------------------------------clearance_to_replace-+
| Examine if a reserved name is allowed to be replaced                         |
| Return NULL_ER if OK, or an appropriate return code.                         |
+-----------------------------------------------------------------------------*/
e_ErrorCode ReferenceString::clearance_to_replace(
   const char *pszName, e_IxRsvnameVarnt * const pix
)
{
   e_ErrorCode msgcode = ER_NULL;

   get_index(pszName, pix);
   if (*pix == NOT_FOUND_VRN_ix) { // not found
      e_IxRsvnameRefer ix;
      get_index(pszName, &ix);
      if (ix == NOT_FOUND_RRN_ix) {
         msgcode = NOREFNAM;       // Not a keyword
      }else {
         switch ((achReferNameValues[ix])[-1]) {
         case SYST_CHARID:
            msgcode = NOVA2NAM;    // Reserved for the SYSTEM declaration
            break;
         default:                  // SGML_CHARID
            msgcode = NOVANAM;     // Reserved for the SGML declaration
            break;
         }
      }
   }
   return msgcode;
}

/*--------------------------------------------------------------ISO8879_banner-+
|                                                                              |
+-----------------------------------------------------------------------------*/
const char * ReferenceString::ISO8879_banner() {
   return iso_bnr;
}

/*--------------------------------------------------------------------token_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
e_LexToken ReferenceString::token_of(const e_IxDlmitmRefer ix)
{
   return dlmv_tab[ix].token;
}

/*--------------------------------------------------------------------token_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
e_LexToken ReferenceString::token_of(const e_IxDDlmitmRefer ix)
{
   return ddlmv_tab[ix].token;
}

/*-------------------------------------------------------------delim1_index_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
e_LexToken ReferenceString::delim1_token_of(const e_IxDDlmitmRefer ix)
{
   return ddlmv_tab[ix].tk1;
}

/*-------------------------------------------------------------delim2_index_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
e_LexToken ReferenceString::delim2_token_of(const e_IxDDlmitmRefer ix)
{
   return ddlmv_tab[ix].tk2;
}

/*--------------------------------------------------------------------value_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
const char *ReferenceString::value_of(const e_IxDlmitmRefer ix)
{
   return dlmv_tab[ix].assigned;
}

/*--------------------------------------------------------recognition_modes_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
unsigned int ReferenceString::recognition_modes_of(const e_IxDlmitmRefer ix)
{
   return dlmv_tab[ix].allmode;
}

/*--------------------------------------------------------recognition_modes_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
unsigned int ReferenceString::recognition_modes_of(const e_IxDDlmitmRefer ix)
{
   return ddlmv_tab[ix].allmode;
}

/*--------------------------------------------------------------delim_index_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
e_IxDlmstrVarnt ReferenceString::delim_index_of(const e_IxDlmitmRefer ix)
{
   return dlmv_tab[ix].ix;
}

/*------------------------------------------------------get_achReferNameValues-+
|                                                                              |
+-----------------------------------------------------------------------------*/
const char * ReferenceString::get_achReferNameValue(const int ix)
{
   return achReferNameValues[ix];
}

/*----------------------------------------------------get_aKywdVarntNameValues-+
|                                                                              |
+-----------------------------------------------------------------------------*/
KeyWord * ReferenceString::get_aKywdVarntNameValues()
{
   return aKywdVarntNameValues;
}

/*-------------------------------------------------------get_achReferDlmValues-+
|                                                                              |
+-----------------------------------------------------------------------------*/
const char * const * ReferenceString::get_achReferDlmValues()
{
   return achReferDlmValues;
}

/*-----------------------------------------------------get_aKywdVarntDlmValues-+
|                                                                              |
+-----------------------------------------------------------------------------*/
KeyWord * ReferenceString::get_aKywdVarntDlmValues()
{
   return aKywdVarntDlmValues;
}
/******************************************************************************/
