/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// rsvname.h
///  Declarations for rsvname
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _RSVNAME_H
#define _RSVNAME_H 1
//******************************************************************************

#include "refstr.h"     // this class is used by inlines

#include "rsvname.eh"   // enums

class StringKeeper;

/*-----------------------------------------------------------------------------+
| ReservedName class                                                           |
|                                                                              |
| This class describes the reserved names of the syntax.                       |
+-----------------------------------------------------------------------------*/
class ReservedName {             // Rsvname
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   ReservedName();                                          // Core constructor
   ReservedName(ApplicationHandle ahlAccess, char *strbuf); // OCS constructor

   static const char * name_of(const e_IxRsvnameRefer ix);  // Reference value
   const char * name_of(const e_IxRsvnameVarnt ix) const;   // Variant value

protected:
   /*----------------------+
   | Protected Definitions |
   +----------------------*/
   union {
      char *p;
      unsigned int ofs;
   }aRsvnameVarnt[VRN_MAX_SIZE];

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   ReservedName(int);            // when building from the SGML declaration
};

/*-----------------------------------------------------------------------------+
| NewReservedName class  (while building it)                                   |
+-----------------------------------------------------------------------------*/
class NewReservedName : public ReservedName
{
public:
   NewReservedName(StringKeeper& sk);
   const char * name_of(const e_IxRsvnameVarnt ix) const;  // Variant value
   unsigned int offset_of(const e_IxRsvnameVarnt ix) const;
   e_ErrorCode clearance_to_replace(
      const char *pszRsvnameReplaced,
      e_IxRsvnameVarnt * const pixRsvnameReplaced
   ) const;
   e_ErrorCode replace(
      const char *pszRsvnameReplacement,
      e_IxRsvnameVarnt * const pixRsvnameReplaced
   );
   void default_non_replaced(const Syntax *pSyntaxCore);
   void unload(ApplicationHandle ahlAccess);
private:
   StringKeeper& sk;
};

/*INLINE---------------------------------------------------------------name_of-+
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline const char * ReservedName::name_of(const e_IxRsvnameRefer ix) {
   return ReferenceString::get_achReferNameValue(ix);
}
inline const char * ReservedName::name_of(const e_IxRsvnameVarnt ix) const {
   return aRsvnameVarnt[ix].p;
}

/*INLINE-------------------------------------------------------------offset_of-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline unsigned int NewReservedName::offset_of(const e_IxRsvnameVarnt ix) const
{
   return aRsvnameVarnt[ix].ofs;
}
#endif /* _RSVNAME_H **********************************************************/
