/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// server.h
///  Declarations for server
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _SERVER_H
#define _SERVER_H 1
//******************************************************************************

#include <stdio.h>

class ExternId;
class Server;

class Document {         // doc
   friend Server;
public:
   _export Document(
      const char *pszPrimFileSpec = "TERMINAL",     // Primary EntityDef File Spec
      const char *pszOcsFileName = 0,               // OCS File to read / write
      const char *pszTerminalFileSpec = "TERMINAL", // Terminal reserved name
      boolean isProduction = FALSE,                 // Use a pre-compiled DTD
      boolean isSortAttributes = FALSE // Sort attribute lists lexicographically
   );
private:
   const char  *pszPrimFileSpec;
   const char  *pszOcsFileName;
   const char  *pszTerminalFileSpec;
   boolean isProduction;
   boolean isSortAttributes;
};

class Server {
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   Server(const Document& doc);    // Constructor
   ~Server();                      // Destructor

   int sr_find(                    // Find an ID-handle for a file
      const ExternId *pExid,
      ApplicationHandle& ahlIdent
   );
   int sr_xenop(                   // Open an External EntityDef (text file)
      const ApplicationHandle ahlIdent,
      ApplicationHandle& ahlAccess
   );
   int sr_xencl(                   // Close External EntityDef (text file)
      ApplicationHandle ahlAccess
   );
   int sr_xenrd(                   // Read an External EntityDef (text file)
      ApplicationHandle ahlAccess,
      char **ppchBuffer,
      int *iLength
   );
   int sr_xengp(                   // Get External EntityDef Position
      ApplicationHandle ahlAccess,
      void **ppSysPosBlock,
      int *iSizeSysPosBlock
   );
   int sr_xensp (                  // Set External EntityDef Position
      ApplicationHandle ahlAccess,
      void *pSysPosBlock
   );
   static int sr_utopen(           // Open an utility file
      const ApplicationHandle ahlIdent,
      ApplicationHandle *pahlAccess,
      const boolean isWriteMode
   );
   static int sr_utrd(             // Read an Utility file
      ApplicationHandle ahlAccess,
      void *pchBuffer,
      int iLength
   );
   static int sr_utwrt(            // Write an Utility file
      ApplicationHandle ahlAccess,
      const void *pchBuffer,
      int iLength
   );
   static int sr_utcl(             // Close an Utility file
      ApplicationHandle ahlAccess
   );
   boolean is_sort_attributes();   // attribute list must be sorted

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   struct FYLE;
   struct POSBLK {                 // Block to get/set a physical position
      friend struct FYLE;
      boolean isFirst     :1;
      boolean isAtEof     :1;
      boolean isTruncated :1;
      fpos_t fpos;                 // system position
   };

   struct FYLE {                   // For bufferizing the I/O
      FYLE *next;                  // Files chain
      FILE *fd;                    // File Descriptor
      char *pszName;               // File name
      int iBufSize;                // Buffer size
      boolean isTruncated :1;      // when a record is greater than BUFSIZE
      boolean isFirst     :1;      // First read is issued
      POSBLK posblk;               // Current record position
      char *buffer;                // Data transfer buffer
   };

   /*------------------+
   | Private Functions |
   +------------------*/
   boolean  set_current(FYLE *);

   /*------------------+
   | Private Variables |
   +------------------*/
   Document doc;                  // base values
   FYLE *pFyleTop;                // first fyle in chain
   FYLE *pFyleCur;                // current fyle in chain
   FYLE *pFyleTerm;               // fyle describing the TERMINAL
   FILE *err_fd;                  // stderr
   char *pchFnChain;              // filename chain
};

/*--- INLINES ---*/
inline boolean Server::is_sort_attributes() { return doc.isSortAttributes; }

#endif /* _SERVER_H **********************************************************/
