/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// sgmldcl.h
///  Declarations for sgmldcl
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _SGMLDCL_H
#define _SGMLDCL_H 1
//******************************************************************************

#include "skmgr.h"      // Part of the NewSGMLdcl class

#include "externid.eh"  // enums
#include "syntax.eh"
#include "rsvname.eh"

class ParseState;
class EventList;
class EntityHandler;
class ParseMode;
class Syntax;
class NewSyntax;
class UtilityControl;
class Capacity;
class NewCapacity;
class ExternId;

/*-----------------------------------------------------------------------------+
| Public structures                                                            |
+-----------------------------------------------------------------------------*/
struct Features {
   boolean isFormal      :1;    // formal public identifiers
   boolean isConcur      :1;    // concurrent document instances
   boolean isExplicit    :1;    // explicit link process
   boolean isImplicit    :1;    // implicit link process
   boolean isSimple      :1;    // simple link process
   boolean isSubdoc      :1;    // nested subdocuments
   boolean isRank        :1;    // omitted rank suffix minimization
   boolean isDatatag     :1;    // datatag minimization
   boolean isOmittag     :1;    // omitted tag minimization
   boolean isShorttag    :1;    // short tag minimization
   unsigned int iMaxSimple;     // Maximum number of simple links
   unsigned int iMaxExplicit;   // Maximum number of explicit links
   unsigned int iMaxConcur;     // Maximum number of concurrent documents
   unsigned int iMaxSubdoc;     // Maximum number of subdocs
};

#if (INT_MAX == SHRT_MAX)
#define MAX_SUBDOC   3
#define MAX_CONCUR   3
#define MAX_SIMPLE   3
#define MAX_EXPLICIT 3

#else
#define MAX_SUBDOC   10
#define MAX_CONCUR   10
#define MAX_SIMPLE   10
#define MAX_EXPLICIT 10
#endif

/*-----------------------------------------------------------------------------+
| NewSGMLdcl Class                                                             |
+-----------------------------------------------------------------------------*/
class NewSGMLdcl
{
public:
   /*-----------------+
   | Public functions |
   +-----------------*/
   NewSGMLdcl(                      // to compile an SGML declaration
      ParseState *pPrsstat, EventList& evlst
   );
   ~NewSGMLdcl();

   boolean compile();

private:
   /*--------------------+
   | Private Definitions |
   +--------------------*/
   struct CharDefinition {          // indexed by the system character position
      char  doc_d_chnb;             // code # in the document
      char  syn_d_chnb;             // code # in the syntax
      boolean isDescribedInDoc : 1; // 1 document code # describes this char
      boolean isDescribedInSyn : 1; // 1 syntax code # describes this char
      boolean isMultipleInDoc  : 1; // > 1 document code # describes this char
      boolean isMultipleInSyn  : 1; // > 1 syntax code # describes this char
      char upper;                   // s_chnb upper case form
   };

   /*-------------+
   | Private Data |
   +-------------*/
   EventList& evlst;
   ParseState *pPrsstat;
   EntityHandler *pEh;
   Syntax *pSyntax;
   ParseMode *pPrsmode;

   StringKeeper sk;

   NewSyntax *pSyntaxNew;
   NewCapacity *pCapacityNew;
   boolean isScopeInstanceNew;
   Features featuresNew;
   char *pszAppinfoNew;

   CharDefinition aChdefTable[2 * MAX_CHARSET];
   char achSwitchTable[MAX_CHARSET];
   char *parmlit[MAX_CHARSET];
   char *pchLitBuffer;

   boolean isSevereErrorFound :1;
   boolean isNotFormalPID     :1;
   boolean isMdoChanged       :1;
   boolean isPioChanged       :1;
   boolean isComChanged       :1;
   boolean isMdcChanged       :1;

   /*------------------+
   | Private Functions |
   +------------------*/
   void parse_function_parameter();
   void parse_naming_parameter();
   void parse_capacity_parameter();
   void parse_quantity_parameter();
   void parse_names_parameter();
   void parse_basesets_and_descsets(boolean isDocument);
   void parse_shunchar_parameter();
   void parse_delim_parameter();
   void unload(ApplicationHandle ahlAccess);
   unsigned int read_parameter_literal(
      char *pchBuffer, boolean is_solely_funchar
   );
   boolean answer_is_yes();
   int check_and_map_to_doc_chnb(char system_chnb);
   int map_to_doc_chnb(char syn_chnb);
   void check_instance_chars();
   void build_reference_shortref_literal(char *pchBuf);
   boolean get_feature(
      e_IxRsvnameRefer ixFeatureName,
      boolean isSupported,
      unsigned int *piNumber = 0,
      unsigned int iMaxNumber = 0
   );
   void process_features();
};

/*-----------------------------------------------------------------------------+
| SGMLdcl Class                                                                |
+-----------------------------------------------------------------------------*/
class SGMLdcl
{
public:
   /*-----------------+
   | Public Variables |
   +-----------------*/
   Syntax *pSyntax;
   Capacity *pCapacity;
   boolean isScopeInstance;         // DOCUMENT: 0, INSTANCE: 1
   Features features;
   char *pszAppinfo;

   /*-----------------+
   | Public Functions |
   +-----------------*/
   SGMLdcl();                              // Build minimal OCS
   SGMLdcl(ApplicationHandle ahlAccess);   // Build from an OCS file
   ~SGMLdcl();
};
#endif /* _SGMLDCL_H **********************************************************/
