/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// shortref.c
///  Shortref class initialization
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include "gendef.h"
#include "lextoken.eh"
#include "enthdlr.h"
#include "shortref.h"
#include "samgr.h"

/*--------------------------------------------------------------------Shortref-+
| Constructor from OCS                                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Shortref::Shortref(ApplicationHandle ahlAccess, char *strbuf)
{
   sr_max = EntityHandler::read_record(
      ahlAccess, (void **)&srtab
   ) / sizeof srtab[0];

   /*
   | Relocate srtab
   */
   {
      for (int i=0; i < sr_max; ++i) {
         srtab[i].p = strbuf + srtab[i].ofs;
      }
   }
   --sr_max;      // because there is an extraneous last entry in srtab
}

/*PROTECTED-----------------------------------------------------------Shortref-+
| Constructor for building shortrefs from the SGML declaration                 |
+-----------------------------------------------------------------------------*/
Shortref::Shortref(int)
{
   sr_max = 0;
   srtab = 0;
}

/*--------------------------------------------------------------------Shortref-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
Shortref::~Shortref()
{
   free(srtab);
}

/*--------------------------------------------------------------------token_of-+
| Get the lex token associated to a Shortref string                            |
|                                                                              |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
|                                                                              |
| Note: there is an extraneous last entry in srtab -- i.e:                     |
|       srtab[sr_max] == <"", NOT_FOUND_TK>                                    |
|       It's unused here.                                                      |
+-----------------------------------------------------------------------------*/
e_LexToken Shortref::token_of(char *pszShortrefString)
{
   for (int i=0; i < sr_max; ++i) {
      if (!strcmp(srtab[i].p, pszShortrefString)) return srtab[i].token;
   }
   return NOT_FOUND_TK;
}

/*-----------------------------------------------------------------NewShortref-+
| NewShortref constructor                                                      |
+-----------------------------------------------------------------------------*/
NewShortref::NewShortref(
   StringKeeper& skArg
) : Shortref(0), sk(skArg)
{
   pSaShortref = new SortArray(
      sizeof(ShortrefItem),      // size of each array item
      0,                         // comparison function
      &sr_max,                   // where to return count of items entered
      (void **)&srtab,           // where to return address of the array
      0
   );
}

/*-------------------------------------------------------------define_shortref-+
| Enter a new shortref in the SaShortref SortedArray                           |
+-----------------------------------------------------------------------------*/
void NewShortref::define_shortref(const e_LexToken token, const char *pszValue)
{
   ShortrefItem srwd;

   srwd.token = token;
   srwd.ofs = sk.park(pszValue, 1 + strlen(pszValue));  // in srtab
   pSaShortref->enter(&srwd);
}

/*-----------------------------------------------------------------------build-+
| Build the table of Shortrefs                                                 |
+-----------------------------------------------------------------------------*/
void NewShortref::build()
{
   ShortrefItem srwd;

   srwd.token = NOT_FOUND_TK;
   srwd.ofs = 0;                   // put a sentinel at the end
   pSaShortref->enter(&srwd);
   delete pSaShortref;
}

/*----------------------------------------------------------------------unload-+
| Write the shortrefs to disk                                                  |
+-----------------------------------------------------------------------------*/
void NewShortref::unload(ApplicationHandle ahlAccess)
{
   EntityHandler::write_record(ahlAccess, srtab, sr_max * sizeof(ShortrefItem));
}
/******************************************************************************/
