/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// shortref.h
///  Declarations for shortref
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _SHORTREF_H
#define _SHORTREF_H 1
/*
| The SHORTREF delimiter table is built by reading the SHORTREF
| parameter from the SGML declaration.  Then, it is used to recognize
| the parameter literal of the DTD shortref mapping declaration.
*/
//******************************************************************************

class StringKeeper;
class SortArray;

class Shortref {
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   Shortref(ApplicationHandle ahlAccess, char *strbuf);     // OCS constructor
   ~Shortref();
   int count();                         // max count of shortrefs
   e_LexToken token_of(char *pszShortrefString);

protected:
   /*----------------------+
   | Protected Definitions |
   +----------------------*/
   struct ShortrefItem {      // sr
      union {
         char *p;             // Shortref String
         unsigned int ofs;    // (offset when in OCS file)
      };
      e_LexToken token;       // associated lexical token
   };

   /*--------------------+
   | Protected Variables |
   +--------------------*/
   ShortrefItem *srtab;       // list of all (shortref string <-> token)
   unsigned int sr_max;       // number of entries in srtab

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   Shortref(int);             // when building from the SGML declaration
};

inline int Shortref::count() {  return sr_max; }

/*-----------------------------------------------------------------------------+
| NewShortref class  (while building it)                                       |
+-----------------------------------------------------------------------------*/
class NewShortref : public Shortref
{
public:
   NewShortref(StringKeeper& sk);
   void define_shortref(const e_LexToken token, const char *pszValue);
   void build();
   void unload(ApplicationHandle ahlAccess);
private:
   StringKeeper& sk;
   SortArray *pSaShortref;
};

#endif /* _SHORTREF_H *********************************************************/
