/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// wqsort.c
///  Hoares Quick Sort
//   Pierre G. Richard 87/06/30 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| Sort an array of elements "in place" using Hoare's Quicksort algorithm.
|
| void quick(num, compare, exchange, base)
|
| where:
| - "num" is the number of elements of the array,
| - "compare" is an user-supplied routine that compares two array elements,
| - "exchange" is an user-supplied routine that exchanges two array elements.
|
| The form of the call to "compare" is:
|
|  compare(element_1_index,element_2_index,base)
|  unsigned int element_1_index, element_2_index;
|
|  where: "element_1_index"/"element_2_index" is the index of the
|          element_1/element_2 within the array.
|
| This must return:
| - 0 if the key in element_1 is equal to the one in element_2,
| - Less than 0 if the key if the key in element_1 is less
|   than the one in element_2,
| - Greather than 0 if the key if the key in element_1 is is greather
|   than the one in element_2.
|
| The form of the call to "exchange" is:
|
|  exchange(element_1_index,element_2_index,base)
|  unsigned int element_1_index, element_2_index;
|
|  where: "element_1_index"/"element_2_index" is the index of the
|          element_1/element_2 within the array.
|
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "sort.h"

/*PRIVATE----------------------------------------------------------hoares_sort-+
|recursive function to sort array[beg..end] using Hoares algorithm             |
+-----------------------------------------------------------------------------*/
void Sort::hoares_sort(int beg, int end, W const * const w_p)
{
   int j, k;

   k = end + 1;
   while (beg < end) {
      j = beg;
      /*-------------------------------------------------+
      |Get "k" such that:                                |
      |in the range beg+1 to k, all values are <= V[beg] |
      +-------------------------------------------------*/
      for (;;) {
         while ((++j < k) && ((*w_p->comp)(j,beg,w_p->base) <= 0));
         while ((--k > j) && ((*w_p->comp)(k,beg,w_p->base) >= 0));
         if (k > j) {
            (*w_p->exchg)(j,k,w_p->base);
            continue;
         }
         if (j == k) --k;
         break;
      }
      if (k != beg) (*w_p->exchg)(beg,k,w_p->base);

      /*--------------------+
      |Sort the lowest range|
      +--------------------*/
      if ((k - beg) >= (end - k)) {
         hoares_sort(k+1, end, w_p);
         end = k-1;
         /* k = end+1; */
      }else {
         hoares_sort(beg, k-1, w_p);
         beg = k+1;
         k = end+1;
      }
   }
}

/*-----------------------------------------------------------------------quick-+
|Main routine:                                                                 |
|Check if we need to sort the array, then call the recursive routine.          |
+-----------------------------------------------------------------------------*/
void Sort::quick(
   int iNum,                                   // number of items to sort
   int(*compare_function)(int, int, void *),   // key comparison function
   void(*exchange_function)(int, int, void *), // record exchange function
   void *pvBase                                // passed as it is (base ?)
)
{
   unsigned int i;
   int j, k;
   W w;

   if (iNum > 1) {                             // else, nothing to sort
      for (i=iNum, j=-1, k=0; --i; ) {
         if ((*compare_function)(++j, ++k, pvBase) > 0) {   // at least one...
            w.comp = compare_function;
            w.exchg = exchange_function;
            w.base = pvBase;
            hoares_sort(0, iNum - 1, &w);
            break;
         }
      }
   }
}
/******************************************************************************/
