/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// srefdcl.c
///  SHORTREF declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "prsstate.h"
#include "quantity.h"
#include "shortref.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "srefdcl.h"
#include "errorh.h"

/*-----------------------------------------------------------------------parse-+
| Process SHORTREF MD.                                                         |
|                                                                              |
| Enter here when SHORTREF has been decoded,                                   |
| Returns with curr_char (supposed to ) == MDC                                 |
|                                                                              |
| This routine build an address list with as many entries as shortref.         |
| The adresses of the mapped shortref point the name of the entities they      |
| are related to.  All shortref not mapped have NULL adresses.                 |
| When all the prolog will have been parsed (and therefore all entities set),  |
| this list will be scanned again and the entities solved.                     |
|                                                                              |
| [150] short reference mapping declaration = mdo, "SHORTREF",                 |
|                    ps+, map name,                                            |
|                    (ps+, parameter literal, ps+, name)+,                     |
|                    ps*, mdc                                                  |
|                                                                              |
| [151] map name = name                                                        |
|                                                                              |
+-----------------------------------------------------------------------------*/
void ShortrefDcl::parse()
{
   ShortrefNode  *pSnodeNew;

   /*---------------------+
   | Decode shortmap name |
   +---------------------*/
   pPrsstat->skip_inferable_required_ps();
   if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {
      errorh(NONAME, pEh, ERH_T_SEV);
   }
   {
      char pszMapName[NAMELEN_MAX+1];

      pSyntax->read_general_name(pEh, pszMapName);
      pSnodeNew = (ShortrefNode *)expnhashSrMap.gimme_value(
        pszMapName, sizeof(ShortrefNode)
      );
      if (pSnodeNew->is_defined()) errorh(DUPSRMD, pEh, ERH_E_SEV);
   }
   pPrsstat->skip_ps();

   /*
   | Get enough room for each mapped entity name + one extra name for INVSREF
   I Construct a temporary snode
   */
   ShortrefNode snodeTemp(iSrMax);
   char *achTempBuffer = (char *)malloc((iSrMax+1) * (NAMELEN+1));
   char *pszName = achTempBuffer;

   do {                                // for each mapping
      int iSrNmbr;
      /*
      | Get the SR number associated to the parameter literal
      */
      {
         pPrsstat->read_parameter_literal(pchLitBuffer, FALSE);
         e_LexToken tk = pShortref->token_of(pchLitBuffer);
         if (tk == NOT_FOUND_TK) {
            errorh(INVSREF, pEh, ERH_E_SEV);
            iSrNmbr = -1;
         }else {
            iSrNmbr = (unsigned int)tk - (unsigned int)SR_first;
         }
      }
      /*
      | Decode mapped entity name
      */
      pPrsstat->skip_ps();
      pSyntax->read_entity_name(pEh, pszName);
      if (iSrNmbr != -1) {
         if (!snodeTemp.set_ent_name(pszName, iSrNmbr)) {
            errorh(DUPSREF, pEh, ERH_E_SEV);
         }else {
            pszName += 1+strlen(pszName);
         }
      }
   }while (
      (pPrsstat->skip_ps()) &&
      (
         (pPrsstat->get_curr_token() == LIT_TK) ||
         (pPrsstat->get_curr_token() == LITA_TK)
      )
   );
   if (!pSnodeNew->is_defined()) {    // if not duplicate (error caught above)
      new(pSnodeNew) ShortrefNode(
         snodeTemp, iSrMax, achTempBuffer, diffptr(pszName, achTempBuffer)
      );
   }
   free(achTempBuffer);               // free temp buffer
}

/*----------------------------------------------------------------ShortrefNode-+
| Regular constructor                                                          |
+-----------------------------------------------------------------------------*/
ShortrefNode::ShortrefNode(const int iMaxShortrefs)
{
   pSrm = (ShortrefMap *)new char[iMaxShortrefs * sizeof(ShortrefMap)];
   memset(pSrm, 0, iMaxShortrefs * sizeof(ShortrefMap));
}

/*----------------------------------------------------------------ShortrefNode-+
| "Smart" copy constructor                                                     |
+-----------------------------------------------------------------------------*/
ShortrefNode::ShortrefNode(
   const ShortrefNode& snodeSource,
   const int iMaxShortrefs,
   const char *pchNamesBufSource,
   const int iNamesBufLength
)
{
   pSrm = (ShortrefMap *)new char[
      (iMaxShortrefs * sizeof(ShortrefMap)) + iNamesBufLength
   ];
   char *pchNamesBufTarget = (char *)(pSrm + iMaxShortrefs);
   memcpy(pchNamesBufTarget, pchNamesBufSource, iNamesBufLength);
   for (int i=0; i < iMaxShortrefs; ++i) {
      if (!snodeSource.pSrm[i].pszName) {
         pSrm[i].pszName = 0;
      }else {
         pSrm[i].pszName = pchNamesBufTarget + diffptr(
            snodeSource.pSrm[i].pszName, pchNamesBufSource
         );
      }
   }
}

/*---------------------------------------------------------------~ShortrefNode-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
ShortrefNode::~ShortrefNode()
{
   delete [] (char *)pSrm;
}

/*----------------------------------------------------------------set_ent_name-+
| Set the entity name of a shortref                                            |
+-----------------------------------------------------------------------------*/
boolean ShortrefNode::set_ent_name(char *pszName, const int iSrNmbr)
{
   if (pSrm[iSrNmbr].pszName) {
      return FALSE;
   }else {
      pSrm[iSrNmbr].pszName = pszName;
      return TRUE;
   }
}

/*---------------------------------------------------------------------pile_up-+
| Build short reference table from the ShortrefNode structure.                 |
|                                                                              |
| Note: Even if there are no shortrefs defined, the EMPTY shortref map         |
|       always needs to be entered in the ShortRef map hash table,             |
|       because, even with no shortref at all, <!USEMAP EMPTY> is allowed.     |
+-----------------------------------------------------------------------------*/
boolean ShortrefNode::pile_up(
   StringKeeper& sk, const char *pszName, const int iMaxShortrefs,
   ExpandedHash& expnhashGenEnt,
   ShortrefMap*& pSrmMP, const ShortrefMap* const aSrmTable
)
{
   if (!pSrm) {                           // Used, but no mapping found
      errorh(NOSRM, ERH_E_SEV, pszName);  // tell user,
      ix = -1;                            // discard it
      return FALSE;                       // do not unload this hash entry
   }
   (pSrmMP++)->ofsName = sk.park(pszName, 1+strlen(pszName));

   const int ixSrm = diffptr(pSrmMP, aSrmTable); // index to first real entry
   for (int i=0; i < iMaxShortrefs; ++i, ++pSrmMP) {
      char *pszNameEnt = pSrm[i].pszName;

      if (pszNameEnt) {
         EntityDef *pEnt = (EntityDef *)expnhashGenEnt.search_value(pszNameEnt);

         if (!pEnt) {
            errorh(NOSRENT, ERH_E_SEV, pszName, pszNameEnt);
            pSrmMP->ixEnt = -1;
         }else {
            /*---------------------------------------------------------+
            | If an associated GenEnt does exist, replace the name ptr |
            | by the value of its index entry in the EntityDef array.  |
            | This assumes all general entities already been piled up. |
            +---------------------------------------------------------*/
            pSrmMP->ixEnt = pEnt->ent_index();
         }
      }else {
         pSrmMP->ixEnt = -1;
      }
   }
   /*
   | Warning: order matters, `ix' and `pSrm' are a same field!
   */
   delete [] (char *)pSrm;     // no more needed
   ix = ixSrm;                 // set index for GIs srm pointer
   return TRUE;
}
/*============================================================================*/
