/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// syntax.c
///  Class for the SYNTAX parameter of the SGML declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "gendef.h"
#include "enthdlr.h"
#include "externid.h"
#include "syntax.h"
#include "dlmvect.h"
#include "dlmstr.h"
#include "quantity.h"
#include "shortref.h"
#include "rsvname.h"
#include "errorh.h"
#include "enthdlr.h"

/*------------------------------------------------------------------TinyValues-+
| Default Constructor: Reference Concrete Syntax                               |
+-----------------------------------------------------------------------------*/
Syntax::TinyValues::TinyValues(
) {
   chRE = '\r';
   chRS = '\n';
   chSPACE = ' ';
   isCaseGeneral = TRUE;
   isCaseEntity = FALSE;
}

/*----------------------------------------------------------------------Syntax-+
| Constructor from scratch (minimal SGMLdcl)                                   |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Syntax::Syntax()
{
   strbuf = 0;               // Unused

   /*---------------------------------------------------------+
   |dsettab is unused                                         |
   |Build gtype, uctab                                        |
   |                                                          |
   |The shunned chars in the reference concrete syntax are:   |
   |CONTROLS + 0-31, 127, 255;                                |
   |Also in ISO 8879-1986 note 1, p50 states these shunned    |
   |char are "character set" independant.                     |
   +---------------------------------------------------------*/
   for (int i=0; i < MAX_CHARSET; ++i) {
      if (islower(i)) {           /* system charset dependent */
         gtype[i] = S_LETTER | S_NMSTRT;
         uctab[i] = (char)toupper(i);
      }else if (isupper(i)) {
         gtype[i] = S_LETTER | S_NMSTRT;
         uctab[i] = (char)i;
      }else if (isdigit(i)) {
         gtype[i] = S_DIGIT;
         uctab[i] = (char)i;         /* an upper digit :-) */
      }else {
         uctab[i] = (char)i;         /* consider upper stays the same */
         gtype[i] = S_NONSGML;       /* can be changed later */
      }
   }
   gtype['.']  = S_NMCHAR | S_SPECIAL;
   gtype['_']  = S_NMCHAR;
   gtype['\t'] = S_SEPCHAR;
   gtype['<']  = S_SGML;
   gtype['!']  = S_SGML;
   gtype['\"'] = S_SGML;
   gtype['&']  = S_SGML;
   gtype['>']  = S_SGML;
   gtype['#']  = S_SGML;
   gtype['%']  = S_SGML;
   gtype['*']  = S_SGML;
   gtype[';']  = S_SGML;
   gtype['@']  = S_SGML;
   gtype['[']  = S_SGML;
   gtype[']']  = S_SGML;
   gtype['{']  = S_SGML;
   gtype['|']  = S_SGML;
   gtype['}']  = S_SGML;
   gtype['~']  = S_SGML;
   gtype['^']  = S_SGML;
   gtype['\''] = S_SPECIAL;
   gtype['-']  = S_SPECIAL;
   gtype['(']  = S_SPECIAL;
   gtype[')']  = S_SPECIAL;
   gtype['+']  = S_SPECIAL;
   gtype[',']  = S_SPECIAL;
   gtype['/']  = S_SPECIAL;
   gtype[':']  = S_SPECIAL;
   gtype['=']  = S_SPECIAL;
   gtype['?']  = S_SPECIAL;
   gtype[' ']  = S_SSPCHAR;
   gtype['\n'] = S_SSPCHAR;
   gtype['\r'] = S_SSPCHAR;

   /*--------------------------------------+
   |Build the FNCR tab, complete the gtype |
   +--------------------------------------*/
   fncr_max = 4;                   // RE, RS, SPACE, TAB
   fncrtab = (FunctionCharAssoc *)malloc(fncr_max * sizeof fncrtab[0]);

   fncrtab[0].p = ReservedName::name_of(RE_RRN_ix);
   fncrtab[0].fctcls = RE_FCTCLS;
   fncrtab[0].ch_nb = tv.chRE;

   fncrtab[1].p = ReservedName::name_of(RS_RRN_ix);
   fncrtab[1].fctcls = RS_FCTCLS;
   fncrtab[1].ch_nb = tv.chRS;

   fncrtab[2].p = ReservedName::name_of(SPACE_RRN_ix);
   fncrtab[2].fctcls = SPACE_FCTCLS;
   fncrtab[2].ch_nb = tv.chSPACE;

   fncrtab[3].p = ReservedName::name_of(TAB_RRN_ix);
   fncrtab[3].fctcls = SEPCHAR_FCTCLS;
   fncrtab[3].ch_nb = '\t';

   pDv = new DelimiterVector();
   pDlmstr = new DelimiterString();
   pShortref = 0;                         // Unused
   pRsvname = new ReservedName();
   pQuantity = new Quantity();
}

/*----------------------------------------------------------------------Syntax-+
| Constructor                                                                  |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Syntax::Syntax(ApplicationHandle ahlAccess)
{
   EntityHandler::read_record(ahlAccess, (void **)&strbuf);
   EntityHandler::read(ahlAccess, &tv,     sizeof tv);
   EntityHandler::read(ahlAccess, uctab,   sizeof uctab);
   EntityHandler::read(ahlAccess, gtype,   sizeof gtype);
   EntityHandler::read(ahlAccess, dsettab, sizeof dsettab);
   fncr_max= EntityHandler::read_record(
      ahlAccess, (void **)&fncrtab
   ) / sizeof (FunctionCharAssoc);
   /*
   | Relocate fncrtab
   */
   for (int i=0; i < fncr_max; ++i) {
      fncrtab[i].p = strbuf + fncrtab[i].ofs;
   }
   pDv = new DelimiterVector(ahlAccess, strbuf);
   pDlmstr = new DelimiterString(ahlAccess, strbuf);
   pShortref = new Shortref(ahlAccess, strbuf);
   pRsvname = new ReservedName(ahlAccess, strbuf);
   pQuantity = new Quantity(ahlAccess);
}

/*PROTECTED-------------------------------------------------------------Syntax-+
| Constructor for building Syntax from the SGML declaration                    |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Syntax::Syntax(int)
{
   pQuantity = 0;
   pRsvname = 0;
   pShortref = 0;
   pDv = 0;
   pDlmstr = 0;
   fncrtab = 0;
   strbuf = 0;
}

/*---------------------------------------------------------------------~Syntax-+
| Destructor                                                                   |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
Syntax::~Syntax()
{
   delete pQuantity;
   delete pRsvname;
   delete pShortref;
   delete pDv;
   delete pDlmstr;
   free(fncrtab);
   free(strbuf);
}

/*------------------------------------------------------read_and_check_charset-+
|Check the Charset used to create the ODT is still compatible with the         |
|one currently active.                                                         |
+-----------------------------------------------------------------------------*/
boolean Syntax::read_and_check_charset(ApplicationHandle ahlAccess)
{
   boolean rc = TRUE;
   CharsetDescriptor aDsetDtd[sizeof dsettab / sizeof dsettab[0]];

   EntityHandler::read(ahlAccess, aDsetDtd, sizeof aDsetDtd);
   for (int i=0; i < sizeof dsettab / sizeof dsettab[0]; ++i) {
      if (
         (dsettab[i].isDescribed != aDsetDtd[i].isDescribed) ||
         (dsettab[i].isBased     != aDsetDtd[i].isBased)     ||
         (dsettab[i].isParmLit   != aDsetDtd[i].isParmLit)   ||
         (dsettab[i].s_chnb      != aDsetDtd[i].s_chnb)
      ) {
         errorh(INVISTDST, ERH_E_SEV, i);
         rc = FALSE;
      }
   }
   return rc;
}

/*---------------------------------------------------------------write_charset-+
|Write the current charset.  This function is also used to save the Charset    |
|the DTD was compiled with inside the ODT.   This permits to check the         |
|compatibility of an ODT with any active OCS (see read_and_check_charset).     |
+-----------------------------------------------------------------------------*/
void Syntax::write_charset(ApplicationHandle ahlAccess)
{
   EntityHandler::write(ahlAccess, dsettab, sizeof dsettab);
}

/*----------------------------------------------------------------match_string-+
| Check if input matches a given string.                                       |
| RETURN: FALSE if no match found,                                             |
|         TRUE, if the input matches.                                          |
|               curr_char is next char on input (as usual)                     |
+-----------------------------------------------------------------------------*/
boolean Syntax::match_string(EntityHandler *pEh, const char *pszString)
{
   for (int i=0; pszString[i]; ++i) {
      if ((pEh->is_Ee()) || (pszString[i] != (char)pEh->curr_char_val())) {
         pEh->unread(i);
         return FALSE;
      }
      pEh->read();
   }
   return TRUE;
}

/*-------------------------------------------------------read_to_end_delimiter-+
| Read a minimum literal, a character data (SDATA) literal, or a PI.           |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the Quantity, DelimString and Charset       |
| classes.                                                                     |
|                                                                              |
| No markup is recognized in a minimum parameter literal, a character literal, |
| or a PI, other than the LIT/LITA delimiter that terminates it.               |
| Hence this function simply reads until the ending delimiter is reached.      |
|                                                                              |
| -- INPUT --                                                                  |
| `pchBuffer' points a target buffer                                           |
| `glc' is one of:                                                             |
|     MINLIT_GLC : to get a minimum literal                                    |
|     LIT_GLC    : to get a character data (SDATA) literal                     |
|     LIT_GLC    : to get a PI                                                 |
|                                                                              |
| -- PRECONDITION --                                                           |
| `glc' is one of the `literal_kind' enum                                      |
| `pchBuffer' is LITLEN+1 long, or PILEN + 1 long for a PI                     |
|                                                                              |
| -- OUTPUT --                                                                 |
| `pchBuffer' contains as much characters as possible read from input,         |
| until (but not including) the ending delimiter.  For minimum literals,       |
| extraneous SPACE's are removed (10.1.7).                                     |
|                                                                              |
| If the maximum length was exceeded, this function exits -- which is not      |
| too nasty for minimum literal and SDATA literals (occur only in prologs).    |
| For PI, we plan to do better.                                                |
|                                                                              |
| The length of the literal is returned.                                       |
|                                                                              |
| -- POSTCONDITION --                                                          |
| The entity manager has the next character to read.                           |
|                                                                              |
| -- NOTE --                                                                   |
| [44] processing instruction = pio, system data, pic                          |
|                                                                              |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::read_to_end_delimiter(
   EntityHandler *pEh,
   char *pchBuffer,
   e_LexToken token,
   e_LiteralType glc
)
{
   int iMaxLength;
   int iLength = 0;
   const char * pszDlm;

   if (glc == PI_GLC) {
      iMaxLength = pQuantity->value_of(PILEN_QTY_ix);
      pszDlm = pDlmstr->value_of(PIC_VDL_ix);
   }else {
      iMaxLength = pQuantity->value_of(LITLEN_QTY_ix);
      switch (token) {
      case LIT_TK:
         pszDlm = pDlmstr->value_of(LIT_VDL_ix);
         break;
      case LITA_TK:
         pszDlm = pDlmstr->value_of(LITA_VDL_ix);
         break;
      default:
         errorh(NOLIT, pEh, ERH_T_SEV);
         pszDlm = 0; // avoid compiler warning (no return from ERH_T)
         break;
      }
   }
   /*--------------------------------------------------+
   | Loop until:                                       |
   | 1) an Ee is found                                 |
   | 2) the ending delimiter is found; or,             |
   | 3) the number of characters read and moved        |
   |    into `pchBuffer' is at iMaxLength +1; or,      |
   | If 1):                                            |
   |    an appropriate message is issued,              |
   |    and the loop breaks as if the ending delimiter |
   |    had been found                                 |
   | If 2):                                            |
   |    iLength == number of characters read           |
   |    pchBuffer[iLength] == '\0'                     |
   | If 3):                                            |
   |    iLength == iMaxLength + 1                      |
   |    pchBuffer[iMaxLength] == 1st char in excess    |
   +--------------------------------------------------*/
   {
      boolean isSpaceWaiting = FALSE;

      for (;;) {                                   // main loop
         if (pEh->is_Ee()) {                       // Ee found:
            switch (glc) {                         // issue appropriate message
            case PI_GLC:
               errorh(UBLNPIC, pEh, ERH_T_SEV);
               break;
            default:  // case MINLIT_GLC, case LIT_GLC:
               errorh(UBLNLIT, pEh, ERH_T_SEV, pszDlm);
               break;
            }
            pchBuffer[iLength] = '\0';             // put an ending 0
            break;                                 // and leave
         }
         if (pszDlm[0] == (char)pEh->curr_char_val()) { // pszDlm[0] is found
            pEh->read();                           // get next character ready
            if (match_string(pEh, pszDlm + 1)) {   // see if the rest matches
               pchBuffer[iLength] = '\0';          // match: put an ending 0
               break;                              // leave
            }
            pEh->unread(1);                        // mismatch: unread
         }
         /*----------------------------+
         |For minimum literal, compress|
         |separators (ISO8879 10.1.7)  |
         +----------------------------*/
         if (glc == MINLIT_GLC) {
            if (!pEh->is_minldat()) {
               if (!pEh->is_sspchar()) {
                  errorh(INVMLCHAR, pEh, ERH_E_SEV, pEh->curr_char_val());
                  pEh->read();                     // ignore non-mindata
               }else if (pEh->is_RSchar()) {
                  pEh->read();                     // ignore  RS
               }else  {                            // RE, TAB, SPACE, etc..
                  while (is_sspchar(pEh->read())); //leading spaces
                  if (iLength) {                   // no space at beginning
                      isSpaceWaiting = TRUE;
                  }
               }
               continue;
            }
            if (isSpaceWaiting) {
               isSpaceWaiting = FALSE;
               pchBuffer[iLength] = (char)get_SPACEchar();
               if (++iLength > iMaxLength) break;  // Overflow...
            }
         }
         #ifdef MSS_UNACTIVATED /*-----------------<MSS_PROCESSING>------*/
         /*--------------------------------------------------+
         | Suppress delimiter recognition if MSO or MSS is   |
         | encountered ?                                     |
         | It seems inappropriate to care about Markup Scan  |
         | Characters for PI (SDATA), MINLT (minimum literal)|
         | or LIT (SDATA).  This is why the code has been    |
         | deactivated...  (but not destroyed, as this is    |
         | not explicitely said in ISO)                      |
         +--------------------------------------------------*/
         if (pEh->is_msrchar()) {              // MSO or MSS
            if (pEh->is_msochar()) {
               do {
                  pchBuffer[iLength] = (char)pEh->curr_char_val();
                  pEh->read();
               }while (
                  (++iLength <= iMaxLength) &&
                  (!pEh->is_Ee()) && (!pEh->is_msichar())
               );
            }else {                            // MSS char
               pchBuffer[iLength] = (char)pEh->curr_char_val();
               pEh->read();
               ++iLength;
            }
            if (iLength > iMaxLength) break;   // Overflow
            continue;
         }
         #endif /*--------------------------------</MSS_PROCESSING>------*/

         /*------------------+
         |Store the character|
         +------------------*/
         pchBuffer[iLength] = (char)pEh->curr_char_val();
         pEh->read();
         if (++iLength > iMaxLength) break;    // break if Overflow
      }
   }
   if (iLength > iMaxLength) {
      if (glc == PI_GLC) {
         errorh(BIGPI, pEh, ERH_E_SEV);
         if (!skip_to_delimiter(pEh, PIC_VDL_ix)) {
            errorh(ABORTED, ERH_T_SEV);
         }
      }else {
         errorh(BIGLIT, pEh, ERH_T_SEV);
      }
   }
   if ((glc == MINLIT_GLC) && (iLength > LITLEN_RCS)) {
      errorh(BIGLITRCS, pEh, ERH_E_SEV, iLength);
   }
   return iLength;
}

/*-----------------------------------------------------------------read_number-+
| Read a (long) integer - At least first char must be a digit                  |
+-----------------------------------------------------------------------------*/
unsigned long Syntax::read_number(EntityHandler *pEh)
{
   unsigned long i = 0;
   unsigned int k = pEh->curr_char_val();

   if (is_digit(k)) {
      while (
         (i < (((unsigned long)~0)/10)) ||
         ((i <= (((unsigned long)~0)/10)) && (k <= '5'))
      ) {
         i = (i*10) + k - '0';
         k = pEh->read();
         if (!is_digit(k)) return i;
      }
   }
   errorh(INVNUM, pEh, ERH_T_SEV);
   return i;
}

/*-------------------------------------------------------read_character_number-+
| Read a character number                                                      |
+-----------------------------------------------------------------------------*/
boolean Syntax::read_character_number(EntityHandler *pEh, unsigned int *pui)
{
   unsigned long il = read_number(pEh);

   if (il >= MAX_CHARSET) {
      errorh(BIGCHDESC, pEh, ERH_I_SEV);
      return FALSE;
   }else {
      *pui = (unsigned int)il;
      return TRUE;
   }
}

/*---------------------------------------------------------------read_namechar-+
|Move all name characters of standard input to "pszName", an ASCIIZ string.    |
|Characters in excess are truncated;                                           |
|Returns the count of characters read.                                         |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::read_namechar(
   EntityHandler *pEh, char *pszName, unsigned int isNamecase
)
{
   char *psz = pszName;
   unsigned int count = 0;
   boolean already_said = FALSE;

   do {
      if (++count > pQuantity->value_of(NAMELEN_QTY_ix)) {
         if (!already_said) {
            errorh(BIGNAMET, pEh, ERH_E_SEV);
            already_said = TRUE;
         }
      }else {
         if (isNamecase) {
            *psz++ = (char)pEh->uppercase();
         }else {
            *psz++ =  (char)pEh->curr_char_val();
         }
      }
   }while (is_name(pEh->read()));
   *psz = '\0';
   return count;
}

/*-----------------------------------------------------------read_general_name-+
|get a name - if impossible, issue error message and leave.                    |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::read_general_name(EntityHandler *pEh, char *pszName)
{
   if (!pEh->is_namestrt()) {
      errorh(NONAME, pEh, ERH_T_SEV);
   }
   return read_namechar(pEh, pszName, is_case_general());
}

/*------------------------------------------------------------read_entity_name-+
|get an entity name if impossible, issue error message and leave.              |
|(as read_general_name, except the case translation)                           |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::read_entity_name(EntityHandler *pEh, char *pszName)
{
   if (!pEh->is_namestrt()) {
      errorh(NONAME, pEh, ERH_T_SEV);
   }
   return read_namechar(pEh, pszName, is_case_entity());
}

/*--------------------------------------------------------------read_nametoken-+
| read a name token                                                            |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::read_nametoken(EntityHandler *pEh, char *pszName)
{
   if (!pEh->is_name()) {
      errorh(NONMTK, pEh, ERH_T_SEV);
   }
   return read_namechar(pEh, pszName, is_case_general());
}

/*-------------------------------------------------------read_refd_entity_name-+
|get a REFC delimited entity name                                              |
+-----------------------------------------------------------------------------*/
void Syntax::read_refd_entity_name(EntityHandler *pEh, char *pszName)
{
   read_entity_name(pEh, pszName);
   eat_refc(pEh);
}

/*--------------------------------------------------------------------eat_refc-+
| Check if current input could end a reference - skip past it if so.           |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the Charset and DelimString classes         |
|                                                                              |
| -- INPUT --                                                                  |
| (none)                                                                       |
|                                                                              |
| -- PRECONDITION --                                                           |
| (none)                                                                       |
|                                                                              |
| -- OUTPUT --                                                                 |
| Return 0 if no refc found, else the length skipped;                          |
|                                                                              |
| -- POSTCONDITION --                                                          |
| The next character to read is the first character that cannot be classified  |
| as being a Reference End delimiter: [61] reference end = (refc | RE)?        |
| Hence, if on entry `curr_char' cannot start a REFC, `curr_char' is           |
| unchanged on return.                                                         |
|                                                                              |
+-----------------------------------------------------------------------------*/
unsigned int Syntax::eat_refc(EntityHandler *pEh)
{
   const char *achRefc = pDlmstr->value_of(REFC_VDL_ix);
   unsigned int iLength;

   if (pEh->curr_char_val() == achRefc[0]) {
      for (iLength=1; achRefc[iLength]; ++iLength) {
         if (achRefc[iLength] != (char)pEh->read()) {
            pEh->unread(iLength);
            return 0;
         }
      }
   }else if (pEh->is_REchar()) {
      iLength = 1;
   }else {
      return 0;
   }
   pEh->read();                // get next char
   return iLength;
}

/*---------------------------------------------------------------match_rsvname-+
| Check if input matches a given concrete reference name.                      |
| RETURN: FALSE if no match found,                                             |
|         TRUE, if the input matches.                                          |
|               curr_char is next char on input (as usual)                     |
+-----------------------------------------------------------------------------*/
boolean Syntax::match_rsvname(
   EntityHandler *pEh, e_IxRsvnameRefer ix
)
{
   const char *pchBase = pRsvname->name_of(ix);
   const char *pch;

   for (pch = pchBase; *pch; pEh->read(), ++pch) {
      if (
         (pEh->is_Ee()) ||
         (*pch != (
               (is_case_general())?
               pEh->uppercase() : pEh->curr_char_val()
            )
         )
      ) {
         pEh->unread(diffptr(pch, pchBase));
         return FALSE;
      }
   }
   return TRUE;
}

/*---------------------------------------------------------------match_rsvname-+
| Check if input matches a given variant syntax name.                          |
| RETURN: FALSE if no match found,                                             |
|         TRUE, if the input matches.                                          |
|               curr_char is next char on input (as usual)                     |
+-----------------------------------------------------------------------------*/
boolean Syntax::match_rsvname(
   EntityHandler *pEh, e_IxRsvnameVarnt ix
)
{
   const char *pchBase = pRsvname->name_of(ix);
   const char *pch;

   for (pch = pchBase; *pch; pEh->read(), ++pch) {
      if (
         (pEh->is_Ee()) ||
         (*pch != (
               (is_case_general())?
               pEh->uppercase() : pEh->curr_char_val()
            )
         )
      ) {
         pEh->unread(diffptr(pch, pchBase));
         return FALSE;
      }
   }
   return TRUE;
}

/*--------------------------------------------------------match_rsvname_or_die-+
| Check if input matches a given concrete reference reserved name.             |
| If not, issue a terminating error (strong check)                             |
+-----------------------------------------------------------------------------*/
void Syntax::match_rsvname_or_die(
   EntityHandler *pEh, e_IxRsvnameRefer ix
)
{
   if (!match_rsvname(pEh, ix)) {
      errorh(NOKYWD, pEh, ERH_T_SEV, pRsvname->name_of(ix));
   }
}

/*--------------------------------------------------------match_rsvname_or_die-+
| Check if input matches a given variant syntax name.                          |
| If not, issue a terminating error (strong check)                             |
+-----------------------------------------------------------------------------*/
void Syntax::match_rsvname_or_die(
   EntityHandler *pEh, e_IxRsvnameVarnt ix
)
{
   if (!match_rsvname(pEh, ix)) {
      errorh(NOKYWD, pEh, ERH_T_SEV, pRsvname->name_of(ix));
   }
}

/*-------------------------------------------------------match_rsvname_in_list-+
| Check if input matches a concrete reference name from a given list.          |
| Return an associated value.                                                  |
|                                                                              |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the                                         |
| Charset, ReservedName, and Quantity classes.                                 |
|                                                                              |
| `e_IxRsvnameRefer' is a key to all reserved names of the Reference,          |
| including the ones found in the System Declaration.                          |
|                                                                              |
| -- INPUT --                                                                  |
| `pChoice' is base of ReferChoiceEntry assoc array,                           |
|    each entry formed by a pair:                                              |
|      - a value from the refer_rsvname_ix enum,                               |
|      - caller defined associated value (int) to return                       |
|    the last entry is always made of (NOT_FOUND_FRN_ix, "not_found" value)    |
|                                                                              |
| -- PRECONDITION --                                                           |
| the `ix' value of the last ReferChoiceEntry assoc array is NOT_FOUND_FRN_ix  |
|                                                                              |
| -- OUTPUT --                                                                 |
| value from assoc pair, if any -- or value from last entry ("not found" value)|                   |
|                                                                              |
| -- POSTCONDITION --                                                          |
| if name is found in assoc array:                                             |
|    next character to read is after name                                      |
| otherwise:                                                                   |
|    next character to read is unchanged                                       |
+-----------------------------------------------------------------------------*/
int Syntax::match_rsvname_in_list(
   EntityHandler *pEh, const ReferChoiceEntry *pChoice
)
{
    char pszName[NAMELEN_MAX + 1];
    int iReadCount;

    /*----------+
    | read name |
    +----------*/
    if (pEh->is_namestrt()) {
       iReadCount = read_namechar(pEh, pszName, is_case_general());
    }else {
       iReadCount = 0;
       pszName[0] = '\0';
    }
    /*-----------------+
    | scan assoc array |
    +-----------------*/
    while (pChoice->ix != NOT_FOUND_RRN_ix) {
       if (!strcmp(pRsvname->name_of(pChoice->ix), pszName)) {
          return pChoice->value;
       }
       ++pChoice;
    }
    /*--------------------------------+
    | not found in assoc array,       |
    | unwind input, return last value |
    +--------------------------------*/
    pEh->unread(iReadCount);
    return pChoice->value;
}

/*-------------------------------------------------------match_rsvname_in_list-+
| Check if input matches a variant syntax name from a given list.              |
| Return an associated value.                                                  |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the                                         |
| Charset, ReservedName, and Quantity classes.                                 |
|                                                                              |
| `e_IxRsvnameVarnt' is a key to get the "actual" reserved names,         |
| taking into account the possible changes implied by the current syntax.      |
| This enum lists only the names that may be found in prologs and document     |
| instances.                                                                   |
|                                                                              |
| -- INPUT --                                                                  |
| `pChoice' is base of VarntChoiceEntry assoc array,                           |
|    each entry formed by a pair:                                              |
|      - a value from the varnt_rsvname_ix enum,                               |
|      - caller defined associated value (int) to return                       |
|    the last entry is always made of (NOT_FOUND_FRN_ix, "not_found" value)    |
|                                                                              |
| -- PRECONDITION --                                                           |
| the `ix' value of the last VarntChoiceEntry assoc array is NOT_FOUND_FRN_ix  |
|                                                                              |
| -- OUTPUT --                                                                 |
| value from assoc pair, if any -- or value from last entry ("not found" value)|                   |
|                                                                              |
| -- POSTCONDITION --                                                          |
| if name is found in assoc array:                                             |
|    next character to read is after name                                      |
| otherwise:                                                                   |
|    next character to read is unchanged                                       |
+-----------------------------------------------------------------------------*/
int Syntax::match_rsvname_in_list(
   EntityHandler *pEh, const VarntChoiceEntry *pChoice
)
{
    char pszName[NAMELEN_MAX + 1];
    int iReadCount;

    /*----------+
    | read name |
    +----------*/
    if (pEh->is_namestrt()) {
       iReadCount = read_namechar(pEh, pszName, is_case_general());
    }else {
       iReadCount = 0;
       pszName[0] = '\0';
    }
    /*-----------------+
    | scan assoc array |
    +-----------------*/
    while (pChoice->ix != NOT_FOUND_VRN_ix) {
       if (!strcmp(pRsvname->name_of(pChoice->ix), pszName)) {
          return pChoice->value;
       }
       ++pChoice;
    }
    /*--------------------------------+
    | not found in assoc array,       |
    | unwind input, return last value |
    +--------------------------------*/
    pEh->unread(iReadCount);
    return pChoice->value;
}

/*-----------------------------------------------------------skip_to_delimiter-+
| Skip to an (ending) delimiter.                                               |
|                                                                              |
| On Entry:                                                                    |
|    vdl_ix describes the delimiter to skip to                                 |
| Returns:                                                                     |
|    TRUE: if the ending delimiter was found,                                  |
|          curr_char is the character that follows the delimiter.              |
|    FALSE: If End-of-EntityDef was found,                                     |
|          curr_char is the Ee character.                                      |
+-----------------------------------------------------------------------------*/
boolean Syntax::skip_to_delimiter(
   EntityHandler *pEh, e_IxDlmstrVarnt vdl_ix
)
{
   const char *pszDelim = pDlmstr->value_of(vdl_ix);

   while (!pEh->is_Ee()) {
      if (pszDelim[0] == pEh->curr_char_val()) { // if 1st char in dlm is found
         pEh->read();                         // get next character ready
         if (match_string(pEh, pszDelim+1)) {    // if the rest matches,
            return TRUE;                         // it's OK
         }
      }else {                                    // first char doesn't match:
         pEh->read();                         // read and loop
      }
   }
   return FALSE;                                 // Ee found: bad return code
}

/*-----------------------------------------------------translate_character_ref-+
| Translate a character reference.                                             |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to Syntax and Quantity classes.                |
|                                                                              |
| -- INPUT --                                                                  |
| `iLengtht' has the length of the CRO delimiter.                              |
|                                                                              |
| -- PRECONDITION --                                                           |
| CRO_TK has been found by `recognize_delimiter'                               |
|                                                                              |
| -- OUTPUT --                                                                 |
| Return INVALID_CHRF (unsuccessful return) when the character cannot be       |
| translated                                                                   |
|                                                                              |
| When the translation has been successful, return code is:                    |
|   - NAMED_CHRF if a named character reference was translated                 |
|   - NUMERIC_CHRF if a numeric character reference was translated             |
|                                                                              |
| -- POSTCONDITION --                                                          |
| When the character cannot be translated (unsuccessful return),               |
| The character currently to read, is after the REFC, if any.                  |
|                                                                              |
| When the translation has been successful, `curr_char' has the                |
| translated character, and the read buffer is updated -- em_replace --        |
| "as though the translated character were entered directly"                   |
| (and unread will work properly)                                              |
|                                                                              |
+-----------------------------------------------------------------------------*/
e_CharRefType Syntax::translate_character_ref(
   EntityHandler *pEh, int iLength
)
{
   if (pEh->is_digit()) {
      /*----------------------------+
      | Numeric Character Reference |
      +----------------------------*/
      e_ErrorCode msgcode = ER_NULL;
      unsigned int iNumber = 0;
      unsigned int iNameLength = 0;
      unsigned int iREFClength;

      do {
         iNumber = (iNumber * 10) + pEh->curr_char_val() - '0';
         if (iNumber >= MAX_CHARSET) {
            msgcode = BIGCHNB;
         }else if (
            (++iNameLength > pQuantity->value_of(NAMELEN_QTY_ix)) && (!msgcode)
         ) {
            msgcode = BIGNAMET;
         }
         pEh->read();
         iREFClength = eat_refc(pEh);   // Try to end the reference
      }while (
         (iREFClength == 0) &&          // reference not ended, and
         (pEh->is_digit())              // still a digit to read
      );

      if (msgcode) {
         errorh(msgcode, pEh, ERH_E_SEV);
         if (msgcode == BIGCHNB) return INVALID_CHRF;
      }
      pEh->replace(
         (char)iNumber, iLength + iNameLength + iREFClength, NUMCHAR_KOC
      );
      return NUMERIC_CHRF;
   }else {
      /*--------------------------+
      | Named Character Reference |
      +--------------------------*/
      char pszName[NAMELEN_MAX + 1]; // see 9.3.1 */

      iLength += read_general_name(pEh, pszName);
      iLength += eat_refc(pEh);          // will "eat" the refc if any
      for (int i=0; i < max_functionchar(); ++i) {
         if (!strcmp(pszName, functionchar_name(i))) {
            e_CharKind koc;

            switch (functionchar_class(i)) {
            case RS_FCTCLS:
               koc = RS_KOC;
               break;
            default:
               koc = REGULAR_KOC;
               break;
            }
            pEh->replace((char)(functionchar_value(i)), iLength, koc);
            return NAMED_CHRF;
         }
      }
      errorh(INVCHRF, pEh, ERH_E_SEV, pszName);
      return INVALID_CHRF;
   }
}

/*-------------------------------------------------------------------NewSyntax-+
| NewSyntax constructor                                                        |
+-----------------------------------------------------------------------------*/
NewSyntax::NewSyntax(
   StringKeeper& skArg
) : Syntax(0), sk(skArg)
{
   pDv = new NewDelimiterVector(sk, *this);
   pDlmstr = new NewDelimiterString(sk);
   pQuantity = new NewQuantity();
   pRsvname = new NewReservedName(sk);
   pShortref = new NewShortref(sk);
}

/*------------------------------------------------------------------~NewSyntax-+
| NewSyntax destructor                                                         |
+-----------------------------------------------------------------------------*/
NewSyntax::~NewSyntax()
{
   delete pShortref;
   delete pRsvname;
   delete pQuantity;
   delete pDlmstr;
   delete pDv;
}

/*--------------------------------------------------reset_functionchar_rsvname-+
| Reset the reserved name: RE, RS and SPACE of the FunctionChar table:         |
| to reflect the changes made by the NAMES parameter of the SGML declaration   |
+-----------------------------------------------------------------------------*/
void NewSyntax::reset_functionchar_rsvname()
{
   fncrtab[0].ofs = pRsvname->offset_of(RE_VRN_ix);
   fncrtab[1].ofs = pRsvname->offset_of(RS_VRN_ix);
   fncrtab[2].ofs = pRsvname->offset_of(SPACE_VRN_ix);
}

/*----------------------------------------------------------check_functionchar-+
| Check redundancies in fncrtab:                                               |
| all function names and codes must be different                               |
+-----------------------------------------------------------------------------*/
boolean NewSyntax::check_functionchar()
{
   const int iMax = fncr_max;
   boolean rc = TRUE;

   FunctionCharAssoc *pFncr1 = fncrtab;
   for (int i1=1; i1 < iMax; ++pFncr1, ++i1) {
      FunctionCharAssoc *pFncr2 = pFncr1 + 1;
      for (int i2=i1+1; i2 <= iMax; ++i2) {
         if (!strcmp(sk.point(pFncr1->ofs), sk.point(pFncr2->ofs))) {
            errorh(DUPFCTREF, ERH_E_SEV, sk.point(pFncr1->ofs));
            rc = FALSE;
         }
         if (pFncr1->ch_nb == pFncr2->ch_nb) {
            errorh(DUPFCTCNB, ERH_E_SEV, pFncr2->ch_nb);
            rc = FALSE;
         }
      }
   }
   return rc;
}

/*----------------------------------------------------------------------unload-+
| Write the syntax to disk                                                     |
+-----------------------------------------------------------------------------*/
void NewSyntax::unload(ApplicationHandle ahlAccess)
{

   /*
   |  Clean up gtype, removing the SHUNCHAR bit
   |  Pure SHUNCHAR are now simply NON-SGML
   */
   {
      for (int i=0; i < MAX_CHARSET; ++i) gtype[i] &= ~S_SHUNNED;
   }
   sk.output(EntityHandler::write, ahlAccess);
   EntityHandler::write(ahlAccess, &tv,     sizeof tv);
   EntityHandler::write(ahlAccess, uctab,   sizeof uctab);
   EntityHandler::write(ahlAccess, gtype,   sizeof gtype);
   EntityHandler::write(ahlAccess, dsettab, sizeof dsettab);
   EntityHandler::write_record(
      ahlAccess, fncrtab, fncr_max * sizeof(FunctionCharAssoc)
   );

   pDv->unload(ahlAccess);
   pDlmstr->unload(ahlAccess);
   pShortref->unload(ahlAccess);
   pRsvname->unload(ahlAccess);
   pQuantity->unload(ahlAccess);
}
/******************************************************************************/
