/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// umapdcl.c
///  USEMAP declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "prsstate.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "umapdcl.h"
#include "srefdcl.h"
#include "errorh.h"

/*STATIC----------------------------------------------------inst_new_mapped_gi-+
| This procedure is called by pPrsstat->parse_name_group for each associated   |
| element name of this USEMAP.                                                 |
+-----------------------------------------------------------------------------*/
static void inst_new_mapped_gi(void *pvBase)
{
   ((UsemapDcl *)pvBase)->new_mapped_gi();
}

/*-----------------------------------------------------------------------parse-+
| Process USEMAP MD.                                                           |
|                                                                              |
| Enter here when USEMAP has been decoded,                                     |
| Returns with curr_char (supposed to ) == MDC                                 |
|                                                                              |
| [152] short reference use declaration = mdo, "USEMAP",                       |
|                    ps+, map specification,                                   |
|                    (ps+, associated element type)?, <= one at least in DTD   |
|                    ps*, mdc                                                  |
|                                                                              |
| [153] map specification = map name | (rni,"EMPTY")                           |
|                                                                              |
+-----------------------------------------------------------------------------*/
void UsemapDcl::parse()
{
   char pszMapName[NAMELEN_MAX+1];

   /*--------------+
   |Decode map name|
   +--------------*/
   pPrsstat->skip_ps();      // 10/30/90: no ambiguity created if no ps
   if (pPrsstat->get_curr_token() == RNI_TK) {
      pSyntax->match_rsvname_or_die(pEh, EMPTY_VRN_ix);
      pszMapName[0] = '\0';
   }else {
      if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {
         errorh(NONAME, pEh, ERH_T_SEV);
      }
      pSyntax->read_general_name(pEh, pszMapName);
   }
   pSnodePV = (ShortrefNode *)expnhashSrMap.gimme_value(
      pszMapName, sizeof(ShortrefNode)
   );

   /*-----------------------------+
   |Decode associated element type|
   +-----------------------------*/
   pPrsstat->skip_ps();      // 10/30/90: no ps is allowed
   switch (pPrsstat->get_curr_token()) {
   case GRPO_TK:             // these gi's associated with a map
      pPrsstat->parse_name_group(inst_new_mapped_gi, this);
      break;
   case NOT_FOUND_TK:
      new_mapped_gi();       // this gi associated with a map
      break;
   default:
      errorh(NOASSELEM, pEh, ERH_E_SEV);
      return;                // If at MDC, recovery is possible
   }
   pPrsstat->skip_ps();
}

/*---------------------------------------------------------------new_mapped_gi-+
| Install a (new ?) element type name in the hash tables,                      |
| that will change the current map.                                            |
+-----------------------------------------------------------------------------*/
void UsemapDcl::new_mapped_gi()
{
   char pszElemName[NAMELEN_MAX+1];

   pSyntax->read_general_name(pEh, pszElemName);
   GenericIdBlock *pGi = (GenericIdBlock *)expnhashElem.gimme_value(
     pszElemName, sizeof(GenericIdBlock)
   );
   if (pGi->pSnode) {                 // already switching to a map
      errorh(DUPSRU, pEh, ERH_I_SEV); // is ignored
   }else {
      pGi->pSnode = pSnodePV;
   }
}
/******************************************************************************/
