/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// utilctl.h
///  Declarations for utilctl
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _UTID_H
#define _UTID_H 1
/*
| The UtilityControl class describes utility files associated to
| SGML declarations, Document Type Declaration or Link Type Declaration.
|
| Main data of this class is a dynamic table of UtilityIdentifier's,
| `aUtidTable':
|
|  - 1st entry always describes the LOCAL.OCS file
|    (i.e. the Reference OCS)
|  - 2nd entry always describes the USER.OCS file
|    (i.e. SCOPE INSTANCE OCS)
|  - 3rd entry always describe the base ODT.
|
| Further entries describes ODT's (CONCUR), followed by OLT's (LINK).
| When a SUBDOC is entered, there is again the base, ODT's and OLT's.
|
| Then:
|   `aUtidTable' = local_ocs, user_ocs, (base_odt, other_odt*, olt*)+
|
|    where the repetitions of the parenthesized group is the main document,
|    followed by 0 or more SUBDOC's.  Each group starts with an utid
|    describing a base (utid.is.base is TRUE) and ends with an
|    UtilityIdentifier pointed by `pUtidLast'.  `pUtidLast' is updated
|    when a new SUBDOC is entered, and restored when it ends.
|
|    Note: when a SUBDOC ends, its odt/olt group is however kept in the utid
|          array to avoid finds for the same file to be issued twice
|          (temp files, for instance).
|
| For each UtilityIdentifier:
|   - name is the name of the DTD or LTD - irrelevant for OCS
|   - id_ahl is the identifier handle, returned by FIND_PFC.
|     This ensures that FIND will not be called twice (used by find)
|   - position is the position within the primary entity of the source
|     -> for an ODT, this points just after "<!DOCTYPE "
|     -> for an OLT, this points after "<!LINKTYPE ... >"
|     This field is unused for the OCS's (two first entries)
|
*/
//******************************************************************************

#include "externid.h"   // part of the class

#include "utilctl.eh"   // enums

class EntityHandler;
class SGMLdcl;
class SortedStringList;

class UtilityIdentifier {        // UTID: associate an Id-AHL to utility file
   friend class UtilityControl;
public:
   char *pszName;
   ExternId exid;
   void *position;               // ODT or OLT: dcl position in #PRIMARY file
   boolean isActive   :1;        // This DOCTYPE or LINKTYPE is active
   e_LinkKind lkkind  :2;        // ODT or OLT kind

   UtilityIdentifier(
      const char *pszNameArg,
      boolean isActiveArg,
      boolean isBaseDtdArg,
      e_LinkKind lkkindArg
   );
   ~UtilityIdentifier();

private:
   boolean isExidSet  :1;        // exid part has been set
   boolean isBase     :1;        // First DTD in base document or SUBDOC
};

class UtilityControl
{
public:
   /*-----------------+
   | Public Functions |
   +-----------------*/
   UtilityControl();
   _export ~UtilityControl();
   UtilityIdentifier * new_utid(
      EntityHandler *pEh,
      SGMLdcl *pSgmldcl,
      SortedStringList& sstrlActiveType,
      boolean isBaseDtd, const char *pszName, e_LinkKind lkkind
   );
   UtilityIdentifier * find_utid_base();
   boolean is_active(char *pszName);
   ExternId * build_compiled_exid(
      e_TextClass xfcCompiled, ExternId *pExidSource = 0
   );
   const unsigned long get_last_utid_ix();
   void  set_last_utid_ix(int ixSaveUtidLast);
private:
   /*------------------+
   | Private Variables |
   +------------------*/
   UtilityIdentifier   *aUtidTable;   // 1st members: OCS Local, OCS User
   UtilityIdentifier   *pUtidLast;    // Last relevant utid entry
   int iUtidCount;                    // Count of utid in aUtidTable
};

/*INLINE------------------------------------------------------get_last_utid_ix-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline const unsigned long UtilityControl::get_last_utid_ix() {
   return pUtidLast - aUtidTable;
}

/*INLINE------------------------------------------------------set_last_utid_ix-+
|                                                                              |
+-----------------------------------------------------------------------------*/
inline void UtilityControl::set_last_utid_ix(int ixSaveUtidLast)
{
   pUtidLast = aUtidTable + ixSaveUtidLast;
}
#endif /* _UTID_H *************************************************************/
