/*
**  pubid.C
**
**  resolve public id to filenames ( or something else )
**
**  Pekka Kataja - 1995
*/


#include "pubid.h"
#include <iostream.h>


map<char *, char *, ResolvePublicId::strLess> ResolvePublicId::idmap ;
int ResolvePublicId::classcounter = 0 ;

int ResolvePublicId::resolve ( const char *id , char *buffer ) 
{
  *buffer = '\0' ;
  map <char *, char *, strLess>::iterator f = idmap.find ( id ) ;
  if ( (*f).first )
    strcpy ( buffer , (*f).second ) ;
  return ( strlen ( buffer ) ) ;
}



ResolvePublicId::ResolvePublicId ()
{
  ++classcounter ;
}


ResolvePublicId::ResolvePublicId ( const char *catalogfile )
{
  ++classcounter ;
  addcatalog ( catalogfile ) ;
}

void ResolvePublicId::addcatalog ( const char *catalogfile )
{

  ifstream catalog ( catalogfile ) ;
  if ( ! catalog.good () ) {
    cerr << "PKH0001: (S) Can't open catalog: " << catalogfile << endl ;
    return ;
  }

  char buffer [MAXSTRLEN] ;
  buffer [0] = '\0' ;
  while ( catalog.get ( buffer , sizeof (buffer) ) ) {
    catalog.get () ;

    if ( ! strlen ( buffer ) )
      continue ;

    char *s = buffer ;
    for ( ; *s && isspace ( *s ) ; ++s ) ;
    if ( *s == '\0' || *s == '#' )
      continue ;

    for ( ; *s && *s != '\"' ; ++s ) ;
    if ( *s == '\0' )
      continue ;

    char *pid = ++s ;                            // public id start here
    for ( ; *s && *s != '\"' ; ++s ) ;
    if ( *s == '\0' )                  
      continue ;
    *s = '\0' ;                                  // and end here
    
    for ( ++s ; *s && isspace ( *s ) ; ++s ) ;
    if ( *s == '\0' )                            // no _filename_
      continue ;

    char *realthing = s ;                      // filename start
    for ( ++s ; *s && ! isspace ( *s ) ; ++s ) ;
    *s = '\0' ;                                  // filename end
    
    char *tmppid = new char [ strlen ( pid ) + 1 ] ;
    strcpy ( tmppid , pid ) ;
    char *tmpreal = new char [ strlen ( realthing ) + 1 ] ;
    strcpy ( tmpreal , realthing ) ;

    map <char *, char *, strLess>::iterator f = idmap.find ( tmppid ) ;
    if ( ! (*f).first )
      idmap [tmppid] = tmpreal ;

  }

  catalog.close () ;
}


ResolvePublicId::~ResolvePublicId ()
{

  if ( --classcounter )
    return ;                                
			       
  map <char *, char *, strLess>::iterator i = idmap.begin () ;
  for ( ; i != idmap.end () ; i++ ) {
    delete (*i).first ;
    delete (*i).second ;
  }
}
