/*-------------------------------------------------------------------------
         PROJECT YAO PORTABLE OBJECT-ORIENTED ENTITY MANAGER (POEM)
                         Version 1.0 Alpha level 2.0
  -------------------------------------------------------------------------
  (C) 1993-1994 Yuan-ze Institute of Technology
 
  Distributed under the terms of the License and Disclaimer of Warranties
  for Project YAO Materials, which includes the following text:
 
  "YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
  worldwide, non-exclusive license to use, execute, reproduce, display,
  perform and distribute copies of, and to prepare derivative works based
  upon these materials; and (2) the right to authorize others to do any of
  the foregoing."
 
  The full text of the License and Disclaimer of Warranties for Project YAO
  Materials should be consulted before using these materials.
  -------------------------------------------------------------------------
               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE
 
     Only a "final" version is subject to the YUAN-ZE license described
       above. A version is final if it is not identified as "Alpha" or
     "Beta", or if no later version is released within 60 days.  Interim
     releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// poem.h
///  Portable Object-oriented Entity Manager
//   Erik Naggum 1994-04-12
  -------------------------------------------------------------------------*/
 
#ifndef _POEM_H_
#define _POEM_H_

#ifndef __GNUG__
#pragma interface
#endif
 
class POEM {
 
public:
 
    class Marker {
    public:
        Marker (long mm1, long mm2): m1(mm1), m2(mm2) { }
        long get (int which) const { return (--which)? m2 : m1; }
    private:
        long m1, m2;
    };
 
    struct Adjust {
        int buf;
        int pos;
        int adj;
    };
 
    class AdjustRing {
    public:
        AdjustRing (int size);
        ~AdjustRing (void);
 
        Adjust& top (void) { return ring[ringtop]; }
        Adjust& item (int index) { return ring[modsize (ringtop - index)]; }
        int bottom (int index) { return modsize (ringtop - index) == ringbot; }
        int size (void) { return ringsize; }
        void reset (int pos);
        void push (int pos, int delta, int buf = 0);
        void push (Adjust old);
    private:
        int modsize (int i) { return i & ringsize - 1; }
        int ringsize;
        Adjust * ring;
        int ringtop;
        int ringbot;
    };
 
    struct InternalLocation {
        InternalLocation (void);
        InternalLocation (InternalLocation *);
 
        // entity location
 
        long chars;                     // characters
        long columns;                   // characters in record
        long records;                   // records
 
        // buffer location
 
        int point;                      // into buffer
 
        // buffer positioning information
 
        struct bufferinfo {
            long position;                      // file position
            int objspec;                                // value at buffer read
            int dimspec;                                        // value at buffer read
        } buffer[2], write, read;
    };
 
    class DimSpec {
    public:
        DimSpec (long m1, long m2): raw (m1, m2), file (-1, -1) { }
        long getraw (int which) const { return raw.get (which); }
        long getfile (int which) const { return file.get (which); }
        void setfile (long m1, long m2) { file = Marker (m1, m2); }
 
    private:
        DimSpec (const DimSpec&);       // not defined
        DimSpec& operator = (const DimSpec&); // not defined
 
        Marker raw, file;
    };
 
 
    typedef unsigned short character;
 
    // abstract class
    class StorageObject {
    public:
        //  template for StorageSystem
        //  static StorageObject * create (const char *);
 
        StorageObject (const char *);
        virtual ~StorageObject (void);
        virtual int open (DimSpec *);
        virtual int read (char *, int);
        virtual int close (void);
        virtual long seek (long, int = 0);
        virtual long tell (void) const;
 
        int valid (void) const { return status != invalid; }
        int eof (void) const { return status == ateof; }
        const char * geterror (void) const { return error; }
 
        enum Status { invalid, closed, opened, ateof };
 
    protected:
        virtual int  sys_open (const char *) = 0;
        virtual long sys_seek (int, long, int) = 0;
        virtual int  sys_read (int, char *, int) = 0;
        virtual int  sys_close (int) = 0;
 
        long position;
        DimSpec * dimspec;
        char * objectname;
        char * error;
        int openhandle;
        Status status;
 
    private:
        StorageObject& operator = (const StorageObject&); // not defined
        StorageObject (const StorageObject&);             // not defined
    };
 
 
    class ObjSpec {
    public:
        ObjSpec (const char * storage, const char * sos,
                 const char * rbi = 0, const char * dimspeclist = 0);
        ~ObjSpec (void);
 
        int valid (void) const { return status != invalid; }
        int RE (void) const { return inRE; }
        int RS (void) const { return inRS; }
        int overrun_error (void) const { return overrun == error; }
        void setRERS (int RE, int RS) { inRE = RE; inRS = RS; }
 
        StorageObject * getobject (void) const { return object; }
        DimSpec * getdimspec (int i = 0) const { return dimspec[i]; }
 
    private:
        ObjSpec operator = (const ObjSpec&); // not defined
        ObjSpec (const ObjSpec&);            // not defined
 
        StorageObject * object;
        DimSpec ** dimspec;             // dynarray
        int inRE, inRS;
        enum Overrun { trunc, error } overrun;
        enum Status { invalid, parsed } status;
    };
 
    class Entity {
    public:
        Entity (const char * entname, const char * object, int literal = 0);
        ~Entity (void);
 
        class Location {
            friend Entity;
        public:
            Location (const Entity *in);
            long charno (void) const { return location->chars; }
            long colno (void) const { return location->columns; }
            long recno (void) const { return location->records; }
            int valid (const Entity *check) const { return check == source; }
        private:
            const Entity * source;
            InternalLocation * location;
        };
 
        int open (int maxlookahead = 1024);
        int open (Location * at, int maxlookahead = 1024);
        int seek (Location * to);
        Location tell (void) { return Location (this); }
        void close (void);
 
        void setsyntax (int RE, int RS, int EE = 0, int QUE = 127);
 
        character peek (int ahead);
        int step (int ahead);
        int get (char * buffer, int length);
        character get (void) { character c = peek (0); step (1); return c; }
 
        enum Status { invalid, closed, opened };
        int valid (void) const { return status != invalid; }
 
    private:
        Entity (const Entity&);         // not defined
        Entity& operator = (const Entity&); // not defined
 
        int swapbuffer (void);
        int fillbuffer (int which);
        int stuff (int which, int where, int what);
 
        Status status;
        int exhausted;                  // entity is exhausted
 
        struct synchars {
            character RE;               // returned record end
            character RS;               // returned record start
            character EE;               // returned entity end
            character QUE;              // replacement character
        } syntax;
 
        const char * name;              // not duplicated
        ObjSpec ** objspec;             // dynarray
        friend Location;
        InternalLocation * location;    // location at point
        character *buffer[2];           // input read buffers
        int bufsize[2];                 // ... and their sizes
        int lookahead;                  // minimum payload in buffer
        int lastahead;                  //
        AdjustRing * adjustring;        // lookahead adjustment
 
        int lastseen;                   // magic
    };
 
    class StorageSystem {
    public:
        typedef StorageObject *(*creator)(const char *);
 
        StorageSystem (const char * nm, creator creat): name(nm), create(creat) { }
        StorageSystem (int): name(0), create(0) { }
        static int addsystem (const char *, creator);
        static StorageObject * (*getsystem(const char *))(const char *);
    private:
        StorageSystem& operator = (StorageSystem&); // not defined
        StorageSystem (const StorageSystem&);         // not defined
        static void initsystem (void);
        const char * name;
        creator create;
        static StorageSystem **list;
        static int systems;
        static int maxsystems;
    };
 
    class FileStorageObject : public StorageObject {
    public:
        // required for StorageSystem::addsystem
        static StorageObject * create (const char *);
 
        FileStorageObject (const char *);
        ~FileStorageObject (void);
 
    protected:
        int  sys_open (const char *);
        long sys_seek (int, long, int);
        int  sys_read (int, char *, int);
        int  sys_close (int);
 
    private:
        FileStorageObject& operator = (FileStorageObject&); // not defined
 
#ifdef OPENFILELIMIT
        static FileStorageObject *openlist[OPENFILELIMIT];
#endif
    };
 
    class StreamStorageObject : public StorageObject {
    public:
        // required for StorageSystem::addsystem
        static StorageObject * create (const char *);
 
        StreamStorageObject (const char *);
        ~StreamStorageObject (void);
 
    protected:
        int  sys_open (const char *);
        long sys_seek (int, long, int);
        int  sys_read (int, char *, int);
        int  sys_close (int);
 
    private:
        StreamStorageObject& operator = (StreamStorageObject&); // not defined
    };
 
    class LiteralStorageObject : public StorageObject {
    public:
        // required for StorageSystem::addsystem
        static StorageObject * create (const char *);
 
        LiteralStorageObject (const char *);
        ~LiteralStorageObject (void);
 
    protected:
        int  sys_open (const char *);
        long sys_seek (int, long, int);
        int  sys_read (int, char *, int);
        int  sys_close (int);
 
    private:
        LiteralStorageObject& operator = (LiteralStorageObject&); // not defined
    };
};
 
#endif
 
// local variables:
// mode: c++
// end:
