/*-------------------------------------------------------------------------
      PROJECT YAO REFERENCE APPLICATION FOR SGML TESTING (RAST)
                   Version 1.0 Alpha level 2.1
 An SGML application conforming to International Standard ISO 8879 --
               Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

            INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------

          Reference Application for SGML Testing (RAST)
            Conforms to X3.190-199X, draft 3/14/1991
//// arg.c
///  Process arguments from the command line
//   Pierre G. Richard 1989-07-18 - Petrus Services and Consulting, Ltd.

CHANGE HISTORY
#   Date       Who?          Change & Reason (brief)
  -------------------------------------------------------------------------*/

/******************************************************************************/

/*-------------+
|Include Files |
+-------------*/
#include <stdio.h>
#include "gendef.h"
#include "message.h"
#include "arg.h"

/*-------------------------------------------------------------------Arguments-+
| Arguments constructor: decode the options and parameters on the command line |
+-----------------------------------------------------------------------------*/
Arguments::Arguments(
   int argc, const char * const * argv, const char * const pszTerminalName
) {
   is.bOk = TRUE;
   is.bDebugMode   = FALSE;
   is.bRact        = FALSE;
   is.bSortRequired = TRUE;
   is.bProduction  = FALSE;
   is.bUsageSaid   = FALSE;
   pszPrimaryFileSpec = pszTerminalName;
   pszCmsgFileSpec = "rast.cmo";
   pszRactFileSpec = "ract.ctr";
   pszRastFileSpec = "rast.ctr";
   pszOcsFileSpec = 0;

   unsigned int parm_pos = 0;         // count for positional parm
   while (--argc) {
      const char *pch;
      if ((*(pch = *++argv) == SEPOPT) || (*pch == SEPOPT2)) {
redo:    switch (*++pch) {            /* keywords parm */
         case 'd':
         case 'D':
            is.bDebugMode = TRUE;
            if (*(pch + 1)) goto redo;
            break;
         case 'h':
         case 'H':
            usage();
            if (*(pch + 1)) goto redo;
            break;
         case 'k':
         case 'K':
            is.bRact = TRUE;
            if (*(pch + 1)) pszRactFileSpec = pch + 1;
            break;
         case 'm':
         case 'M':
            is.bCanonicalMsg = TRUE;
            if (*(pch + 1)) pszCmsgFileSpec = pch + 1;
            break;
         case 's':
         case 'S':
            if (*(pch + 1)) pszOcsFileSpec = pch + 1;
            break;
         case 'u':
         case 'U':
            is.bSortRequired = FALSE;
            if (*(pch + 1)) goto redo;
            break;
         case 'p':
         case 'P':
            is.bProduction = TRUE;
            break;
         default :
            is.bOk = FALSE;
            break;
         }
      }else {
         if (*pch == '?') {
            usage();
            break;
         }
         switch(++parm_pos) {            // Positional parms
         case 1:
            pszPrimaryFileSpec = pch;    // get the file name
            break;
         case 2:
            pszRastFileSpec = pch;       // get the rast output filename
            break;
         default:
            is.bOk = FALSE;              // no more than 2 allowed
            break;
         }
      }
   }
   if (!is.bOk) {
      RastMessage::bad_syntax();         // send explanations
      usage();
   }
}

/*-----------------------------------------------------------------------usage-+
| Issue a message on how to use this code, if not already done                 |
+-----------------------------------------------------------------------------*/
void Arguments::usage()
{
   if (!is.bUsageSaid) {
      is.bUsageSaid = TRUE;
      RastMessage::usage();
   }
}
/*============================================================================*/
