/*-------------------------------------------------------------------------
      PROJECT YAO REFERENCE APPLICATION FOR SGML TESTING (RAST)
                   Version 1.0 Alpha level 2.1
 An SGML application conforming to International Standard ISO 8879 --
               Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

            INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------

          Reference Application for SGML Testing (RAST)
            Conforms to X3.190-199X, draft 3/14/1991

//// files.c
///  File IO, and formatting functions
//   Pierre G. Richard 1989-07-18 - Petrus Services and Consulting, Ltd.

CHANGE HISTORY
#   Date       Who?          Change & Reason (brief)
  -------------------------------------------------------------------------*/

/******************************************************************************/

/*-------------+
|Include Files |
+-------------*/
#include <ctype.h>
#include "gendef.h"
#include "arg.h"
#include "files.h"
#include "message.h"

/*-------------------------------------------------------------------RastFiles-+
| Open the rast files                                                          |
+-----------------------------------------------------------------------------*/
RastFiles::RastFiles(Arguments* pArg) : arg(*pArg)
{
   isOk = FALSE;
   if (!arg.is_arg_ok()) return;

   isMarkupData = FALSE;
   isValidOutput = TRUE;
   iColPos = 0;
   if (arg.is_debug_mode()) {
      fpCmsg = fpRast = fpRact = stdout;
      isOk = TRUE;
   }else {
      const char *pch;
      if (fpRast = fopen(pch=arg.rast_file_spec(), "w"), !fpRast) {
         /* ... */
      }else if (
         (arg.is_canonical_msg()) && (
            fpCmsg = fpRast, strcmp(arg.rast_file_spec(), arg.cmsg_file_spec())
         ) && (
            fpCmsg = fopen(pch=arg.cmsg_file_spec(), "w"), !fpCmsg
         )
      ) {
         fclose(fpRast);
      }else if (
         (arg.is_ract()) && (
            fpRact = fpRast, strcmp(arg.rast_file_spec(), arg.ract_file_spec())
         ) && (
            fpRact = fopen(pch=arg.ract_file_spec(), "w"), !fpRact
         )
      ) {
         fclose(fpRast);
         fclose(fpCmsg);
      }else {
         isOk = TRUE;
      }
   }
}

/*-------------------------------------------------------------------RastFiles-+
| Close all rast files                                                         |
+-----------------------------------------------------------------------------*/
RastFiles::~RastFiles()
{
   if (!isOk) return;
   if (!arg.is_debug_mode()) {           // don't close stdout!
      if (isValidOutput) fclose(fpRast);
      if ((arg.is_canonical_msg()) && (fpCmsg != fpRast)) fclose(fpCmsg);
      if ((arg.is_ract()) && (fpRact != fpRast)) fclose(fpRact);
   }
}

/*-----------------------------------------------------------reset_rast_output-+
| Restart the RAST output file from the beginning                              |
+-----------------------------------------------------------------------------*/
void RastFiles::reset_rast_output()
{
   boolean isCmsgAsRast = (
      (arg.is_canonical_msg()) && (fpCmsg==fpRast)
   )? TRUE : FALSE;
   boolean isRactAsRast = (
      (arg.is_ract()) && (fpRact==fpRast)
   )? TRUE : FALSE;

   if (fpRast = freopen(arg.rast_file_spec(), "w", fpRast), !fpRast) {
      RastMessage::open_failed(arg.rast_file_spec());
      isValidOutput = FALSE;
   }else {
      if (isCmsgAsRast) fpCmsg = fpRast;
      if (isRactAsRast) fpRact = fpRast;
   }
}

/*-------------------------------------------------------------------box_print-+
| Displays a buffer: no more than LENGTH_ALLOWED characters per line,          |
| surrounded by delimiters.  Delimiters depends on markup/data mode.           |
+-----------------------------------------------------------------------------*/
void RastFiles::box_print(const char *pchText, int iLen)
{
   int i;

   while (iLen) {
      if (!iColPos) {
         if (isMarkupData) {
            print(RastText::somd());
         }else {
            print(RastText::sodl());
         }
      }
      if (iLen >= (i = LENGTH_ALLOWED - iColPos)) {
         print(pchText, i);
         if (isMarkupData) {
            print(RastText::eomd());
         }else {
            print(RastText::eodl());
         }
         iColPos = 0;
         iLen -= i;
         pchText += i;
      }else {
         print(pchText, iLen);
         iColPos += iLen;
         break;
      }
   }
}

/*---------------------------------------------------------[19]data_line_print-+
|[19] data_line =                                                              |
|   ("|", DATA CHARACTER+, "|", LE) | special character                        |
|                                                                              |
|[21] special character =                                                      |
|   ("#RS", LE) | ("#RE", LE) | ("#TAB", LE) | ("#", character_number, LE)     |
|                                                                              |
|[22] character_number = nonnegative integer                                   |
|                                                                              |
|[23] nonnegative integer =                                                    |
|   "0" | (non-zero digit, ("0" | non-zero digit)*)                            |
|                                                                              |
|[24] non-zero digit = "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"     |
|                                                                              |
+-----------------------------------------------------------------------------*/
void RastFiles::data_line_print(const char *pchText, int iLen)
{
   const char *pch = pchText;

   while (iLen--) {
      if (isprint(*pch)) {
         ++pch;
      }else {
         if (pch > pchText) box_print(pchText, pch - pchText);
         close_box();
         switch (*pch) {
         case '\r':
            print(RastText::re());
            break;
         case '\n':
            print(RastText::rs());
            break;
         case '\t':
            print(RastText::tab());
            break;
         default:
            fmt_print("#%d\n", (int)(*pch));
            break;
         }
         pchText = ++pch;
      }
   }
   if (pch > pchText) box_print(pchText, pch - pchText);
}

/*-------------------------------------------------------------[20]markup_data-+
|[20] markup data =                                                            |
|  ("!", DATA CHARACTER+, "!", LE) | special character                         |
+-----------------------------------------------------------------------------*/
void RastFiles::markup_data_print(const char *pszData)
{
   MarkupDataSession session(*this);
   data_line_print(pszData);
}

/*-------------------------------------------------------------------close_box-+
| Terminate displayed line, if any (add the ending delimiter)                  |
+-----------------------------------------------------------------------------*/
void RastFiles::close_box()
{
   if (iColPos) {
      if (isMarkupData) {
         print(RastText::eomd());
      }else {
         print(RastText::eodl());
      }
      iColPos = 0;
   }
}
/******************************************************************************/
