/*-------------------------------------------------------------------------
      PROJECT YAO REFERENCE APPLICATION FOR SGML TESTING (RAST)
                   Version 1.0 Alpha level 2.1
 An SGML application conforming to International Standard ISO 8879 --
               Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

            INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------

          Reference Application for SGML Testing (RAST)
            Conforms to X3.190-199X, draft 3/14/1991

//// files.h
///  Declarations for files
//   Pierre G. Richard 1989-07-18 - Petrus Services and Consulting, Ltd.

CHANGE HISTORY
#   Date       Who?          Change & Reason (brief)
  -------------------------------------------------------------------------*/
#ifndef _FILES_H
#define _FILES_H 1

/******************************************************************************/

#include <stdio.h>
#include <stdarg.h>

/*-----------------------------------------------------------------------------+
| Rast files                                                                   |
+-----------------------------------------------------------------------------*/
class RastFiles {
   friend class CanonicalMsgSession;
   friend class MarkupDataSession;
public:
   RastFiles(Arguments* pArg);
   ~RastFiles();
   boolean is_files_ok();
   void reset_rast_output();
   void print(const char *pchText, int iLen);
   void print(const char *pszText);
   void box_print(const char *pchText, int iLen);
   void box_print(const char *pszText);
   void data_line_print(const char *pszText);
   void data_line_print(const char *pszText, int iLen);
   void markup_data_print(const char *pszData);
   void close_box();
   void fmt_print(const char *pszFmt, ...);
   void ract_print(const char * pszCapaName, unsigned long ulCapaValue);
private:
   enum { LENGTH_ALLOWED = 60 };  // Maximum line-length on output (box)
   int iColPos;
   boolean isValidOutput;
   boolean isMarkupData;
   boolean isOk;
   Arguments &arg;
   FILE   *fpRast;
   FILE   *fpRact;
   FILE   *fpCmsg;
};

/*--- INLINES ---*/
inline boolean RastFiles::is_files_ok() { return isOk; }
inline void RastFiles::print(const char *pchText, int iLen) {
   if (isValidOutput) fwrite(pchText, iLen, 1, fpRast);
}
inline void RastFiles::print(const char *pchText) {
   if (isValidOutput) fwrite(pchText, strlen(pchText), 1, fpRast);
}
inline void RastFiles::box_print(const char *pszText) {
   box_print(pszText, strlen(pszText));
}
inline void RastFiles::data_line_print(const char *pszText) {
   data_line_print(pszText, strlen(pszText));
}
inline void RastFiles::fmt_print(const char *pszFmt, ...)  {
   va_list vaArg;
   va_start(vaArg, pszFmt);
   if (isValidOutput) vfprintf(fpRast, pszFmt, vaArg);
   va_end(vaArg);
}
inline void RastFiles::ract_print(
   const char * pszCapaName, unsigned long ulCapaValue
) {
   fprintf(fpRact, "%s %lu\n", pszCapaName, ulCapaValue);
}

/*-----------------------------------------------------------------------------+
| Redirected Output (session for canonical messages)                           |
+-----------------------------------------------------------------------------*/
class CanonicalMsgSession {
public:
   CanonicalMsgSession(RastFiles& rfArg);
   ~CanonicalMsgSession();
private:
   RastFiles &rf;
   FILE *fpRast;
};

/*--- INLINES ---*/
inline CanonicalMsgSession::CanonicalMsgSession(RastFiles& rfArg) : rf(rfArg) {
   fpRast = rf.fpRast;
   if (rf.fpCmsg == rf.fpRast) {      // Set rast_output to cmsg_output
      rf.close_box();                 // this can break a line...
   }else {
      rf.fpRast = rf.fpCmsg;
   }
}

inline CanonicalMsgSession::~CanonicalMsgSession()
{
   rf.fpRast = fpRast;
}

/*-----------------------------------------------------------------------------+
| Markup Data session                                                          |
+-----------------------------------------------------------------------------*/
class MarkupDataSession {
public:
   MarkupDataSession(RastFiles& rfArg);
   ~MarkupDataSession();
private:
   RastFiles &rf;
   FILE *fpRast;
};

/*--- INLINES ---*/
inline MarkupDataSession::MarkupDataSession(RastFiles& rfArg) : rf(rfArg) {
   rf.isMarkupData = TRUE;
}

inline MarkupDataSession::~MarkupDataSession()
{
   rf.close_box();
   rf.isMarkupData = FALSE;
}
#endif /* _FILES_H ************************************************************/
