/*-------------------------------------------------------------------------
      PROJECT YAO REFERENCE APPLICATION FOR SGML TESTING (RAST)
                   Version 1.0 Alpha level 2.1
 An SGML application conforming to International Standard ISO 8879 --
               Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

            INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------

          Reference Application for SGML Testing (RAST)
            Conforms to X3.190-199X, draft 3/14/1991

//// message.c
///  Issue messages - keep it apart for NLS reasons
//   Pierre G. Richard 1989-07-18 - Petrus Services and Consulting, Ltd.

CHANGE HISTORY
#   Date       Who?          Change & Reason (brief)
  -------------------------------------------------------------------------*/

/******************************************************************************/

/*-------------+
|Include Files |
+-------------*/
#include <iostream.h>
#include "gendef.h"
#include "message.h"
#include "parser.h"

/*-----------------------------------------------------------------------usage-+
| Explain how to use this code                                                 |
+-----------------------------------------------------------------------------*/
void RastMessage::usage()
{
   static const char * const msg[] = {
      " rast [<file_in>] [<rast_out>] [options]",
      " \t(C) Copyright IBM Corporation 1992-1994.",
      " \tLicensed to Yuan-ze Institute of Technology.",
      " \tReference Application for SGML Testing.",
      " \tConforms to the draft from X3V1.8, dated 3/14/1991.",
      " \tThe SGML Parser is:",
      "%",
      "",
      " <file_in>:  name of file to be parsed, default to \"stdin\"",
      " <rast_out>: name of file to contain the RAST output,",
      "             default is: \"rast.ctr\"",
      " options: (d, m, s, and u do not conform to the ANSI standard)",
      "  -d: traces the RAST output and parser messages-id on \"stdout\"",
      "      ** this debug option voids all output user specifications **",
      "  -h: displays this help text",
      "  -k[<ract_out>]: produces in the file <ract_out> the RACT result;",
      "      can be same as <rast.out>; deflt: \"ract.ctr\"",
      "  -m[<pmsg_out>]: outputs in the file <pmsg_out> a canonical form",
      "    of parser messages; can be same as <rast_out); deflt: \"rast.cmo\"",
      "  -s<ocsname>: uses the OCS file \"<ocsname>.ocs\"",
      "  -u: suppresses sorting of attribute names",
      "  -p: production mode: OK to use pre-compiled DTD, and OCS",
      "",
      "Version 2.00 - Pierre G. Richard, Petrus Services & Consulting"
   };
   unsigned int count = sizeof(msg)/sizeof(char *);
   const char * const * ppch =msg;
   while (count--) {
      if (**ppch == '%') {
         cerr << SGMLparser::identify() << endl;
      }else {
         cerr << *ppch << endl;
      }
      ++ppch;
   }
};

/*-----------------------------------------------------------------open_failed-+
| Open failed                                                                  |
+-----------------------------------------------------------------------------*/
void RastMessage::open_failed(const char *pszFileName) {
   cerr << "RST0001: Cannot open file: \""
        << pszFileName << "\"; ended" << endl;
}

/*------------------------------------------------------------------bad_syntax-+
| Invalid command syntax                                                       |
+-----------------------------------------------------------------------------*/
void RastMessage::bad_syntax() {
   cerr << "RST0002: Command syntax invalid; ended" << endl;
}

/*-------------------------------------------------------incompatible_versions-+
| Parser has obsolete version                                                  |
+-----------------------------------------------------------------------------*/
void RastMessage::incompatible_versions(
   const char *pszExpected, const char *pszFound
) {
   cerr << "RST0003: Mismatch in parser versions:  OSGML:\""
        << pszFound << " RAST: " << pszExpected << "\"; ended" << endl;
}

/*-----------------------------------------------------------unrecognizable_PI-+
| A RAST PI is not conforming                                                  |
+-----------------------------------------------------------------------------*/
void RastMessage::unrecognizable_PI() {
   cerr << "RST0004: Not a RAST processing instruction; discarded" << endl;
}
/*============================================================================*/
