/*-------------------------------------------------------------------------
      PROJECT YAO REFERENCE APPLICATION FOR SGML TESTING (RAST)
                   Version 1.0 Alpha level 2.1
 An SGML application conforming to International Standard ISO 8879 --
               Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

            INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------

          Reference Application for SGML Testing (RAST)
            Conforms to X3.190-199X, draft 3/14/1991

//// rast.c
///  RAST Application Object
//   Pierre G. Richard 1989-07-18 - Petrus Services and Consulting, Ltd.

CHANGE HISTORY
#   Date       Who?          Change & Reason (brief)
001 1992-09-20 P.G.Richard   Migrated from CTR to RAST
002 1992-10-30 P.G.Richard   Added the -m option
003 1994-02-28 C.F.Goldfarb  Converted to C++ (with const) and ObjectSGML
004 1994-03-27 P.G.Richard   Reworked a few methods
  -------------------------------------------------------------------------*/

/******************************************************************************/

/*-------------+
|Include Files |
+-------------*/
#include "gendef.h"
#include "rast.h"
#include "message.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*------------------------------------------------------------------------Rast-+
| Constructor for Rast Application                                             |
+-----------------------------------------------------------------------------*/
Rast::Rast(
   int argc, const char * const * argv, const char * pszTerminalName
) :
   Arguments(argc, argv, pszTerminalName),  // Process command line arguments
   RastFiles(this),                         // Get the files
   doc(                                     // Establish parsing context:
      Arguments::primary_file_spec(),       // - Primary EntityDef File Spec
      Arguments::ocs_file_spec(),           // - Compiled SGML decl (if any)
      pszTerminalName,                      // - Reserved name for terminal
      Arguments::is_production(),           // - OK to use / produce an ODT
      Arguments::is_sort_required()         // - Parser must sort attlists
   ),
   pSgml(0)                                 // useless if bad arguments
{
   if (is_files_ok()) {
      pSgml = new SGMLparser(doc);
   }
}

/*---------------------------------------------------------------------process-+
| Main loop                                                                    |
+-----------------------------------------------------------------------------*/
int Rast::process()
{
   if (!pSgml) return 8;

   (*pSgml)                       // filter events: remove non-ESIS
   >> SGML_START_EV
   >> SGML_END_EV
   >> DOCTYPE_START_EV
   >> DOCTYPE_END_EV
   >> LINKTYPE_START_EV
   >> LINKTYPE_END_EV
   >> SECTION_START_EV
   >> SECTION_END_EV
   >> ENTITY_END_EV
   >> COMMENT_EV
   >> ERROR_EV
   >> ATT_VAL_INVALID_EV
   >> ATT_VAL_TOKEN_IDREF_EV      // ATT_VAL_TOKEN_ENTITY_EV only
   >> ATT_VAL_TOKEN_NAME_EV
   >> ATT_VAL_TOKEN_NMTOKEN_EV
   >> ATT_VAL_TOKEN_NUMBER_EV
   >> ATT_VAL_TOKEN_NUTOKEN_EV
   >> ENT_INT_SGML_EV
   >> ENT_INT_PI_EV
   >> ENT_INT_STARTTAG_EV
   >> ENT_INT_ENDTAG_EV
   >> ENT_INT_MS_EV
   >> ENT_INT_MD_EV
   >> ENT_EXT_SGML_EV;
   next_event();

   for (;;) {
      switch (evType) {
      case ELEMENT_START_EV:  process_element_start();          break;
      case ELEMENT_END_EV:    process_element_end();            break;
      case DATA_EV:           process_data();                   break;
      case RE_EV:             process_record_end();             break;
      case ENTITY_START_EV:   process_entity_start();           break;
      case PI_EV:             process_processing_instruction(); break;
      case EOP_EV:
         if (is_ract()) ract_result();
         return 0;
      default:                next_event();   break;
      }
   }
}

/*-------------------------------------------------[13]external_id_information-+
| On entry to this routine, current event is one of:                           |
|         ENT_EXT_SUBDOC_EV                                                    |
|         ENT_EXT_CDATA_EV                                                     |
|         ENT_EXT_NDATA_EV                                                     |
|         ENT_EXT_SDATA_EV                                                     |
|         ENT_NOTATION_EV                                                      |
|         ATT_VAL_NOTATION_EV                                                  |
|                                                                              |
|[13] external id information =                                                |
|   public identifier information?,                                            |
|   system identifier information?                                             |
|                                                                              |
|[14] public identifier information =                                          |
|   "#PUBLIC", LE, (markup data+ | (#EMPTY, LE))                               |
|                                                                              |
|[15] system identifier information =                                          |
|   "#SYSTEM", LE, (markup data+ | ((#EMPTY | #NONE), LE))?                    |
+-----------------------------------------------------------------------------*/
void Rast::external_id_information()
{
   const char *pszValue;
   for (;;) {
      switch (next_event()) {
      case PUBLIC_ID_EV:
         print(RastText::public_st());
         pszValue = pSgml->zstring();
         if (*pszValue) {
            markup_data_print(pszValue);
         }else {
            print(RastText::empty());
         }
         break;
      case SYSTEM_ID_EV:
         print(RastText::system());
         pszValue = pSgml->zstring();
         if (*pszValue) {
            markup_data_print(pszValue);
         }else {
            print(RastText::empty());
         }
         break;
      case SYSTEM_ID_OMITTED_EV:
         print(RastText::system());
         print(RastText::none());
         break;
      default:                       // if not for this function, go back
         return;
      }
   }
}

/*---------------------------------------------------------[16]external_entity-+
| On entry to this routine, current event is one of:                           |
|    ENT_EXT_SUBDOC_EV                                                         |
|    ENT_EXT_CDATA_EV                                                          |
|    ENT_EXT_NDATA_EV                                                          |
|    ENT_EXT_SDATA_EV                                                          |
|                                                                              |
| [16]external_entity =                                                        |
|    "[&", NAME, LE, external entity information, "]", LE                      |
+-----------------------------------------------------------------------------*/
void Rast::external_entity(const char *pszEntName)
{
   print(RastText::ero());
   print(pszEntName);
   print(RastText::le());
   external_entity_information();
   print(RastText::refc());
}

/*---------------------------------------------------[18]internal_sdata_entity-+
| On entry to this routine, current event is:                                  |
|    ENT_INT_SDATA_EV                                                          |
|    ATT_VAL_COMP_SDATA_EV                                                     |
|                                                                              |
|[18] internal sdata entity =                                                  |
|   "#SDATA-TEXT", LE, markup data+, "#END-SDATA", LE                          |
+-----------------------------------------------------------------------------*/
void Rast::internal_sdata_entity()
{
   print(RastText::sdata_start());
   markup_data_print(pSgml->zstring());
   print(RastText::sdata_end());
   next_event();
}

/*-------------------------------------------------------[9]entity_information-+
| On entry to this routine, current event is:                                  |
|    ATT_VAL_TOKEN_ENTITY_EV                                                   |
|    ATT_VAL_ENTITY_EV                                                         |
|                                                                              |
|[9] entity information =                                                      |
|   (external entity information | internal entity information),               |
|   "#END-ENTITY"                                                              |
+-----------------------------------------------------------------------------*/
void Rast::entity_information()
{
   switch (next_event()) {
   case ENT_EXT_SUBDOC_EV:
   case ENT_EXT_CDATA_EV:
   case ENT_EXT_NDATA_EV:
   case ENT_EXT_SDATA_EV:
      external_entity_information();
      break;
   case ENT_INT_CDATA_EV:
   case ENT_INT_SDATA_EV:
      internal_entity_information();
      break;
   default:                          // if not for this function, go back
      break;
   }
   print(RastText::entity_end());
}

/*---------------------------------------------[10]external_entity_information-+
| On entry to this routine, current event is one of:                           |
|    ENT_EXT_SUBDOC_EV                                                         |
|    ENT_EXT_CDATA_EV                                                          |
|    ENT_EXT_NDATA_EV                                                          |
|    ENT_EXT_SDATA_EV                                                          |
|                                                                              |
|[10] external entity information =                                            |
|   ("#SUBDOC", LE, external id information) |                                 |
|   ("#CDATA-EXTERNAL", LE, entity information details) |                      |
|   ("#SDATA-EXTERNAL", LE, entity information details) |                      |
|   ("#NDATA-EXTERNAL", LE, entity information details)                        |
|                                                                              |
|[11] entity information details =                                             |
|   external id information,                                                   |
|   "#NOTATION=", NAME, LE,                                                    |
|   external id information,                                                   |
|   attribute information*                                                     |
+-----------------------------------------------------------------------------*/
void Rast::external_entity_information()
{
   switch (evType) {
   case ENT_EXT_SUBDOC_EV: print(RastText::subdoc());     break;
   case ENT_EXT_CDATA_EV:  print(RastText::cdata_ext());  break;
   case ENT_EXT_NDATA_EV:  print(RastText::sdata_ext());  break;
   case ENT_EXT_SDATA_EV:  print(RastText::ndata_ext());  break;
   default:                print(RastText::unkwn_ext());  break;
   }
   external_id_information();
   if (evType == ENT_NOTATION_EV) {
      print(RastText::notation());
      print(pSgml->zstring());
      print(RastText::le());
      external_id_information();
      if (evType == DATA_ATT_NAME_EV) attribute_information();
   }
}

/*---------------------------------------------[12]internal_entity_information-+
| On entry to this routine, current event is one of:                           |
|    ENT_INT_CDATA_EV                                                          |
|    ENT_INT_SDATA_EV                                                          |
|                                                                              |
|[12] internal entity information =                                            |
|   ("#CDATA-INTERNAL", LE, markup data*) |                                    |
|   ("#SDATA-INTERNAL", LE, markup data*)                                      |
+-----------------------------------------------------------------------------*/
void Rast::internal_entity_information()
{
   switch (evType) {
   case ENT_INT_CDATA_EV: print(RastText::cdata_int());  break;
   case ENT_INT_SDATA_EV: print(RastText::sdata_int());  break;
   default:               print(RastText::unkwn_int());  break;
   }
   markup_data_print(pSgml->zstring());
   next_event();
}

/*-----------------------------------------------------------composite_att_val-+
| Describe the SDATA entities composing a CDATA attribute value                |
+-----------------------------------------------------------------------------*/
void Rast::composite_att_val()
{
   MarkupDataSession session(*this);
   next_event();
   for (;;) {
      switch (evType) {
      case ATT_VAL_COMP_DATA_EV:
      case ATT_VAL_COMP_CDATA_EV:
         box_print(pSgml->zstring());
         next_event();
         break;
      case ATT_VAL_COMP_SDATA_EV:
         close_box();
         internal_sdata_entity();
         break;
      default:                       // if not for this function, go back
         return;
      }
   }
}

/*----------------------------------------------------[8]attribute_information-+
| On entry to this routine, current event is:                                  |
|    ATT_NAME_EV                                                               |
|    DATA_ATT_NAME_EV                                                          |
|                                                                              |
|[8] attribute information =                                                   |
|   NAME, "=" LE,                                                              |
|   (                                                                          |
|      ("#IMPLIED | LE)                                                        |
|      |                                                                       |
|      (                                                                       |
|         (markup data | internal sdata entity)*,                              |
|         (external id information | entity information+)?                     |
|      )                                                                       |
|   )                                                                          |
|                                                                              |
| Notes:                                                                       |
|                                                                              |
| 1) CDATA attribute values may be the result of resolving references to       |
|    SDATA entities.  ESIS requires that we unravel these references.          |
|    This is done by the ObjectSGML parser, which returns each                 |
|    component as a separate event.                                            |
|                                                                              |
| 2) Sorting was performed by parser if requested by application               |
|    when parsing context is established.                                      |
|                                                                              |
+-----------------------------------------------------------------------------*/
void Rast::attribute_information()
{
   const e_Event evTypeAtt = evType;
   do {
      print(pSgml->zstring());
      print(RastText::equal());
      next_event();
      switch (evType) {
      case ATT_VAL_IMPLIED_EV:
         print(RastText::implied());
         next_event();
         break;
      case ATT_VAL_COMPOSITE_EV:
      case ATT_VAL_COMPOSITE_SDATA_EV:
         composite_att_val();
         break;
      case ATT_VAL_ENTITY_EV:
         markup_data_print(pSgml->zstring());
         entity_information();
         break;
      case ATT_VAL_TOKENS_EV:
         markup_data_print(pSgml->zstring());
         while (next_event() == ATT_VAL_TOKEN_ENTITY_EV) {
            entity_information();
         };
         break;
      case ATT_VAL_NOTATION_EV:
         markup_data_print(pSgml->zstring());
         external_id_information();
         break;
      default:
         markup_data_print(pSgml->zstring());
         next_event();
         break;
      }
   }while (evType == evTypeAtt);
}

/*----------------------------------------------------[6]process_element_start-+
| Notification of a Start-Tag                                                  |
|                                                                              |
| [6] element start =                                                          |
|    "[", NAME, (LE, attribute information+)?,  link information?,  "]", LE    |
+-----------------------------------------------------------------------------*/
void Rast::process_element_start()
{
   close_box();
   print(RastText::stago());
   print(pSgml->zstring());
   if (next_event() == ATT_NAME_EV) {
      print(RastText::le());
      attribute_information();
   }
   print(RastText::tagc());
}

/*------------------------------------------------------[7]process_element_end-+
| Notification of an End-Tag                                                   |
|                                                                              |
| [7] element end = "[/", NAME, "]", LE                                        |
+-----------------------------------------------------------------------------*/
void Rast::process_element_end()
{
   close_box();
   print(RastText::etago());
   print(pSgml->zstring());
   print(RastText::tagc());
   next_event();
}

/*----------------------------------------------------------process_record_end-+
| Notification of a Record End                                                 |
+-----------------------------------------------------------------------------*/
void Rast::process_record_end()
{
   close_box();
   print(RastText::re());
   next_event();
}

/*----------------------------------------------------------------process_data-+
| Notification of Data                                                         |
+-----------------------------------------------------------------------------*/
void Rast::process_data()
{
   data_line_print(pSgml->zstring());
   next_event();
}

/*----------------------------------------------process_processing_instruction-+
| Notification of Processing Instruction                                       |
|                                                                              |
|[17]processing instrution data =                                              |
|  "[?", (LE, data line+)?, "]", LE                                            |
+-----------------------------------------------------------------------------*/
void Rast::process_processing_instruction()
{
   close_box();
   const char *pszValue = pSgml->zstring();
   if (*pszValue) {
      print(RastText::pio());
      data_line_print(pszValue);
      close_box();
      print(RastText::pic());
   }else {
      print(RastText::empty_pi());
   }
   next_event();
}

/*--------------------------------------------------------process_entity_start-+
| Notification of an Entity                                                    |
+-----------------------------------------------------------------------------*/
void Rast::process_entity_start()
{
   const char * pszEntName = pSgml->zstring();
   switch (next_event()) {
   case ENT_INT_CDATA_EV:
      data_line_print(pSgml->zstring());
      next_event();
      break;
   case ENT_INT_SDATA_EV:
      close_box();
      internal_sdata_entity();
      break;
   case ENT_EXT_SUBDOC_EV:
   case ENT_EXT_CDATA_EV:
   case ENT_EXT_NDATA_EV:
   case ENT_EXT_SDATA_EV:
      close_box();
      external_entity(pszEntName);  // Entity name
      break;
   default:                         // if not for this function, go back
      break;
   }
}

/*-------------------------------------------------------------[33]ract_result-+
| [33] RACT result =                                                           |
|    totalcap, entcap, entchcap, elemcap, grpcap, exgrpcap,                    |
|    exnmcap, attcap, attchcap, avrgcap, notcap, notchcap,                     |
|    idcap, idrefcap, mapcap, lksetcap, lknmcap                                |
+-----------------------------------------------------------------------------*/
void Rast::ract_result()
{
   ract_print("TOTALCAP" , pSgml->capacity(TOTALCAP_ix));
   ract_print("ENTCAP"   , pSgml->capacity(ENTCAP_ix));
   ract_print("ENTCHCAP" , pSgml->capacity(ENTCHCAP_ix));
   ract_print("ELEMCAP"  , pSgml->capacity(ELEMCAP_ix));
   ract_print("GRPCAP"   , pSgml->capacity(GRPCAP_ix));
   ract_print("EXGRPCAP" , pSgml->capacity(EXGRPCAP_ix));
   ract_print("EXNMCAP"  , pSgml->capacity(EXNMCAP_ix));
   ract_print("ATTCAP"   , pSgml->capacity(ATTCAP_ix));
   ract_print("ATTCHCAP" , pSgml->capacity(ATTCHCAP_ix));
   ract_print("AVGRPCAP" , pSgml->capacity(AVGRPCAP_ix));
   ract_print("NOTCAP"   , pSgml->capacity(NOTCAP_ix));
   ract_print("NOTCHCAP" , pSgml->capacity(NOTCHCAP_ix));
   ract_print("IDCAP"    , pSgml->capacity(IDCAP_ix));
   ract_print("IDREFCAP" , pSgml->capacity(IDREFCAP_ix));
   ract_print("MAPCAP"   , pSgml->capacity(MAPCAP_ix));
   ract_print("LKSETCAP" , pSgml->capacity(LKSETCAP_ix));
   ract_print("LKNMCAP"  , pSgml->capacity(LKNMCAP_ix));
}
/******************************************************************************/
